@@ -23,16 +23,7 @@
 #include <windows.h>
 #endif
 
-#include <pthread.h> /* PTHREAD_* pthread_*() */
-
 #include <fcntl.h> /* open() */
-#include <unistd.h> /* select() */
-
-#include <assert.h> /* assert() */
-#include <errno.h> /* errno */
-#include <stddef.h> /* wchar_t NULL */
-#include <stdlib.h> /* EXIT_FAILURE _Exit() free() malloc() */
-#include <string.h>
 #include <sys/stat.h> /* O_RDONLY */
 #include <sys/types.h> /* pid_t ssize_t */
 #ifndef _WIN32
@@ -40,11 +31,21 @@
 #include <sys/time.h> /* timeval */
 #include <sys/wait.h> /* WEXITSTATUS() waitpid() */
 #endif
+#include <signal.h> /* kill() */
+#include <unistd.h> /* select() */
+
+#include <assert.h> /* assert() */
+#include <errno.h> /* errno */
+#include <stddef.h> /* NULL wchar_t */
+#include <stdlib.h> /* EXIT_FAILURE _Exit() free() malloc() */
+#include <string.h> /* memcpy() */
 
 #include "cfg/config.h"
+#include "compat/pthread.h"
 #include "modes/dialogs/msg_dialog.h"
 #include "ui/cancellation.h"
 #include "ui/statusline.h"
+#include "utils/cancellation.h"
 #include "utils/env.h"
 #include "utils/fs.h"
 #include "utils/log.h"
@@ -68,6 +69,16 @@
  * UI.
  *
  * Operations are displayed on designated job bar.
+ *
+ * On non-Windows systems background thread reads data from error streams of
+ * external applications, which are then displayed by main thread.  This thread
+ * maintains its own list of jobs (via err_next field), which is added to by
+ * building a temporary list with new_err_jobs pointing to its head.  Every job
+ * that has associated external process has the following life cycle:
+ *  1. Created by main thread and passed to error thread through new_err_jobs.
+ *  2. Either gets marked by signal handler or its stream reaches EOF.
+ *  3. It's in_use flag is reset.
+ *  4. Main thread frees corresponding entry.
  */
 
 /* Turns pointer (P) to field (F) of a structure (S) to address of that
@@ -94,48 +105,71 @@ typedef struct
 {
 	bg_task_func func; /* Function to execute in a background thread. */
 	void *args;        /* Argument to pass. */
-	job_t *job;        /* Job identifier that corresponds to the task. */
+	bg_job_t *job;     /* Job identifier that corresponds to the task. */
 }
 background_task_args;
 
-static void job_check(job_t *const job);
-static void job_free(job_t *const job);
+static void job_check(bg_job_t *const job);
+static void job_free(bg_job_t *const job);
 #ifndef _WIN32
-static job_t * add_background_job(pid_t pid, const char cmd[], int fd,
+static void * error_thread(void *p);
+static int update_job_list(bg_job_t **jobs, fd_set *set);
+static void report_error_msg(const char title[], const char text[]);
+static void append_error_msg(bg_job_t *job, const char err_msg[]);
+static bg_job_t * add_background_job(pid_t pid, const char cmd[], int fd,
 		BgJobType type);
 #else
-static job_t * add_background_job(pid_t pid, const char cmd[], HANDLE hprocess,
-		BgJobType type);
+static bg_job_t * add_background_job(pid_t pid, const char cmd[],
+		HANDLE hprocess, BgJobType type);
 #endif
 static void * background_task_bootstrap(void *arg);
-static void set_current_job(job_t *job);
+static void set_current_job(bg_job_t *job);
 static void make_current_job_key(void);
+static int bg_op_cancel(bg_op_t *bg_op);
 
-job_t *jobs;
+bg_job_t *bg_jobs = NULL;
 
+#ifndef _WIN32
+/* Head of list of newly started jobs. */
+static bg_job_t *new_err_jobs;
+/* Mutex to protect new_err_jobs. */
+static pthread_mutex_t new_err_jobs_lock = PTHREAD_MUTEX_INITIALIZER;
+/* Conditional variable to signal availability of new jobs in new_err_jobs. */
+static pthread_cond_t new_err_jobs_cond = PTHREAD_COND_INITIALIZER;
+#endif
+
+/* Thread local storage for bg_job_t associated with active thread. */
 static pthread_key_t current_job;
-static pthread_once_t current_job_once = PTHREAD_ONCE_INIT;
 
 void
-init_background(void)
+bg_init(void)
 {
+#ifndef _WIN32
+	pthread_t id;
+	const int err = pthread_create(&id, NULL, &error_thread, NULL);
+	assert(err == 0);
+	(void)err;
+#endif
+
 	/* Initialize state for the main thread. */
 	set_current_job(NULL);
 }
 
 void
-add_finished_job(pid_t pid, int exit_code)
+bg_process_finished_cb(pid_t pid, int exit_code)
 {
-	job_t *job;
+	bg_job_t *job;
 
-	/* Mark any finished jobs */
-	job = jobs;
+	/* Mark any finished jobs. */
+	job = bg_jobs;
 	while(job != NULL)
 	{
 		if(job->pid == pid)
 		{
+			pthread_spin_lock(&job->status_lock);
 			job->running = 0;
 			job->exit_code = exit_code;
+			pthread_spin_unlock(&job->status_lock);
 			break;
 		}
 		job = job->next;
@@ -143,11 +177,11 @@ add_finished_job(pid_t pid, int exit_code)
 }
 
 void
-check_background_jobs(void)
+bg_check(void)
 {
-	job_t *head = jobs;
-	job_t *prev;
-	job_t *p;
+	bg_job_t *head = bg_jobs;
+	bg_job_t *prev;
+	bg_job_t *p;
 
 	/* Quit if there is no jobs or list is unavailable (e.g. used by another
 	 * invocation of this function). */
@@ -161,19 +195,25 @@ check_background_jobs(void)
 		return;
 	}
 
-	head = jobs;
-	jobs = NULL;
+	head = bg_jobs;
+	bg_jobs = NULL;
 
 	p = head;
 	prev = NULL;
 	while(p != NULL)
 	{
+		int can_remove;
+
 		job_check(p);
 
+		pthread_spin_lock(&p->status_lock);
+		can_remove = (!p->running && !p->in_use);
+		pthread_spin_unlock(&p->status_lock);
+
 		/* Remove job if it is finished now. */
-		if(!p->running)
+		if(can_remove)
 		{
-			job_t *j = p;
+			bg_job_t *j = p;
 			if(prev != NULL)
 				prev->next = p->next;
 			else
@@ -195,8 +235,8 @@ check_background_jobs(void)
 		}
 	}
 
-	assert(jobs == NULL && "Job list shouldn't be used by anyone.");
-	jobs = head;
+	assert(bg_jobs == NULL && "Job list shouldn't be used by anyone.");
+	bg_jobs = head;
 
 	bg_jobs_unfreeze();
 }
@@ -204,47 +244,28 @@ check_background_jobs(void)
 /* Checks status of the job.  Processes error stream or checks whether process
  * is still running. */
 static void
-job_check(job_t *const job)
+job_check(bg_job_t *const job)
 {
 #ifndef _WIN32
-	fd_set ready;
-	int max_fd = 0;
-	struct timeval ts = { .tv_sec = 0, .tv_usec = 1000 };
+	char *new_errors;
 
-	/* Setup pipe for reading */
-	FD_ZERO(&ready);
-	if(job->fd >= 0)
+	/* Display portions of errors from the job while there are any. */
+	do
 	{
-		FD_SET(job->fd, &ready);
-		max_fd = job->fd;
-	}
+		pthread_spin_lock(&job->errors_lock);
+		new_errors = job->new_errors;
+		job->new_errors = NULL;
+		job->new_errors_len = 0U;
+		pthread_spin_unlock(&job->errors_lock);
 
-	if(job->error != NULL)
-	{
-		if(!job->skip_errors)
+		if(new_errors != NULL && !job->skip_errors)
 		{
 			job->skip_errors = prompt_error_msg("Background Process Error",
-					job->error);
-		}
-		free(job->error);
-		job->error = NULL;
-	}
-
-	while(select(max_fd + 1, &ready, NULL, NULL, &ts) > 0)
-	{
-		char err_msg[ERR_MSG_LEN];
-
-		const ssize_t nread = read(job->fd, err_msg, sizeof(err_msg) - 1);
-		if(nread == 0)
-		{
-			break;
-		}
-		else if(nread > 0 && !job->skip_errors)
-		{
-			err_msg[nread] = '\0';
-			job->skip_errors = prompt_error_msg("Background Process Error", err_msg);
+					new_errors);
 		}
+		free(new_errors);
 	}
+	while(new_errors != NULL);
 #else
 	DWORD retcode;
 	if(GetExitCodeProcess(job->hprocess, &retcode) != 0)
@@ -257,19 +278,21 @@ job_check(job_t *const job)
 #endif
 }
 
-/* Frees resources allocated by the job as well as the job_t structure itself.
- * The job can be NULL. */
+/* Frees resources allocated by the job as well as the bg_job_t structure
+ * itself.  The job can be NULL. */
 static void
-job_free(job_t *const job)
+job_free(bg_job_t *const job)
 {
 	if(job == NULL)
 	{
 		return;
 	}
 
+	pthread_spin_destroy(&job->errors_lock);
+	pthread_spin_destroy(&job->status_lock);
 	if(job->type != BJT_COMMAND)
 	{
-		pthread_mutex_destroy(&job->bg_op_guard);
+		pthread_spin_destroy(&job->bg_op_lock);
 	}
 
 #ifndef _WIN32
@@ -285,18 +308,21 @@ job_free(job_t *const job)
 #endif
 	free(job->bg_op.descr);
 	free(job->cmd);
+	free(job->new_errors);
+	free(job->errors);
 	free(job);
 }
 
 /* Used for FUSE mounting and unmounting only. */
 int
-background_and_wait_for_status(char cmd[], int cancellable, int *cancelled)
+bg_and_wait_for_status(char cmd[],
+		const struct cancellation_t *cancellation, int *cancelled)
 {
 #ifndef _WIN32
 	pid_t pid;
 	int status;
 
-	if(cancellable)
+	if(cancellation_possible(cancellation))
 	{
 		*cancelled = 0;
 	}
@@ -327,11 +353,6 @@ background_and_wait_for_status(char cmd[], int cancellable, int *cancelled)
 		_Exit(127);
 	}
 
-	if(cancellable)
-	{
-		ui_cancellation_enable();
-	}
-
 	while(waitpid(pid, &status, 0) == -1)
 	{
 		if(errno != EINTR)
@@ -341,16 +362,12 @@ background_and_wait_for_status(char cmd[], int cancellable, int *cancelled)
 			status = -1;
 			break;
 		}
-		process_cancel_request(pid);
+		process_cancel_request(pid, cancellation);
 	}
 
-	if(cancellable)
+	if(cancellation_requested(cancellation))
 	{
-		if(ui_cancellation_requested())
-		{
-			*cancelled = 1;
-		}
-		ui_cancellation_disable();
+		*cancelled = 1;
 	}
 
 	(void)set_sigchld(0);
@@ -362,24 +379,8 @@ background_and_wait_for_status(char cmd[], int cancellable, int *cancelled)
 #endif
 }
 
-#ifndef _WIN32
-static void
-error_msg(const char *title, const char *text)
-{
-	job_t *job = pthread_getspecific(current_job);
-	if(job == NULL)
-	{
-		show_error_msg(title, text);
-	}
-	else
-	{
-		(void)replace_string(&job->error, text);
-	}
-}
-#endif
-
 int
-background_and_wait_for_errors(char cmd[], int cancellable)
+bg_and_wait_for_errors(char cmd[], const struct cancellation_t *cancellation)
 {
 #ifndef _WIN32
 	pid_t pid;
@@ -388,7 +389,7 @@ background_and_wait_for_errors(char cmd[], int cancellable)
 
 	if(pipe(error_pipe) != 0)
 	{
-		error_msg("File pipe error", "Error creating pipe");
+		report_error_msg("File pipe error", "Error creating pipe");
 		return -1;
 	}
 
@@ -413,12 +414,7 @@ background_and_wait_for_errors(char cmd[], int cancellable)
 
 		close(error_pipe[1]); /* Close write end of pipe. */
 
-		if(cancellable)
-		{
-			ui_cancellation_enable();
-		}
-
-		wait_for_data_from(pid, NULL, error_pipe[0]);
+		wait_for_data_from(pid, NULL, error_pipe[0], cancellation);
 
 		buf[0] = '\0';
 		while((nread = read(error_pipe[0], linebuf, sizeof(linebuf) - 1)) > 0)
@@ -432,18 +428,13 @@ background_and_wait_for_errors(char cmd[], int cancellable)
 				strncat(buf, linebuf, sizeof(buf) - strlen(buf) - 1);
 			}
 
-			wait_for_data_from(pid, NULL, error_pipe[0]);
+			wait_for_data_from(pid, NULL, error_pipe[0], cancellation);
 		}
 		close(error_pipe[0]);
 
-		if(cancellable)
-		{
-			ui_cancellation_disable();
-		}
-
 		if(result != 0)
 		{
-			error_msg("Background Process Error", buf);
+			report_error_msg("Background Process Error", buf);
 		}
 		else
 		{
@@ -464,8 +455,174 @@ background_and_wait_for_errors(char cmd[], int cancellable)
 }
 
 #ifndef _WIN32
+/* Entry point of a thread which reads input from input of active background
+ * programs.  Does not return. */
+static void *
+error_thread(void *p)
+{
+	bg_job_t *jobs = NULL;
+
+	(void)pthread_detach(pthread_self());
+	block_all_thread_signals();
+
+	while(1)
+	{
+		fd_set active, ready;
+		const int max_fd = update_job_list(&jobs, &active);
+		struct timeval timeout = { .tv_sec = 0, .tv_usec = 250*1000 }, ts = timeout;
+
+		while(select(max_fd + 1, memcpy(&ready, &active, sizeof(active)), NULL,
+					NULL, &ts) > 0)
+		{
+			int need_update_list = (jobs == NULL);
+
+			bg_job_t **job = &jobs;
+			while(*job != NULL)
+			{
+				bg_job_t *const j = *job;
+				char err_msg[ERR_MSG_LEN];
+				ssize_t nread;
+
+				if(!FD_ISSET(j->fd, &ready))
+				{
+					goto next_job;
+				}
+
+				nread = read(j->fd, err_msg, sizeof(err_msg) - 1U);
+				if(nread < 0)
+				{
+					need_update_list = 1;
+					goto next_job;
+				}
+
+				if(nread == 0)
+				{
+					/* Reached EOF, cut this job out of our list and allow its
+					 * deletion. */
+					pthread_spin_lock(&j->status_lock);
+					j->in_use = 0;
+					*job = j->err_next;
+					pthread_spin_unlock(&j->status_lock);
+					continue;
+				}
+
+				err_msg[nread] = '\0';
+				append_error_msg(j, err_msg);
+
+			next_job:
+				job = &j->err_next;
+			}
+
+			if(!need_update_list)
+			{
+				pthread_mutex_lock(&new_err_jobs_lock);
+				need_update_list = (new_err_jobs != NULL);
+				pthread_mutex_unlock(&new_err_jobs_lock);
+			}
+			if(need_update_list)
+			{
+				break;
+			}
+
+			ts = timeout;
+		}
+	}
+	return NULL;
+}
+
+/* Updates *jobs by removing finished tasks and adding new ones.  Initializes
+ * *set in the process.  Returns max file descriptor number of the set. */
+static int
+update_job_list(bg_job_t **jobs, fd_set *set)
+{
+	int max_fd;
+	bg_job_t *new_jobs;
+	bg_job_t **job;
+
+	/* Add new tasks to internal list, wait if there are no jobs. */
+	pthread_mutex_lock(&new_err_jobs_lock);
+	while(*jobs == NULL && new_err_jobs == NULL)
+	{
+		pthread_cond_wait(&new_err_jobs_cond, &new_err_jobs_lock);
+	}
+	new_jobs = new_err_jobs;
+	new_err_jobs = NULL;
+	pthread_mutex_unlock(&new_err_jobs_lock);
+
+	/* Prepend new jobs to the list. */
+	while(new_jobs != NULL)
+	{
+		bg_job_t *const new_job = new_jobs;
+		new_jobs = new_jobs->err_next;
+
+		assert(new_job->type == BJT_COMMAND &&
+				"Only external commands should be here.");
+
+		new_job->err_next = *jobs;
+		*jobs = new_job;
+	}
+
+	FD_ZERO(set);
+
+	max_fd = 0;
+	job = jobs;
+	while(*job != NULL)
+	{
+		bg_job_t *const j = *job;
+
+		pthread_spin_lock(&j->status_lock);
+		/* If finished, reset in_use mark and drop it from the list. */
+		if(!j->running)
+		{
+			j->in_use = 0;
+			*job = j->err_next;
+			pthread_spin_unlock(&j->status_lock);
+			continue;
+		}
+		pthread_spin_unlock(&j->status_lock);
+
+		FD_SET(j->fd, set);
+		if(j->fd > max_fd)
+		{
+			max_fd = j->fd;
+		}
+
+		job = &j->err_next;
+	}
+
+	return max_fd;
+}
+
+/* Either displays error message to the user for foreground operations or saves
+ * it for displaying on the next invocation of bg_check(). */
+static void
+report_error_msg(const char title[], const char text[])
+{
+	bg_job_t *const job = pthread_getspecific(current_job);
+	if(job == NULL)
+	{
+		const int cancellation_state = ui_cancellation_pause();
+		show_error_msg(title, text);
+		ui_cancellation_resume(cancellation_state);
+	}
+	else
+	{
+		append_error_msg(job, text);
+	}
+}
+
+/* Appends message to error-related fields of the job. */
+static void
+append_error_msg(bg_job_t *job, const char err_msg[])
+{
+	pthread_spin_lock(&job->errors_lock);
+	(void)strappend(&job->errors, &job->errors_len, err_msg);
+	(void)strappend(&job->new_errors, &job->new_errors_len, err_msg);
+	pthread_spin_unlock(&job->errors_lock);
+}
+
 pid_t
-background_and_capture(char *cmd, int user_sh, FILE **out, FILE **err)
+bg_run_and_capture(char cmd[], int user_sh, FILE **out, FILE **err)
 {
 	pid_t pid;
 	int out_pipe[2];
@@ -566,6 +723,10 @@ background_and_capture_internal(char cmd[], int user_sh, FILE **out, FILE **err,
 		args[3] = NULL;
 	}
 
+	/* XXX: using unicode version of spawn breaks commands like `:!echo hi%S` when
+	 *      'shell' is "bash".  Windows version of bash seems to process some
+	 *      strings in unicode as ASCII.  Using narrow version on the other hand
+	 *      will break use of non-ASCII paths on English version of Windows... */
 	code = _wspawnvp(P_NOWAIT, args[0], (const wchar_t **)args);
 
 	free(wide_sh);
@@ -593,7 +754,7 @@ background_and_capture_internal(char cmd[], int user_sh, FILE **out, FILE **err,
 }
 
 pid_t
-background_and_capture(char cmd[], int user_sh, FILE **out, FILE **err)
+bg_run_and_capture(char cmd[], int user_sh, FILE **out, FILE **err)
 {
 	int out_fd, out_pipe[2];
 	int err_fd, err_pipe[2];
@@ -636,9 +797,9 @@ background_and_capture(char cmd[], int user_sh, FILE **out, FILE **err)
 #endif
 
 int
-start_background_job(const char *cmd, int skip_errors)
+bg_run_external(const char cmd[], int skip_errors)
 {
-	job_t *job = NULL;
+	bg_job_t *job = NULL;
 #ifndef _WIN32
 	pid_t pid;
 	int error_pipe[2];
@@ -697,7 +858,8 @@ start_background_job(const char *cmd, int skip_errors)
 
 		setpgid(0, 0);
 
-		execve(get_execv_path(cfg.shell), make_execv_array(cfg.shell, command), environ);
+		execve(get_execv_path(cfg.shell), make_execv_array(cfg.shell, command),
+				environ);
 		_Exit(127);
 	}
 	else
@@ -766,7 +928,6 @@ bg_execute(const char descr[], const char op_descr[], int total, int important,
 		bg_task_func task_func, void *args)
 {
 	pthread_t id;
-	pthread_attr_t attr;
 	int ret;
 
 	background_task_args *const task_args = malloc(sizeof(*task_args));
@@ -786,19 +947,6 @@ bg_execute(const char descr[], const char op_descr[], int total, int important,
 		return 1;
 	}
 
-	if(pthread_attr_init(&attr) != 0)
-	{
-		free(task_args);
-		return 1;
-	}
-
-	if(pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED) != 0)
-	{
-		free(task_args);
-		(void)pthread_attr_destroy(&attr);
-		return 1;
-	}
-
 	replace_string(&task_args->job->bg_op.descr, op_descr);
 	task_args->job->bg_op.total = total;
 
@@ -808,31 +956,32 @@ bg_execute(const char descr[], const char op_descr[], int total, int important,
 	}
 
 	ret = 0;
-	if(pthread_create(&id, &attr, &background_task_bootstrap, task_args) != 0)
+	if(pthread_create(&id, NULL, &background_task_bootstrap, task_args) != 0)
 	{
 		/* Mark job as finished with error. */
+		pthread_spin_lock(&task_args->job->status_lock);
 		task_args->job->running = 0;
 		task_args->job->exit_code = 1;
+		pthread_spin_unlock(&task_args->job->status_lock);
 
 		free(task_args);
 		ret = 1;
 	}
 
-	(void)pthread_attr_destroy(&attr);
 	return ret;
 }
 
 /* Creates structure that describes background job and registers it in the list
  * of jobs. */
 #ifndef _WIN32
-static job_t *
+static bg_job_t *
 add_background_job(pid_t pid, const char cmd[], int fd, BgJobType type)
 #else
-static job_t *
+static bg_job_t *
 add_background_job(pid_t pid, const char cmd[], HANDLE hprocess, BgJobType type)
 #endif
 {
-	job_t *new = malloc(sizeof(*new));
+	bg_job_t *new = malloc(sizeof(*new));
 	if(new == NULL)
 	{
 		show_error_msg("Memory error", "Unable to allocate enough memory");
@@ -841,26 +990,45 @@ add_background_job(pid_t pid, const char cmd[], HANDLE hprocess, BgJobType type)
 	new->type = type;
 	new->pid = pid;
 	new->cmd = strdup(cmd);
-	new->next = jobs;
+	new->next = bg_jobs;
+	new->skip_errors = 0;
+	new->new_errors = NULL;
+	new->new_errors_len = 0U;
+	new->errors = NULL;
+	new->errors_len = 0U;
+	new->cancelled = 0;
+
+	pthread_spin_init(&new->errors_lock, PTHREAD_PROCESS_PRIVATE);
+	pthread_spin_init(&new->status_lock, PTHREAD_PROCESS_PRIVATE);
+	new->running = 1;
+	new->in_use = (type == BJT_COMMAND);
+	new->exit_code = -1;
+
 #ifndef _WIN32
 	new->fd = fd;
+	if(fd != -1)
+	{
+		pthread_mutex_lock(&new_err_jobs_lock);
+		new->err_next = new_err_jobs;
+		new_err_jobs = new;
+		pthread_mutex_unlock(&new_err_jobs_lock);
+		pthread_cond_signal(&new_err_jobs_cond);
+	}
 #else
 	new->hprocess = hprocess;
 #endif
-	new->skip_errors = 0;
-	new->running = 1;
-	new->error = NULL;
 
 	if(type != BJT_COMMAND)
 	{
-		pthread_mutex_init(&new->bg_op_guard, NULL);
+		pthread_spin_init(&new->bg_op_lock, PTHREAD_PROCESS_PRIVATE);
 	}
 	new->bg_op.total = 0;
 	new->bg_op.done = 0;
 	new->bg_op.progress = -1;
 	new->bg_op.descr = NULL;
+	new->bg_op.cancelled = 0;
 
-	jobs = new;
+	bg_jobs = new;
 	return new;
 }
 
@@ -872,13 +1040,17 @@ background_task_bootstrap(void *arg)
 {
 	background_task_args *const task_args = arg;
 
+	(void)pthread_detach(pthread_self());
+	block_all_thread_signals();
 	set_current_job(task_args->job);
 
 	task_args->func(&task_args->job->bg_op, task_args->args);
 
 	/* Mark task as finished normally. */
+	pthread_spin_lock(&task_args->job->status_lock);
 	task_args->job->running = 0;
 	task_args->job->exit_code = 0;
+	pthread_spin_unlock(&task_args->job->status_lock);
 
 	free(task_args);
 
@@ -887,9 +1059,11 @@ background_task_bootstrap(void *arg)
 
 /* Stores pointer to the job in a thread-local storage. */
 static void
-set_current_job(job_t *job)
+set_current_job(bg_job_t *job)
 {
-	pthread_once(&current_job_once, &make_current_job_key);
+	static pthread_once_t once = PTHREAD_ONCE_INIT;
+	pthread_once(&once, &make_current_job_key);
+
 	(void)pthread_setspecific(current_job, job);
 }
 
@@ -903,8 +1077,8 @@ make_current_job_key(void)
 int
 bg_has_active_jobs(void)
 {
-	const job_t *job;
-	int bg_op_count;
+	bg_job_t *job;
+	int running;
 
 	if(bg_jobs_freeze() != 0)
 	{
@@ -913,18 +1087,18 @@ bg_has_active_jobs(void)
 		return 1;
 	}
 
-	bg_op_count = 0;
-	for(job = jobs; job != NULL; job = job->next)
+	running = 0;
+	for(job = bg_jobs; job != NULL && !running; job = job->next)
 	{
-		if(job->running && job->type == BJT_OPERATION)
+		if(job->type == BJT_OPERATION)
 		{
-			++bg_op_count;
+			running |= bg_job_is_running(job);
 		}
 	}
 
 	bg_jobs_unfreeze();
 
-	return bg_op_count > 0;
+	return running;
 }
 
 int
@@ -944,18 +1118,73 @@ bg_jobs_unfreeze(void)
 	(void)set_sigchld(0);
 }
 
+int
+bg_job_cancel(bg_job_t *job)
+{
+	int was_cancelled;
+
+	if(job->type != BJT_COMMAND)
+	{
+		return !bg_op_cancel(&job->bg_op);
+	}
+
+	was_cancelled = job->cancelled;
+#ifndef _WIN32
+	if(kill(job->pid, SIGINT) == 0)
+	{
+		job->cancelled = 1;
+	}
+	else
+	{
+		LOG_SERROR_MSG(errno, "Failed to send SIGINT to %" PRINTF_ULL,
+				(unsigned long long)job->pid);
+	}
+#else
+	if(win_cancel_process(job->pid, job->hprocess) == 0)
+	{
+		job->cancelled = 1;
+	}
+	else
+	{
+		LOG_SERROR_MSG(errno, "Failed to send WM_CLOSE to %" PRINTF_ULL,
+				(unsigned long long)job->pid);
+	}
+#endif
+	return !was_cancelled;
+}
+
+int
+bg_job_cancelled(bg_job_t *job)
+{
+	if(job->type != BJT_COMMAND)
+	{
+		return bg_op_cancelled(&job->bg_op);
+	}
+	return job->cancelled;
+}
+
+int
+bg_job_is_running(bg_job_t *job)
+{
+	int running;
+	pthread_spin_lock(&job->status_lock);
+	running = job->running;
+	pthread_spin_unlock(&job->status_lock);
+	return running;
+}
+
 void
 bg_op_lock(bg_op_t *bg_op)
 {
-	job_t *const job = STRUCT_FROM_FIELD(job_t, bg_op, bg_op);
-	pthread_mutex_lock(&job->bg_op_guard);
+	bg_job_t *const job = STRUCT_FROM_FIELD(bg_job_t, bg_op, bg_op);
+	pthread_spin_lock(&job->bg_op_lock);
 }
 
 void
 bg_op_unlock(bg_op_t *bg_op)
 {
-	job_t *const job = STRUCT_FROM_FIELD(job_t, bg_op, bg_op);
-	pthread_mutex_unlock(&job->bg_op_guard);
+	bg_job_t *const job = STRUCT_FROM_FIELD(bg_job_t, bg_op, bg_op);
+	pthread_spin_unlock(&job->bg_op_lock);
 }
 
 void
@@ -974,5 +1203,33 @@ bg_op_set_descr(bg_op_t *bg_op, const char descr[])
 	bg_op_changed(bg_op);
 }
 
+/* Convenience method to cancel background job.  Returns previous version of the
+ * cancellation flag. */
+static int
+bg_op_cancel(bg_op_t *bg_op)
+{
+	int was_cancelled;
+
+	bg_op_lock(bg_op);
+	was_cancelled = bg_op->cancelled;
+	bg_op->cancelled = 1;
+	bg_op_unlock(bg_op);
+
+	bg_op_changed(bg_op);
+	return was_cancelled;
+}
+
+int
+bg_op_cancelled(bg_op_t *bg_op)
+{
+	int cancelled;
+
+	bg_op_lock(bg_op);
+	cancelled = bg_op->cancelled;
+	bg_op_unlock(bg_op);
+
+	return cancelled;
+}
+
 /* vim: set tabstop=2 softtabstop=2 shiftwidth=2 noexpandtab cinoptions-=(0 : */
 /* vim: set cinoptions+=t0 filetype=c : */
