@@ -56,6 +56,7 @@
 #include "../utils/utils.h"
 #include "../cmd_core.h"
 #include "../filelist.h"
+#include "../flist_hist.h"
 #include "../marks.h"
 #include "../opt_handlers.h"
 #include "../status.h"
@@ -123,7 +124,6 @@ cfg_init(void)
 	cfg.auto_execute = 0;
 	cfg.time_format = strdup(" %m/%d %H:%M");
 	cfg.wrap_quick_view = 1;
-	cfg.use_iec_prefixes = 0;
 	cfg.undo_levels = 100;
 	cfg.sort_numbers = 0;
 	cfg.follow_links = 1;
@@ -216,6 +216,13 @@ cfg_init(void)
 
 	cfg.fast_file_cloning = 0;
 	cfg.cvoptions = 0;
+
+	cfg.case_override = 0;
+	cfg.case_ignore = 0;
+
+	cfg.sizefmt.base = 1024;
+	cfg.sizefmt.precision = 0;
+	cfg.sizefmt.ieci_prefixes = 0;
 }
 
 void
@@ -553,7 +560,7 @@ setup_dirs(void)
 {
 	LOG_FUNC_ENTER;
 
-	char rc_file[PATH_MAX];
+	char rc_file[PATH_MAX + 8];
 
 	if(is_dir(cfg.config_dir))
 	{
@@ -588,8 +595,8 @@ copy_help_file(void)
 {
 	LOG_FUNC_ENTER;
 
-	char src[PATH_MAX];
-	char dst[PATH_MAX];
+	char src[PATH_MAX + 16];
+	char dst[PATH_MAX + 16];
 
 	io_args_t args = {
 		.arg1.src = src,
@@ -609,7 +616,7 @@ copy_help_file(void)
 static void
 create_scripts_dir(void)
 {
-	char scripts[PATH_MAX];
+	char scripts[PATH_MAX + 16];
 	char readme[PATH_MAX];
 	FILE *fp;
 
@@ -642,8 +649,8 @@ copy_rc_file(void)
 {
 	LOG_FUNC_ENTER;
 
-	char src[PATH_MAX];
-	char dst[PATH_MAX];
+	char src[PATH_MAX + 16];
+	char dst[PATH_MAX + 16];
 
 	io_args_t args = {
 		.arg1.src = src,
@@ -841,8 +848,8 @@ cfg_resize_histories(int new_len)
 
 	if(old_len == 0)
 	{
-		save_view_history(&lwin, NULL, NULL, -1);
-		save_view_history(&rwin, NULL, NULL, -1);
+		flist_hist_save(&lwin, NULL, NULL, -1);
+		flist_hist_save(&rwin, NULL, NULL, -1);
 	}
 }
 
@@ -902,9 +909,9 @@ reduce_view_history(FileView *view, int size)
 	memmove(view->history, view->history + delta,
 			sizeof(history_t)*(view->history_num - delta));
 
-	if(view->history_num >= size)
+	if(view->history_num > size)
 	{
-		view->history_num = size - 1;
+		view->history_num = size;
 	}
 	view->history_pos -= delta;
 }
