@@ -49,6 +49,15 @@ typedef enum
 }
 DotDirs;
 
+/* Tweaks of case sensitivity. */
+typedef enum
+{
+	CO_PATH_COMPL = 1 << 0, /* Completion of paths. */
+	CO_GOTO_FILE  = 1 << 1, /* Navigation to files via f/F/,/;. */
+	NUM_CASE_OPTS =      2  /* Number of case options for compile checks. */
+}
+CaseOpts;
+
 /* Possible flags that regulate key suggestions. */
 typedef enum
 {
@@ -105,10 +114,11 @@ file_dec_t;
 
 typedef struct config_t
 {
-	char home_dir[PATH_MAX];   /* Ends with a slash. */
-	char config_dir[PATH_MAX]; /* Where local configuration files are stored. */
-	char colors_dir[PATH_MAX]; /* Where local color files are stored. */
-	char data_dir[PATH_MAX];   /* Where to store data files. */
+	char home_dir[PATH_MAX];       /* Ends with a slash. */
+	char config_dir[PATH_MAX];     /* Where local configuration files are
+	                                  stored. */
+	char colors_dir[PATH_MAX + 8]; /* Where local color files are stored. */
+	char data_dir[PATH_MAX];       /* Where to store data files. */
 
 	/* This one should be set using set_trash_dir() function. */
 	char trash_dir[PATH_MAX];
@@ -126,7 +136,6 @@ typedef struct config_t
 	int history_len;
 
 	int auto_execute;
-	int use_iec_prefixes;
 	int wrap_quick_view;
 	char *time_format;
 	/* This one should be set using cfg_set_fuse_home() function. */
@@ -248,6 +257,23 @@ typedef struct config_t
 
 	/* Whether various things should be reset on entering/leaving custom views. */
 	int cvoptions;
+
+	/* Tweaks of case sensitivity.  Values of these variables change the default
+	 * behaviour with regard to case sensitivity of various aspects.  These are
+	 * bit sets of CO_* flags. */
+	int case_override; /* Flag set here means the fact of the override. */
+	int case_ignore;   /* Flag here means case should be either always ignored or
+	                      always respected. */
+
+	/* Settings of size formatting. */
+	struct
+	{
+		int base;          /* Base of units to use (either 1000 or 1024). */
+		int precision;     /* Number of digits after dot to consider (0 for old
+		                      behaviour). */
+		int ieci_prefixes; /* When base is 1024, whether to say KiB instead of K. */
+	}
+	sizefmt;
 }
 config_t;
 
