@@ -34,6 +34,7 @@
 #include "../ui/fileview.h"
 #include "../ui/ui.h"
 #include "../utils/file_streams.h"
+#include "../utils/filemon.h"
 #include "../utils/filter.h"
 #include "../utils/fs.h"
 #include "../utils/log.h"
@@ -47,7 +48,9 @@
 #include "../cmd_core.h"
 #include "../dir_stack.h"
 #include "../filelist.h"
+#include "../flist_hist.h"
 #include "../filetype.h"
+#include "../filtering.h"
 #include "../marks.h"
 #include "../opt_handlers.h"
 #include "../registers.h"
@@ -61,12 +64,12 @@ static void get_sort_info(FileView *view, const char line[]);
 static void append_to_history(hist_t *hist, void (*saver)(const char[]),
 		const char item[]);
 static void ensure_history_not_full(hist_t *hist);
-static void get_history(FileView *view, int reread, const char *dir,
-		const char *file, int pos);
+static void get_history(FileView *view, int reread, const char dir[],
+		const char file[], int rel_pos);
 static void set_view_property(FileView *view, char type, const char value[]);
 static int copy_file(const char src[], const char dst[]);
 static int copy_file_internal(FILE *const src, FILE *const dst);
-static void update_info_file(const char filename[]);
+static void update_info_file(const char filename[], int merge);
 static void process_hist_entry(FileView *view, const char dir[],
 		const char file[], int pos, char ***lh, int *nlh, int **lhp, size_t *nlhp);
 static char * convert_old_trash_path(const char trash_path[]);
@@ -88,7 +91,8 @@ static void write_view_history(FILE *fp, FileView *view, const char str[],
 static void write_history(FILE *fp, const char str[], char mark, int prev_count,
 		char *prev[], const hist_t *hist);
 static void write_registers(FILE *const fp, char *regs[], int nregs);
-static void write_dir_stack(FILE *const fp, char *dir_stack[], int ndir_stack);
+static void write_dir_stack(FILE *const fp, char *old_dir_stack[],
+		int nold_dir_stack);
 static void write_trash(FILE *const fp, char *trash[], int ntrash);
 static void write_general_state(FILE *const fp);
 static char * read_vifminfo_line(FILE *fp, char buffer[]);
@@ -99,13 +103,16 @@ static int read_optional_number(FILE *f);
 static int read_number(const char line[], long *value);
 static size_t add_to_int_array(int **array, size_t len, int what);
 
+/* Monitor to check for changes of vifminfo file. */
+static filemon_t vifminfo_mon;
+
 void
 read_info_file(int reread)
 {
 	/* TODO: refactor this function read_info_file() */
 
 	FILE *fp;
-	char info_file[PATH_MAX];
+	char info_file[PATH_MAX + 16];
 	char *line = NULL, *line2 = NULL, *line3 = NULL, *line4 = NULL;
 
 	snprintf(info_file, sizeof(info_file), "%s/vifminfo", cfg.config_dir);
@@ -113,6 +120,8 @@ read_info_file(int reread)
 	if((fp = os_fopen(info_file, "r")) == NULL)
 		return;
 
+	(void)filemon_from_file(info_file, &vifminfo_mon);
+
 	while((line = read_vifminfo_line(fp, line)) != NULL)
 	{
 		const char type = line[0];
@@ -256,7 +265,7 @@ read_info_file(int reread)
 		}
 		else if(type == LINE_TYPE_LWIN_HIST || type == LINE_TYPE_RWIN_HIST)
 		{
-			FileView *const view = (type == LINE_TYPE_LWIN_HIST ) ? &lwin : &rwin;
+			FileView *const view = (type == LINE_TYPE_LWIN_HIST) ? &lwin : &rwin;
 			if(line_val[0] == '\0')
 			{
 				if(!reread && view->history_num > 0)
@@ -267,8 +276,8 @@ read_info_file(int reread)
 			}
 			else if((line2 = read_vifminfo_line(fp, line2)) != NULL)
 			{
-				const int pos = read_optional_number(fp);
-				get_history(view, reread, line_val, line2, pos);
+				const int rel_pos = read_optional_number(fp);
+				get_history(view, reread, line_val, line2, rel_pos);
 			}
 		}
 		else if(type == LINE_TYPE_CMDLINE_HIST)
@@ -295,7 +304,7 @@ read_info_file(int reread)
 				{
 					if((line4 = read_vifminfo_line(fp, line4)) != NULL)
 					{
-						push_to_dirstack(line_val, line2, line3 + 1, line4);
+						dir_stack_push(line_val, line2, line3 + 1, line4);
 					}
 				}
 			}
@@ -364,7 +373,7 @@ static void
 get_sort_info(FileView *view, const char line[])
 {
 	char *const sort = curr_stats.restart_in_progress
-	                 ? ui_view_sort_list_get(view)
+	                 ? ui_view_sort_list_get(view, view->sort)
 	                 : view->sort;
 
 	int j = 0;
@@ -416,8 +425,8 @@ ensure_history_not_full(hist_t *hist)
 
 /* Loads single history entry from vifminfo into the view. */
 static void
-get_history(FileView *view, int reread, const char *dir, const char *file,
-		int pos)
+get_history(FileView *view, int reread, const char dir[], const char file[],
+		int rel_pos)
 {
 	const int list_rows = view->list_rows;
 
@@ -430,7 +439,7 @@ get_history(FileView *view, int reread, const char *dir, const char *file,
 	{
 		view->list_rows = 1;
 	}
-	save_view_history(view, dir, file, pos);
+	flist_hist_save(view, dir, file, rel_pos);
 	if(!reread)
 	{
 		view->list_rows = list_rows;
@@ -443,8 +452,7 @@ set_view_property(FileView *view, char type, const char value[])
 {
 	if(type == PROP_TYPE_DOTFILES)
 	{
-		const int bool_val = atoi(value);
-		view->hide_dot = bool_val;
+		set_dot_files_visible(view, !atoi(value));
 	}
 	else if(type == PROP_TYPE_AUTO_FILTER)
 	{
@@ -463,15 +471,22 @@ set_view_property(FileView *view, char type, const char value[])
 void
 write_info_file(void)
 {
-	char info_file[PATH_MAX];
-	char tmp_file[PATH_MAX];
+	char info_file[PATH_MAX + 16];
+	char tmp_file[PATH_MAX + 16];
 
 	(void)snprintf(info_file, sizeof(info_file), "%s/vifminfo", cfg.config_dir);
 	(void)snprintf(tmp_file, sizeof(tmp_file), "%s_%u", info_file, get_pid());
 
 	if(os_access(info_file, R_OK) != 0 || copy_file(info_file, tmp_file) == 0)
 	{
-		update_info_file(tmp_file);
+		filemon_t current_vifminfo_mon;
+		int vifminfo_changed;
+
+		vifminfo_changed = filemon_from_file(info_file, &current_vifminfo_mon) != 0
+		                || !filemon_equal(&vifminfo_mon, &current_vifminfo_mon);
+
+		update_info_file(tmp_file, vifminfo_changed);
+		(void)filemon_from_file(tmp_file, &vifminfo_mon);
 
 		if(rename_file(tmp_file, info_file) != 0)
 		{
@@ -535,7 +550,7 @@ copy_file_internal(FILE *const src, FILE *const dst)
 /* Reads contents of the filename file as an info file and updates it with the
  * state of current instance. */
 static void
-update_info_file(const char filename[])
+update_info_file(const char filename[], int merge)
 {
 	/* TODO: refactor this function update_info_file() */
 
@@ -562,7 +577,7 @@ update_info_file(const char filename[])
 
 	non_conflicting_marks = strdup(valid_marks);
 
-	if((fp = os_fopen(filename, "r")) != NULL)
+	if(merge && (fp = os_fopen(filename, "r")) != NULL)
 	{
 		size_t nlhp = 0UL, nrhp = 0UL, nbt = 0UL, nbmt = 0UL;
 		char *line = NULL, *line2 = NULL, *line3 = NULL, *line4 = NULL;
@@ -702,7 +717,7 @@ update_info_file(const char filename[])
 				if((line2 = read_vifminfo_line(fp, line2)) != NULL)
 				{
 					char *const trash_name = convert_old_trash_path(line_val);
-					if(exists_in_trash(trash_name) && !is_in_trash(trash_name))
+					if(!trash_includes(line2))
 					{
 						ntrash = add_to_string_array(&trash, ntrash, 2, trash_name, line2);
 					}
@@ -891,7 +906,7 @@ process_hist_entry(FileView *view, const char dir[], const char file[], int pos,
 		char ***lh, int *nlh, int **lhp, size_t *nlhp)
 {
 	if(view->history_pos + *nlh/2 == cfg.history_len - 1 ||
-			is_in_view_history(view, dir) || !is_dir(dir))
+			flist_hist_contains(view, dir) || !is_dir(dir))
 	{
 		return;
 	}
@@ -913,7 +928,7 @@ convert_old_trash_path(const char trash_path[])
 	if(!is_path_absolute(trash_path) && is_dir_writable(cfg.trash_dir))
 	{
 		char *const full_path = format_str("%s/%s", cfg.trash_dir, trash_path);
-		if(path_exists(full_path, DEREF))
+		if(path_exists(full_path, NODEREF))
 		{
 			return full_path;
 		}
@@ -951,7 +966,7 @@ write_options(FILE *const fp)
 	fprintf(fp, "=grepprg=%s\n", escape_spaces(cfg.grep_prg));
 	fprintf(fp, "=history=%d\n", cfg.history_len);
 	fprintf(fp, "=%shlsearch\n", cfg.hl_search ? "" : "no");
-	fprintf(fp, "=%siec\n", cfg.use_iec_prefixes ? "" : "no");
+	fprintf(fp, "=%siec\n", cfg.sizefmt.ieci_prefixes ? "" : "no");
 	fprintf(fp, "=%signorecase\n", cfg.ignore_case ? "" : "no");
 	fprintf(fp, "=%sincsearch\n", cfg.inc_search ? "" : "no");
 	fprintf(fp, "=%slaststatus\n", cfg.display_statusline ? "" : "no");
@@ -966,6 +981,8 @@ write_options(FILE *const fp)
 	fprintf(fp, "=shell=%s\n", escape_spaces(cfg.shell));
 	fprintf(fp, "=shortmess=%s\n",
 			escape_spaces(get_option_value("shortmess", OPT_GLOBAL)));
+	fprintf(fp, "=sizefmt=%s\n",
+			escape_spaces(get_option_value("sizefmt", OPT_GLOBAL)));
 #ifndef _WIN32
 	fprintf(fp, "=slowfs=%s\n", escape_spaces(cfg.slow_fs_list));
 #endif
@@ -997,6 +1014,8 @@ write_options(FILE *const fp)
 	fprintf(fp, "=]numberwidth=%d\n", rwin.num_width_g);
 	fprintf(fp, "=[%srelativenumber\n", (lwin.num_type_g & NT_REL) ? "" : "no");
 	fprintf(fp, "=]%srelativenumber\n", (rwin.num_type_g & NT_REL) ? "" : "no");
+	fprintf(fp, "=[%sdotfiles\n", lwin.hide_dot_g ? "" : "no");
+	fprintf(fp, "=]%sdotfiles\n", rwin.hide_dot_g ? "" : "no");
 
 	fprintf(fp, "%s", "=confirm=");
 	if(cfg.confirm & CONFIRM_DELETE)
@@ -1012,6 +1031,13 @@ write_options(FILE *const fp)
 		fprintf(fp, "%s", "nonrootparent,");
 	fprintf(fp, "\n");
 
+	fprintf(fp, "%s", "=caseoptions=");
+	if(cfg.case_override & CO_PATH_COMPL)
+		fprintf(fp, "%c", (cfg.case_ignore & CO_PATH_COMPL) ? 'p' : 'P');
+	if(cfg.case_override & CO_GOTO_FILE)
+		fprintf(fp, "%c", (cfg.case_ignore & CO_GOTO_FILE) ? 'g' : 'G');
+	fprintf(fp, "\n");
+
 	fprintf(fp, "%s", "=suggestoptions=");
 	if(cfg.sug.flags & SF_NORMAL)
 		fprintf(fp, "%s", "normal,");
@@ -1262,7 +1288,7 @@ write_view_history(FILE *fp, FileView *view, const char str[], char mark,
 		int prev_count, char *prev[], int pos[])
 {
 	int i;
-	save_view_history(view, NULL, NULL, -1);
+	flist_hist_save(view, NULL, NULL, -1);
 	fprintf(fp, "\n# %s window history (oldest to newest):\n", str);
 	for(i = 0; i < prev_count; i += 2)
 	{
@@ -1325,28 +1351,29 @@ write_registers(FILE *const fp, char *regs[], int nregs)
 	}
 }
 
-/* Writes directory stack to vifminfo file.  dir_stack is a list of length
- * ndir_stack entries (4 lines per entry) read from vifminfo. */
+/* Writes directory stack to vifminfo file.  old_dir_stack is a list of length
+ * nold_dir_stack entries (4 lines per entry) read from vifminfo. */
 static void
-write_dir_stack(FILE *const fp, char *dir_stack[], int ndir_stack)
+write_dir_stack(FILE *const fp, char *old_dir_stack[], int nold_dir_stack)
 {
 	fputs("\n# Directory stack (oldest to newest):\n", fp);
 	if(dir_stack_changed())
 	{
 		unsigned int i;
-		for(i = 0U; i < stack_top; ++i)
+		for(i = 0U; i < dir_stack_top; ++i)
 		{
-			fprintf(fp, "S%s\n\t%s\n", stack[i].lpane_dir, stack[i].lpane_file);
-			fprintf(fp, "S%s\n\t%s\n", stack[i].rpane_dir, stack[i].rpane_file);
+			dir_stack_entry_t *const entry = &dir_stack[i];
+			fprintf(fp, "S%s\n\t%s\n", entry->lpane_dir, entry->lpane_file);
+			fprintf(fp, "S%s\n\t%s\n", entry->rpane_dir, entry->rpane_file);
 		}
 	}
 	else
 	{
 		int i;
-		for(i = 0; i < ndir_stack; i += 4)
+		for(i = 0; i < nold_dir_stack; i += 4)
 		{
-			fprintf(fp, "S%s\n\t%s\n", dir_stack[i], dir_stack[i + 1]);
-			fprintf(fp, "S%s\n\t%s\n", dir_stack[i + 2], dir_stack[i + 3]);
+			fprintf(fp, "S%s\n\t%s\n", old_dir_stack[i], old_dir_stack[i + 1]);
+			fprintf(fp, "S%s\n\t%s\n", old_dir_stack[i + 2], old_dir_stack[i + 3]);
 		}
 	}
 }
@@ -1436,7 +1463,7 @@ static void
 put_sort_info(FILE *fp, char leading_char, const FileView *view)
 {
 	int i = -1;
-	const char *const sort = ui_view_sort_list_get(view);
+	const char *const sort = ui_view_sort_list_get(view, view->sort_g);
 
 	fputc(leading_char, fp);
 	while(++i < SK_COUNT && abs(sort[i]) <= SK_LAST)
