@@ -59,10 +59,12 @@
 #include "ui/color_scheme.h"
 #include "ui/colors.h"
 #include "ui/statusbar.h"
+#include "utils/env.h"
 #include "utils/fs.h"
 #include "utils/macros.h"
 #include "utils/path.h"
 #include "utils/str.h"
+#include "utils/string_array.h"
 #include "utils/utils.h"
 #include "bmarks.h"
 #include "cmd_core.h"
@@ -72,6 +74,7 @@
 
 static int earg_num(int argc, const char cmdline[]);
 static int cmd_ends_with_space(const char cmdline[]);
+static void complete_compare(const char str[]);
 static void complete_selective_sync(const char str[]);
 static void complete_wincmd(const char str[]);
 static void complete_help(const char *str);
@@ -95,6 +98,8 @@ static int is_dirent_targets_exec(const struct dirent *d);
 #ifdef _WIN32
 static void complete_with_shared(const char *server, const char *file);
 #endif
+static int file_matches(const char fname[], const char prefix[],
+		size_t prefix_len);
 
 int
 complete_args(int id, const cmd_info_t *cmd_info, int arg_pos, void *extra_arg)
@@ -144,7 +149,7 @@ complete_args(int id, const cmd_info_t *cmd_info, int arg_pos, void *extra_arg)
 
 			complete_real_option_names(start, scope);
 		}
-		else if(dollar == NULL)
+		else if(dollar == NULL && !starts_with_lit(arg, "v:"))
 		{
 			function_complete_name(arg, &start);
 		}
@@ -240,6 +245,10 @@ complete_args(int id, const cmd_info_t *cmd_info, int arg_pos, void *extra_arg)
 	{
 		bmarks_complete(argc, argv, arg);
 	}
+	else if(id == COM_COMPARE)
+	{
+		complete_compare(arg);
+	}
 	else if(id == COM_SYNC && cmd_info->emark)
 	{
 		complete_selective_sync(arg);
@@ -389,6 +398,31 @@ cmd_ends_with_space(const char cmdline[])
 	return (cmdline[0] == ' ');
 }
 
+/* Completes properties for directory comparison. */
+static void
+complete_compare(const char str[])
+{
+	static const char *lines[][2] = {
+		{ "byname",     "compare by file name" },
+		{ "bysize",     "compare by file size" },
+		{ "bycontents", "compare by file size and hash" },
+
+		{ "ofboth",     "use files of two views" },
+		{ "ofone",      "use files of two current view only" },
+
+		{ "listall",    "list all files" },
+		{ "listunique", "list only unique files" },
+		{ "listdups",   "list only duplicated files" },
+
+		{ "groupids",   "group files in two panes by ids" },
+		{ "grouppaths", "group files in two panes by paths" },
+
+		{ "skipempty",  "exclude empty files from comparison" },
+	};
+
+	complete_from_string_list(str, lines, ARRAY_LEN(lines), 0);
+}
+
 /* Completes properties for selective synchronization. */
 static void
 complete_selective_sync(const char str[])
@@ -399,6 +433,7 @@ complete_selective_sync(const char str[])
 		{ "localopts", "values of view-specific options" },
 		{ "filters",   "all filters" },
 		{ "filelist",  "list of files for custom views" },
+		{ "tree",      "tree structure for tree views" },
 		{ "all",       "as much as possible" },
 	};
 
@@ -531,8 +566,17 @@ complete_chown(const char *str)
 static void
 complete_filetype(const char *str)
 {
-	char *const typed_fname = get_typed_entry_fpath(get_current_entry(curr_view));
-	assoc_records_t ft = ft_get_all_programs(typed_fname);
+	dir_entry_t *const curr = get_current_entry(curr_view);
+	char *typed_fname;
+	assoc_records_t ft;
+
+	if(fentry_is_fake(curr))
+	{
+		return;
+	}
+
+	typed_fname = get_typed_entry_fpath(curr);
+	ft = ft_get_all_programs(typed_fname);
 
 	complete_progs(str, ft);
 	ft_assoc_records_free(&ft);
@@ -554,7 +598,7 @@ complete_progs(const char *str, assoc_records_t records)
 	for(i = 0; i < records.count; i++)
 	{
 		const assoc_record_t *const assoc = &records.list[i];
-		char cmd[NAME_MAX];
+		char cmd[NAME_MAX + 1];
 
 		(void)extract_cmd_name(assoc->command, 1, sizeof(cmd), cmd);
 
@@ -680,27 +724,22 @@ complete_highlight_arg(const char *str)
 static void
 complete_envvar(const char str[])
 {
-	extern char **environ;
-	char **p = environ;
 	const size_t len = strlen(str);
 
-	while(*p != NULL)
+	int env_count;
+	char **const env_lst = env_list(&env_count);
+
+	int i;
+	for(i = 0; i < env_count; ++i)
 	{
-		if(strncmp(*p, str, len) == 0)
+		if(strncmp(env_lst[i], str, len) == 0)
 		{
-			char *const equal = strchr(*p, '=');
-			/* Actually equal shouldn't be NULL unless environ content is corrupted.
-			 * But the check below won't harm. */
-			if(equal != NULL)
-			{
-				*equal = '\0';
-				vle_compl_add_match(*p, equal + 1);
-				*equal = '=';
-			}
+			vle_compl_add_match(env_lst[i], env_get(env_lst[i]));
 		}
-		p++;
 	}
 
+	free_string_array(env_lst, env_count);
+
 	vle_compl_finish_group();
 	vle_compl_add_last_match(str);
 }
@@ -746,7 +785,7 @@ fast_run_complete(const char cmd[])
 {
 	char *result = NULL;
 	const char *args;
-	char command[NAME_MAX];
+	char command[NAME_MAX + 1];
 	char *completed;
 
 	args = extract_cmd_name(cmd, 0, sizeof(command), command);
@@ -954,7 +993,7 @@ filename_completion_internal(DIR *dir, const char filename[],
 	{
 		if(filename[0] == '\0' && d->d_name[0] == '.')
 			continue;
-		if(strnoscmp(d->d_name, filename, filename_len) != 0)
+		if(!file_matches(d->d_name, filename, filename_len))
 			continue;
 
 		if(type == CT_DIRONLY && !is_dirent_targets_dir(d))
@@ -1071,7 +1110,7 @@ complete_with_shared(const char *server, const char *file)
 				WideCharToMultiByte(CP_ACP, 0, (LPCWSTR)p->shi502_netname, -1, buf,
 						sizeof(buf), NULL, NULL);
 				strcat(buf, "/");
-				if(strnoscmp(buf, file, len) == 0)
+				if(file_matches(buf, file, len))
 				{
 					vle_compl_put_match(shell_like_escape(buf, 1), "");
 				}
@@ -1085,6 +1124,26 @@ complete_with_shared(const char *server, const char *file)
 
 #endif
 
+/* Checks whether file name has specified prefix accounting for system kind and
+ * case sensitivity options.  Returns non-zero if so, otherwise zero is
+ * returned. */
+static int
+file_matches(const char fname[], const char prefix[], size_t prefix_len)
+{
+	int cmp;
+	if(cfg.case_override & CO_PATH_COMPL)
+	{
+		cmp = (cfg.case_ignore & CO_PATH_COMPL)
+		    ? strncasecmp(fname, prefix, prefix_len)
+		    : strncmp(fname, prefix, prefix_len);
+	}
+	else
+	{
+		cmp = strnoscmp(fname, prefix, prefix_len);
+	}
+	return (cmp == 0);
+}
+
 int
 external_command_exists(const char cmd[])
 {
@@ -1107,13 +1166,13 @@ get_cmd_path(const char cmd[], size_t path_len, char path[])
 
 	if(contains_slash(cmd))
 	{
-		copy_str(path, path_len, cmd);
+		char *const expanded = replace_tilde(expand_envvars(cmd, 0));
+		copy_str(path, path_len, expanded);
+		free(expanded);
 		return 0;
 	}
-	else
-	{
-		return find_cmd_in_path(cmd, path_len, path);
-	}
+
+	return find_cmd_in_path(cmd, path_len, path);
 }
 
 /* vim: set tabstop=2 softtabstop=2 shiftwidth=2 noexpandtab cinoptions-=(0 : */
