@@ -69,6 +69,7 @@
 #include "cmd_handlers.h"
 #include "filelist.h"
 #include "filtering.h"
+#include "flist_sel.h"
 #include "macros.h"
 #include "marks.h"
 #include "opt_handlers.h"
@@ -108,7 +109,7 @@ static char * get_file_first_line(const char path[]);
 static void execute_extcmd(const char command[], CmdInputType type);
 static void save_extcmd(const char command[], CmdInputType type);
 static void post(int id);
-TSTATIC void select_range(int id, const cmd_info_t *cmd_info);
+TSTATIC void cmds_select_range(int id, const cmd_info_t *cmd_info);
 static int skip_at_beginning(int id, const char args[]);
 static char * pattern_expand_hook(const char pattern[]);
 static int cmd_should_be_processed(int cmd_id);
@@ -130,7 +131,7 @@ static cmds_conf_t cmds_conf = {
 	.expand_macros = &cmds_expand_macros,
 	.expand_envvars = &cmds_expand_envvars,
 	.post = &post,
-	.select_range = &select_range,
+	.select_range = &cmds_select_range,
 	.skip_at_beginning = &skip_at_beginning,
 };
 
@@ -371,67 +372,13 @@ post(int id)
 }
 
 TSTATIC void
-select_range(int id, const cmd_info_t *cmd_info)
+cmds_select_range(int id, const cmd_info_t *cmd_info)
 {
-	/* TODO: refactor this function select_range() */
-
-	int x;
-	int y = 0;
-
-	/* Both a starting range and an ending range are given. */
-	if(cmd_info->begin > -1)
+	if(flist_sel_range(curr_view, cmd_info->begin, cmd_info->end,
+				(id != COM_FIND && id != COM_GREP)))
 	{
-		clean_selected_files(curr_view);
-
-		for(x = cmd_info->begin; x <= cmd_info->end; x++)
-		{
-			if(is_parent_dir(curr_view->dir_entry[x].name) &&
-					cmd_info->begin != cmd_info->end)
-				continue;
-			curr_view->dir_entry[x].selected = 1;
-			y++;
-		}
-		curr_view->selected_files = y;
-	}
-	else if(curr_view->selected_files == 0)
-	{
-		if(cmd_info->end > -1)
-		{
-			clean_selected_files(curr_view);
-
-			y = 0;
-			for(x = cmd_info->end; x < curr_view->list_rows; x++)
-			{
-				if(y == 1)
-					break;
-				curr_view->dir_entry[x].selected = 1;
-				y++;
-			}
-			curr_view->selected_files = y;
-		}
-		else if(id != COM_FIND && id != COM_GREP)
-		{
-			clean_selected_files(curr_view);
-
-			y = 0;
-			for(x = curr_view->list_pos; x < curr_view->list_rows; x++)
-			{
-				if(y == 1)
-					break;
-
-				curr_view->dir_entry[x].selected = 1;
-				y++;
-			}
-			curr_view->selected_files = y;
-		}
-	}
-	else
-	{
-		return;
-	}
-
-	if(curr_view->selected_files > 0)
 		curr_view->user_selection = 0;
+	}
 }
 
 /* Command prefix remover for command parsing unit.  Returns < 0 to do nothing
@@ -487,17 +434,6 @@ pattern_expand_hook(const char pattern[])
 	return expanded_pattern;
 }
 
-static void
-remove_selection(FileView *view)
-{
-	/* TODO: maybe move this to filelist.c */
-	if(view->selected_files == 0)
-		return;
-
-	clean_selected_files(view);
-	redraw_view(view);
-}
-
 /* Returns negative value in case of error */
 static int
 execute_command(FileView *view, const char command[], int menu)
@@ -507,7 +443,7 @@ execute_command(FileView *view, const char command[], int menu)
 
 	if(command == NULL)
 	{
-		remove_selection(view);
+		flist_sel_stash_if_nonempty(view);
 		return 0;
 	}
 
@@ -518,7 +454,7 @@ execute_command(FileView *view, const char command[], int menu)
 
 	if(command[0] == '\0' && !menu)
 	{
-		remove_selection(view);
+		flist_sel_stash_if_nonempty(view);
 		return 0;
 	}
 
@@ -614,7 +550,7 @@ execute_command(FileView *view, const char command[], int menu)
 
 	if(!menu && vle_mode_is(NORMAL_MODE))
 	{
-		remove_selection(view);
+		flist_sel_stash_if_nonempty(view);
 	}
 
 	return -1;
@@ -858,7 +794,7 @@ break_cmdline(const char cmdline[], int for_menu)
 	 * - cmdline   -- points to the start of the last command. */
 	while(*cmdline != '\0')
 	{
-		if(args_kind == CAT_REGULAR && *raw == '\\')
+		if(args_kind == CAT_REGULAR && raw[0] == '\\' && raw[1] != '\0')
 		{
 			if(*(raw + 1) == '|')
 			{
@@ -998,6 +934,8 @@ get_cmd_args_type(const char cmd[])
 		case COM_CMAP:
 		case COM_CNOREMAP:
 		case COM_COMMAND:
+		case COM_DMAP:
+		case COM_DNOREMAP:
 		case COM_FILETYPE:
 		case COM_FILEVIEWER:
 		case COM_FILEXTYPE:
@@ -1113,13 +1051,20 @@ exec_command(const char cmd[], FileView *view, CmdInputType type)
 			return execute_command(view, cmd, menu);
 
 		case CIT_FILTER_PATTERN:
-			local_filter_apply(view, cmd);
+			if(view->custom.type != CV_DIFF)
+			{
+				local_filter_apply(view, cmd);
+			}
+			else
+			{
+				show_error_msg("Filtering", "No local filter for diff views");
+			}
 			return 0;
 
 		default:
 			assert(0 && "Command execution request of unknown/unexpected type.");
 			return 0;
-	};
+	}
 }
 
 /* Repeats last command of the specified type.  Returns 0 on success if no
