@@ -173,6 +173,15 @@ vle_keys_reset(void)
 	def_handlers = NULL;
 }
 
+void
+vle_keys_user_clear(void)
+{
+	free_forest(user_cmds_root, max_modes);
+
+	user_cmds_root = calloc(max_modes, sizeof(*user_cmds_root));
+	assert(user_cmds_root != NULL);
+}
+
 /* Releases array of trees of length size including trees. */
 static void
 free_forest(key_chunk_t *forest, size_t size)
@@ -185,19 +194,6 @@ free_forest(key_chunk_t *forest, size_t size)
 	free(forest);
 }
 
-void
-vle_keys_user_clear(void)
-{
-	int i;
-
-	for(i = 0; i < max_modes; i++)
-		free_tree(&user_cmds_root[i]);
-	free(user_cmds_root);
-
-	user_cmds_root = calloc(max_modes, sizeof(*user_cmds_root));
-	assert(user_cmds_root != NULL);
-}
-
 static void
 free_tree(key_chunk_t *root)
 {
@@ -1049,7 +1045,7 @@ add_keys_inner(key_chunk_t *root, const wchar_t *keys)
 }
 
 void
-vle_keys_list(int mode, vle_keys_list_cb cb)
+vle_keys_list(int mode, vle_keys_list_cb cb, int user_only)
 {
 	const key_chunk_t *user = &user_cmds_root[mode];
 	const key_chunk_t *builtin = &builtin_cmds_root[mode];
@@ -1060,13 +1056,13 @@ vle_keys_list(int mode, vle_keys_list_cb cb)
 		traverse_children(user, L"", &list_chunk, cb);
 
 		/* Put separator only when it's needed. */
-		if(builtin->children_count != 0U)
+		if(!user_only && builtin->children_count != 0U)
 		{
 			cb(L"", L"", "");
 		}
 	}
 
-	if(builtin->children_count != 0U)
+	if(!user_only && builtin->children_count != 0U)
 	{
 		traverse_children(builtin, L"", &list_chunk, cb);
 	}
