@@ -44,6 +44,7 @@
 #include "ui/ui.h"
 #include "utils/log.h"
 #include "utils/macros.h"
+#include "utils/test_helpers.h"
 #include "utils/utf8.h"
 #include "utils/utils.h"
 #include "background.h"
@@ -57,7 +58,8 @@
 static int ensure_term_is_ready(void);
 static int get_char_async_loop(WINDOW *win, wint_t *c, int timeout);
 static void process_scheduled_updates(void);
-static int process_scheduled_updates_of_view(FileView *view);
+TSTATIC int process_scheduled_updates_of_view(FileView *view);
+static void update_hardware_cursor(void);
 static int should_check_views_for_changes(void);
 static void check_view_for_changes(FileView *view);
 static void reset_input_buf(wchar_t curr_input_buf[],
@@ -107,12 +109,6 @@ event_loop(const int *quit)
 		size_t counter;
 		int got_input;
 
-		if(!ensure_term_is_ready())
-		{
-			wait_for_enter = 0;
-			continue;
-		}
-
 		lwin.user_selection = 1;
 		rwin.user_selection = 1;
 
@@ -126,19 +122,18 @@ event_loop(const int *quit)
 			                         ? MIN(timeout, cfg.sug.delay)
 			                         : timeout;
 
+			if(!ensure_term_is_ready())
+			{
+				wait_for_enter = 0;
+				continue;
+			}
+
 			modes_periodic();
 
-			check_background_jobs();
+			bg_check();
 
 			got_input = (get_char_async_loop(status_bar, &c, actual_timeout) != ERR);
 
-			/* This loop can be infinite if terminal becomes not available, so force
-			 * checking terminal state here. */
-			if(!got_input && !ui_term_is_alive())
-			{
-				vifm_finish("Terminal is not available.");
-			}
-
 			/* If suggestion delay timed out, reset it and wait the rest of the
 			 * timeout. */
 			if(!got_input && wait_for_suggestion)
@@ -155,6 +150,13 @@ event_loop(const int *quit)
 				timeout = cfg.timeout_len;
 				continue;
 			}
+
+			if(got_input && c == K(KEY_RESIZE))
+			{
+				modes_redraw();
+				continue;
+			}
+
 			break;
 		}
 		while(1);
@@ -162,7 +164,7 @@ event_loop(const int *quit)
 		suggestions_are_visible = 0;
 
 		/* Ensure that current working directory is set correctly (some pieces of
-		 * code rely on this). */
+		 * code rely on this, e.g. %c macro in current directory). */
 		(void)vifm_chdir(flist_get_dir(curr_view));
 
 		if(got_input)
@@ -171,7 +173,7 @@ event_loop(const int *quit)
 			{
 				wait_for_enter = 0;
 				curr_stats.save_msg = 0;
-				clean_status_bar();
+				ui_sb_clear();
 				if(c == WC_CR)
 				{
 					continue;
@@ -331,8 +333,8 @@ ensure_term_is_ready(void)
  *  - checks for new IPC messages;
  *  - checks whether contents of displayed directories changed;
  *  - redraws UI if requested.
- * Returns KEY_CODE_YES for functional keys, OK for wide character and ERR
- * otherwise (e.g. after timeout). */
+ * Returns KEY_CODE_YES for functional keys (preprocesses *c in this case), OK
+ * for wide character and ERR otherwise (e.g. after timeout). */
 static int
 get_char_async_loop(WINDOW *win, wint_t *c, int timeout)
 {
@@ -371,9 +373,18 @@ get_char_async_loop(WINDOW *win, wint_t *c, int timeout)
 				display_suggestion_box(curr_input_buf);
 			}
 
+			/* Update cursor before waiting for input.  Modes set cursor correctly
+			 * within corresponding windows, but we need to call refresh on one of
+			 * them to make it active. */
+			update_hardware_cursor();
+
 			result = compat_wget_wch(win, c);
 			if(result != ERR)
 			{
+				if(result == KEY_CODE_YES)
+				{
+					*c = K(*c);
+				}
 				return result;
 			}
 
@@ -409,7 +420,7 @@ process_scheduled_updates(void)
 
 /* Performs postponed updates for the view, if any.  Returns non-zero if
  * something was indeed updated, and zero otherwise. */
-static int
+TSTATIC int
 process_scheduled_updates_of_view(FileView *view)
 {
 	if(!window_shows_dirlist(view))
@@ -429,7 +440,7 @@ process_scheduled_updates_of_view(FileView *view)
 			load_saving_pos(view, 1);
 			if(view == curr_view && !is_status_bar_multiline())
 			{
-				ui_ruler_update(view);
+				ui_ruler_update(view, 1);
 			}
 			return 1;
 		case UUE_FULL_RELOAD:
@@ -442,6 +453,39 @@ process_scheduled_updates_of_view(FileView *view)
 	}
 }
 
+/* Updates hardware cursor to be on currently active area of the interface,
+ * which depends mainly on current mode.. */
+static void
+update_hardware_cursor(void)
+{
+	if(is_status_bar_multiline())
+	{
+		checked_wmove(status_bar, 0, 0);
+		wrefresh(status_bar);
+		return;
+	}
+
+	switch(vle_mode_get())
+	{
+		case MENU_MODE:
+		case FILE_INFO_MODE:
+		case MORE_MODE: wrefresh(menu_win);       break;
+		case CHANGE_MODE:
+		case ATTR_MODE: wrefresh(change_win);     break;
+		case MSG_MODE:  wrefresh(error_win);      break;
+		case VIEW_MODE: wrefresh(curr_view->win); break;
+		case SORT_MODE: wrefresh(sort_win);       break;
+
+		case NORMAL_MODE:
+		case VISUAL_MODE:
+			if(should_check_views_for_changes())
+			{
+				wrefresh(curr_view->win);
+			}
+			break;
+	}
+}
+
 /* Checks whether views should be checked against external changes.  Returns
  * non-zero is so, otherwise zero is returned. */
 static int
@@ -596,7 +640,6 @@ prepare_suggestion_box(int *height)
 		win = stat_win;
 	}
 
-
 	wbkgdset(win, COLOR_PAIR(colmgr_get_pair(col.fg, col.bg)) | col.attr);
 	werase(win);
 
