@@ -27,20 +27,15 @@
 
 #include "ui/ui.h"
 #include "utils/test_helpers.h"
-#include "registers.h"
-
-/* Type of contiguous area of file list. */
-typedef enum
-{
-	FLS_SELECTION, /* Of selected entries. */
-	FLS_MARKING,   /* Of marked entries. */
-}
-FileListScope;
 
 /* Type of filter function for zapping list of entries.  Should return non-zero
  * if entry is to be kept and zero otherwise. */
 typedef int (*zap_filter)(FileView *view, const dir_entry_t *entry, void *arg);
 
+/* Type of predicate functions to reason about entries.  Should return non-zero
+ * if particular property holds and zero otherwise. */
+typedef int (*entry_predicate)(const dir_entry_t *entry);
+
 /* Initialization/termination functions. */
 
 /* Prepares views for the first time. */
@@ -50,52 +45,6 @@ void reset_views(void);
 /* Loads view file list for the first time. */
 void load_initial_directory(FileView *view, const char dir[]);
 
-/* Position related functions. */
-
-/* Find index of the file within list of currently visible files of the view.
- * Returns file entry index or -1, if file wasn't found. */
-int find_file_pos_in_list(const FileView *const view, const char file[]);
-/* Find index of the file within list of currently visible files of the view.
- * Always matches file name and can optionally match directory if dir is not
- * NULL.  Returns file entry index or -1, if file wasn't found. */
-int flist_find_entry(const FileView *view, const char file[], const char dir[]);
-/* Tries to move cursor by pos_delta positions.  A wrapper for
- * correct_list_pos_on_scroll_up() and correct_list_pos_on_scroll_down()
- * functions. */
-void correct_list_pos(FileView *view, ssize_t pos_delta);
-/* Returns non-zero if doing something makes sense. */
-int correct_list_pos_on_scroll_down(FileView *view, size_t lines_count);
-/* Returns non-zero if doing something makes sense. */
-int correct_list_pos_on_scroll_up(FileView *view, size_t lines_count);
-/* Moves cursor to specified position. */
-void flist_set_pos(FileView *view, int pos);
-/* Ensures that position in the list doesn't exceed its bounds. */
-void flist_ensure_pos_is_valid(FileView *view);
-/* Ensures that cursor is moved outside of entries of certain type. */
-void move_cursor_out_of(FileView *view, FileListScope scope);
-/* Returns non-zero if cursor is on the first line. */
-int at_first_line(const FileView *view);
-/* Returns non-zero if cursor is on the last line. */
-int at_last_line(const FileView *view);
-/* Returns non-zero if cursor is on the first column. */
-int at_first_column(const FileView *view);
-/* Returns non-zero if cursor is on the last column. */
-int at_last_column(const FileView *view);
-/* Moves cursor to the first file in a row. */
-void go_to_start_of_line(FileView *view);
-/* Returns position of the first file in current line. */
-int get_start_of_line(const FileView *view);
-/* Returns position of the last file in current line. */
-int get_end_of_line(const FileView *view);
-/* Finds position of the next/previous group defined by primary sorting key.
- * Returns determined position (might point to the last/first entry in corner
- * cases). */
-int flist_find_group(FileView *view, int next);
-/* Finds position of the next/previous group defined by entries being files or
- * directories.  Returns determined position (might point to the last/first
- * entry in corner cases). */
-int flist_find_dir_group(FileView *view, int next);
-
 /* Appearance related functions. */
 
 /* Inverts primary key sorting order. */
@@ -141,44 +90,10 @@ void leave_invalid_dir(FileView *view);
 /* Checks if the view is the directory specified by the path.  Returns non-zero
  * if so, otherwise zero is returned. */
 int pane_in_dir(const FileView *view, const char path[]);
-
-/* Selection related functions. */
-
-/* Cleans selection possibly saving it for later use. */
-void clean_selected_files(FileView *view);
-/* Erases selection not saving anything. */
-void erase_selection(FileView *view);
-/* Inverts selection of files in the view. */
-void invert_selection(FileView *view);
-/* Reselects previously selected entries.  When reg is NULL, saved selection is
- * restored, otherwise list of files to restore is taken from the register. */
-void flist_sel_restore(FileView *view, reg_t *reg);
-/* Counts number of selected files and writes saves the number in
- * view->selected_files. */
-void recount_selected_files(FileView *view);
-/* Remove dot and regexp filters if it's needed to make file visible.  Returns
- * non-zero if file was found. */
-int ensure_file_is_selected(FileView *view, const char name[]);
-
-/* Directory history related functions. */
-
-/* Changes current directory of the view to next location backward in
- * history, if available. */
-void navigate_backward_in_history(FileView *view);
-/* Changes current directory of the view to next location forward in history, if
- * available. */
-void navigate_forward_in_history(FileView *view);
-/* Adds new entry to directory history of the view or updates an existing entry.
- * If path is NULL, current path is used.  If file is NULL current file is used.
- * If pos is negative, current position is used.  Empty file name signifies
- * visiting directory, which shouldn't reset name of previously active file in
- * it. */
-void save_view_history(FileView *view, const char path[], const char file[],
-		int pos);
-int is_in_view_history(FileView *view, const char *path);
-void clean_positions_in_history(FileView *view);
-/* Looks up history in the source to update cursor position in the view. */
-void flist_hist_lookup(FileView *view, const FileView *source);
+/* Navigates to next/previous sibling directory (does nothing for root or custom
+ * view) with optional wrapping.  Returns non-zero if statusbar message should
+ * be preserved. */
+int go_to_sibling_dir(FileView *view, int next, int wrap);
 
 /* Typed (with trailing slash for directories) file name function. */
 
@@ -193,15 +108,35 @@ char * get_typed_entry_fpath(const dir_entry_t *entry);
 int flist_custom_active(const FileView *view);
 /* Prepares list of files for it to be filled with entries. */
 void flist_custom_start(FileView *view, const char title[]);
-/* Adds an entry to list of files. */
-void flist_custom_add(FileView *view, const char path[]);
+/* Adds an entry to custom list of files.  Returns pointer to just added entry
+ * or NULL on error. */
+dir_entry_t * flist_custom_add(FileView *view, const char path[]);
+/* Puts an entry to custom list of files, contents of the entry gets stolen.
+ * Returns pointer to just added entry or NULL on error. */
+dir_entry_t * flist_custom_put(FileView *view, dir_entry_t *entry);
+/* Parses line to extract path and adds it to custom view or does nothing. */
+void flist_custom_add_spec(FileView *view, const char line[]);
+/* Appends entry separator to the list with specified id. */
+void flist_custom_add_separator(FileView *view, int id);
 /* Finishes file list population, handles empty resulting list corner case.
+ * Non-zero allow_empty makes a single-entry (..) view instead of aborting.
  * Returns zero on success, otherwise (on empty list) non-zero is returned. */
-int flist_custom_finish(FileView *view, int very);
-/* Removes selected files from custom view. */
-void flist_custom_exclude(FileView *view);
+int flist_custom_finish(FileView *view, CVType type, int allow_empty);
+/* A more high level version of flist_custom_finish(), which takes care of error
+ * handling and cursor position. */
+void flist_custom_end(FileView *view, int very);
+/* Loads list of paths (absolute or relative to the path) into custom view.
+ * Exists with error message on failed attempt. */
+void flist_custom_set(FileView *view, const char title[], const char path[],
+		char *lines[], int nlines);
+/* Removes selected files or current one from custom view.  Zero selection_only
+ * enables excluding files that share ids with selected items. */
+void flist_custom_exclude(FileView *view, int selection_only);
 /* Clones list of files from from view to to view. */
 void flist_custom_clone(FileView *to, const FileView *from);
+/* Adds missing parent directories to the tree.  Such entries are marked as
+ * "temporary", which can be used to remove them later. */
+void flist_custom_uncompress_tree(FileView *view);
 
 /* Other functions. */
 
@@ -223,7 +158,7 @@ void load_saving_pos(FileView *view, int reload);
 char * get_current_file_name(FileView *view);
 /* Gets current entry of the view.  Returns the entry or NULL if view doesn't
  * contain any. */
-dir_entry_t * get_current_entry(FileView *view);
+dir_entry_t * get_current_entry(const FileView *view);
 /* Checks whether content in the current directory of the view changed and
  * reloads the view if so. */
 void check_if_filelist_have_changed(FileView *view);
@@ -239,9 +174,6 @@ int view_needs_cd(const FileView *view, const char path[]);
 void set_view_path(FileView *view, const char path[]);
 /* Returns possible cached or calculated value of file size. */
 uint64_t get_file_size_by_entry(const FileView *view, size_t pos);
-/* Checks whether entry corresponds to a directory.  Returns non-zero if so,
- * otherwise zero is returned. */
-int is_directory_entry(const dir_entry_t *entry);
 /* Loads pointer to the next selected entry in file list of the view.  *entry
  * should be NULL for the first call and result of previous call otherwise.
  * Returns zero when there is no more entries to supply, otherwise non-zero is
@@ -260,68 +192,97 @@ int iter_selection_or_current(FileView *view, dir_entry_t **entry);
 /* Maps one of file list entries to its position in the list.  Returns the
  * position or -1 on wrong entry. */
 int entry_to_pos(const FileView *view, const dir_entry_t *entry);
-/* Fills the buffer with the full path to file under cursor. */
+/* Fills the buffer with the full path to file under cursor.  Set to empty if
+ * there are no files in the view. */
 void get_current_full_path(const FileView *view, size_t buf_len, char buf[]);
-/* Fills the buffer with the full path to file at specified position. */
+/* Fills the buffer with the full path to file at specified position.  Set to
+ * empty if there are no files in the view. */
 void get_full_path_at(const FileView *view, int pos, size_t buf_len,
 		char buf[]);
 /* Fills the buffer with the full path to file of specified file list entry. */
 void get_full_path_of(const dir_entry_t *entry, size_t buf_len, char buf[]);
 /* Fills the buffer with short path of specified file list entry.  The
  * shortening occurs for files under original directory of custom views.
- * Non-zero format enables file type specific decoration. */
+ * Non-zero format enables file type specific decoration.  Non-zero drop_prefix
+ * requests omitting file prefix for trees. */
 void get_short_path_of(const FileView *view, const dir_entry_t *entry,
-		int format, size_t buf_len, char buf[]);
+		int format, int drop_prefix, size_t buf_len, char buf[]);
 /* Ensures that either entries at specified positions, selected entries or file
  * under cursor is marked. */
 void check_marking(FileView *view, int count, const int indexes[]);
 /* Marks files at positions specified in the indexes array of size count. */
 void mark_files_at(FileView *view, int count, const int indexes[]);
-/* Marks selected files of the view. */
-void mark_selected(FileView *view);
+/* Marks selected files of the view.  Returns number of marked files. */
+int mark_selected(FileView *view);
 /* Same as mark_selected() function, but when selection is absent current file
- * is marked. */
-void mark_selection_or_current(FileView *view);
+ * is marked.  Returns number of marked files. */
+int mark_selection_or_current(FileView *view);
 /* Counts number of marked files that can be processed (thus excluding parent
  * directory, which is "..").  Returns the count. */
 int flist_count_marked(FileView *const view);
 /* Removes dead entries (those that refer to non-existing files) or those that
- * do not match local filter from the view.  Returns number of erased
+ * do not match local filter from the view.  remove_subtrees flag controls
+ * whether child nodes should be removed as well.  Returns number of erased
  * entries. */
 int zap_entries(FileView *view, dir_entry_t *entries, int *count,
-		zap_filter filter, void *arg, int allow_empty_list);
+		zap_filter filter, void *arg, int allow_empty_list, int remove_subtrees);
+/* Leaves only those entries in compare view, for which filter returns non-zero.
+ * Properly updates the other pane.  Returns non-zero if views were left,
+ * because they became empty. */
+int filter_in_compare(FileView *view, void *arg, zap_filter filter);
 /* Finds directory entry in the list of entries by the path.  Returns pointer to
  * the found entry or NULL. */
-dir_entry_t * entry_from_path(dir_entry_t *entries, int count,
+dir_entry_t * entry_from_path(FileView *view, dir_entry_t *entries, int count,
 		const char path[]);
 /* Retrieves number of items in a directory specified by the entry.  Returns the
  * number, which is zero for files. */
-uint64_t entry_get_nitems(FileView *view, const dir_entry_t *entry);
+uint64_t entry_get_nitems(const FileView *view, const dir_entry_t *entry);
 /* Calculates number of items at path specified by the entry.  No check for file
  * type is performed.  Returns the number, which is zero for files. */
 uint64_t entry_calc_nitems(const dir_entry_t *entry);
+/* Checks whether entry is selected.  Returns non-zero if so, otherwise zero is
+ * returned. */
+int is_entry_selected(const dir_entry_t *entry);
+/* Checks whether entry is marked.  Returns non-zero if so, otherwise zero is
+ * returned. */
+int is_entry_marked(const dir_entry_t *entry);
 /* Replaces all entries of the *entries with copy of with_entries elements. */
 void replace_dir_entries(FileView *view, dir_entry_t **entries, int *count,
 		const dir_entry_t *with_entries, int with_count);
 /* Adds new entry to the *list of length *list_size and updates them
- * appropriately.  Returns zero on success, otherwise non-zero is returned. */
-int add_dir_entry(dir_entry_t **list, size_t *list_size,
+ * appropriately.  Returns NULL on error, otherwise pointer to the entry is
+ * returned. */
+dir_entry_t * add_dir_entry(dir_entry_t **list, size_t *list_size,
 		const dir_entry_t *entry);
+/* Adds new entry to the list and fills it with data.  Returns pointer to the
+ * entry or NULL on error. */
+dir_entry_t * entry_list_add(FileView *view, dir_entry_t **list, int *list_size,
+		const char path[]);
+/* Frees list of directory entries related to the view.  Sets *entries and
+ * *count to safe values. */
+void free_dir_entries(FileView *view, dir_entry_t **entries, int *count);
 /* Frees single directory entry. */
-void free_dir_entry(const FileView *view, dir_entry_t *entry);
+void fentry_free(const FileView *view, dir_entry_t *entry);
 /* Adds parent directory entry (..) to filelist. */
 void add_parent_dir(FileView *view);
-/* Loads list of paths (absolute or relative to the path) into custom view.
- * Exists with error message on failed attempt. */
-void flist_set(FileView *view, const char title[], const char path[],
-		char *lines[], int nlines);
-/* Parses line to extract path and adds it to custom view or does nothing. */
-void flist_add_custom_line(FileView *view, const char line[]);
-/* A more high level version of flist_custom_finish(), which takes care of error
- * handling and cursor position. */
-void flist_end_custom(FileView *view, int very);
 /* Changes name of a file entry, performing additional required updates. */
 void fentry_rename(FileView *view, dir_entry_t *entry, const char to[]);
+/* Checks whether this is fake entry for internal purposes, which should not be
+ * processed as a file. */
+int fentry_is_fake(const dir_entry_t *entry);
+/* Checks whether this is valid entry, which can be selected and processed by
+ * file operations or even just selected.  Currently this checks for entry not
+ * being ".." nor fake. */
+int fentry_is_valid(const dir_entry_t *entry);
+/* Checks whether entry corresponds to a directory (including symbolic links to
+ * directories).  Returns non-zero if so, otherwise zero is returned. */
+int fentry_is_dir(const dir_entry_t *entry);
+/* Loads directory tree specified by its path into the view.  Considers various
+ * filters.  Returns zero on success, otherwise non-zero is returned. */
+int flist_load_tree(FileView *view, const char path[]);
+/* Makes to contain tree with the same root as from including copying list of
+ * excluded files.  Returns zero on success, otherwise non-zero is returned. */
+int flist_clone_tree(FileView *to, const FileView *from);
 
 TSTATIC_DEFS(
 	TSTATIC void pick_cd_path(FileView *view, const char base_dir[],
