@@ -31,14 +31,21 @@
 #include "utils/str.h"
 #include "utils/utils.h"
 #include "filelist.h"
+#include "flist_pos.h"
+#include "flist_sel.h"
+#include "opt_handlers.h"
 
 static void reset_filter(filter_t *filter);
 static int is_newly_filtered(FileView *view, const dir_entry_t *entry,
 		void *arg);
+static int file_is_filtered(FileView *view, const char filename[], int is_dir,
+		int apply_local_filter);
 static int get_unfiltered_pos(const FileView *const view, int pos);
 static int load_unfiltered_list(FileView *const view);
+static int list_is_incomplete(FileView *const view);
 static void store_local_filter_position(FileView *const view, int pos);
 static int update_filtering_lists(FileView *view, int add, int clear);
+static void reparent_tree_node(dir_entry_t *original, dir_entry_t *filtered);
 static void ensure_filtered_list_not_empty(FileView *view,
 		dir_entry_t *parent_entry);
 static int extract_previously_selected_pos(FileView *const view);
@@ -83,26 +90,42 @@ reset_filter(filter_t *filter)
 void
 set_dot_files_visible(FileView *view, int visible)
 {
-	view->hide_dot = !visible;
+	view->hide_dot_g = view->hide_dot = !visible;
 	ui_view_schedule_reload(view);
+	if(view == curr_view)
+	{
+		load_dot_filter_option(view);
+	}
 	if(curr_stats.global_local_settings)
 	{
 		FileView *other = (view == curr_view) ? other_view : curr_view;
-		other->hide_dot = !visible;
+		other->hide_dot_g = other->hide_dot = !visible;
 		ui_view_schedule_reload(other);
+		if(other == curr_view)
+		{
+			load_dot_filter_option(other);
+		}
 	}
 }
 
 void
 toggle_dot_files(FileView *view)
 {
-	view->hide_dot = !view->hide_dot;
+	view->hide_dot_g = view->hide_dot = !view->hide_dot;
 	ui_view_schedule_reload(view);
+	if(view == curr_view)
+	{
+		load_dot_filter_option(view);
+	}
 	if(curr_stats.global_local_settings)
 	{
 		FileView *other = (view == curr_view) ? other_view : curr_view;
-		other->hide_dot = !other->hide_dot;
+		other->hide_dot_g = other->hide_dot = !other->hide_dot;
 		ui_view_schedule_reload(other);
+		if(other == curr_view)
+		{
+			load_dot_filter_option(other);
+		}
 	}
 }
 
@@ -122,7 +145,7 @@ filter_selected_files(FileView *view)
 		const char *name = entry->name;
 		char name_with_slash[NAME_MAX + 1 + 1];
 
-		if(is_directory_entry(entry))
+		if(fentry_is_dir(entry))
 		{
 			append_slash(entry->name, name_with_slash, sizeof(name_with_slash));
 			name = name_with_slash;
@@ -132,14 +155,30 @@ filter_selected_files(FileView *view)
 		(void)filter_append(&filter, name);
 	}
 
+	/* Even current file might be unavailable for filtering.  In this case, just
+	 * do nothing. */
+	if(filter_is_empty(&filter))
+	{
+		filter_dispose(&filter);
+		return;
+	}
+
+	if(view->custom.type == CV_DIFF)
+	{
+		(void)filter_in_compare(view, &filter, &is_newly_filtered);
+		ui_view_schedule_redraw(view);
+		filter_dispose(&filter);
+		return;
+	}
+
 	/* Update entry lists to remove entries that must be filtered out now.  No
 	 * view reload is needed. */
 	filtered = zap_entries(view, view->dir_entry, &view->list_rows,
-			&is_newly_filtered, &filter, 0);
+			&is_newly_filtered, &filter, 0, 1);
 	if(flist_custom_active(view))
 	{
-		(void)zap_entries(view, view->custom.entries, &view->custom.entry_count,
-				&is_newly_filtered, &filter, 1);
+		(void)zap_entries(view, view->local_filter.entries,
+				&view->local_filter.entry_count, &is_newly_filtered, &filter, 1, 1);
 	}
 	else
 	{
@@ -164,7 +203,7 @@ is_newly_filtered(FileView *view, const dir_entry_t *entry, void *arg)
 	char name_with_slash[NAME_MAX + 1 + 1];
 	const char *filename = entry->name;
 
-	if(is_directory_entry(entry))
+	if(fentry_is_dir(entry))
 	{
 		append_slash(filename, name_with_slash, sizeof(name_with_slash));
 		filename = name_with_slash;
@@ -209,6 +248,11 @@ filename_filter_clear(FileView *view)
 void
 restore_filename_filter(FileView *view)
 {
+	if(view->prev_manual_filter[0] == '\0' && view->prev_auto_filter[0] == '\0')
+	{
+		return;
+	}
+
 	(void)filter_set(&view->manual_filter, view->prev_manual_filter);
 	(void)filter_set(&view->auto_filter, view->prev_auto_filter);
 	view->invert = view->prev_invert;
@@ -224,7 +268,23 @@ toggle_filter_inversion(FileView *view)
 }
 
 int
-file_is_visible(FileView *view, const char filename[], int is_dir)
+filters_file_is_visible(FileView *view, const char filename[], int is_dir)
+{
+	return file_is_filtered(view, filename, is_dir, 1);
+}
+
+int
+filters_file_is_filtered(FileView *view, const char filename[], int is_dir)
+{
+	return file_is_filtered(view, filename, is_dir, 0);
+}
+
+/* Checks whether file/directory passes filename filters of the view.  Returns
+ * non-zero if given filename passes filter and should be visible, otherwise
+ * zero is returned, in which case the file should be hidden. */
+static int
+file_is_filtered(FileView *view, const char filename[], int is_dir,
+		int apply_local_filter)
 {
 	/* FIXME: some very long file names won't be matched against some regexps. */
 	char name_with_slash[NAME_MAX + 1 + 1];
@@ -239,7 +299,8 @@ file_is_visible(FileView *view, const char filename[], int is_dir)
 		return 0;
 	}
 
-	if(filter_matches(&view->local_filter.filter, filename) == 0)
+	if(apply_local_filter &&
+			filter_matches(&view->local_filter.filter, filename) == 0)
 	{
 		return 0;
 	}
@@ -265,6 +326,45 @@ filters_dir_updated(FileView *view)
 	filter_clear(&view->local_filter.filter);
 }
 
+void
+filter_temporary_nodes(FileView *view, dir_entry_t *list)
+{
+	/* This is basically a simplified version of update_filtering_lists().  Not
+	 * sure if it's worth merging them. */
+
+	int i;
+	size_t list_size = 0U;
+
+	for(i = 0; i < view->list_rows; ++i)
+	{
+		dir_entry_t *new_entry;
+		dir_entry_t *const entry = &view->dir_entry[i];
+
+		/* tag links to position of nodes passed through filter in list of
+		 * visible files.  Removed nodes have -1. */
+		entry->tag = -1;
+
+		if(entry->temporary)
+		{
+			fentry_free(view, entry);
+			continue;
+		}
+
+		new_entry = add_dir_entry(&list, &list_size, entry);
+		if(new_entry != NULL)
+		{
+			entry->tag = list_size - 1U;
+			/* We basically grow the tree node by node while performing
+			 * reparenting. */
+			reparent_tree_node(entry, new_entry);
+		}
+	}
+
+	dynarray_free(view->dir_entry);
+	view->dir_entry = list;
+	view->list_rows = list_size;
+}
+
 int
 local_filter_set(FileView *view, const char filter[])
 {
@@ -318,7 +418,7 @@ load_unfiltered_list(FileView *const view)
 
 	view->local_filter.saved = strdup(view->local_filter.filter.raw);
 
-	if(view->filtered > 0)
+	if(list_is_incomplete(view))
 	{
 		char full_path[PATH_MAX];
 		dir_entry_t *entry;
@@ -329,7 +429,7 @@ load_unfiltered_list(FileView *const view)
 		populate_dir_list(view, 1);
 
 		/* Resolve current file position in updated list. */
-		entry = entry_from_path(view->dir_entry, view->list_rows, full_path);
+		entry = entry_from_path(view, view->dir_entry, view->list_rows, full_path);
 		if(entry != NULL)
 		{
 			current_file_pos = entry_to_pos(view, entry);
@@ -343,8 +443,8 @@ load_unfiltered_list(FileView *const view)
 	else
 	{
 		/* Save unfiltered (by local filter) list for further use. */
-		replace_dir_entries(view, &view->custom.entries,
-				&view->custom.entry_count, view->dir_entry, view->list_rows);
+		replace_dir_entries(view, &view->local_filter.entries,
+				&view->local_filter.entry_count, view->dir_entry, view->list_rows);
 	}
 
 	view->local_filter.unfiltered = view->dir_entry;
@@ -355,6 +455,39 @@ load_unfiltered_list(FileView *const view)
 	return current_file_pos;
 }
 
+/* Checks whether list of files is incomplete.  Returns non-zero if so,
+ * otherwise zero is returned. */
+static int
+list_is_incomplete(FileView *const view)
+{
+	int i;
+
+	if(view->filtered > 0)
+	{
+		return 1;
+	}
+
+	if(flist_custom_active(view) && view->custom.type != CV_TREE)
+	{
+		return 0;
+	}
+
+	/* Check if there is any directories without leaf nodes.  They aren't counted
+	 * as filtered out, so need to check separately (or start counting them in
+	 * some way). */
+	for(i = 0; i < view->list_rows; ++i)
+	{
+		dir_entry_t *const entry = &view->dir_entry[i];
+		if(entry->type == FT_DIR && entry->child_count == 0 &&
+				!is_parent_dir(entry->name))
+		{
+			return 1;
+		}
+	}
+
+	return 0;
+}
+
 /* Adds local filter position (in unfiltered list) to position history. */
 static void
 store_local_filter_position(FileView *const view, int pos)
@@ -377,6 +510,8 @@ store_local_filter_position(FileView *const view, int pos)
 static int
 update_filtering_lists(FileView *view, int add, int clear)
 {
+	/* filter_temporary_nodes() is similar function. */
+
 	size_t i;
 	size_t list_size = 0U;
 	dir_entry_t *parent_entry = NULL;
@@ -393,38 +528,69 @@ update_filtering_lists(FileView *view, int add, int clear)
 
 		if(is_parent_dir(name))
 		{
-			parent_entry = entry;
-			if(add && cfg_parent_dir_is_visible(is_root_dir(view->curr_dir)))
+			if(entry->child_pos == 0)
 			{
-				(void)add_dir_entry(&view->dir_entry, &list_size, entry);
+				parent_entry = entry;
+				if(add && cfg_parent_dir_is_visible(is_root_dir(view->curr_dir)))
+				{
+					(void)add_dir_entry(&view->dir_entry, &list_size, entry);
 
-				parent_added = 1;
+					parent_added = 1;
+				}
+				continue;
+			}
+			else if(!filter_is_empty(&view->local_filter.filter))
+			{
+				if(clear)
+				{
+					fentry_free(view, entry);
+				}
+				continue;
 			}
-			continue;
 		}
 
-		if(is_directory_entry(entry))
+		if(fentry_is_dir(entry))
 		{
 			append_slash(name, name_with_slash, sizeof(name_with_slash));
 			name = name_with_slash;
 		}
 
+		/* tag links to position of nodes passed through filter in list of visible
+		 * files.  Nodes that didn't pass have -1. */
+		entry->tag = -1;
 		if(filter_matches(&view->local_filter.filter, name) != 0)
 		{
 			if(add)
 			{
-				(void)add_dir_entry(&view->dir_entry, &list_size, entry);
+				dir_entry_t *e = add_dir_entry(&view->dir_entry, &list_size, entry);
+				if(e != NULL)
+				{
+					entry->tag = list_size - 1U;
+					/* We basically grow the tree node by node while performing
+					 * reparenting. */
+					reparent_tree_node(entry, e);
+				}
 			}
 		}
 		else
 		{
 			if(clear)
 			{
-				free_dir_entry(view, entry);
+				fentry_free(view, entry);
 			}
 		}
 	}
 
+	if(clear)
+	{
+		/* XXX: the check of name pointer is horrible, but is needed to prevent
+		 *      freeing of entry in use. */
+		if(!parent_added && parent_entry != NULL && list_size != 0U &&
+				view->dir_entry[0].name != parent_entry->name)
+		{
+			fentry_free(view, parent_entry);
+		}
+	}
 	if(add)
 	{
 		view->list_rows = list_size;
@@ -438,6 +604,42 @@ update_filtering_lists(FileView *view, int add, int clear)
 	return 0;
 }
 
+/* Reparents *filtered node by attaching it to the closes ancestor of *original
+ * mapped onto the list of filtered nodes.  tag field of entries is used to
+ * perform the mapping. */
+static void
+reparent_tree_node(dir_entry_t *original, dir_entry_t *filtered)
+{
+	dir_entry_t *parent, *child;
+
+	filtered->child_pos = 0;
+	filtered->child_count = 0;
+
+	/* Go through items in unfiltered list looking for the closest ancestor, which
+	 * wasn't filtered out and make it the parent. */
+	child = original;
+	parent = child - child->child_pos;
+	while(parent != child)
+	{
+		if(parent->tag >= 0)
+		{
+			filtered->child_pos = original->tag - parent->tag;
+			parent = filtered - filtered->child_pos;
+			while(parent != child)
+			{
+				++parent->child_count;
+
+				child = parent;
+				parent -= parent->child_pos;
+			}
+			break;
+		}
+
+		child = parent;
+		parent -= parent->child_pos;
+	}
+}
+
 /* Use parent_entry to make filtered list not empty, or create such entry (if
  * parent_entry is NULL) and put it to original list. */
 static void
@@ -479,9 +681,9 @@ local_filter_update_view(FileView *view, int rel_pos)
 	if(pos >= 0)
 	{
 		if(pos == 0 && is_parent_dir(view->dir_entry[0].name) &&
-				view->list_rows > 0)
+				view->list_rows > 1 && !filter_is_empty(&view->local_filter.filter))
 		{
-			pos++;
+			++pos;
 		}
 
 		view->list_pos = pos;
@@ -559,7 +761,7 @@ local_filter_accept(FileView *view)
 
 	/* Some of previously selected files could be filtered out, update number of
 	 * selected files. */
-	recount_selected_files(view);
+	flist_sel_recount(view);
 }
 
 void
@@ -573,6 +775,14 @@ local_filter_apply(FileView *view, const char filter[])
 
 	(void)filter_set(&view->local_filter.filter, filter);
 	cfg_save_filter_history(view->local_filter.filter.raw);
+
+	if(flist_custom_active(view) && view->custom.type != CV_TREE &&
+			view->local_filter.entry_count == 0)
+	{
+		/* Save unfiltered (by local filter) list for further use. */
+		replace_dir_entries(view, &view->local_filter.entries,
+				&view->local_filter.entry_count, view->dir_entry, view->list_rows);
+	}
 }
 
 void
@@ -628,7 +838,7 @@ local_filter_matches(FileView *view, const dir_entry_t *entry)
 	 * regexps. */
 	char name_with_slash[NAME_MAX + 1 + 1];
 	const char *filename = entry->name;
-	if(is_directory_entry(entry))
+	if(fentry_is_dir(entry))
 	{
 		append_slash(filename, name_with_slash, sizeof(name_with_slash));
 		filename = name_with_slash;
