@@ -35,6 +35,7 @@ void filters_view_reset(FileView *view);
 
 /* Generic filters related functions. */
 
+/* Sets new value of dot files filter. */
 void set_dot_files_visible(FileView *view, int visible);
 
 void toggle_dot_files(FileView *view);
@@ -59,19 +60,26 @@ void toggle_filter_inversion(FileView *view);
 /* Checks whether file/directory passes filename filters of the view.  Returns
  * non-zero if given filename passes filter and should be visible, otherwise
  * zero is returned, in which case the file should be hidden. */
-int file_is_visible(FileView *view, const char filename[], int is_dir);
+int filters_file_is_visible(FileView *view, const char filename[], int is_dir);
+
+/* Same as filters_file_is_visible(), but ignores local filter. */
+int filters_file_is_filtered(FileView *view, const char filename[], int is_dir);
 
 /* Callback-like function which triggers some view-specific updates after
  * directory of the view changes. */
 void filters_dir_updated(FileView *view);
 
+/* Filters out nodes that bear "temporary" mark.  The list can be NULL,
+ * otherwise it's a memory to be used for new list. */
+void filter_temporary_nodes(FileView *view, dir_entry_t *list);
+
 /* Local filter related functions. */
 
 /* Sets regular expression of the local filter for the view.  First call of this
  * function initiates filter set process, which should be ended by call to
  * local_filter_accept() or local_filter_cancel().  Returns zero if not all
  * files are filtered out, -1 if filter expression is incorrect and 1 if all
- * files were filtered. */
+ * files were filtered out. */
 int local_filter_set(FileView *view, const char filter[]);
 
 /* Updates cursor position and top line of the view according to interactive
