@@ -178,6 +178,9 @@ expand_macros_i(const char command[], const char args[], MacroFlags *flags,
 		switch(filter(&quotes, command[x],
 					command[x] == '\0' ? '\0' : command[x + 1]))
 		{
+			int well_formed;
+			char key;
+
 			case 'a': /* user arguments */
 				if(args != NULL)
 				{
@@ -235,6 +238,9 @@ expand_macros_i(const char command[], const char args[], MacroFlags *flags,
 			case 'S': /* Show command output in the status bar */
 				set_flags(flags, MF_STATUSBAR_OUTPUT);
 				break;
+			case 'q': /* Show command output in the preview */
+				set_flags(flags, MF_PREVIEW_OUTPUT);
+				break;
 			case 's': /* Split in new screen region and execute command there. */
 				set_flags(flags, MF_SPLIT);
 				break;
@@ -247,33 +253,40 @@ expand_macros_i(const char command[], const char args[], MacroFlags *flags,
 			case 'i': /* Ignore output. */
 				set_flags(flags, MF_IGNORE);
 				break;
-			case 'r': /* Registers' content. */
+			case 'I': /* Interactive custom views. */
+				switch(command[x + 1])
 				{
-					int well_formed;
-					expanded = expand_register(flist_get_dir(curr_view), expanded, quotes,
-							command + x + 2, command[x + 1], &well_formed, for_shell);
-					len = strlen(expanded);
-					if(well_formed)
-					{
-						x++;
-					}
+					case 'u':
+						++x;
+						set_flags(flags, MF_CUSTOMVIEW_IOUTPUT);
+						break;
+					case 'U':
+						++x;
+						set_flags(flags, MF_VERYCUSTOMVIEW_IOUTPUT);
+						break;
+				}
+				break;
+			case 'r': /* Registers' content. */
+				expanded = expand_register(flist_get_dir(curr_view), expanded, quotes,
+						command + x + 2, command[x + 1], &well_formed, for_shell);
+				len = strlen(expanded);
+				if(well_formed)
+				{
+					++x;
 				}
 				break;
 			case 'p': /* Preview pane properties. */
+				key = command[x + 1];
+				if(key == 'c')
 				{
-					int well_formed;
-					const char key = command[x + 1];
-					if(key == 'c')
-					{
-						return expanded;
-					}
+					return expanded;
+				}
 
-					expanded = expand_preview(expanded, key, &well_formed);
-					len = strlen(expanded);
-					if(well_formed)
-					{
-						++x;
-					}
+				expanded = expand_preview(expanded, key, &well_formed);
+				len = strlen(expanded);
+				if(well_formed)
+				{
+					++x;
 				}
 				break;
 			case '%':
@@ -369,8 +382,12 @@ append_selected_files(FileView *view, char expanded[], int under_cursor,
 	}
 	else
 	{
-		expanded = append_entry(view, expanded, type, get_current_entry(view),
-				quotes, mod, for_shell);
+		dir_entry_t *const curr = get_current_entry(view);
+		if(!fentry_is_fake(curr))
+		{
+			expanded = append_entry(view, expanded, type, curr, quotes, mod,
+					for_shell);
+		}
 	}
 
 #ifdef _WIN32
@@ -398,7 +415,7 @@ append_entry(FileView *view, char expanded[], PathType type, dir_entry_t *entry,
 			copy_str(path, sizeof(path), entry->name);
 			break;
 		case PT_REL:
-			get_short_path_of(view, entry, 0, sizeof(path), path);
+			get_short_path_of(view, entry, 0, 0, sizeof(path), path);
 			break;
 		case PT_FULL:
 			get_full_path_of(entry, sizeof(path), path);
@@ -588,17 +605,17 @@ append_to_expanded(char expanded[], const char str[])
 }
 
 const char *
-ma_get_clean_cmd(const char cmd[])
+ma_get_clear_cmd(const char cmd[])
 {
-	const char *clean_cmd;
+	const char *clear_cmd;
 	const char *const break_point = strstr(cmd, "%pc");
 	if(break_point == NULL)
 	{
 		return NULL;
 	}
 
-	clean_cmd = break_point + 3;
-	return is_null_or_empty(clean_cmd) ? NULL : clean_cmd;
+	clear_cmd = break_point + 3;
+	return is_null_or_empty(clear_cmd) ? NULL : clear_cmd;
 }
 
 char *
@@ -691,8 +708,11 @@ macros_to_str(MacroFlags flags)
 		case MF_MENU_OUTPUT: return "%m";
 		case MF_MENU_NAV_OUTPUT: return "%M";
 		case MF_STATUSBAR_OUTPUT: return "%S";
+		case MF_PREVIEW_OUTPUT: return "%q";
 		case MF_CUSTOMVIEW_OUTPUT: return "%u";
 		case MF_VERYCUSTOMVIEW_OUTPUT: return "%U";
+		case MF_CUSTOMVIEW_IOUTPUT: return "%Iu";
+		case MF_VERYCUSTOMVIEW_IOUTPUT: return "%IU";
 
 		case MF_SPLIT: return "%s";
 		case MF_IGNORE: return "%i";
