@@ -26,6 +26,7 @@
 
 #include "../compat/fs_limits.h"
 #include "../int/file_magic.h"
+#include "../modes/dialogs/msg_dialog.h"
 #include "../modes/cmdline.h"
 #include "../modes/menu.h"
 #include "../ui/fileview.h"
@@ -34,6 +35,7 @@
 #include "../utils/str.h"
 #include "../utils/string_array.h"
 #include "../filelist.h"
+#include "../flist_sel.h"
 #include "../filetype.h"
 #include "../running.h"
 #include "../types.h"
@@ -42,26 +44,36 @@
 static const char * form_filetype_menu_entry(assoc_record_t prog,
 		int descr_width);
 static const char * form_filetype_data_entry(assoc_record_t prog);
-static int execute_filetype_cb(FileView *view, menu_info *m);
-static KHandlerResponse filetypes_khandler(menu_info *m, const wchar_t keys[]);
-static void fill_menu_from_records(menu_info *m,
+static int execute_filetype_cb(FileView *view, menu_data_t *m);
+static KHandlerResponse filetypes_khandler(FileView *view, menu_data_t *m,
+		const wchar_t keys[]);
+static void fill_menu_from_records(menu_data_t *m,
 		const assoc_records_t *records);
 static int max_desc_len(const assoc_records_t *records);
 
 int
 show_file_menu(FileView *view, int background)
 {
-	static menu_info m;
+	static menu_data_t m;
 
 	int i;
 	int max_len;
+	assoc_records_t ft, magic;
+	char *typed_name;
 
-	char *const typed_name = get_typed_entry_fpath(get_current_entry(view));
-	assoc_records_t ft = ft_get_all_programs(typed_name);
-	assoc_records_t magic = get_magic_handlers(typed_name);
+	dir_entry_t *const entry = get_current_entry(view);
+	if(fentry_is_fake(entry))
+	{
+		show_error_msg("File menu", "Entry doesn't correspond to a file.");
+		return 0;
+	}
+
+	typed_name = get_typed_entry_fpath(entry);
+	ft = ft_get_all_programs(typed_name);
+	magic = get_magic_handlers(typed_name);
 	free(typed_name);
 
-	init_menu_info(&m, strdup("Filetype associated commands"),
+	init_menu_data(&m, view, strdup("Filetype associated commands"),
 			strdup("No programs set for this filetype"));
 
 	m.execute_handler = &execute_filetype_cb;
@@ -70,7 +82,7 @@ show_file_menu(FileView *view, int background)
 
 	max_len = MAX(max_desc_len(&ft), max_desc_len(&magic));
 
-	for(i = 0; i < ft.count; i++)
+	for(i = 0; i < ft.count; ++i)
 	{
 		(void)add_to_string_array(&m.data, m.len, 1,
 				form_filetype_data_entry(ft.list[i]));
@@ -78,15 +90,17 @@ show_file_menu(FileView *view, int background)
 				form_filetype_menu_entry(ft.list[i], max_len));
 	}
 
+	/* Add empty line separator if there is at least one other item. */
+	if(ft.count > 0 || magic.count > 0)
+	{
+		(void)add_to_string_array(&m.data, m.len, 1,
+				form_filetype_data_entry(NONE_PSEUDO_PROG));
+		m.len = add_to_string_array(&m.items, m.len, 1, "");
+	}
+
 	ft_assoc_records_free(&ft);
 
-#ifdef ENABLE_DESKTOP_FILES
-	(void)add_to_string_array(&m.data, m.len, 1,
-			form_filetype_data_entry(NONE_PSEUDO_PROG));
-	m.len = add_to_string_array(&m.items, m.len, 1, "");
-#endif
-
-	for(i = 0; i < magic.count; i++)
+	for(i = 0; i < magic.count; ++i)
 	{
 		(void)add_to_string_array(&m.data, m.len, 1,
 				form_filetype_data_entry(magic.list[i]));
@@ -94,7 +108,7 @@ show_file_menu(FileView *view, int background)
 				form_filetype_menu_entry(magic.list[i], max_len));
 	}
 
-	return display_menu(&m, view);
+	return display_menu(m.state, view);
 }
 
 /* Returns pointer to a statically allocated buffer */
@@ -134,15 +148,15 @@ form_filetype_data_entry(assoc_record_t prog)
 /* Callback that is invoked when menu item is selected.  Should return non-zero
  * to stay in menu mode. */
 static int
-execute_filetype_cb(FileView *view, menu_info *m)
+execute_filetype_cb(FileView *view, menu_data_t *m)
 {
-	if(view->dir_entry[view->list_pos].type == FT_DIR && m->pos == 0)
+	if(get_current_entry(view)->type == FT_DIR && m->pos == 0)
 	{
 		open_dir(view);
 	}
 	else
 	{
-		const char *prog_str = strchr(m->data[m->pos], '|') + 1;
+		const char *const prog_str = strchr(m->data[m->pos], '|') + 1;
 		if(prog_str[0] != '\0')
 		{
 			int background = m->extra_data & 1;
@@ -150,7 +164,7 @@ execute_filetype_cb(FileView *view, menu_info *m)
 		}
 	}
 
-	clean_selected_files(view);
+	flist_sel_stash(view);
 	redraw_view(view);
 	return 0;
 }
@@ -158,11 +172,11 @@ execute_filetype_cb(FileView *view, menu_info *m)
 /* Menu-specific shortcut handler.  Returns code that specifies both taken
  * actions and what should be done next. */
 static KHandlerResponse
-filetypes_khandler(menu_info *m, const wchar_t keys[])
+filetypes_khandler(FileView *view, menu_data_t *m, const wchar_t keys[])
 {
 	if(wcscmp(keys, L"c") == 0)
 	{
-		const char *prog_str = after_first(m->data[m->pos], '|');
+		const char *const prog_str = after_first(m->data[m->pos], '|');
 		if(prog_str[0] != '\0')
 		{
 			menu_morph_into_cmdline(CLS_COMMAND, prog_str, 1);
@@ -175,40 +189,40 @@ filetypes_khandler(menu_info *m, const wchar_t keys[])
 int
 show_fileprograms_menu(FileView *view, const char fname[])
 {
-	static menu_info m;
+	static menu_data_t m;
 
 	assoc_records_t file_programs;
 
-	init_menu_info(&m, format_str("Programs that match %s", fname),
+	init_menu_data(&m, view, format_str("Programs that match %s", fname),
 			format_str("No programs match %s", fname));
 
 	file_programs = ft_get_all_programs(fname);
 	fill_menu_from_records(&m, &file_programs);
 	ft_assoc_records_free(&file_programs);
 
-	return display_menu(&m, view);
+	return display_menu(m.state, view);
 }
 
 int
 show_fileviewers_menu(FileView *view, const char fname[])
 {
-	static menu_info m;
+	static menu_data_t m;
 
 	assoc_records_t file_viewers;
 
-	init_menu_info(&m, format_str("Viewers that match %s", fname),
+	init_menu_data(&m, view, format_str("Viewers that match %s", fname),
 			format_str("No viewers match %s", fname));
 
 	file_viewers = ft_get_all_viewers(fname);
 	fill_menu_from_records(&m, &file_viewers);
 	ft_assoc_records_free(&file_viewers);
 
-	return display_menu(&m, view);
+	return display_menu(m.state, view);
 }
 
 /* Fills the menu with commands from association records. */
 static void
-fill_menu_from_records(menu_info *m, const assoc_records_t *records)
+fill_menu_from_records(menu_data_t *m, const assoc_records_t *records)
 {
 	int i;
 	const int max_len = max_desc_len(records);
