@@ -28,6 +28,7 @@
 
 #include "../compat/reallocarray.h"
 #include "../engine/keys.h"
+#include "../modes/modes.h"
 #include "../ui/ui.h"
 #include "../utils/str.h"
 #include "../utils/string_array.h"
@@ -38,7 +39,7 @@ static void add_mapping_item(const wchar_t lhs[], const wchar_t rhs[],
 		const char descr[]);
 
 /* Menu object is global to make it available in add_mapping_item(). */
-static menu_info m;
+static menu_data_t m;
 
 /* Prefix to check LHS against. */
 static const wchar_t *prefix;
@@ -49,15 +50,28 @@ int
 show_map_menu(FileView *view, const char mode_str[], int mode,
 		const wchar_t start[])
 {
-	init_menu_info(&m, format_str("Mappings for %s mode", mode_str),
+	const int dialogs = mode == SORT_MODE || mode == ATTR_MODE
+	                 || mode == CHANGE_MODE || mode == FILE_INFO_MODE;
+
+	init_menu_data(&m, view,
+			format_str("Mappings for %s mode%s", mode_str, dialogs ? "s" : ""),
 			strdup("No mappings found"));
 
 	prefix = start;
 	prefix_len = wcslen(prefix);
 
-	vle_keys_list(mode, &add_mapping_item);
+	vle_keys_list(mode, &add_mapping_item, dialogs);
 
-	return display_menu(&m, view);
+	/* If we filtered out all meaningful lines, clear the menu. */
+	if(m.len == 1 && *m.items[0] == '\0')
+	{
+		free(m.items[0]);
+		free(m.items);
+		m.items = NULL;
+		m.len = 0;
+	}
+
+	return display_menu(m.state, view);
 }
 
 /* Adds matching key information to the menu after pre-formatting. */
@@ -68,22 +82,30 @@ add_mapping_item(const wchar_t lhs[], const wchar_t rhs[], const char descr[])
 
 	char *mb_lhs;
 
-	if(wcsncmp(prefix, lhs, prefix_len) != 0)
+	const int is_separator = (lhs[0] == L'\0');
+	if(!is_separator)
 	{
-		return;
-	}
+		if(wcsncmp(prefix, lhs, prefix_len) != 0)
+		{
+			return;
+		}
 
-	/* Handle empty RHS, but don't affect separator line. */
-	if(rhs[0] == L'\0' && lhs[0] != L'\0' && descr[0] == '\0')
-	{
-		rhs = L"<nop>";
+		/* Handle empty RHS. */
+		if(rhs[0] == L'\0' && descr[0] == '\0')
+		{
+			rhs = L"<nop>";
+		}
 	}
 
 	mb_lhs = wstr_to_spec(lhs);
 
 	m.items = reallocarray(m.items, m.len + 1, sizeof(char *));
 
-	if(rhs[0] == L'\0')
+	if(is_separator)
+	{
+		m.items[m.len++] = strdup("");
+	}
+	else if(rhs[0] == L'\0')
 	{
 		m.items[m.len++] = format_str("%-*s %s", MAP_WIDTH, mb_lhs, descr);
 	}
