@@ -33,30 +33,38 @@ show_undolist_menu(FileView *view, int with_details)
 {
 	char **p;
 
-	static menu_info m;
-	init_menu_info(&m, strdup("Undolist"), strdup("Undolist is empty"));
-	m.current = get_undolist_pos(with_details) + 1;
-	m.pos = m.current - 1;
+	static menu_data_t m;
+	init_menu_data(&m, view, strdup("Undolist"), strdup("Undolist is empty"));
 
 	m.items = undolist(with_details);
 	p = m.items;
 	while(*p++ != NULL)
-		m.len++;
-
-	if(m.len > 0)
 	{
-		size_t len;
-
-		m.len = add_to_string_array(&m.items, m.len, 1, "list end");
-
-		/* Add current position mark to menu item. */
-		len = (m.items[m.pos] != NULL) ? strlen(m.items[m.pos]) : 0;
-		m.items[m.pos] = realloc(m.items[m.pos], len + 1 + 1);
-		memmove(m.items[m.pos] + 1, m.items[m.pos], len + 1);
-		m.items[m.pos][0] = '*';
+		++m.len;
 	}
 
-	return display_menu(&m, view);
+	/* Add additional entry before setting position. */
+	if(m.len > 0)
+	{
+		m.len = add_to_string_array(&m.items, m.len, 1, "list end");
+	}
+
+	move_to_menu_pos(get_undolist_pos(with_details), m.state);
+
+	/* Add current position mark to menu item. */
+	if(m.len > 0)
+	{
+		const size_t len = (m.items[m.pos] != NULL) ? strlen(m.items[m.pos]) : 0;
+		char *const new_line = realloc(m.items[m.pos], len + 1U + 1U);
+		if(new_line != NULL)
+		{
+			m.items[m.pos] = new_line;
+			memmove(m.items[m.pos] + 1, m.items[m.pos], len + 1U);
+			m.items[m.pos][0] = '*';
+		}
+	}
+
+	return display_menu(m.state, view);
 }
 
 /* vim: set tabstop=2 softtabstop=2 shiftwidth=2 noexpandtab cinoptions-=(0 : */
