@@ -30,9 +30,10 @@
 #include "../utils/str.h"
 #include "../utils/string_array.h"
 #include "../filelist.h"
+#include "../flist_pos.h"
 #include "menus.h"
 
-static int execute_volumes_cb(FileView *view, menu_info *m);
+static int execute_volumes_cb(FileView *view, menu_data_t *m);
 
 int
 show_volumes_menu(FileView *view)
@@ -41,8 +42,9 @@ show_volumes_menu(FileView *view)
 	char vol_name[MAX_PATH];
 	char file_buf[MAX_PATH];
 
-	static menu_info m;
-	init_menu_info(&m, strdup("Mounted Volumes"), strdup("No volumes mounted"));
+	static menu_data_t m;
+	init_menu_data(&m, view, strdup("Mounted Volumes"),
+			strdup("No volumes mounted"));
 	m.execute_handler = &execute_volumes_cb;
 
 	for(c = 'a'; c <= 'z'; ++c)
@@ -60,13 +62,13 @@ show_volumes_menu(FileView *view)
 		}
 	}
 
-	return display_menu(&m, view);
+	return display_menu(m.state, view);
 }
 
 /* Callback that is called when menu item is selected.  Should return non-zero
  * to stay in menu mode. */
 static int
-execute_volumes_cb(FileView *view, menu_info *m)
+execute_volumes_cb(FileView *view, menu_data_t *m)
 {
 	char path_buf[4];
 	copy_str(path_buf, sizeof(path_buf), m->items[m->pos]);
