@@ -58,6 +58,8 @@
 #include "../cmd_core.h"
 #include "../filelist.h"
 #include "../filtering.h"
+#include "../flist_pos.h"
+#include "../flist_sel.h"
 #include "../marks.h"
 #include "../search.h"
 #include "../status.h"
@@ -92,21 +94,21 @@ PromptState;
 /* Holds state of the command-line editing mode. */
 typedef struct
 {
-	wchar_t *line;            /* the line reading */
-	wchar_t *initial_line;    /* initial state of the line */
-	int index;                /* index of the current character in cmdline */
-	int curs_pos;             /* position of the cursor in status bar*/
-	int len;                  /* length of the string */
-	int cmd_pos;              /* position in the history */
-	wchar_t prompt[NAME_MAX]; /* prompt */
-	int prompt_wid;           /* width of prompt */
-	int complete_continue;    /* if non-zero, continue the previous completion */
-	int dot_pos;              /* history position for dot completion, or < 0 */
-	size_t dot_index;         /* dot completion line index */
-	size_t dot_len;           /* dot completion previous completion len */
-	HIST history_search;      /* HIST_* */
-	int hist_search_len;      /* length of history search pattern */
-	wchar_t *line_buf;        /* content of line before using history */
+	wchar_t *line;                /* the line reading */
+	wchar_t *initial_line;        /* initial state of the line */
+	int index;                    /* index of the current character in cmdline */
+	int curs_pos;                 /* position of the cursor in status bar*/
+	int len;                      /* length of the string */
+	int cmd_pos;                  /* position in the history */
+	wchar_t prompt[NAME_MAX + 1]; /* prompt */
+	int prompt_wid;               /* width of prompt */
+	int complete_continue;        /* if non-zero, continue previous completion */
+	int dot_pos;                  /* history pos for dot completion, or < 0 */
+	size_t dot_index;             /* dot completion line index */
+	size_t dot_len;               /* dot completion previous completion len */
+	HIST history_search;          /* HIST_* */
+	int hist_search_len;          /* length of history search pattern */
+	wchar_t *line_buf;            /* content of line before using history */
 	int reverse_completion;
 	complete_cmd_func complete;
 	int search_mode;
@@ -132,6 +134,7 @@ static int sub_mode_allows_ee;
 
 static int def_handler(wchar_t key);
 static void update_cmdline_text(line_stats_t *stat);
+static void draw_cmdline_text(line_stats_t *stat);
 static void input_line_changed(void);
 static void handle_empty_input(void);
 static void handle_nonempty_input(void);
@@ -285,15 +288,15 @@ static keys_add_info_t builtin_cmds[] = {
 	{{ALT_PERIOD},    {{&cmd_meta_dot}, .descr = "start/continue last arg completion"}},
 #endif
 #ifdef ENABLE_EXTENDED_KEYS
-	{{KEY_BACKSPACE}, {{&cmd_ctrl_h}, .descr = "remove char to the left"}},
-	{{KEY_DOWN},      {{&cmd_down},   .descr = "prefix-complete next history item"}},
-	{{KEY_UP},        {{&cmd_up},     .descr = "prefix-complete previous history item"}},
-	{{KEY_LEFT},      {{&cmd_left},   .descr = "move cursor to the left"}},
-	{{KEY_RIGHT},     {{&cmd_right},  .descr = "move cursor to the right"}},
-	{{KEY_HOME},      {{&cmd_home},   .descr = "move cursor to the beginning"}},
-	{{KEY_END},       {{&cmd_end},    .descr = "move cursor to the end"}},
-	{{KEY_DC},        {{&cmd_delete}, .descr = "delete current character"}},
-	{{KEY_BTAB},      {{&cmd_shift_tab}, .descr = "complete in reverse order"}},
+	{{K(KEY_BACKSPACE)}, {{&cmd_ctrl_h}, .descr = "remove char to the left"}},
+	{{K(KEY_DOWN)},      {{&cmd_down},   .descr = "prefix-complete next history item"}},
+	{{K(KEY_UP)},        {{&cmd_up},     .descr = "prefix-complete previous history item"}},
+	{{K(KEY_LEFT)},      {{&cmd_left},   .descr = "move cursor to the left"}},
+	{{K(KEY_RIGHT)},     {{&cmd_right},  .descr = "move cursor to the right"}},
+	{{K(KEY_HOME)},      {{&cmd_home},   .descr = "move cursor to the beginning"}},
+	{{K(KEY_END)},       {{&cmd_end},    .descr = "move cursor to the end"}},
+	{{K(KEY_DC)},        {{&cmd_delete}, .descr = "delete current character"}},
+	{{K(KEY_BTAB)},      {{&cmd_shift_tab}, .descr = "complete in reverse order"}},
 #endif /* ENABLE_EXTENDED_KEYS */
 };
 
@@ -329,7 +332,7 @@ def_handler(wchar_t key)
 
 	/* There is no need for resetting completion because of terminal dimensions
 	 * change (note also that iswprint(KEY_RESIZE) might return false). */
-	if(key == KEY_RESIZE)
+	if(key == K(KEY_RESIZE))
 	{
 		return 0;
 	}
@@ -367,15 +370,21 @@ def_handler(wchar_t key)
 	return 0;
 }
 
-/* Update test displayed on the command line and cursor. */
+/* Processes input change and redraw of resulting command-line. */
 static void
 update_cmdline_text(line_stats_t *stat)
+{
+	input_line_changed();
+	draw_cmdline_text(stat);
+}
+
+/* Updates text displayed on the command line and cursor within it. */
+static void
+draw_cmdline_text(line_stats_t *stat)
 {
 	int pair = -1;
 	int attr = 0;
 
-	input_line_changed();
-
 	werase(status_bar);
 
 	switch(input_stat.state)
@@ -418,7 +427,7 @@ input_line_changed(void)
 	 * wrong place. */
 	curs_set(0);
 
-	set_view_port();
+	/* set_view_port() should not be called if none of the conditions are true. */
 
 	input_stat.state = PS_NORMAL;
 	if(is_input_line_empty())
@@ -426,12 +435,14 @@ input_line_changed(void)
 		free(previous);
 		previous = NULL;
 
+		set_view_port();
 		handle_empty_input();
 	}
 	else if(previous == NULL || wcscmp(previous, input_stat.line) != 0)
 	{
 		(void)replace_wstring(&previous, input_stat.line);
 
+		set_view_port();
 		handle_nonempty_input();
 	}
 
@@ -456,7 +467,7 @@ handle_empty_input(void)
 		/* Clear selection. */
 		if(prev_mode != MENU_MODE)
 		{
-			clean_selected_files(curr_view);
+			flist_sel_stash(curr_view);
 		}
 		else
 		{
@@ -499,7 +510,7 @@ handle_nonempty_input(void)
 		case CLS_MENU_FSEARCH:
 		case CLS_MENU_BSEARCH:
 			result = search_menu_list(mbinput, sub_mode_ptr, 0);
-			update_state(result, ((menu_info *)sub_mode_ptr)->matching_entries);
+			update_state(result, menu_get_matches(sub_mode_ptr));
 			break;
 		case CLS_FILTER:
 			set_local_filter(mbinput);
@@ -576,6 +587,12 @@ enter_cmdline_mode(CmdLineSubmode cl_sub_mode, const char cmd[], void *ptr)
 	const wchar_t *wprompt;
 	complete_cmd_func complete_func;
 
+	if(cl_sub_mode == CLS_FILTER && curr_view->custom.type == CV_DIFF)
+	{
+		show_error_msg("Filtering", "No local filter for diff views");
+		return;
+	}
+
 	wcmd = to_wide_force(cmd);
 	if(wcmd == NULL)
 	{
@@ -649,7 +666,7 @@ redraw_cmdline(void)
 	}
 	else
 	{
-		update_screen(UT_FULL);
+		update_screen(UT_REDRAW);
 		if(prev_mode == SORT_MODE)
 		{
 			redraw_sort_dialog();
@@ -662,7 +679,7 @@ redraw_cmdline(void)
 
 	line_width = getmaxx(stdscr);
 	update_cmdline_size();
-	update_cmdline_text(&input_stat);
+	draw_cmdline_text(&input_stat);
 	curs_set(1);
 
 	if(input_stat.complete_continue && cfg.wild_menu)
@@ -715,7 +732,10 @@ prepare_cmdline_mode(const wchar_t prompt[], const wchar_t cmd[],
 
 	update_cmdline_size();
 	update_cmdline_text(&input_stat);
-	curs_set(1);
+	if(curr_stats.load_stage > 0)
+	{
+		curs_set(1);
+	}
 
 	curr_stats.save_msg = 1;
 
@@ -799,9 +819,12 @@ leave_cmdline_mode(void)
 	input_stat.initial_line = NULL;
 	input_stat.line_buf = NULL;
 
-	curs_set(0);
+	if(curr_stats.load_stage > 0)
+	{
+		curs_set(0);
+	}
 	curr_stats.save_msg = 0;
-	clean_status_bar();
+	ui_sb_clear();
 
 	if(vle_mode_is(CMDLINE_MODE))
 	{
@@ -810,7 +833,7 @@ leave_cmdline_mode(void)
 
 	if(!vle_mode_is(MENU_MODE))
 	{
-		ui_ruler_update(curr_view);
+		ui_ruler_update(curr_view, 1);
 	}
 
 	attr = cfg.cs.color[CMD_LINE_COLOR].attr;
@@ -1737,7 +1760,7 @@ paste_short_path(FileView *view)
 	if(flist_custom_active(view))
 	{
 		char short_path[PATH_MAX];
-		get_short_path_of(view, get_current_entry(view), 0, sizeof(short_path),
+		get_short_path_of(view, get_current_entry(view), 0, 0, sizeof(short_path),
 				short_path);
 		paste_str(short_path, 1);
 		return;
@@ -1808,7 +1831,7 @@ static void
 paste_short_path_root(FileView *view)
 {
 	char short_path[PATH_MAX];
-	get_short_path_of(view, get_current_entry(view), 0, sizeof(short_path),
+	get_short_path_of(view, get_current_entry(view), 0, 0, sizeof(short_path),
 			short_path);
 	paste_name_part(short_path, 1);
 }
@@ -2640,5 +2663,11 @@ stop_regular_completion(void)
 	}
 }
 
+line_stats_t *
+get_line_stats(void)
+{
+	return &input_stat;
+}
+
 /* vim: set tabstop=2 softtabstop=2 shiftwidth=2 noexpandtab cinoptions-=(0 : */
 /* vim: set cinoptions+=t0 filetype=c : */
