@@ -31,10 +31,10 @@
 #include <stdio.h> /* snprintf() */
 #include <string.h> /* strncat() strlen() */
 
+#include "../../compat/curses.h"
 #include "../../compat/fs_limits.h"
 #include "../../engine/keys.h"
 #include "../../engine/mode.h"
-#include "../../modes/dialogs/msg_dialog.h"
 #include "../../ui/cancellation.h"
 #include "../../ui/fileview.h"
 #include "../../ui/ui.h"
@@ -43,11 +43,13 @@
 #include "../../utils/path.h"
 #include "../../utils/test_helpers.h"
 #include "../../filelist.h"
+#include "../../flist_sel.h"
 #include "../../ops.h"
 #include "../../status.h"
 #include "../../undo.h"
 #include "../modes.h"
 #include "../wk.h"
+#include "msg_dialog.h"
 
 static const char * get_title(void);
 static int is_one_file_selected(int first_file_index);
@@ -100,11 +102,11 @@ static keys_add_info_t builtin_cmds[] = {
 	{WK_q,      {{&cmd_ctrl_c}, .descr = "close the dialog"}},
 	{WK_t,      {{&cmd_space},  .descr = "toggle current item"}},
 #ifdef ENABLE_EXTENDED_KEYS
-	{{KEY_HOME},  {{&cmd_gg},     .descr = "go to the first item"}},
-	{{KEY_END},   {{&cmd_G},      .descr = "go to the last item"}},
-	{{KEY_UP},    {{&cmd_k},      .descr = "go to item above"}},
-	{{KEY_DOWN},  {{&cmd_j},      .descr = "go to item below"}},
-	{{KEY_RIGHT}, {{&cmd_return}, .descr = "update permissions"}},
+	{{K(KEY_HOME)},  {{&cmd_gg},     .descr = "go to the first item"}},
+	{{K(KEY_END)},   {{&cmd_G},      .descr = "go to the last item"}},
+	{{K(KEY_UP)},    {{&cmd_k},      .descr = "go to item above"}},
+	{{K(KEY_DOWN)},  {{&cmd_j},      .descr = "go to item below"}},
+	{{K(KEY_RIGHT)}, {{&cmd_return}, .descr = "update permissions"}},
 #endif /* ENABLE_EXTENDED_KEYS */
 };
 
@@ -122,10 +124,11 @@ init_attr_dialog_mode(void)
 void
 enter_attr_mode(FileView *active_view)
 {
-	int i;
-	mode_t fmode;
+	mode_t fmode = 0;
 	mode_t diff;
 	uid_t uid = geteuid();
+	dir_entry_t *entry;
+	int first;
 
 	if(curr_stats.load_stage < 2)
 		return;
@@ -133,42 +136,31 @@ enter_attr_mode(FileView *active_view)
 	view = active_view;
 	memset(perms, 0, sizeof(perms));
 
-	diff = 0;
-	i = 0;
-	while(i < view->list_rows && !view->dir_entry[i].selected)
-		i++;
+	first = 1;
 	file_is_dir = 0;
-	if(i == view->list_rows)
+	diff = 0;
+	entry = NULL;
+	while(iter_selection_or_current(view, &entry))
 	{
-		i = view->list_pos;
-		file_is_dir = is_dir(view->dir_entry[i].name);
-	}
-	fmode = view->dir_entry[i].mode;
-	if(uid != 0 && view->dir_entry[i].uid != uid)
-	{
-		show_error_msgf("Access error", "You are not owner of %s",
-				view->dir_entry[i].name);
-		clean_selected_files(view);
-		load_dir_list(view, 1);
-		fview_cursor_redraw(view);
-		return;
-	}
-	while(i < view->list_rows)
-	{
-		if(view->dir_entry[i].selected)
+		if(first)
 		{
-			diff |= (view->dir_entry[i].mode ^ fmode);
-			file_is_dir = file_is_dir || is_dir(view->dir_entry[i].name);
-
-			if(uid != 0 && view->dir_entry[i].uid != uid)
-			{
-				show_error_msgf("Access error", "You are not owner of %s",
-						view->dir_entry[i].name);
-				return;
-			}
+			fmode = entry->mode;
+			first = 0;
 		}
 
-		i++;
+		diff |= (entry->mode ^ fmode);
+		file_is_dir |= fentry_is_dir(entry);
+
+		if(uid != 0 && entry->uid != uid)
+		{
+			show_error_msgf("Access error", "You are not owner of %s", entry->name);
+			return;
+		}
+	}
+	if(first)
+	{
+		show_error_msg("Permissions", "No files to process");
+		return;
 	}
 
 	vle_mode_set(ATTR_MODE, VMT_SECONDARY);
@@ -301,8 +293,8 @@ redraw_attr_dialog(void)
 		waddstr(change_win, "...");
 	}
 
-	curs_set(1);
 	checked_wmove(change_win, curr, col);
+	curs_set(1);
 	wrefresh(change_win);
 }
 
@@ -311,7 +303,7 @@ redraw_attr_dialog(void)
 static const char *
 get_title(void)
 {
-	static char title[NAME_MAX];
+	static char title[NAME_MAX + 1];
 
 	const int first_file_index = get_first_file_index();
 
@@ -384,11 +376,8 @@ leave_attr_mode(void)
 	curs_set(0);
 	curr_stats.use_input_bar = 1;
 
-	clean_selected_files(view);
-	load_dir_list(view, 1);
-	fview_cursor_redraw(view);
-
-	update_all_windows();
+	flist_sel_stash(view);
+	ui_view_schedule_reload(view);
 }
 
 static void
@@ -432,8 +421,9 @@ set_perm_string(FileView *view, const int perms[13], const int origin_perms[13],
 		adv_perms[1] = -1;
 		adv_perms[2] = -1;
 
-		perm_str_len += snprintf(perm_str + perm_str_len,
-				sizeof(perm_str) - perm_str_len, "a-x+X,");
+		snprintf(perm_str + perm_str_len, sizeof(perm_str) - perm_str_len,
+				"a-x+X,");
+		perm_str_len += strlen(perm_str + perm_str_len);
 	}
 
 	for(i = 0; i < 12; i++)
@@ -465,8 +455,9 @@ set_perm_string(FileView *view, const int perms[13], const int origin_perms[13],
 			perm = add_perm[i];
 		}
 
-		perm_str_len += snprintf(perm_str + perm_str_len,
-				sizeof(perm_str) - perm_str_len, "%s,", perm);
+		snprintf(perm_str + perm_str_len, sizeof(perm_str) - perm_str_len, "%s,",
+				perm);
+		perm_str_len += strlen(perm_str + perm_str_len);
 	}
 	perm_str[strlen(perm_str) - 1] = '\0'; /* Remove last comma (','). */
 
@@ -474,63 +465,41 @@ set_perm_string(FileView *view, const int perms[13], const int origin_perms[13],
 }
 
 void
-files_chmod(FileView *view, const char *mode, int recurse_dirs)
+files_chmod(FileView *view, const char mode[], int recurse_dirs)
 {
-	int i;
+	char undo_msg[COMMAND_GROUP_INFO_LEN];
+	dir_entry_t *entry;
+	size_t len;
+
+	snprintf(undo_msg, sizeof(undo_msg), "chmod in %s: ",
+			replace_home_part(flist_get_dir(view)));
+	len = strlen(undo_msg);
 
 	ui_cancellation_reset();
 
-	i = 0;
-	while(i < view->list_rows && !view->dir_entry[i].selected)
-		i++;
+	entry = NULL;
+	while(iter_selection_or_current(view, &entry) && !ui_cancellation_requested())
+	{
+		if(len >= 2U && undo_msg[len - 2U] != ':')
+		{
+			strncat(undo_msg + len, ", ", sizeof(undo_msg) - len - 1);
+			len += strlen(undo_msg + len);
+		}
+		strncat(undo_msg + len, entry->name, sizeof(undo_msg) - len - 1);
+		len += strlen(undo_msg + len);
+	}
 
-	if(i == view->list_rows)
+	cmd_group_begin(undo_msg);
+
+	entry = NULL;
+	while(iter_selection_or_current(view, &entry) && !ui_cancellation_requested())
 	{
-		char buf[COMMAND_GROUP_INFO_LEN];
 		char inv[16];
-		snprintf(buf, sizeof(buf), "chmod in %s: %s",
-				replace_home_part(flist_get_dir(view)),
-				view->dir_entry[view->list_pos].name);
-		cmd_group_begin(buf);
-		snprintf(inv, sizeof(inv), "0%o",
-				view->dir_entry[view->list_pos].mode & 0xff);
-		chmod_file_in_list(view, view->list_pos, mode, inv, recurse_dirs);
+		snprintf(inv, sizeof(inv), "0%o", entry->mode & 0xff);
+		chmod_file_in_list(view, entry_to_pos(view, entry), mode, inv,
+				recurse_dirs);
 	}
-	else
-	{
-		char buf[COMMAND_GROUP_INFO_LEN];
-		size_t len;
-		int j = i;
-		len = snprintf(buf, sizeof(buf), "chmod in %s: ",
-				replace_home_part(flist_get_dir(view)));
 
-		while(i < view->list_rows && len < sizeof(buf))
-		{
-			if(view->dir_entry[i].selected)
-			{
-				if(len >= 2 && buf[len - 2] != ':')
-				{
-					strncat(buf + len, ", ", sizeof(buf) - len - 1);
-					len += strlen(buf + len);
-				}
-				strncat(buf + len, view->dir_entry[i].name, sizeof(buf) - len - 1);
-				len += strlen(buf + len);
-			}
-			i++;
-		}
-
-		cmd_group_begin(buf);
-		while(j < view->list_rows && !ui_cancellation_requested())
-		{
-			if(view->dir_entry[j].selected)
-			{
-				char inv[16];
-				snprintf(inv, sizeof(inv), "0%o", view->dir_entry[j].mode & 0xff);
-				chmod_file_in_list(view, j, mode, inv, recurse_dirs);
-			}
-			j++;
-		}
-	}
 	cmd_group_end();
 }
 
