@@ -190,7 +190,7 @@ handle_response(Result r)
 	};
 
 	(void)def_handler(r_to_c[r]);
-	/* Default handler might have requested quitting. */
+	/* Default handler might have already requested quitting. */
 	if(!quit)
 	{
 		if(accept_mask & MASK(r))
@@ -237,7 +237,18 @@ show_error_msgf(const char title[], const char format[], ...)
 int
 prompt_error_msg(const char title[], const char message[])
 {
-	return prompt_error_msg_internal(title, message, 1);
+	char portion[1024];
+	do
+	{
+		snprintf(portion, sizeof(portion), "%s", message);
+		message += strlen(portion);
+		if(prompt_error_msg_internal(title, portion, 1))
+		{
+			return 1;
+		}
+	}
+	while(*message != '\0');
+	return 0;
 }
 
 int
@@ -327,7 +338,7 @@ prompt_msg_internal(const char title[], const char message[],
 
 	redraw_error_msg(title, message, 0, 0);
 
-	enter(MASK(R_YES, R_NO));
+	enter(variants == NULL ? MASK(R_YES, R_NO) : 0);
 
 	touchwin(stdscr);
 
@@ -472,10 +483,13 @@ draw_msg(const char title[], const char msg[], const char ctrl_msg[],
 
 	int sw, sh;
 	int w, h;
+	int max_h;
 	int len;
 	size_t ctrl_msg_n;
 	size_t wctrl_msg;
 	size_t i;
+	int first_line_x = 1;
+	const int first_line_y = 2;
 
 	curs_set(0);
 
@@ -483,7 +497,8 @@ draw_msg(const char title[], const char msg[], const char ctrl_msg[],
 
 	ctrl_msg_n = MAX(measure_sub_lines(ctrl_msg, &wctrl_msg), 1U);
 
-	h = sh - 2 - ctrl_msg_n + !cfg.display_statusline;
+	max_h = sh - 2 - ctrl_msg_n + !cfg.display_statusline;
+	h = max_h;
 	/* The outermost condition is for VLA below (to calm static analyzers). */
 	w = MAX(2 + 2*margin, MIN(sw - 2,
 	        MAX(MAX(recommended_width, sw/3),
@@ -495,17 +510,17 @@ draw_msg(const char title[], const char msg[], const char ctrl_msg[],
 	len = strlen(msg);
 	if(len <= w - 2 && strchr(msg, '\n') == NULL)
 	{
+		first_line_x = (w - len)/2;
 		h = 5 + ctrl_msg_n;
 		wresize(error_win, h, w);
 		mvwin(error_win, (sh - h)/2, (sw - w)/2);
-		checked_wmove(error_win, 2, (w - len)/2);
+		checked_wmove(error_win, first_line_y, first_line_x);
 		wprint(error_win, msg);
 	}
 	else
 	{
-		int i;
-		int cy = 2;
-		i = 0;
+		int i = 0;
+		int cy = first_line_y;
 		while(i < len)
 		{
 			int j;
@@ -526,11 +541,27 @@ draw_msg(const char title[], const char msg[], const char ctrl_msg[],
 			if(buf[0] == '\0')
 				continue;
 
-			h = cy + 4;
+			if(cy >= max_h - (int)ctrl_msg_n - 3)
+			{
+				/* Skip trailing part of the message if it's too long, just print how
+				 * many lines we're omitting. */
+				size_t max_len;
+				const int more_lines = 1U + measure_sub_lines(msg + i, &max_len);
+				snprintf(buf, sizeof(buf), "<<%d more line%s not shown>>", more_lines,
+						(more_lines == 1) ? "" : "s");
+				/* Make sure this is the last iteration of the loop. */
+				i = len;
+			}
+
+			h = 1 + cy + 1 + ctrl_msg_n + 1;
 			wresize(error_win, h, w);
 			mvwin(error_win, (sh - h)/2, (sw - w)/2);
 
 			cx = centered ? (w - utf8_strsw(buf))/2 : (1 + margin);
+			if(cy == first_line_y)
+			{
+				first_line_x = cx;
+			}
 			checked_wmove(error_win, cy++, cx);
 			wprint(error_win, buf);
 		}
@@ -549,6 +580,8 @@ draw_msg(const char title[], const char msg[], const char ctrl_msg[],
 		mvwaddnstr(error_win, h - i - 1, MAX(0, (w - (int)len)/2), ctrl_msg, len);
 		ctrl_msg = skip_char(ctrl_msg + len + 1U, '/');
 	}
+
+	checked_wmove(error_win, first_line_y, first_line_x);
 }
 
 /* Counts number of sub-lines (separated by new-line character in the msg.  Sets
