@@ -29,6 +29,7 @@
 #include <string.h> /* strerror() */
 
 #include "../cfg/config.h"
+#include "../compat/curses.h"
 #include "../compat/reallocarray.h"
 #include "../engine/cmds.h"
 #include "../engine/keys.h"
@@ -45,7 +46,7 @@
 #include "../utils/utils.h"
 #include "../cmd_core.h"
 #include "../cmd_completion.h"
-#include "../filelist.h"
+#include "../flist_sel.h"
 #include "../status.h"
 #include "cmdline.h"
 #include "modes.h"
@@ -64,20 +65,21 @@ static int skip_at_beginning(int id, const char *args);
 
 static int key_handler(wchar_t key);
 static void cmd_ctrl_b(key_info_t key_info, keys_info_t *keys_info);
-static int can_scroll_menu_up(const menu_info *menu);
+static int can_scroll_menu_up(const menu_data_t *menu);
 static void cmd_ctrl_c(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_ctrl_d(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_ctrl_e(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_ctrl_f(key_info_t key_info, keys_info_t *keys_info);
-static int can_scroll_menu_down(const menu_info *menu);
-static void change_menu_top(menu_info *const menu, int delta);
+static int can_scroll_menu_down(const menu_data_t *menu);
+static void change_menu_top(menu_data_t *const menu, int delta);
 static void cmd_ctrl_l(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_return(key_info_t key_info, keys_info_t *keys_info);
 static void update_ui_on_leaving(void);
 static void cmd_ctrl_u(key_info_t key_info, keys_info_t *keys_info);
-static int get_effective_menu_scroll_offset(const menu_info *menu);
+static int get_effective_menu_scroll_offset(const menu_data_t *menu);
 static void cmd_ctrl_y(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_slash(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_percent(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_colon(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_qmark(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_B(key_info_t key_info, keys_info_t *keys_info);
@@ -103,7 +105,7 @@ static void cmd_zh(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_zl(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_zt(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_zz(key_info_t key_info, keys_info_t *keys_info);
-static int all_lines_visible(const menu_info *const menu);
+static int all_lines_visible(const menu_data_t *const menu);
 static int goto_cmd(const cmd_info_t *cmd_info);
 static int nohlsearch_cmd(const cmd_info_t *cmd_info);
 static int quit_cmd(const cmd_info_t *cmd_info);
@@ -111,62 +113,63 @@ static int write_cmd(const cmd_info_t *cmd_info);
 static void leave_menu_mode(int reset_selection);
 
 static FileView *view;
-static menu_info *menu;
+static menu_data_t *menu;
 static int last_search_backward;
 static int was_redraw;
 static int saved_top, saved_pos;
 
 static keys_add_info_t builtin_cmds[] = {
-	{WK_C_b,    {{&cmd_ctrl_b}, .descr = "scroll page up"}},
-	{WK_C_c,    {{&cmd_ctrl_c}, .descr = "leave menu mode"}},
-	{WK_C_d,    {{&cmd_ctrl_d}, .descr = "scroll half-page down"}},
-	{WK_C_e,    {{&cmd_ctrl_e}, .descr = "scroll one line down"}},
-	{WK_C_f,    {{&cmd_ctrl_f}, .descr = "scroll page down"}},
-	{WK_C_l,    {{&cmd_ctrl_l}, .descr = "redraw"}},
-	{WK_CR,     {{&cmd_return}, .descr = "pick current item"}},
-	{WK_C_n,    {{&cmd_j},      .descr = "go to item below"}},
-	{WK_C_p,    {{&cmd_k},      .descr = "go to item above"}},
-	{WK_C_u,    {{&cmd_ctrl_u}, .descr = "scroll half-page up"}},
-	{WK_C_y,    {{&cmd_ctrl_y}, .descr = "scroll one line up"}},
-	{WK_ESC,    {{&cmd_ctrl_c}, .descr = "leave menu mode"}},
-	{WK_SLASH,  {{&cmd_slash},  .descr = "search forward"}},
-	{WK_COLON,  {{&cmd_colon},  .descr = "go to cmdline mode"}},
-	{WK_QM,     {{&cmd_qmark},  .descr = "search backward"}},
-	{WK_B,      {{&cmd_B},      .descr = "make unsorted custom view"}},
-	{WK_G,      {{&cmd_G},      .descr = "go to the last item"}},
-	{WK_H,      {{&cmd_H},      .descr = "go to top of viewport"}},
-	{WK_L,      {{&cmd_L},      .descr = "go to bottom of viewport"}},
-	{WK_M,      {{&cmd_M},      .descr = "go to middle of viewport"}},
-	{WK_N,      {{&cmd_N},      .descr = "go to previous search match"}},
-	{WK_Z WK_Z, {{&cmd_ctrl_c}, .descr = "leave menu mode"}},
-	{WK_Z WK_Q, {{&cmd_ctrl_c}, .descr = "leave menu mode"}},
-	{WK_b,      {{&cmd_b},      .descr = "make custom view"}},
-	{WK_d WK_d, {{&cmd_dd},     .descr = "remove files"}},
-	{WK_g WK_f, {{&cmd_gf},     .descr = "navigate to file location"}},
-	{WK_g WK_g, {{&cmd_gg},     .descr = "go to the first item"}},
-	{WK_j,      {{&cmd_j},      .descr = "go to item below"}},
-	{WK_k,      {{&cmd_k},      .descr = "go to item above"}},
-	{WK_l,      {{&cmd_return}, .descr = "pick current item"}},
-	{WK_n,      {{&cmd_n},      .descr = "go to next search match"}},
-	{WK_q,      {{&cmd_ctrl_c}, .descr = "leave menu mode"}},
-	{WK_v,      {{&cmd_v},      .descr = "use items as Vim quickfix list"}},
-	{WK_z WK_b, {{&cmd_zb},     .descr = "push cursor to the bottom"}},
-	{WK_z WK_H, {{&cmd_zH},     .descr = "scroll page left"}},
-	{WK_z WK_L, {{&cmd_zL},     .descr = "scroll page right"}},
-	{WK_z WK_h, {{&cmd_zh},     .descr = "scroll one column left"}},
-	{WK_z WK_l, {{&cmd_zl},     .descr = "scroll one column right"}},
-	{WK_z WK_t, {{&cmd_zt},     .descr = "push cursor to the top"}},
-	{WK_z WK_z, {{&cmd_zz},     .descr = "center cursor position"}},
+	{WK_C_b,     {{&cmd_ctrl_b},  .descr = "scroll page up"}},
+	{WK_C_c,     {{&cmd_ctrl_c},  .descr = "leave menu mode"}},
+	{WK_C_d,     {{&cmd_ctrl_d},  .descr = "scroll half-page down"}},
+	{WK_C_e,     {{&cmd_ctrl_e},  .descr = "scroll one line down"}},
+	{WK_C_f,     {{&cmd_ctrl_f},  .descr = "scroll page down"}},
+	{WK_C_l,     {{&cmd_ctrl_l},  .descr = "redraw"}},
+	{WK_CR,      {{&cmd_return},  .descr = "pick current item"}},
+	{WK_C_n,     {{&cmd_j},       .descr = "go to item below"}},
+	{WK_C_p,     {{&cmd_k},       .descr = "go to item above"}},
+	{WK_C_u,     {{&cmd_ctrl_u},  .descr = "scroll half-page up"}},
+	{WK_C_y,     {{&cmd_ctrl_y},  .descr = "scroll one line up"}},
+	{WK_ESC,     {{&cmd_ctrl_c},  .descr = "leave menu mode"}},
+	{WK_SLASH,   {{&cmd_slash},   .descr = "search forward"}},
+	{WK_PERCENT, {{&cmd_percent}, .descr = "go to [count]% position"}},
+	{WK_COLON,   {{&cmd_colon},   .descr = "go to cmdline mode"}},
+	{WK_QM,      {{&cmd_qmark},   .descr = "search backward"}},
+	{WK_B,       {{&cmd_B},       .descr = "make unsorted custom view"}},
+	{WK_G,       {{&cmd_G},       .descr = "go to the last item"}},
+	{WK_H,       {{&cmd_H},       .descr = "go to top of viewport"}},
+	{WK_L,       {{&cmd_L},       .descr = "go to bottom of viewport"}},
+	{WK_M,       {{&cmd_M},       .descr = "go to middle of viewport"}},
+	{WK_N,       {{&cmd_N},       .descr = "go to previous search match"}},
+	{WK_Z WK_Z,  {{&cmd_ctrl_c},  .descr = "leave menu mode"}},
+	{WK_Z WK_Q,  {{&cmd_ctrl_c},  .descr = "leave menu mode"}},
+	{WK_b,       {{&cmd_b},       .descr = "make custom view"}},
+	{WK_d WK_d,  {{&cmd_dd},      .descr = "remove files"}},
+	{WK_g WK_f,  {{&cmd_gf},      .descr = "navigate to file location"}},
+	{WK_g WK_g,  {{&cmd_gg},      .descr = "go to the first item"}},
+	{WK_j,       {{&cmd_j},       .descr = "go to item below"}},
+	{WK_k,       {{&cmd_k},       .descr = "go to item above"}},
+	{WK_l,       {{&cmd_return},  .descr = "pick current item"}},
+	{WK_n,       {{&cmd_n},       .descr = "go to next search match"}},
+	{WK_q,       {{&cmd_ctrl_c},  .descr = "leave menu mode"}},
+	{WK_v,       {{&cmd_v},       .descr = "use items as Vim quickfix list"}},
+	{WK_z WK_b,  {{&cmd_zb},      .descr = "push cursor to the bottom"}},
+	{WK_z WK_H,  {{&cmd_zH},      .descr = "scroll page left"}},
+	{WK_z WK_L,  {{&cmd_zL},      .descr = "scroll page right"}},
+	{WK_z WK_h,  {{&cmd_zh},      .descr = "scroll one column left"}},
+	{WK_z WK_l,  {{&cmd_zl},      .descr = "scroll one column right"}},
+	{WK_z WK_t,  {{&cmd_zt},      .descr = "push cursor to the top"}},
+	{WK_z WK_z,  {{&cmd_zz},      .descr = "center cursor position"}},
 #ifdef ENABLE_EXTENDED_KEYS
-	{{KEY_PPAGE},       {{&cmd_ctrl_b}, .descr = "scroll page up"}},
-	{{KEY_NPAGE},       {{&cmd_ctrl_f}, .descr = "scroll page down"}},
-	{{KEY_UP},          {{&cmd_k},      .descr = "go to item above"}},
-	{{KEY_DOWN},        {{&cmd_j},      .descr = "go to item below"}},
-	{{KEY_RIGHT},       {{&cmd_return}, .descr = "pick current item"}},
-	{{KEY_HOME},        {{&cmd_gg},     .descr = "go to the first item"}},
-	{{KEY_END},         {{&cmd_G},      .descr = "go to the last item"}},
-	{{WC_z, KEY_LEFT},  {{&cmd_zh},     .descr = "scroll one column left"}},
-	{{WC_z, KEY_RIGHT}, {{&cmd_zl},     .descr = "scroll one column right"}},
+	{{K(KEY_PPAGE)},       {{&cmd_ctrl_b}, .descr = "scroll page up"}},
+	{{K(KEY_NPAGE)},       {{&cmd_ctrl_f}, .descr = "scroll page down"}},
+	{{K(KEY_UP)},          {{&cmd_k},      .descr = "go to item above"}},
+	{{K(KEY_DOWN)},        {{&cmd_j},      .descr = "go to item below"}},
+	{{K(KEY_RIGHT)},       {{&cmd_return}, .descr = "pick current item"}},
+	{{K(KEY_HOME)},        {{&cmd_gg},     .descr = "go to the first item"}},
+	{{K(KEY_END)},         {{&cmd_G},      .descr = "go to the last item"}},
+	{{WC_z, K(KEY_LEFT)},  {{&cmd_zh},     .descr = "scroll one column left"}},
+	{{WC_z, K(KEY_RIGHT)}, {{&cmd_zl},     .descr = "scroll one column right"}},
 #endif /* ENABLE_EXTENDED_KEYS */
 };
 
@@ -262,10 +265,14 @@ init_menu_mode(void)
 
 	(void)ret_code;
 
-	vle_keys_set_def_handler(MENU_MODE, key_handler);
+	vle_keys_set_def_handler(MENU_MODE, &key_handler);
 
-	init_cmds(0, &cmds_conf);
-	add_builtin_commands((const cmd_add_t *)&commands, ARRAY_LEN(commands));
+	/* Double initialization can happen in tests. */
+	if(cmds_conf.inner == NULL)
+	{
+		init_cmds(0, &cmds_conf);
+		add_builtin_commands((const cmd_add_t *)&commands, ARRAY_LEN(commands));
+	}
 }
 
 static int
@@ -283,10 +290,12 @@ key_handler(wchar_t key)
 }
 
 void
-enter_menu_mode(menu_info *m, FileView *active_view)
+enter_menu_mode(menu_data_t *m, FileView *active_view)
 {
 	if(curr_stats.load_stage < 2)
+	{
 		return;
+	}
 
 	assert(m->len > 0 && "Menu cannot be empty.");
 
@@ -301,6 +310,17 @@ enter_menu_mode(menu_info *m, FileView *active_view)
 	init_cmds(0, &cmds_conf);
 }
 
+void
+reenter_menu_mode(menu_data_t *m)
+{
+	assert(vle_mode_is(MENU_MODE) && "Can't reenter if not in menu mode.");
+	assert(m->len > 0 && "Menu cannot be empty.");
+
+	menus_replace_menu(m);
+	redraw_menu(m->state);
+	menu = m;
+}
+
 void
 menu_pre(void)
 {
@@ -323,7 +343,7 @@ void
 menu_redraw(void)
 {
 	was_redraw = 1;
-	redraw_menu(menu);
+	redraw_menu(menu->state);
 }
 
 static void
@@ -332,11 +352,14 @@ cmd_ctrl_b(key_info_t key_info, keys_info_t *keys_info)
 	if(can_scroll_menu_up(menu))
 	{
 		const int s = get_effective_menu_scroll_offset(menu);
-		const int off = (menu->win_rows - 2) - SCROLL_GAP;
+		const int off = (getmaxy(menu_win) - 2) - SCROLL_GAP;
 		menu->pos = get_last_visible_line(menu) - off;
 		change_menu_top(menu, -off);
-		if(cfg.scroll_off > 0 && menu->top + (menu->win_rows - 3) - menu->pos < s)
-			menu->pos -= s - (menu->top + (menu->win_rows - 3) - menu->pos);
+		if(cfg.scroll_off > 0 &&
+				menu->top + (getmaxy(menu_win) - 3) - menu->pos < s)
+		{
+			menu->pos -= s - (menu->top + (getmaxy(menu_win) - 3) - menu->pos);
+		}
 
 		update_menu();
 	}
@@ -344,7 +367,7 @@ cmd_ctrl_b(key_info_t key_info, keys_info_t *keys_info)
 
 /* Returns non-zero if menu can be scrolled up. */
 static int
-can_scroll_menu_up(const menu_info *menu)
+can_scroll_menu_up(const menu_data_t *menu)
 {
 	return menu->top > 0;
 }
@@ -359,11 +382,13 @@ static void
 cmd_ctrl_d(key_info_t key_info, keys_info_t *keys_info)
 {
 	const int s = get_effective_menu_scroll_offset(menu);
-	clean_menu_position(menu);
-	menu->top += DIV_ROUND_UP(menu->win_rows - 3, 2);
-	menu->pos += DIV_ROUND_UP(menu->win_rows - 3, 2);
+	menu_current_line_erase(menu->state);
+	menu->top += DIV_ROUND_UP(getmaxy(menu_win) - 3, 2);
+	menu->pos += DIV_ROUND_UP(getmaxy(menu_win) - 3, 2);
 	if(cfg.scroll_off > 0 && menu->pos - menu->top < s)
+	{
 		menu->pos += s - (menu->pos - menu->top);
+	}
 
 	update_menu();
 }
@@ -375,9 +400,11 @@ cmd_ctrl_e(key_info_t key_info, keys_info_t *keys_info)
 	{
 		int off = MAX(cfg.scroll_off, 0);
 		if(menu->pos <= menu->top + off)
+		{
 			menu->pos = menu->top + 1 + off;
+		}
 
-		menu->top++;
+		++menu->top;
 		update_menu();
 	}
 }
@@ -388,11 +415,13 @@ cmd_ctrl_f(key_info_t key_info, keys_info_t *keys_info)
 	if(can_scroll_menu_down(menu))
 	{
 		const int s = get_effective_menu_scroll_offset(menu);
-		const int off = (menu->win_rows - 2) - SCROLL_GAP;
+		const int off = (getmaxy(menu_win) - 2) - SCROLL_GAP;
 		menu->pos = menu->top + off;
 		change_menu_top(menu, off);
 		if(cfg.scroll_off > 0 && menu->pos - menu->top < s)
+		{
 			menu->pos += s - (menu->pos - menu->top);
+		}
 
 		update_menu();
 	}
@@ -400,22 +429,23 @@ cmd_ctrl_f(key_info_t key_info, keys_info_t *keys_info)
 
 /* Returns non-zero if menu can be scrolled down. */
 static int
-can_scroll_menu_down(const menu_info *menu)
+can_scroll_menu_down(const menu_data_t *menu)
 {
 	return get_last_visible_line(menu) < menu->len - 1;
 }
 
 /* Moves top line of the menu ensuring that its value is correct. */
 static void
-change_menu_top(menu_info *const menu, int delta)
+change_menu_top(menu_data_t *const menu, int delta)
 {
-	menu->top = MAX(MIN(menu->top + delta, menu->len - (menu->win_rows - 2)), 0);
+	menu->top =
+		MAX(MIN(menu->top + delta, menu->len - (getmaxy(menu_win) - 2)), 0);
 }
 
 int
-get_last_visible_line(const menu_info *menu)
+get_last_visible_line(const menu_data_t *menu)
 {
-	return menu->top + (menu->win_rows - 2) - 1;
+	return menu->top + (getmaxy(menu_win) - 2) - 1;
 }
 
 /* Redraw TUI. */
@@ -428,7 +458,7 @@ cmd_ctrl_l(key_info_t key_info, keys_info_t *keys_info)
 static void
 cmd_return(key_info_t key_info, keys_info_t *keys_info)
 {
-	static menu_info *saved_menu;
+	static menu_data_t *saved_menu;
 
 	vle_mode_set(NORMAL_MODE, VMT_PRIMARY);
 	saved_menu = menu;
@@ -441,11 +471,11 @@ cmd_return(key_info_t key_info, keys_info_t *keys_info)
 
 	if(!vle_mode_is(MENU_MODE))
 	{
-		reset_popup_menu(saved_menu);
+		reset_menu_data(saved_menu);
 	}
 	else if(menu != saved_menu)
 	{
-		reset_popup_menu(saved_menu);
+		reset_menu_data(saved_menu);
 		update_menu();
 	}
 
@@ -470,24 +500,28 @@ static void
 cmd_ctrl_u(key_info_t key_info, keys_info_t *keys_info)
 {
 	const int s = get_effective_menu_scroll_offset(menu);
-	clean_menu_position(menu);
+	menu_current_line_erase(menu->state);
 
-	if(cfg.scroll_off > 0 && menu->top + menu->win_rows - menu->pos < s)
-		menu->pos -= s - (menu->top + (menu->win_rows - 3) - menu->pos);
+	if(cfg.scroll_off > 0 && menu->top + getmaxy(menu_win) - menu->pos < s)
+	{
+		menu->pos -= s - (menu->top + (getmaxy(menu_win) - 3) - menu->pos);
+	}
 
-	menu->top -= DIV_ROUND_UP(menu->win_rows - 3, 2);
+	menu->top -= DIV_ROUND_UP(getmaxy(menu_win) - 3, 2);
 	if(menu->top < 0)
+	{
 		menu->top = 0;
-	menu->pos -= DIV_ROUND_UP(menu->win_rows - 3, 2);
+	}
+	menu->pos -= DIV_ROUND_UP(getmaxy(menu_win) - 3, 2);
 
 	update_menu();
 }
 
 /* Returns scroll offset value for the menu taking menu height into account. */
 static int
-get_effective_menu_scroll_offset(const menu_info *menu)
+get_effective_menu_scroll_offset(const menu_data_t *menu)
 {
-	return MIN(DIV_ROUND_UP(menu->win_rows - 3, 2) - 1, cfg.scroll_off);
+	return MIN(DIV_ROUND_UP(getmaxy(menu_win) - 3, 2) - 1, cfg.scroll_off);
 }
 
 static void
@@ -496,10 +530,12 @@ cmd_ctrl_y(key_info_t key_info, keys_info_t *keys_info)
 	if(can_scroll_menu_up(menu))
 	{
 		int off = MAX(cfg.scroll_off, 0);
-		if(menu->pos >= menu->top + menu->win_rows - 3 - off)
-			menu->pos = menu->top - 1 + menu->win_rows - 3 - off;
+		if(menu->pos >= menu->top + getmaxy(menu_win) - 3 - off)
+		{
+			menu->pos = menu->top - 1 + getmaxy(menu_win) - 3 - off;
+		}
 
-		menu->top--;
+		--menu->top;
 		update_menu();
 	}
 }
@@ -507,14 +543,25 @@ cmd_ctrl_y(key_info_t key_info, keys_info_t *keys_info)
 static void
 cmd_slash(key_info_t key_info, keys_info_t *keys_info)
 {
-	menu->search_repeat = def_count(key_info.count);
 	last_search_backward = 0;
-	menu->backward_search = 0;
-	free(menu->regexp);
-	menu->regexp = NULL;
+	menu_new_search(menu->state, last_search_backward, def_count(key_info.count));
 	enter_cmdline_mode(CLS_MENU_FSEARCH, "", menu);
 }
 
+/* Jump to percent of list. */
+static void
+cmd_percent(key_info_t key_info, keys_info_t *keys_info)
+{
+	int line;
+	if(key_info.count == NO_COUNT_GIVEN || key_info.count > 100)
+	{
+		return;
+	}
+	line = (key_info.count*menu->len)/100;
+	move_to_menu_pos(line, menu->state);
+	wrefresh(menu_win);
+}
+
 static void
 cmd_colon(key_info_t key_info, keys_info_t *keys_info)
 {
@@ -527,10 +574,8 @@ cmd_colon(key_info_t key_info, keys_info_t *keys_info)
 static void
 cmd_qmark(key_info_t key_info, keys_info_t *keys_info)
 {
-	menu->search_repeat = def_count(key_info.count);
 	last_search_backward = 1;
-	menu->backward_search = 1;
-	free(menu->regexp);
+	menu_new_search(menu->state, last_search_backward, def_count(key_info.count));
 	enter_cmdline_mode(CLS_MENU_BSEARCH, "", menu);
 }
 
@@ -545,10 +590,12 @@ static void
 cmd_G(key_info_t key_info, keys_info_t *keys_info)
 {
 	if(key_info.count == NO_COUNT_GIVEN)
+	{
 		key_info.count = menu->len;
+	}
 
-	clean_menu_position(menu);
-	move_to_menu_pos(key_info.count - 1, menu);
+	menu_current_line_erase(menu->state);
+	move_to_menu_pos(key_info.count - 1, menu->state);
 	wrefresh(menu_win);
 }
 
@@ -557,16 +604,15 @@ cmd_H(key_info_t key_info, keys_info_t *keys_info)
 {
 	int top;
 	int off = MAX(cfg.scroll_off, 0);
-	if(off > menu->win_rows/2)
+	if(off > getmaxy(menu_win)/2)
+	{
 		return;
+	}
 
-	if(menu->top == 0)
-		top = 0;
-	else
-		top = menu->top + off;
+	top = (menu->top == 0) ? 0 : (menu->top + off);
 
-	clean_menu_position(menu);
-	move_to_menu_pos(top, menu);
+	menu_current_line_erase(menu->state);
+	move_to_menu_pos(top, menu->state);
 	wrefresh(menu_win);
 }
 
@@ -584,16 +630,19 @@ cmd_L(key_info_t key_info, keys_info_t *keys_info)
 	}
 
 	off = MAX(cfg.scroll_off, 0);
-	if(off > menu->win_rows/2)
+	if(off > getmaxy(menu_win)/2)
+	{
 		return;
+	}
 
-	if(menu->top + menu->win_rows < menu->len - 1)
-		top = menu->top + menu->win_rows - off;
-	else
-		top = menu->top + menu->win_rows;
+	top = menu->top + getmaxy(menu_win);
+	if(menu->top + getmaxy(menu_win) < menu->len - 1)
+	{
+		top -= off;
+	}
 
-	clean_menu_position(menu);
-	move_to_menu_pos(top - 3, menu);
+	menu_current_line_erase(menu->state);
+	move_to_menu_pos(top - 3, menu->state);
 	wrefresh(menu_win);
 }
 
@@ -602,13 +651,17 @@ static void
 cmd_M(key_info_t key_info, keys_info_t *keys_info)
 {
 	int new_pos;
-	if(menu->len < menu->win_rows)
+	if(menu->len < getmaxy(menu_win))
+	{
 		new_pos = DIV_ROUND_UP(menu->len, 2);
+	}
 	else
-		new_pos = menu->top + DIV_ROUND_UP(menu->win_rows - 3, 2);
+	{
+		new_pos = menu->top + DIV_ROUND_UP(getmaxy(menu_win) - 3, 2);
+	}
 
-	clean_menu_position(menu);
-	move_to_menu_pos(MAX(0, new_pos - 1), menu);
+	menu_current_line_erase(menu->state);
+	move_to_menu_pos(MAX(0, new_pos - 1), menu->state);
 	wrefresh(menu_win);
 }
 
@@ -618,7 +671,7 @@ cmd_N(key_info_t key_info, keys_info_t *keys_info)
 	key_info.count = def_count(key_info.count);
 	while(key_info.count-- > 0)
 	{
-		menus_search(menu, !last_search_backward);
+		menus_search(menu->state, !last_search_backward);
 	}
 }
 
@@ -633,7 +686,7 @@ cmd_b(key_info_t key_info, keys_info_t *keys_info)
 static void
 dump_into_custom_view(int very)
 {
-	if(menu_to_custom_view(menu, view, very) != 0)
+	if(menu_to_custom_view(menu->state, view, very) != 0)
 	{
 		show_error_msg("Menu transformation",
 				"No valid paths discovered in menu content");
@@ -672,7 +725,7 @@ pass_combination_to_khandler(const wchar_t keys[])
 		return 0;
 	}
 
-	handler_response = menu->key_handler(menu, keys);
+	handler_response = menu->key_handler(view, menu, keys);
 
 	switch(handler_response)
 	{
@@ -697,40 +750,33 @@ pass_combination_to_khandler(const wchar_t keys[])
 static void
 cmd_gg(key_info_t key_info, keys_info_t *keys_info)
 {
-	if(key_info.count == NO_COUNT_GIVEN)
-		key_info.count = 1;
-
-	clean_menu_position(menu);
-	move_to_menu_pos(key_info.count - 1, menu);
+	menu_current_line_erase(menu->state);
+	move_to_menu_pos(def_count(key_info.count) - 1, menu->state);
 	wrefresh(menu_win);
 }
 
 static void
 cmd_j(key_info_t key_info, keys_info_t *keys_info)
 {
-	if(menu->pos == menu->len - 1)
-		return;
-	if(key_info.count == NO_COUNT_GIVEN)
-		key_info.count = 1;
-
-	clean_menu_position(menu);
-	menu->pos += key_info.count;
-	move_to_menu_pos(menu->pos, menu);
-	wrefresh(menu_win);
+	if(menu->pos != menu->len - 1)
+	{
+		menu_current_line_erase(menu->state);
+		menu->pos += def_count(key_info.count);
+		move_to_menu_pos(menu->pos, menu->state);
+		wrefresh(menu_win);
+	}
 }
 
 static void
 cmd_k(key_info_t key_info, keys_info_t *keys_info)
 {
-	if(menu->pos == 0)
-		return;
-	if(key_info.count == NO_COUNT_GIVEN)
-		key_info.count = 1;
-
-	clean_menu_position(menu);
-	menu->pos -= key_info.count;
-	move_to_menu_pos(menu->pos, menu);
-	wrefresh(menu_win);
+	if(menu->pos != 0)
+	{
+		menu_current_line_erase(menu->state);
+		menu->pos -= def_count(key_info.count);
+		move_to_menu_pos(menu->pos, menu->state);
+		wrefresh(menu_win);
+	}
 }
 
 static void
@@ -739,7 +785,7 @@ cmd_n(key_info_t key_info, keys_info_t *keys_info)
 	key_info.count = def_count(key_info.count);
 	while(key_info.count-- > 0)
 	{
-		menus_search(menu, last_search_backward);
+		menus_search(menu->state, last_search_backward);
 	}
 }
 
@@ -810,10 +856,14 @@ cmd_zb(key_info_t key_info, keys_info_t *keys_info)
 {
 	if(can_scroll_menu_up(menu))
 	{
-		if(menu->pos < menu->win_rows)
+		if(menu->pos < getmaxy(menu_win))
+		{
 			menu->top = 0;
+		}
 		else
-			menu->top = menu->pos - (menu->win_rows - 3);
+		{
+			menu->top = menu->pos - (getmaxy(menu_win) - 3);
+		}
 		update_menu();
 	}
 }
@@ -821,10 +871,11 @@ cmd_zb(key_info_t key_info, keys_info_t *keys_info)
 static void
 cmd_zH(key_info_t key_info, keys_info_t *keys_info)
 {
-	if(menu->hor_pos == 0)
-		return;
-	menu->hor_pos = MAX(0, menu->hor_pos - (getmaxx(menu_win) - 4));
-	update_menu();
+	if(menu->hor_pos != 0)
+	{
+		menu->hor_pos = MAX(0, menu->hor_pos - (getmaxx(menu_win) - 4));
+		update_menu();
+	}
 }
 
 static void
@@ -837,20 +888,17 @@ cmd_zL(key_info_t key_info, keys_info_t *keys_info)
 static void
 cmd_zh(key_info_t key_info, keys_info_t *keys_info)
 {
-	if(menu->hor_pos == 0)
-		return;
-	if(key_info.count == NO_COUNT_GIVEN)
-		key_info.count = 1;
-	menu->hor_pos = MAX(0, menu->hor_pos - key_info.count);
-	update_menu();
+	if(menu->hor_pos != 0)
+	{
+		menu->hor_pos = MAX(0, menu->hor_pos - def_count(key_info.count));
+		update_menu();
+	}
 }
 
 static void
 cmd_zl(key_info_t key_info, keys_info_t *keys_info)
 {
-	if(key_info.count == NO_COUNT_GIVEN)
-		key_info.count = 1;
-	menu->hor_pos += key_info.count;
+	menu->hor_pos += def_count(key_info.count);
 	update_menu();
 }
 
@@ -859,10 +907,10 @@ cmd_zt(key_info_t key_info, keys_info_t *keys_info)
 {
 	if(can_scroll_menu_down(menu))
 	{
-		if(menu->len - menu->pos >= menu->win_rows - 3 + 1)
+		if(menu->len - menu->pos >= getmaxy(menu_win) - 3 + 1)
 			menu->top = menu->pos;
 		else
-			menu->top = menu->len - (menu->win_rows - 3 + 1);
+			menu->top = menu->len - (getmaxy(menu_win) - 3 + 1);
 		update_menu();
 	}
 }
@@ -872,12 +920,12 @@ cmd_zz(key_info_t key_info, keys_info_t *keys_info)
 {
 	if(!all_lines_visible(menu))
 	{
-		if(menu->pos <= (menu->win_rows - 3)/2)
+		if(menu->pos <= (getmaxy(menu_win) - 3)/2)
 			menu->top = 0;
-		else if(menu->pos > menu->len - DIV_ROUND_UP(menu->win_rows - 3, 2))
-			menu->top = menu->len - (menu->win_rows - 3 + 1);
+		else if(menu->pos > menu->len - DIV_ROUND_UP(getmaxy(menu_win) - 3, 2))
+			menu->top = menu->len - (getmaxy(menu_win) - 3 + 1);
 		else
-			menu->top = menu->pos - DIV_ROUND_UP(menu->win_rows - 3, 2);
+			menu->top = menu->pos - DIV_ROUND_UP(getmaxy(menu_win) - 3, 2);
 
 		update_menu();
 	}
@@ -885,27 +933,28 @@ cmd_zz(key_info_t key_info, keys_info_t *keys_info)
 
 /* Returns non-zero if all menu lines are visible, so no scrolling is needed. */
 static int
-all_lines_visible(const menu_info *const menu)
+all_lines_visible(const menu_data_t *const menu)
 {
-	return menu->len <= menu->win_rows - 2;
+	return menu->len <= getmaxy(menu_win) - 2;
 }
 
 void
 update_menu(void)
 {
-	draw_menu(menu);
-	move_to_menu_pos(menu->pos, menu);
+	draw_menu(menu->state);
+	move_to_menu_pos(menu->pos, menu->state);
 	wrefresh(menu_win);
 }
 
 static int
 goto_cmd(const cmd_info_t *cmd_info)
 {
-	if(cmd_info->end == NOT_DEF)
-		return 0;
-	clean_menu_position(menu);
-	move_to_menu_pos(cmd_info->end, menu);
-	wrefresh(menu_win);
+	if(cmd_info->end != NOT_DEF)
+	{
+		menu_current_line_erase(menu->state);
+		move_to_menu_pos(cmd_info->end, menu->state);
+		wrefresh(menu_win);
+	}
 	return 0;
 }
 
@@ -913,7 +962,7 @@ goto_cmd(const cmd_info_t *cmd_info)
 static int
 nohlsearch_cmd(const cmd_info_t *cmd_info)
 {
-	menus_reset_search_highlight(menu);
+	menus_reset_search_highlight(menu->state);
 	return 0;
 }
 
@@ -988,11 +1037,11 @@ leave_menu_mode(int reset_selection)
 		return;
 	}
 
-	reset_popup_menu(menu);
+	reset_menu_data(menu);
 
 	if(reset_selection)
 	{
-		clean_selected_files(view);
+		flist_sel_stash(view);
 		redraw_view(view);
 	}
 
