@@ -130,7 +130,7 @@ modes_pre(void)
 	}
 	else if(!curr_stats.save_msg)
 	{
-		clean_status_bar();
+		ui_sb_clear();
 		wrefresh(status_bar);
 	}
 }
@@ -174,7 +174,7 @@ modes_post(void)
 		if(!is_status_bar_multiline())
 		{
 			update_stat_window(curr_view, 0);
-			ui_ruler_update(curr_view);
+			ui_ruler_update(curr_view, 1);
 		}
 	}
 
@@ -201,7 +201,7 @@ modes_statusbar_update(void)
 	}
 	else
 	{
-		clean_status_bar();
+		ui_sb_clear();
 	}
 }
 
