@@ -30,12 +30,13 @@
 #include <stddef.h> /* size_t wchar_t */
 #include <stdio.h>  /* snprintf() */
 #include <stdlib.h> /* free() */
-#include <string.h> /* strncmp() */
+#include <string.h> /* strncmp() strspn() */
 #include <wctype.h> /* towupper() */
 #include <wchar.h> /* wcscpy() */
 
 #include "../cfg/config.h"
 #include "../cfg/hist.h"
+#include "../compat/curses.h"
 #include "../compat/fs_limits.h"
 #include "../compat/reallocarray.h"
 #include "../engine/keys.h"
@@ -51,10 +52,17 @@
 #include "../utils/str.h"
 #include "../utils/utf8.h"
 #include "../utils/utils.h"
+#include "../compare.h"
 #include "../cmd_core.h"
 #include "../filelist.h"
-#include "../fileops.h"
+#include "../flist_hist.h"
 #include "../filtering.h"
+#include "../flist_pos.h"
+#include "../flist_sel.h"
+#include "../fops_common.h"
+#include "../fops_misc.h"
+#include "../fops_put.h"
+#include "../fops_rename.h"
 #include "../marks.h"
 #include "../registers.h"
 #include "../running.h"
@@ -157,6 +165,8 @@ static void cmd_D_selector(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_d_selector(key_info_t key_info, keys_info_t *keys_info);
 static void call_delete(key_info_t key_info, keys_info_t *keys_info,
 		int use_trash);
+static void cmd_do(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_dp(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_e(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_f(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_gA(key_info_t key_info, keys_info_t *keys_info);
@@ -211,6 +221,16 @@ static void cmd_zo(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_zr(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_left_paren(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_right_paren(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_z_k(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_z_j(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_lb_c(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_rb_c(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_lb_d(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_rb_d(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_lb_s(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_rb_s(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_lb_z(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_rb_z(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_left_curly_bracket(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_right_curly_bracket(key_info_t key_info,
 		keys_info_t *keys_info);
@@ -316,6 +336,8 @@ static keys_add_info_t builtin_cmds[] = {
 	{WK_c WK_w,        {{&cmd_cw}, .descr = "rename files"}},
 	{WK_D WK_D,        {{&cmd_DD}, .nim = 1, .descr = "remove files permanently"}},
 	{WK_d WK_d,        {{&cmd_dd}, .nim = 1, .descr = "remove files"}},
+	{WK_d WK_o,        {{&cmd_do}, .descr = "obtain current file"}},
+	{WK_d WK_p,        {{&cmd_dp}, .descr = "put current file"}},
 	{WK_D,             {{&cmd_D_selector}, FOLLOWED_BY_SELECTOR, .descr = "remove files permanently"}},
 	{WK_d,             {{&cmd_d_selector}, FOLLOWED_BY_SELECTOR, .descr = "remove files"}},
 	{WK_e,             {{&cmd_e}, .descr = "explore file contents"}},
@@ -370,8 +392,18 @@ static keys_add_info_t builtin_cmds[] = {
 	{WK_z WK_z,        {{&normal_cmd_zz},   .descr = "center cursor position"}},
 	{WK_LP,            {{&cmd_left_paren},  .descr = "go to previous group of files"}},
 	{WK_RP,            {{&cmd_right_paren}, .descr = "go to next group of files"}},
-	{WK_LCB,           {{&cmd_left_curly_bracket},  .descr = "go to previous file/dir"}},
-	{WK_RCB,           {{&cmd_right_curly_bracket}, .descr = "go to next file/dir"}},
+	{WK_z WK_k,        {{&cmd_z_k},  .descr = "go to previous sibling dir"}},
+	{WK_z WK_j,        {{&cmd_z_j},  .descr = "go to next sibling dir"}},
+	{WK_LB WK_c,       {{&cmd_lb_c}, .descr = "go to previous mismatch"}},
+	{WK_RB WK_c,       {{&cmd_rb_c}, .descr = "go to next mismatch"}},
+	{WK_LB WK_d,       {{&cmd_lb_d}, .descr = "go to previous dir"}},
+	{WK_RB WK_d,       {{&cmd_rb_d}, .descr = "go to next dir"}},
+	{WK_LB WK_s,       {{&cmd_lb_s}, .descr = "go to previous selected entry"}},
+	{WK_RB WK_s,       {{&cmd_rb_s}, .descr = "go to next selected entry"}},
+	{WK_LB WK_z,       {{&cmd_lb_z}, .descr = "go to first sibling"}},
+	{WK_RB WK_z,       {{&cmd_rb_z}, .descr = "go to last sibling"}},
+	{WK_LCB,           {{&cmd_left_curly_bracket},  .descr = "go to previous file/dir group"}},
+	{WK_RCB,           {{&cmd_right_curly_bracket}, .descr = "go to next file/dir group"}},
 #ifndef _WIN32
 	{WK_c WK_g,        {{&cmd_cg}, .descr = "change group"}},
 	{WK_c WK_o,        {{&cmd_co}, .descr = "change owner"}},
@@ -379,16 +411,16 @@ static keys_add_info_t builtin_cmds[] = {
 	{WK_g WK_r,        {{&cmd_gr}, .descr = "open file with rights elevation"}},
 #endif
 #ifdef ENABLE_EXTENDED_KEYS
-	{{WC_C_w, KEY_BACKSPACE}, {{&cmd_ctrl_wh}, .descr = "go to left window"}},
-	{{KEY_PPAGE},             {{&cmd_ctrl_b},  .descr = "scroll page up"}},
-	{{KEY_NPAGE},             {{&cmd_ctrl_f},  .descr = "scroll page down"}},
-	{{KEY_LEFT},              {{&cmd_h},       .descr = "go to parent directory/item to the left"}},
-	{{KEY_DOWN},              {{&cmd_j},       .descr = "go to item below"}},
-	{{KEY_UP},                {{&cmd_k},       .descr = "go to item above"}},
-	{{KEY_RIGHT},             {{&cmd_l},       .descr = "open file/go to item to the right"}},
-	{{KEY_HOME},              {{&cmd_gg},      .descr = "go to the first item"}},
-	{{KEY_END},               {{&cmd_G},       .descr = "go to the last item"}},
-	{{KEY_BTAB},              {{&cmd_shift_tab}, .descr = "switch to view pane"}},
+	{{WC_C_w, K(KEY_BACKSPACE)}, {{&cmd_ctrl_wh}, .descr = "go to left window"}},
+	{{K(KEY_PPAGE)},             {{&cmd_ctrl_b},  .descr = "scroll page up"}},
+	{{K(KEY_NPAGE)},             {{&cmd_ctrl_f},  .descr = "scroll page down"}},
+	{{K(KEY_LEFT)},              {{&cmd_h},       .descr = "go to parent directory/item to the left"}},
+	{{K(KEY_DOWN)},              {{&cmd_j},       .descr = "go to item below"}},
+	{{K(KEY_UP)},                {{&cmd_k},       .descr = "go to item above"}},
+	{{K(KEY_RIGHT)},             {{&cmd_l},       .descr = "open file/go to item to the right"}},
+	{{K(KEY_HOME)},              {{&cmd_gg},      .descr = "go to the first item"}},
+	{{K(KEY_END)},               {{&cmd_G},       .descr = "go to the last item"}},
+	{{K(KEY_BTAB)},              {{&cmd_shift_tab}, .descr = "switch to view pane"}},
 #else
 	{WK_ESC L"[Z",            {{&cmd_shift_tab}, .descr = "switch to view pane"}},
 #endif /* ENABLE_EXTENDED_KEYS */
@@ -420,13 +452,23 @@ static keys_add_info_t selectors[] = {
 	{WK_s,       {{&selector_s},      .descr = "selected files"}},
 	{WK_LP,      {{&cmd_left_paren},  .descr = "to previous group of files"}},
 	{WK_RP,      {{&cmd_right_paren}, .descr = "to next group of files"}},
-	{WK_LCB,     {{&cmd_left_curly_bracket},  .descr = "to previous file/dir"}},
-	{WK_RCB,     {{&cmd_right_curly_bracket}, .descr = "to next file/dir"}},
+	{WK_z WK_k,  {{&cmd_z_k},  .descr = "to previous sibling dir"}},
+	{WK_z WK_j,  {{&cmd_z_j},  .descr = "to next sibling dir"}},
+	{WK_LB WK_c, {{&cmd_lb_c}, .descr = "to previous mismatch"}},
+	{WK_RB WK_c, {{&cmd_rb_c}, .descr = "to next mismatch"}},
+	{WK_LB WK_d, {{&cmd_lb_d}, .descr = "to previous dir"}},
+	{WK_RB WK_d, {{&cmd_rb_d}, .descr = "to next dir"}},
+	{WK_LB WK_s, {{&cmd_lb_s}, .descr = "to previous selected entry"}},
+	{WK_RB WK_s, {{&cmd_rb_s}, .descr = "to next selected entry"}},
+	{WK_LB WK_z, {{&cmd_lb_z}, .descr = "to first sibling"}},
+	{WK_RB WK_z, {{&cmd_rb_z}, .descr = "to last sibling"}},
+	{WK_LCB,     {{&cmd_left_curly_bracket},  .descr = "to previous file/dir group"}},
+	{WK_RCB,     {{&cmd_right_curly_bracket}, .descr = "to next file/dir group"}},
 #ifdef ENABLE_EXTENDED_KEYS
-	{{KEY_DOWN}, {{&cmd_j},  .descr = "to item below"}},
-	{{KEY_UP},   {{&cmd_k},  .descr = "to item above"}},
-	{{KEY_HOME}, {{&cmd_gg}, .descr = "to the first item"}},
-	{{KEY_END},  {{&cmd_G},  .descr = "go to the last item"}},
+	{{K(KEY_DOWN)}, {{&cmd_j},  .descr = "to item below"}},
+	{{K(KEY_UP)},   {{&cmd_k},  .descr = "to item above"}},
+	{{K(KEY_HOME)}, {{&cmd_gg}, .descr = "to the first item"}},
+	{{K(KEY_END)},  {{&cmd_G},  .descr = "to the last item"}},
 #endif /* ENABLE_EXTENDED_KEYS */
 };
 
@@ -451,7 +493,7 @@ static void
 cmd_ctrl_a(key_info_t key_info, keys_info_t *keys_info)
 {
 	check_marking(curr_view, 0, NULL);
-	curr_stats.save_msg = incdec_names(curr_view, def_count(key_info.count));
+	curr_stats.save_msg = fops_incdec(curr_view, def_count(key_info.count));
 }
 
 static void
@@ -468,7 +510,7 @@ static void
 cmd_ctrl_c(key_info_t key_info, keys_info_t *keys_info)
 {
 	ui_view_reset_search_highlight(curr_view);
-	clean_selected_files(curr_view);
+	flist_sel_stash(curr_view);
 	redraw_current_view();
 }
 
@@ -587,7 +629,7 @@ cmd_ctrl_i(key_info_t key_info, keys_info_t *keys_info)
 	}
 	else
 	{
-		navigate_forward_in_history(curr_view);
+		flist_hist_go_forward(curr_view);
 	}
 }
 
@@ -603,14 +645,14 @@ static void
 cmd_return(key_info_t key_info, keys_info_t *keys_info)
 {
 	open_file(curr_view, FHE_RUN);
-	clean_selected_files(curr_view);
+	flist_sel_stash(curr_view);
 	redraw_current_view();
 }
 
 static void
 cmd_ctrl_o(key_info_t key_info, keys_info_t *keys_info)
 {
-	navigate_backward_in_history(curr_view);
+	flist_hist_go_back(curr_view);
 }
 
 static void
@@ -860,7 +902,7 @@ cmd_ctrl_wx(key_info_t key_info, keys_info_t *keys_info)
 static void
 cmd_ctrl_wz(key_info_t key_info, keys_info_t *keys_info)
 {
-	preview_close();
+	qv_hide();
 }
 
 /* Decrements first number in names of marked files of the view [count=1]
@@ -869,7 +911,7 @@ static void
 cmd_ctrl_x(key_info_t key_info, keys_info_t *keys_info)
 {
 	check_marking(curr_view, 0, NULL);
-	curr_stats.save_msg = incdec_names(curr_view, -def_count(key_info.count));
+	curr_stats.save_msg = fops_incdec(curr_view, -def_count(key_info.count));
 }
 
 static void
@@ -923,92 +965,34 @@ static void
 cmd_C(key_info_t key_info, keys_info_t *keys_info)
 {
 	check_marking(curr_view, 0, NULL);
-	curr_stats.save_msg = clone_files(curr_view, NULL, 0, 0,
+	curr_stats.save_msg = fops_clone(curr_view, NULL, 0, 0,
 			def_count(key_info.count));
 }
 
+/* Navigates to previous word which starts with specified character. */
 static void
 cmd_F(key_info_t key_info, keys_info_t *keys_info)
 {
 	last_fast_search_char = key_info.multi;
 	last_fast_search_backward = 1;
 
-	if(key_info.count == NO_COUNT_GIVEN)
-		key_info.count = 1;
-	find_goto(key_info.multi, key_info.count, 1, keys_info);
-}
-
-/* Returns negative number if nothing was found */
-int
-ffind(int ch, int backward, int wrap)
-{
-	int x;
-	int upcase = cfg.ignore_case && !(cfg.smart_case && iswupper(ch));
-
-	if(upcase)
-		ch = towupper(ch);
-
-	x = curr_view->list_pos;
-	do
-	{
-		if(backward)
-		{
-			x--;
-			if(x < 0)
-			{
-				if(wrap)
-					x = curr_view->list_rows - 1;
-				else
-					return -1;
-			}
-		}
-		else
-		{
-			x++;
-			if(x > curr_view->list_rows - 1)
-			{
-				if(wrap)
-					x = 0;
-				else
-					return -1;
-			}
-		}
-
-		if(ch > 255)
-		{
-			wchar_t wc = get_first_wchar(curr_view->dir_entry[x].name);
-			if(upcase)
-				wc = towupper(wc);
-			if(wc == ch)
-				break;
-		}
-		else
-		{
-			int c = curr_view->dir_entry[x].name[0];
-			if(upcase)
-				c = towupper(c);
-			if(c == ch)
-				break;
-		}
-	}
-	while(x != curr_view->list_pos);
-
-	return x;
+	find_goto(key_info.multi, def_count(key_info.count), 1, keys_info);
 }
 
+/* Navigates to next/previous file which starts with given character. */
 static void
 find_goto(int ch, int count, int backward, keys_info_t *keys_info)
 {
 	int pos;
 	int old_pos = curr_view->list_pos;
 
-	pos = ffind(ch, backward, !keys_info->selector);
+	pos = flist_find_by_ch(curr_view, ch, backward, !keys_info->selector);
 	if(pos < 0 || pos == curr_view->list_pos)
 		return;
 	while(--count > 0)
 	{
 		curr_view->list_pos = pos;
-		pos = ffind(ch, backward, !keys_info->selector);
+		pos = flist_find_by_ch(curr_view, ch, backward, !keys_info->selector);
 	}
 
 	if(keys_info->selector)
@@ -1039,20 +1023,20 @@ cmd_G(key_info_t key_info, keys_info_t *keys_info)
 static void
 cmd_gA(key_info_t key_info, keys_info_t *keys_info)
 {
-	calculate_size_bg(curr_view, 1);
+	fops_size_bg(curr_view, 1);
 }
 
 /* Calculate size of selected directories taking cached sizes into account. */
 static void
 cmd_ga(key_info_t key_info, keys_info_t *keys_info)
 {
-	calculate_size_bg(curr_view, 0);
+	fops_size_bg(curr_view, 0);
 }
 
 static void
 cmd_gf(key_info_t key_info, keys_info_t *keys_info)
 {
-	clean_selected_files(curr_view);
+	flist_sel_stash(curr_view);
 	follow_file(curr_view);
 	redraw_current_view();
 }
@@ -1079,7 +1063,7 @@ static void
 cmd_gr(key_info_t key_info, keys_info_t *keys_info)
 {
 	open_file(curr_view, FHE_ELEVATE_AND_RUN);
-	clean_selected_files(curr_view);
+	flist_sel_stash(curr_view);
 	redraw_current_view();
 }
 #endif
@@ -1154,7 +1138,7 @@ do_gu(key_info_t key_info, keys_info_t *keys_info, int upper)
 	}
 
 	check_marking(curr_view, keys_info->count, keys_info->indexes);
-	curr_stats.save_msg = change_case(curr_view, upper);
+	curr_stats.save_msg = fops_case(curr_view, upper);
 	free_list_of_file_indexes(keys_info);
 }
 
@@ -1307,16 +1291,16 @@ cmd_equal(key_info_t key_info, keys_info_t *keys_info)
 	enter_cmdline_mode(CLS_FILTER, curr_view->local_filter.filter.raw, NULL);
 }
 
+/* Continues navigation to word which starts with specified character in
+ * opposite direction. */
 static void
 cmd_comma(key_info_t key_info, keys_info_t *keys_info)
 {
-	if(last_fast_search_backward == -1)
-		return;
-
-	if(key_info.count == NO_COUNT_GIVEN)
-		key_info.count = 1;
-	find_goto(last_fast_search_char, key_info.count, !last_fast_search_backward,
-			keys_info);
+	if(last_fast_search_backward != -1)
+	{
+		find_goto(last_fast_search_char, def_count(key_info.count),
+				!last_fast_search_backward, keys_info);
+	}
 }
 
 /* Repeat last change. */
@@ -1350,16 +1334,16 @@ cmd_colon(key_info_t key_info, keys_info_t *keys_info)
 	enter_cmdline_mode(CLS_COMMAND, prefix, NULL);
 }
 
+/* Continues navigation to word which starts with specified character in initial
+ * direction. */
 static void
 cmd_semicolon(key_info_t key_info, keys_info_t *keys_info)
 {
-	if(last_fast_search_backward == -1)
-		return;
-
-	if(key_info.count == NO_COUNT_GIVEN)
-		key_info.count = 1;
-	find_goto(last_fast_search_char, key_info.count, last_fast_search_backward,
-			keys_info);
+	if(last_fast_search_backward != -1)
+	{
+		find_goto(last_fast_search_char, def_count(key_info.count),
+				last_fast_search_backward, keys_info);
+	}
 }
 
 /* Search forward. */
@@ -1394,7 +1378,7 @@ cmd_av(key_info_t key_info, keys_info_t *keys_info)
 static void
 cmd_cW(key_info_t key_info, keys_info_t *keys_info)
 {
-	rename_current_file(curr_view, 1);
+	fops_rename_current(curr_view, 1);
 }
 
 #ifndef _WIN32
@@ -1402,7 +1386,7 @@ cmd_cW(key_info_t key_info, keys_info_t *keys_info)
 static void
 cmd_cg(key_info_t key_info, keys_info_t *keys_info)
 {
-	change_group();
+	fops_chgroup();
 }
 #endif
 
@@ -1410,7 +1394,7 @@ cmd_cg(key_info_t key_info, keys_info_t *keys_info)
 static void
 cmd_cl(key_info_t key_info, keys_info_t *keys_info)
 {
-	curr_stats.save_msg = change_link(curr_view);
+	curr_stats.save_msg = fops_retarget(curr_view);
 }
 
 #ifndef _WIN32
@@ -1418,7 +1402,7 @@ cmd_cl(key_info_t key_info, keys_info_t *keys_info)
 static void
 cmd_co(key_info_t key_info, keys_info_t *keys_info)
 {
-	change_owner();
+	fops_chuser();
 }
 #endif
 
@@ -1426,7 +1410,38 @@ cmd_co(key_info_t key_info, keys_info_t *keys_info)
 static void
 cmd_cp(key_info_t key_info, keys_info_t *keys_info)
 {
-	enter_attr_mode(curr_view);
+	normal_cmd_cp(curr_view, key_info);
+}
+
+void
+normal_cmd_cp(FileView *view, key_info_t key_info)
+{
+#ifndef _WIN32
+	char mode[32];
+	int len;
+
+	if(key_info.count == NO_COUNT_GIVEN)
+	{
+		enter_attr_mode(view);
+		return;
+	}
+
+	len = snprintf(mode, sizeof(mode), "%04d", key_info.count);
+	if(len != 4 || ((int)strspn(mode, "01234567") != len))
+	{
+		/* Update view because there might be not reflected changes of cursor
+		 * position (when called from visual mode). */
+		redraw_view(view);
+
+		status_bar_errorf("Incorrect file mode: %s", mode);
+		curr_stats.save_msg = 1;
+		return;
+	}
+
+	files_chmod(view, mode, 0);
+#else
+	enter_attr_mode(view);
+#endif
 }
 
 /* Change word (rename file or files). */
@@ -1436,11 +1451,11 @@ cmd_cw(key_info_t key_info, keys_info_t *keys_info)
 	if(curr_view->selected_files > 1)
 	{
 		check_marking(curr_view, 0, NULL);
-		rename_files(curr_view, NULL, 0, 0);
+		fops_rename(curr_view, NULL, 0, 0);
 		return;
 	}
 
-	rename_current_file(curr_view, 0);
+	fops_rename_current(curr_view, 0);
 }
 
 /* Delete file. */
@@ -1464,7 +1479,7 @@ delete(key_info_t key_info, int use_trash)
 {
 	keys_info_t keys_info = {};
 
-	if(!can_change_view_files(curr_view))
+	if(!fops_view_can_be_changed(curr_view))
 	{
 		return;
 	}
@@ -1487,7 +1502,7 @@ delete(key_info_t key_info, int use_trash)
 static void
 cmd_D_selector(key_info_t key_info, keys_info_t *keys_info)
 {
-	if(can_change_view_files(curr_view))
+	if(fops_view_can_be_changed(curr_view))
 	{
 		call_delete(key_info, keys_info, 0);
 	}
@@ -1510,12 +1525,26 @@ call_delete(key_info_t key_info, keys_info_t *keys_info, int use_trash)
 
 	if(count != 0 && confirm_deletion(count, use_trash))
 	{
-		curr_stats.save_msg = delete_files(curr_view, def_reg(key_info.reg),
+		curr_stats.save_msg = fops_delete(curr_view, def_reg(key_info.reg),
 				use_trash);
 		free_list_of_file_indexes(keys_info);
 	}
 }
 
+/* Applies change from other compare to current one. */
+static void
+cmd_do(key_info_t key_info, keys_info_t *keys_info)
+{
+	curr_stats.save_msg = (compare_move(other_view, curr_view) != 0);
+}
+
+/* Applies change from current compare to other one. */
+static void
+cmd_dp(key_info_t key_info, keys_info_t *keys_info)
+{
+	curr_stats.save_msg = (compare_move(curr_view, other_view) != 0);
+}
+
 static void
 cmd_e(key_info_t key_info, keys_info_t *keys_info)
 {
@@ -1528,24 +1557,36 @@ cmd_e(key_info_t key_info, keys_info_t *keys_info)
 	enter_view_mode(curr_view, 1);
 }
 
+/* Navigates to next word which starts with specified character. */
 static void
 cmd_f(key_info_t key_info, keys_info_t *keys_info)
 {
 	last_fast_search_char = key_info.multi;
 	last_fast_search_backward = 0;
 
-	if(key_info.count == NO_COUNT_GIVEN)
-		key_info.count = 1;
-	find_goto(key_info.multi, key_info.count, 0, keys_info);
+	find_goto(key_info.multi, def_count(key_info.count), 0, keys_info);
 }
 
 /* Updir or one file left in less-like sub-mode. */
 static void
 cmd_h(key_info_t key_info, keys_info_t *keys_info)
 {
-	if(curr_view->ls_view)
+	const dir_entry_t *entry = get_current_entry(curr_view);
+
+	if(!ui_view_displays_columns(curr_view))
 	{
 		go_to_prev(key_info, keys_info, 1);
+		return;
+	}
+
+	if(entry->child_pos != 0)
+	{
+		key_info.count = def_count(key_info.count);
+		while(key_info.count-- > 0)
+		{
+			entry -= entry->child_pos;
+		}
+		pick_or_move(keys_info, entry_to_pos(curr_view, entry));
 	}
 	else
 	{
@@ -1557,7 +1598,7 @@ static void
 cmd_i(key_info_t key_info, keys_info_t *keys_info)
 {
 	open_file(curr_view, FHE_NO_RUN);
-	clean_selected_files(curr_view);
+	flist_sel_stash(curr_view);
 	redraw_current_view();
 }
 
@@ -1592,13 +1633,13 @@ go_to_prev(key_info_t key_info, keys_info_t *keys_info, int step)
 static void
 cmd_l(key_info_t key_info, keys_info_t *keys_info)
 {
-	if(curr_view->ls_view)
+	if(ui_view_displays_columns(curr_view))
 	{
-		go_to_next(key_info, keys_info, 1);
+		cmd_return(key_info, keys_info);
 	}
 	else
 	{
-		cmd_return(key_info, keys_info);
+		go_to_next(key_info, keys_info, 1);
 	}
 }
 
@@ -1614,9 +1655,12 @@ go_to_next(key_info_t key_info, keys_info_t *keys_info, int step)
 static void
 cmd_m(key_info_t key_info, keys_info_t *keys_info)
 {
-	const dir_entry_t *const entry = &curr_view->dir_entry[curr_view->list_pos];
-	curr_stats.save_msg = set_user_mark(key_info.multi, entry->origin,
-			entry->name);
+	const dir_entry_t *const curr = get_current_entry(curr_view);
+	if(!fentry_is_fake(curr))
+	{
+		curr_stats.save_msg = set_user_mark(key_info.multi, curr->origin,
+				curr->name);
+	}
 }
 
 static void
@@ -1646,7 +1690,7 @@ search(key_info_t key_info, int backward)
 		const char *const pattern = cfg_get_last_search_pattern();
 		curr_stats.save_msg = (find_pattern(curr_view, pattern, backward, 1, &found,
 				0) != 0);
-		key_info.count--;
+		--key_info.count;
 	}
 
 	while(key_info.count-- > 0)
@@ -1677,7 +1721,7 @@ cmd_p(key_info_t key_info, keys_info_t *keys_info)
 static void
 call_put_files(key_info_t key_info, int move)
 {
-	curr_stats.save_msg = put_files(curr_view, def_reg(key_info.reg), move);
+	curr_stats.save_msg = fops_put(curr_view, -1, def_reg(key_info.reg), move);
 	ui_views_reload_filelists();
 }
 
@@ -1692,7 +1736,8 @@ cmd_rl(key_info_t key_info, keys_info_t *keys_info)
 static void
 call_put_links(key_info_t key_info, int relative)
 {
-	curr_stats.save_msg = put_links(curr_view, def_reg(key_info.reg), relative);
+	curr_stats.save_msg =
+		fops_put_links(curr_view, def_reg(key_info.reg), relative);
 	ui_views_reload_filelists();
 }
 
@@ -1744,23 +1789,26 @@ cmd_q_equals(key_info_t key_info, keys_info_t *keys_info)
 	get_and_execute_command("", 0U, CIT_FILTER_PATTERN);
 }
 
-/* Tag file. */
+/* Toggles selection of the current file. */
 static void
 cmd_t(key_info_t key_info, keys_info_t *keys_info)
 {
-	if(curr_view->dir_entry[curr_view->list_pos].selected == 0)
+	dir_entry_t *const curr = get_current_entry(curr_view);
+	/* Special purpose kind of entries can't be selected. */
+	if(!fentry_is_valid(curr))
 	{
-		/* The ../ dir cannot be selected */
-		if(is_parent_dir(curr_view->dir_entry[curr_view->list_pos].name))
-			return;
+		return;
+	}
 
-		curr_view->dir_entry[curr_view->list_pos].selected = 1;
-		curr_view->selected_files++;
+	if(curr->selected == 0)
+	{
+		curr->selected = 1;
+		++curr_view->selected_files;
 	}
 	else
 	{
-		curr_view->dir_entry[curr_view->list_pos].selected = 0;
-		curr_view->selected_files--;
+		curr->selected = 0;
+		--curr_view->selected_files;
 	}
 
 	fview_cursor_redraw(curr_view);
@@ -1871,10 +1919,10 @@ cmd_y_selector(key_info_t key_info, keys_info_t *keys_info)
 static void
 yank(key_info_t key_info, keys_info_t *keys_info)
 {
-	curr_stats.save_msg = yank_files(curr_view, def_reg(key_info.reg));
+	curr_stats.save_msg = fops_yank(curr_view, def_reg(key_info.reg));
 	free_list_of_file_indexes(keys_info);
 
-	clean_selected_files(curr_view);
+	flist_sel_stash(curr_view);
 	ui_view_schedule_redraw(curr_view);
 }
 
@@ -1924,7 +1972,7 @@ cmd_za(key_info_t key_info, keys_info_t *keys_info)
 static void
 cmd_zd(key_info_t key_info, keys_info_t *keys_info)
 {
-	flist_custom_exclude(curr_view);
+	flist_custom_exclude(curr_view, key_info.count == 1);
 }
 
 /* Redraw with file in bottom of list. */
@@ -1983,6 +2031,84 @@ cmd_right_paren(key_info_t key_info, keys_info_t *keys_info)
 	pick_or_move(keys_info, flist_find_group(curr_view, 1));
 }
 
+/* Go to or pick files until and including previous sibling directory entry or
+ * do nothing. */
+static void
+cmd_z_k(key_info_t key_info, keys_info_t *keys_info)
+{
+	pick_or_move(keys_info, flist_prev_dir_sibling(curr_view));
+}
+
+/* Go to or pick files until and including next sibling directory entry or do
+ * nothing. */
+static void
+cmd_z_j(key_info_t key_info, keys_info_t *keys_info)
+{
+	pick_or_move(keys_info, flist_next_dir_sibling(curr_view));
+}
+
+/* Go to or pick files until and including previous mismatched entry or do
+ * nothing. */
+static void
+cmd_lb_c(key_info_t key_info, keys_info_t *keys_info)
+{
+	pick_or_move(keys_info, flist_prev_mismatch(curr_view));
+}
+
+/* Go to or pick files until and including next mismatched entry or do
+ * nothing. */
+static void
+cmd_rb_c(key_info_t key_info, keys_info_t *keys_info)
+{
+	pick_or_move(keys_info, flist_next_mismatch(curr_view));
+}
+
+/* Go to or pick files until and including previous directory entry or do
+ * nothing. */
+static void
+cmd_lb_d(key_info_t key_info, keys_info_t *keys_info)
+{
+	pick_or_move(keys_info, flist_prev_dir(curr_view));
+}
+
+/* Go to or pick files until and including next directory entry or do
+ * nothing. */
+static void
+cmd_rb_d(key_info_t key_info, keys_info_t *keys_info)
+{
+	pick_or_move(keys_info, flist_next_dir(curr_view));
+}
+
+/* Go to or pick files until and including previous selected entry or do
+ * nothing. */
+static void
+cmd_lb_s(key_info_t key_info, keys_info_t *keys_info)
+{
+	pick_or_move(keys_info, flist_prev_selected(curr_view));
+}
+
+/* Go to or pick files until and including next selected entry or do nothing. */
+static void
+cmd_rb_s(key_info_t key_info, keys_info_t *keys_info)
+{
+	pick_or_move(keys_info, flist_next_selected(curr_view));
+}
+
+/* Go to or pick files until and including first sibling of the current
+ * entry. */
+static void
+cmd_lb_z(key_info_t key_info, keys_info_t *keys_info)
+{
+	pick_or_move(keys_info, flist_first_sibling(curr_view));
+}
+
+/* Go to or pick files until and including last sibling of the current entry. */
+static void
+cmd_rb_z(key_info_t key_info, keys_info_t *keys_info)
+{
+	pick_or_move(keys_info, flist_last_sibling(curr_view));
+}
+
 /* Moves cursor to the beginning of the previous group of files defined by them
  * being directory or files. */
 static void
@@ -2141,24 +2267,20 @@ find_npattern(FileView *view, const char pattern[], int backward,
 	const int nrepeats = search_repeat - 1;
 	int i;
 	int save_msg;
-	int msg;
+	int found;
 
 	/* Reset number of repeats so that future calls are not affected by the
 	 * previous ones. */
 	search_repeat = 1;
 
-	msg = find_pattern(view, pattern, backward, 1, &save_msg, print_errors);
-	if(!print_errors && msg < 0)
+	save_msg = find_pattern(view, pattern, backward, 1, &found, print_errors);
+	if(!print_errors && save_msg < 0)
 	{
 		/* If we're not printing messages, we might be interested in broken
 		 * pattern. */
 		return -1;
 	}
 
-	/* Take wrong regular expression message into account, otherwise we can't
-	 * distinguish "no files matched" situation from "wrong regexp". */
-	save_msg += msg;
-
 	for(i = 0; i < nrepeats; ++i)
 	{
 		save_msg += (goto_search_match(view, backward) != 0);
