@@ -26,6 +26,9 @@
 /* Initializes normal mode. */
 void init_normal_mode(void);
 
+/* Change file attributes (permissions or properties). */
+void normal_cmd_cp(FileView *view, key_info_t key_info);
+
 void normal_cmd_zb(key_info_t key_info, keys_info_t *keys_info);
 
 void normal_cmd_zt(key_info_t key_info, keys_info_t *keys_info);
@@ -45,8 +48,6 @@ void normal_cmd_ctrl_wminus(key_info_t key_info, keys_info_t *keys_info);
 
 void normal_cmd_ctrl_wpipe(key_info_t key_info, keys_info_t *keys_info);
 
-int ffind(int ch, int backward, int wrap);
-
 /* Kind of callback to allow starting searches from the module and rely on other
  * modules.  Returns new value for status bar message flag, but when
  * print_errors isn't requested can return -1 to indicate issues with the
