@@ -30,6 +30,7 @@
 #include <stdlib.h> /* free() */
 
 #include "../cfg/config.h"
+#include "../compat/curses.h"
 #include "../compat/fs_limits.h"
 #include "../compat/os.h"
 #include "../compat/reallocarray.h"
@@ -99,6 +100,8 @@ typedef struct
 	/* The rest of the state. */
 	FileView *view; /* File view association with the view. */
 	char *filename; /* Full path to the file being viewed. */
+	char *viewer;   /* When non-NULL, specifies custom preview command (no
+	                   implicit %c). */
 	int abandoned;  /* Whether view mode was abandoned. */
 	int graphics;   /* Whether viewer presumably displays graphics. */
 	int wrap;       /* Whether lines are wrapped. */
@@ -279,16 +282,16 @@ static keys_add_info_t builtin_cmds[] = {
 	{{ALT_V},          {{&cmd_b}, .descr = "scroll page up"}},
 #endif
 #ifdef ENABLE_EXTENDED_KEYS
-	{{WC_C_w, KEY_BACKSPACE}, {{&cmd_ctrl_wh}, .descr = "go to left window"}},
-	{{KEY_PPAGE},             {{&cmd_b}, .descr = "scroll page up"}},
-	{{KEY_NPAGE},             {{&cmd_f}, .descr = "scroll page down"}},
-	{{KEY_DOWN},              {{&cmd_j}, .descr = "scroll one line down"}},
-	{{KEY_UP},                {{&cmd_k}, .descr = "scroll one line up"}},
-	{{KEY_HOME},              {{&cmd_g}, .descr = "scroll to the beginning"}},
-	{{KEY_END},               {{&cmd_G}, .descr = "scroll to the end"}},
-	{{KEY_BTAB},              {{&cmd_q}, .descr = "leave view mode"}},
+	{{WC_C_w, K(KEY_BACKSPACE)}, {{&cmd_ctrl_wh}, .descr = "go to left window"}},
+	{{K(KEY_PPAGE)},             {{&cmd_b}, .descr = "scroll page up"}},
+	{{K(KEY_NPAGE)},             {{&cmd_f}, .descr = "scroll page down"}},
+	{{K(KEY_DOWN)},              {{&cmd_j}, .descr = "scroll one line down"}},
+	{{K(KEY_UP)},                {{&cmd_k}, .descr = "scroll one line up"}},
+	{{K(KEY_HOME)},              {{&cmd_g}, .descr = "scroll to the beginning"}},
+	{{K(KEY_END)},               {{&cmd_G}, .descr = "scroll to the end"}},
+	{{K(KEY_BTAB)},              {{&cmd_q}, .descr = "leave view mode"}},
 #else
-	{WK_ESCAPE L"[Z",         {{&cmd_q}, .descr = "leave view mode"}},
+	{WK_ESCAPE L"[Z",            {{&cmd_q}, .descr = "leave view mode"}},
 #endif /* ENABLE_EXTENDED_KEYS */
 };
 
@@ -328,13 +331,14 @@ enter_view_mode(FileView *view, int explore)
 	pick_vi(explore);
 
 	vi->view = view;
+	vi->filename = strdup(full_path);
+
 	if(load_view_data(vi, "File exploring", full_path, NOSILENT) != 0)
 	{
+		update_string(&vi->filename, NULL);
 		return;
 	}
 
-	vi->filename = strdup(full_path);
-
 	vle_mode_set(VIEW_MODE, VMT_SECONDARY);
 
 	if(explore)
@@ -348,7 +352,35 @@ enter_view_mode(FileView *view, int explore)
 	}
 
 	ui_views_update_titles();
+	view_redraw();
+}
 
+void
+make_abandoned_view(FileView *view, const char cmd[])
+{
+	char full_path[PATH_MAX];
+
+	if(get_file_to_explore(curr_view, full_path, sizeof(full_path)) != 0)
+	{
+		show_error_msg("File exploring", "The file cannot be explored");
+		return;
+	}
+
+	pick_vi(0);
+	reset_view_info(vi);
+
+	vi->view = view;
+	vi->viewer = strdup(cmd);
+	vi->filename = strdup(full_path);
+	vi->abandoned = 1;
+
+	if(load_view_data(vi, "File viewing", full_path, NOSILENT) != 0)
+	{
+		reset_view_info(vi);
+		return;
+	}
+
+	ui_views_update_titles();
 	view_redraw();
 }
 
@@ -447,6 +479,8 @@ try_redraw_explore_view(const FileView *const view, int vi_index)
 void
 leave_view_mode(void)
 {
+	reset_view_info(vi);
+
 	vle_mode_set(NORMAL_MODE, VMT_PRIMARY);
 
 	if(curr_view->explore_mode)
@@ -456,13 +490,11 @@ leave_view_mode(void)
 	}
 	else
 	{
-		quick_view_file(curr_view);
+		qv_draw(curr_view);
 	}
 
 	ui_view_title_update(curr_view);
 
-	reset_view_info(vi);
-
 	if(curr_view->explore_mode || other_view->explore_mode)
 	{
 		view_redraw();
@@ -504,6 +536,11 @@ init_view_info(view_info_t *vi)
 	vi->width = -1;
 	vi->last_search_backward = -1;
 	vi->search_repeat = NO_COUNT_GIVEN;
+	vi->nlines = 0;
+	vi->lines = NULL;
+	vi->widths = NULL;
+	vi->filename = NULL;
+	vi->viewer = NULL;
 }
 
 /* Frees all resources allocated by view_info_t structure instance. */
@@ -517,6 +554,7 @@ free_view_info(view_info_t *vi)
 		regfree(&vi->re);
 	}
 	free(vi->filename);
+	free(vi->viewer);
 }
 
 /* Updates line width and redraws the view. */
@@ -593,7 +631,7 @@ draw(void)
 	if(vi->graphics)
 	{
 		const char *cmd = qv_get_viewer(vi->filename);
-		cmd = (cmd != NULL) ? ma_get_clean_cmd(cmd) : NULL;
+		cmd = (cmd != NULL) ? ma_get_clear_cmd(cmd) : NULL;
 		qv_cleanup(vi->view, cmd);
 
 		free_string_array(vi->lines, vi->nlines);
@@ -628,6 +666,8 @@ draw(void)
 		}
 	}
 	refresh_view_win(vi->view);
+
+	checked_wmove(vi->view->win, ui_qv_top(vi->view), ui_qv_left(vi->view));
 }
 
 int
@@ -847,7 +887,7 @@ cmd_ctrl_ww(key_info_t key_info, keys_info_t *keys_info)
 	ui_views_update_titles();
 	if(curr_stats.view)
 	{
-		quick_view_file(curr_view);
+		qv_draw(curr_view);
 	}
 }
 
@@ -869,7 +909,7 @@ static void
 cmd_ctrl_wz(key_info_t key_info, keys_info_t *keys_info)
 {
 	leave_view_mode();
-	preview_close();
+	qv_hide();
 }
 
 static void
@@ -1036,11 +1076,14 @@ get_view_data(view_info_t *vi, const char file_to_view[])
 	FILE *fp;
 	const char *const viewer = qv_get_viewer(file_to_view);
 
-	if(is_null_or_empty(viewer))
+	if(vi->viewer == NULL && is_null_or_empty(viewer))
 	{
 		if(is_dir(file_to_view))
 		{
+			ui_cancellation_reset();
+			ui_cancellation_enable();
 			fp = qv_view_dir(file_to_view);
+			ui_cancellation_disable();
 		}
 		else
 		{
@@ -1056,7 +1099,8 @@ get_view_data(view_info_t *vi, const char file_to_view[])
 	}
 	else
 	{
-		const int graphics = is_graphics_viewer(viewer);
+		const char *const v = (vi->viewer != NULL) ? vi->viewer : viewer;
+		const int graphics = is_graphics_viewer(v);
 		FileView *const curr = curr_view;
 		curr_view = curr_stats.view ? curr_view
 		          : (vi->view != NULL) ? vi->view : curr_view;
@@ -1068,7 +1112,17 @@ get_view_data(view_info_t *vi, const char file_to_view[])
 			 * of them need this). */
 			usleep(50000);
 		}
-		fp = use_info_prog(viewer);
+		if(vi->viewer == NULL)
+		{
+			fp = qv_execute_viewer(viewer);
+		}
+		else
+		{
+			/* Don't add implicit %c to a command with %e macro. */
+			char *const cmd = expand_macros(vi->viewer, NULL, NULL, 1);
+			fp = read_cmd_output(cmd);
+			free(cmd);
+		}
 
 		curr_view = curr;
 		curr_stats.preview_hint = NULL;
@@ -1085,7 +1139,7 @@ get_view_data(view_info_t *vi, const char file_to_view[])
 
 		ui_cancellation_reset();
 		ui_cancellation_enable();
-		vi->lines = read_stream_lines(fp, &vi->nlines, 0);
+		vi->lines = read_stream_lines(fp, &vi->nlines, 0, NULL, NULL);
 		ui_cancellation_disable();
 	}
 
@@ -1107,6 +1161,9 @@ replace_vi(view_info_t *const orig, view_info_t *const new)
 	new->filename = orig->filename;
 	orig->filename = NULL;
 
+	new->viewer = orig->viewer;
+	orig->viewer = NULL;
+
 	if(orig->last_search_backward != -1)
 	{
 		new->last_search_backward = orig->last_search_backward;
@@ -1505,11 +1562,15 @@ is_trying_the_same_file(void)
 static int
 get_file_to_explore(const FileView *view, char buf[], size_t buf_len)
 {
-	const dir_entry_t *const entry = &view->dir_entry[view->list_pos];
+	const dir_entry_t *const curr = get_current_entry(view);
+	if(fentry_is_fake(curr))
+	{
+		return 1;
+	}
 
-	get_full_path_of(entry, buf_len, buf);
+	qv_get_path_to_explore(curr, buf, buf_len);
 
-	switch(entry->type)
+	switch(curr->type)
 	{
 		case FT_CHAR_DEV:
 		case FT_BLOCK_DEV:
@@ -1519,7 +1580,7 @@ get_file_to_explore(const FileView *view, char buf[], size_t buf_len)
 #endif
 			return 1;
 		case FT_LINK:
-			if(get_link_target_abs(buf, entry->origin, buf, buf_len) != 0)
+			if(get_link_target_abs(buf, curr->origin, buf, buf_len) != 0)
 			{
 				return 1;
 			}
@@ -1602,8 +1663,9 @@ reload_view(view_info_t *vi, int silent)
 	view_info_t new_vi;
 
 	init_view_info(&new_vi);
-	/* The view field is used in get_view_data(). */
+	/* These fields are used in get_view_data(). */
 	new_vi.view = vi->view;
+	new_vi.viewer = vi->viewer;
 
 	if(load_view_data(&new_vi, "File exploring reload", vi->filename, silent)
 			== 0)
