@@ -27,6 +27,7 @@
 
 #include "../cfg/config.h"
 #include "../cfg/hist.h"
+#include "../compat/curses.h"
 #include "../engine/keys.h"
 #include "../engine/mode.h"
 #include "../modes/dialogs/msg_dialog.h"
@@ -39,8 +40,11 @@
 #include "../utils/utils.h"
 #include "../cmd_core.h"
 #include "../filelist.h"
-#include "../fileops.h"
 #include "../filtering.h"
+#include "../flist_pos.h"
+#include "../flist_sel.h"
+#include "../fops_misc.h"
+#include "../fops_rename.h"
 #include "../marks.h"
 #include "../registers.h"
 #include "../running.h"
@@ -141,6 +145,16 @@ static void cmd_zd(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_zf(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_left_paren(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_right_paren(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_z_k(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_z_j(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_lb_c(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_rb_c(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_lb_d(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_rb_d(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_lb_s(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_rb_s(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_lb_z(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_rb_z(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_left_curly_bracket(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_right_curly_bracket(key_info_t key_info,
 		keys_info_t *keys_info);
@@ -149,7 +163,6 @@ static void select_up_one(FileView *view, int start_pos);
 static void select_down_one(FileView *view, int start_pos);
 static void apply_selection(int pos);
 static void revert_selection(int pos);
-static int is_parent_dir_at(int pos);
 static void update(void);
 static int find_update(FileView *view, int backward);
 static void goto_pos_force_update(int pos);
@@ -240,14 +253,24 @@ static keys_add_info_t builtin_cmds[] = {
 	{WK_z WK_z,     {{&normal_cmd_zz},   .descr = "center cursor position"}},
 	{WK_LP,         {{&cmd_left_paren},  .descr = "go to previous group of files"}},
 	{WK_RP,         {{&cmd_right_paren}, .descr = "go to next group of files"}},
-	{WK_LCB,        {{&cmd_left_curly_bracket},  .descr = "go to previous file/dir"}},
-	{WK_RCB,        {{&cmd_right_curly_bracket}, .descr = "go to next file/dir"}},
+	{WK_z WK_k,     {{&cmd_z_k},  .descr = "go to previous sibling dir"}},
+	{WK_z WK_j,     {{&cmd_z_j},  .descr = "go to next sibling dir"}},
+	{WK_LB WK_c,    {{&cmd_lb_c}, .descr = "go to previous mismatch"}},
+	{WK_RB WK_c,    {{&cmd_rb_c}, .descr = "go to next mismatch"}},
+	{WK_LB WK_d,    {{&cmd_lb_d}, .descr = "go to previous dir"}},
+	{WK_RB WK_d,    {{&cmd_rb_d}, .descr = "go to next dir"}},
+	{WK_LB WK_s,    {{&cmd_lb_s}, .descr = "go to previous selected entry"}},
+	{WK_RB WK_s,    {{&cmd_rb_s}, .descr = "go to next selected entry"}},
+	{WK_LB WK_z,    {{&cmd_lb_z}, .descr = "go to first sibling"}},
+	{WK_RB WK_z,    {{&cmd_rb_z}, .descr = "go to last sibling"}},
+	{WK_LCB,        {{&cmd_left_curly_bracket},  .descr = "go to previous file/dir group"}},
+	{WK_RCB,        {{&cmd_right_curly_bracket}, .descr = "go to next file/dir group"}},
 #ifdef ENABLE_EXTENDED_KEYS
-	{{KEY_PPAGE},   {{&cmd_ctrl_b}, .descr = "scroll page up"}},
-	{{KEY_NPAGE},   {{&cmd_ctrl_f}, .descr = "scroll page down"}},
-	{{KEY_DOWN},    {{&cmd_j},      .descr = "go to item below"}},
-	{{KEY_UP},      {{&cmd_k},      .descr = "go to item above"}},
-	{{KEY_RIGHT},   {{&cmd_l},      .descr = "open selection/go to item to the right"}},
+	{{K(KEY_PPAGE)},   {{&cmd_ctrl_b}, .descr = "scroll page up"}},
+	{{K(KEY_NPAGE)},   {{&cmd_ctrl_f}, .descr = "scroll page down"}},
+	{{K(KEY_DOWN)},    {{&cmd_j},      .descr = "go to item below"}},
+	{{K(KEY_UP)},      {{&cmd_k},      .descr = "go to item above"}},
+	{{K(KEY_RIGHT)},   {{&cmd_l},      .descr = "open selection/go to item to the right"}},
 #endif /* ENABLE_EXTENDED_KEYS */
 };
 
@@ -281,13 +304,13 @@ enter_visual_mode(VisualSubmodes sub_mode)
 	{
 		case VS_NORMAL:
 			amend_type = AT_NONE;
-			clean_selected_files(view);
+			flist_sel_stash(view);
 			backup_selection_flags(view);
 			select_first_one();
 			break;
 		case VS_RESTORE:
 			amend_type = AT_NONE;
-			clean_selected_files(view);
+			flist_sel_stash(view);
 			backup_selection_flags(view);
 			restore_previous_selection();
 			break;
@@ -302,7 +325,7 @@ enter_visual_mode(VisualSubmodes sub_mode)
 }
 
 void
-leave_visual_mode(int save_msg, int goto_top, int clean_selection)
+leave_visual_mode(int save_msg, int goto_top, int clear_selection)
 {
 	if(goto_top)
 	{
@@ -311,7 +334,7 @@ leave_visual_mode(int save_msg, int goto_top, int clean_selection)
 			view->list_pos = ub;
 	}
 
-	if(clean_selection)
+	if(clear_selection)
 	{
 		reset_search_results(view);
 		restore_selection_flags(view);
@@ -489,7 +512,7 @@ call_incdec(int count)
 {
 	int save_msg;
 	check_marking(view, 0, NULL);
-	save_msg = incdec_names(view, count);
+	save_msg = fops_incdec(view, count);
 	accept_and_leave(save_msg);
 }
 
@@ -511,7 +534,7 @@ cmd_C(key_info_t key_info, keys_info_t *keys_info)
 {
 	int save_msg;
 	check_marking(curr_view, 0, NULL);
-	save_msg = clone_files(view, NULL, 0, 0, def_count(key_info.count));
+	save_msg = fops_clone(view, NULL, 0, 0, def_count(key_info.count));
 	accept_and_leave(save_msg);
 }
 
@@ -521,15 +544,14 @@ cmd_D(key_info_t key_info, keys_info_t *keys_info)
 	delete(key_info, 0);
 }
 
+/* Navigates to previous word which starts with specified character. */
 static void
 cmd_F(key_info_t key_info, keys_info_t *keys_info)
 {
 	last_fast_search_char = key_info.multi;
 	last_fast_search_backward = 1;
 
-	if(key_info.count == NO_COUNT_GIVEN)
-		key_info.count = 1;
-	find_goto(key_info.multi, key_info.count, 1);
+	find_goto(key_info.multi, def_count(key_info.count), 1);
 }
 
 static void
@@ -624,15 +646,16 @@ cmd_percent(key_info_t key_info, keys_info_t *keys_info)
 	goto_pos(line - 1);
 }
 
+/* Continues navigation to word which starts with specified character in
+ * opposite direction. */
 static void
 cmd_comma(key_info_t key_info, keys_info_t *keys_info)
 {
-	if(last_fast_search_backward == -1)
-		return;
-
-	if(key_info.count == NO_COUNT_GIVEN)
-		key_info.count = 1;
-	find_goto(last_fast_search_char, key_info.count, !last_fast_search_backward);
+	if(last_fast_search_backward != -1)
+	{
+		find_goto(last_fast_search_char, def_count(key_info.count),
+				!last_fast_search_backward);
+	}
 }
 
 /* Move cursor to the first column in ls-view sub-mode selecting or unselecting
@@ -654,15 +677,16 @@ cmd_colon(key_info_t key_info, keys_info_t *keys_info)
 	enter_cmdline_mode(CLS_COMMAND, "", NULL);
 }
 
+/* Continues navigation to word which starts with specified character in initial
+ * direction. */
 static void
 cmd_semicolon(key_info_t key_info, keys_info_t *keys_info)
 {
-	if(last_fast_search_backward == -1)
-		return;
-
-	if(key_info.count == NO_COUNT_GIVEN)
-		key_info.count = 1;
-	find_goto(last_fast_search_char, key_info.count, last_fast_search_backward);
+	if(last_fast_search_backward != -1)
+	{
+		find_goto(last_fast_search_char, def_count(key_info.count),
+				last_fast_search_backward);
+	}
 }
 
 /* Search forward. */
@@ -711,27 +735,26 @@ delete(key_info_t key_info, int use_trash)
 
 	if(count != 0 && confirm_deletion(count, use_trash))
 	{
-		const int save_msg = delete_files(view, def_reg(key_info.reg), use_trash);
+		const int save_msg = fops_delete(view, def_reg(key_info.reg), use_trash);
 		accept_and_leave(save_msg);
 	}
 }
 
+/* Navigates to next word which starts with specified character. */
 static void
 cmd_f(key_info_t key_info, keys_info_t *keys_info)
 {
 	last_fast_search_char = key_info.multi;
 	last_fast_search_backward = 0;
 
-	if(key_info.count == NO_COUNT_GIVEN)
-		key_info.count = 1;
-	find_goto(key_info.multi, key_info.count, 0);
+	find_goto(key_info.multi, def_count(key_info.count), 0);
 }
 
 /* Calculate size of selected directories ignoring cached sizes. */
 static void
 cmd_gA(key_info_t key_info, keys_info_t *keys_info)
 {
-	calculate_size_bg(curr_view, 1);
+	fops_size_bg(curr_view, 1);
 	accept_and_leave(0);
 }
 
@@ -739,7 +762,7 @@ cmd_gA(key_info_t key_info, keys_info_t *keys_info)
 static void
 cmd_ga(key_info_t key_info, keys_info_t *keys_info)
 {
-	calculate_size_bg(curr_view, 0);
+	fops_size_bg(curr_view, 0);
 	accept_and_leave(0);
 }
 
@@ -753,9 +776,11 @@ cmd_cp(key_info_t key_info, keys_info_t *keys_info)
 	update_marks(view);
 	ub = check_mark_directory(view, '<');
 	if(ub != -1)
+	{
 		view->list_pos = ub;
+	}
 
-	enter_attr_mode(view);
+	normal_cmd_cp(view, key_info);
 }
 
 /* Renames selected files of the view. */
@@ -766,7 +791,7 @@ cmd_cw(key_info_t key_info, keys_info_t *keys_info)
 	leave_visual_mode(0, 1, 0);
 
 	check_marking(view, 0, NULL);
-	(void)rename_files(view, NULL, 0, 0);
+	(void)fops_rename(view, NULL, 0, 0);
 }
 
 static void
@@ -783,7 +808,7 @@ cmd_gl(key_info_t key_info, keys_info_t *keys_info)
 	update_marks(view);
 	leave_visual_mode(curr_stats.save_msg, 1, 0);
 	open_file(view, FHE_RUN);
-	clean_selected_files(view);
+	flist_sel_stash(view);
 	redraw_view(view);
 }
 
@@ -808,7 +833,7 @@ do_gu(int upper)
 	int save_msg;
 
 	check_marking(view, 0, NULL);
-	save_msg = change_case(view, upper);
+	save_msg = fops_case(view, upper);
 	accept_and_leave(save_msg);
 }
 
@@ -830,7 +855,7 @@ restore_previous_selection(void)
 	if(ub < 0 || lb < 0)
 		return;
 
-	erase_selection(view);
+	flist_sel_drop(view);
 
 	start_pos = ub;
 	view->list_pos = ub;
@@ -854,20 +879,29 @@ restore_previous_selection(void)
 static void
 select_first_one(void)
 {
-	if(!is_parent_dir_at(view->list_pos))
-	{
-		apply_selection(view->list_pos);
-	}
+	apply_selection(view->list_pos);
 }
 
 /* Go backwards [count] (one by default) files in ls-like sub-mode. */
 static void
 cmd_h(key_info_t key_info, keys_info_t *keys_info)
 {
-	if(view->ls_view)
+	if(!ui_view_displays_columns(view))
 	{
 		go_to_prev(key_info, keys_info, 1, 1);
 	}
+	else if(get_current_entry(view)->child_pos != 0)
+	{
+		const dir_entry_t *entry = get_current_entry(view);
+		key_info.count = def_count(key_info.count);
+		while(key_info.count-- > 0)
+		{
+			entry -= entry->child_pos;
+		}
+		key_info.count = 1;
+		go_to_prev(key_info, keys_info, 1,
+				view->list_pos - entry_to_pos(curr_view, entry));
+	}
 }
 
 static void
@@ -880,7 +914,7 @@ cmd_i(key_info_t key_info, keys_info_t *keys_info)
 static void
 cmd_j(key_info_t key_info, keys_info_t *keys_info)
 {
-	if(!view->ls_view || !at_last_line(view))
+	if(ui_view_displays_columns(view) || !at_last_line(view))
 	{
 		go_to_next(key_info, keys_info, 1, view->column_count);
 	}
@@ -889,7 +923,7 @@ cmd_j(key_info_t key_info, keys_info_t *keys_info)
 static void
 cmd_k(key_info_t key_info, keys_info_t *keys_info)
 {
-	if(!view->ls_view || !at_first_line(view))
+	if(ui_view_displays_columns(view) || !at_first_line(view))
 	{
 		go_to_prev(key_info, keys_info, 1, view->column_count);
 	}
@@ -908,13 +942,13 @@ go_to_prev(key_info_t key_info, keys_info_t *keys_info, int def, int step)
 static void
 cmd_l(key_info_t key_info, keys_info_t *keys_info)
 {
-	if(view->ls_view)
+	if(ui_view_displays_columns(view))
 	{
-		go_to_next(key_info, keys_info, 1, 1);
+		cmd_gl(key_info, keys_info);
 	}
 	else
 	{
-		cmd_gl(key_info, keys_info);
+		go_to_next(key_info, keys_info, 1, 1);
 	}
 }
 
@@ -931,8 +965,11 @@ go_to_next(key_info_t key_info, keys_info_t *keys_info, int def, int step)
 static void
 cmd_m(key_info_t key_info, keys_info_t *keys_info)
 {
-	const dir_entry_t *const entry = &view->dir_entry[view->list_pos];
-	set_user_mark(key_info.multi, entry->origin, entry->name);
+	const dir_entry_t *const curr = get_current_entry(view);
+	if(!fentry_is_fake(curr))
+	{
+		set_user_mark(key_info.multi, curr->origin, curr->name);
+	}
 }
 
 static void
@@ -1044,7 +1081,7 @@ change_amend_type(AmendType new_amend_type)
 
 	if(new_amend_type == AT_NONE)
 	{
-		clean_selected_files(view);
+		flist_sel_stash(view);
 	}
 	else
 	{
@@ -1064,7 +1101,7 @@ cmd_y(key_info_t key_info, keys_info_t *keys_info)
 	int save_msg;
 
 	check_marking(view, 0, NULL);
-	save_msg = yank_files(view, def_reg(key_info.reg));
+	save_msg = fops_yank(view, def_reg(key_info.reg));
 
 	accept_and_leave(save_msg);
 }
@@ -1104,6 +1141,7 @@ update_marks(FileView *view)
 {
 	char start_mark, end_mark;
 	const dir_entry_t *start_entry, *end_entry;
+	int delta;
 
 	if(start_pos >= view->list_rows)
 	{
@@ -1111,12 +1149,22 @@ update_marks(FileView *view)
 	}
 
 	upwards_range = view->list_pos < start_pos;
+	delta = upwards_range ? +1 : -1;
 
 	start_mark = upwards_range ? '<' : '>';
 	end_mark = upwards_range ? '>' : '<';
 
 	start_entry = &view->dir_entry[view->list_pos];
 	end_entry = &view->dir_entry[start_pos];
+	/* Fake entries can't be marked, so skip them on both ends. */
+	while(start_entry != end_entry && fentry_is_fake(start_entry))
+	{
+		start_entry += delta;
+	}
+	while(start_entry != end_entry && fentry_is_fake(end_entry))
+	{
+		end_entry -= delta;
+	}
 
 	set_spec_mark(start_mark, start_entry->origin, start_entry->name);
 	set_spec_mark(end_mark, end_entry->origin, end_entry->name);
@@ -1126,7 +1174,7 @@ update_marks(FileView *view)
 static void
 cmd_zd(key_info_t key_info, keys_info_t *keys_info)
 {
-	flist_custom_exclude(curr_view);
+	flist_custom_exclude(curr_view, key_info.count == 1);
 	accept_and_leave(0);
 }
 
@@ -1153,6 +1201,76 @@ cmd_right_paren(key_info_t key_info, keys_info_t *keys_info)
 	goto_pos(flist_find_group(view, 1));
 }
 
+/* Go to previous sibling directory entry or do nothing. */
+static void
+cmd_z_k(key_info_t key_info, keys_info_t *keys_info)
+{
+	goto_pos(flist_prev_dir_sibling(view));
+}
+
+/* Go to next sibling directory entry or do nothing. */
+static void
+cmd_z_j(key_info_t key_info, keys_info_t *keys_info)
+{
+	goto_pos(flist_next_dir_sibling(view));
+}
+
+/* Go to previous mismatched entry or do nothing. */
+static void
+cmd_lb_c(key_info_t key_info, keys_info_t *keys_info)
+{
+	goto_pos(flist_prev_mismatch(view));
+}
+
+/* Go to next mismatched entry or do nothing. */
+static void
+cmd_rb_c(key_info_t key_info, keys_info_t *keys_info)
+{
+	goto_pos(flist_next_mismatch(view));
+}
+
+/* Go to previous directory entry or do nothing. */
+static void
+cmd_lb_d(key_info_t key_info, keys_info_t *keys_info)
+{
+	goto_pos(flist_prev_dir(view));
+}
+
+/* Go to next directory entry or do nothing. */
+static void
+cmd_rb_d(key_info_t key_info, keys_info_t *keys_info)
+{
+	goto_pos(flist_next_dir(view));
+}
+
+/* Go to previous selected entry or do nothing. */
+static void
+cmd_lb_s(key_info_t key_info, keys_info_t *keys_info)
+{
+	goto_pos(flist_prev_selected(view));
+}
+
+/* Go to next selected entry or do nothing. */
+static void
+cmd_rb_s(key_info_t key_info, keys_info_t *keys_info)
+{
+	goto_pos(flist_next_selected(view));
+}
+
+/* Go to first sibling of the current entry. */
+static void
+cmd_lb_z(key_info_t key_info, keys_info_t *keys_info)
+{
+	goto_pos(flist_first_sibling(view));
+}
+
+/* Go to last sibling of the current entry. */
+static void
+cmd_rb_z(key_info_t key_info, keys_info_t *keys_info)
+{
+	goto_pos(flist_last_sibling(view));
+}
+
 /* Moves cursor to the beginning of the previous group of files defined by them
  * being directory or files. */
 static void
@@ -1169,17 +1287,20 @@ cmd_right_curly_bracket(key_info_t key_info, keys_info_t *keys_info)
 	goto_pos(flist_find_dir_group(view, 1));
 }
 
+/* Navigates to next/previous file which starts with given character. */
 static void
 find_goto(int ch, int count, int backward)
 {
-	int pos;
-	pos = ffind(ch, backward, 1);
+	int pos = flist_find_by_ch(view, ch, backward, 1);
 	if(pos < 0 || pos == view->list_pos)
+	{
 		return;
-	while(--count > 0)
+	}
+
+	while(count-- > 0)
 	{
 		goto_pos(pos);
-		pos = ffind(ch, backward, 1);
+		pos = flist_find_by_ch(view, ch, backward, 1);
 	}
 }
 
@@ -1190,19 +1311,8 @@ select_up_one(FileView *view, int start_pos)
 	view->list_pos--;
 	if(view->list_pos < 0)
 	{
-		if(is_parent_dir_at(start_pos))
-		{
-			--view->selected_files;
-		}
 		view->list_pos = 0;
 	}
-	else if(view->list_pos == 0 && is_parent_dir_at(0))
-	{
-		if(start_pos == 0)
-		{
-			revert_selection(1);
-		}
-	}
 	else if(view->list_pos < start_pos)
 	{
 		apply_selection(view->list_pos);
@@ -1222,16 +1332,12 @@ select_up_one(FileView *view, int start_pos)
 static void
 select_down_one(FileView *view, int start_pos)
 {
-	view->list_pos++;
+	++view->list_pos;
 
 	if(view->list_pos >= view->list_rows)
 	{
 		view->list_pos = view->list_rows - 1;
 	}
-	else if(view->list_pos == 1 && start_pos != 0 && is_parent_dir_at(0))
-	{
-		/* do nothing */
-	}
 	else if(view->list_pos > start_pos)
 	{
 		apply_selection(view->list_pos);
@@ -1252,6 +1358,11 @@ static void
 apply_selection(int pos)
 {
 	dir_entry_t *const entry = &view->dir_entry[pos];
+	if(!fentry_is_valid(entry))
+	{
+		return;
+	}
+
 	switch(amend_type)
 	{
 		case AT_NONE:
@@ -1327,28 +1438,19 @@ revert_selection(int pos)
 	}
 }
 
-/* Checks whether file at specified position in file list refers to parent
- * directory. */
-static int
-is_parent_dir_at(int pos)
-{
-	/* Don't allow the ../ dir to be selected. */
-	return is_parent_dir(view->dir_entry[pos].name);
-}
-
 static void
 update(void)
 {
 	redraw_view(view);
-	ui_ruler_update(view);
+	ui_ruler_update(view, 1);
 }
 
 void
 update_visual_mode(void)
 {
-	int pos = view->list_pos;
+	const int pos = view->list_pos;
 
-	clean_selected_files(view);
+	flist_sel_stash(view);
 	view->dir_entry[start_pos].selected = 1;
 	view->selected_files = 1;
 
