@@ -38,11 +38,13 @@
 #include "cfg/config.h"
 #include "compat/fs_limits.h"
 #include "compat/os.h"
+#include "compat/reallocarray.h"
 #include "io/ioeta.h"
 #include "io/iop.h"
 #include "io/ior.h"
 #include "modes/dialogs/msg_dialog.h"
 #include "ui/cancellation.h"
+#include "utils/cancellation.h"
 #include "utils/fs.h"
 #include "utils/log.h"
 #include "utils/macros.h"
@@ -115,14 +117,20 @@ static int op_symlink(ops_t *ops, void *data, const char *src, const char *dst);
 static int op_mkdir(ops_t *ops, void *data, const char *src, const char *dst);
 static int op_rmdir(ops_t *ops, void *data, const char *src, const char *dst);
 static int op_mkfile(ops_t *ops, void *data, const char *src, const char *dst);
+static int ops_uses_syscalls(const ops_t *ops);
 static int exec_io_op(ops_t *ops, int (*func)(io_args_t *const),
-		io_args_t *const args);
+		io_args_t *const args, int cancellable);
 static int confirm_overwrite(io_args_t *args, const char src[],
 		const char dst[]);
 static char * pretty_dir_path(const char path[]);
 static IoErrCbResult dispatch_error(io_args_t *args, const ioe_err_t *err);
 static char prompt_user(const io_args_t *args, const char title[],
 		const char msg[], const response_variant variants[]);
+static int ui_cancellation_hook(void *arg);
+#ifndef _WIN32
+static int run_operation_command(ops_t *ops, char cmd[], int cancellable);
+#endif
+static int bg_cancellation_hook(void *arg);
 
 /* List of functions that implement operations. */
 static op_func op_funcs[] = {
@@ -167,6 +175,10 @@ ops_alloc(OPS main_op, int bg, const char descr[], const char base_dir[],
 	ops_t *const ops = calloc(1, sizeof(*ops));
 	ops->main_op = main_op;
 	ops->descr = descr;
+	update_string(&ops->slow_fs_list, cfg.slow_fs_list);
+	update_string(&ops->delete_prg, cfg.delete_prg);
+	ops->use_system_calls = cfg.use_system_calls;
+	ops->fast_file_cloning = cfg.fast_file_cloning;
 	ops->base_dir = strdup(base_dir);
 	ops->target_dir = strdup(target_dir);
 	ops->bg = bg;
@@ -220,7 +232,7 @@ ops_enqueue(ops_t *ops, const char src[], const char dst[])
 				break;
 		}
 
-		if(is_on_slow_fs(src))
+		if(is_on_slow_fs(src, ops->slow_fs_list))
 		{
 			ops->shallow_eta = 1;
 		}
@@ -251,6 +263,8 @@ ops_free(ops_t *ops)
 
 	ioeta_free(ops->estim);
 	free(ops->errors);
+	free(ops->slow_fs_list);
+	free(ops->delete_prg);
 	free(ops->base_dir);
 	free(ops->target_dir);
 	free(ops);
@@ -288,7 +302,10 @@ op_remove(ops_t *ops, void *data, const char *src, const char *dst)
 static int
 op_removesl(ops_t *ops, void *data, const char *src, const char *dst)
 {
-	if(cfg.delete_prg[0] != '\0')
+	const char *const delete_prg = (ops == NULL)
+	                             ? cfg.delete_prg
+	                             : ops->delete_prg;
+	if(delete_prg[0] != '\0')
 	{
 #ifndef _WIN32
 		char *escaped;
@@ -301,21 +318,21 @@ op_removesl(ops_t *ops, void *data, const char *src, const char *dst)
 			return -1;
 		}
 
-		snprintf(cmd, sizeof(cmd), "%s %s", cfg.delete_prg, escaped);
+		snprintf(cmd, sizeof(cmd), "%s %s", delete_prg, escaped);
 		free(escaped);
 
 		LOG_INFO_MSG("Running trash command: \"%s\"", cmd);
-		return background_and_wait_for_errors(cmd, cancellable);
+		return run_operation_command(ops, cmd, cancellable);
 #else
 		char cmd[PATH_MAX*2 + 1];
-		snprintf(cmd, sizeof(cmd), "%s \"%s\"", cfg.delete_prg, src);
+		snprintf(cmd, sizeof(cmd), "%s \"%s\"", delete_prg, src);
 		to_back_slash(cmd);
 
 		return os_system(cmd);
 #endif
 	}
 
-	if(!cfg.use_system_calls)
+	if(!ops_uses_syscalls(ops))
 	{
 #ifndef _WIN32
 		char *escaped;
@@ -329,7 +346,7 @@ op_removesl(ops_t *ops, void *data, const char *src, const char *dst)
 
 		snprintf(cmd, sizeof(cmd), "rm -rf %s", escaped);
 		LOG_INFO_MSG("Running rm command: \"%s\"", cmd);
-		result = background_and_wait_for_errors(cmd, cancellable);
+		result = run_operation_command(ops, cmd, cancellable);
 
 		free(escaped);
 		return result;
@@ -342,7 +359,12 @@ op_removesl(ops_t *ops, void *data, const char *src, const char *dst)
 			copy_str(path, sizeof(path), src);
 			to_back_slash(path);
 
-			wchar_t *const utf16_path = utf8_to_utf16(path);
+			wchar_t *utf16_path = utf8_to_utf16(path);
+
+			/* SHFileOperationW requires pFrom to be double-nul terminated. */
+			const size_t len = wcslen(utf16_path);
+			utf16_path = reallocarray(utf16_path, len + 1U + 1U, sizeof(*utf16_path));
+			utf16_path[len + 1U] = L'\0';
 
 			SHFILEOPSTRUCTW fo = {
 				.hwnd = NULL,
@@ -382,10 +404,8 @@ op_removesl(ops_t *ops, void *data, const char *src, const char *dst)
 
 	io_args_t args = {
 		.arg1.path = src,
-
-		.cancellable = data == NULL,
 	};
-	return exec_io_op(ops, &ior_rm, &args);
+	return exec_io_op(ops, &ior_rm, &args, data == NULL);
 }
 
 /* OP_COPY operation handler.  Copies file/directory without overwriting
@@ -419,7 +439,11 @@ static int
 op_cp(ops_t *ops, void *data, const char src[], const char dst[],
 		ConflictAction conflict_action)
 {
-	if(!cfg.use_system_calls)
+	const int fast_file_cloning = (ops == NULL)
+	                             ? cfg.fast_file_cloning
+	                             : ops->fast_file_cloning;
+
+	if(!ops_uses_syscalls(ops))
 	{
 #ifndef _WIN32
 		char *escaped_src, *escaped_dst;
@@ -439,10 +463,10 @@ op_cp(ops_t *ops, void *data, const char src[], const char dst[],
 		snprintf(cmd, sizeof(cmd),
 				"cp %s %s -R " PRESERVE_FLAGS " %s %s",
 				(conflict_action == CA_FAIL) ? NO_CLOBBER : "",
-				cfg.fast_file_cloning ? REFLINK_AUTO : "",
+				fast_file_cloning ? REFLINK_AUTO : "",
 				escaped_src, escaped_dst);
 		LOG_INFO_MSG("Running cp command: \"%s\"", cmd);
-		result = background_and_wait_for_errors(cmd, cancellable);
+		result = run_operation_command(ops, cmd, cancellable);
 
 		free(escaped_dst);
 		free(escaped_src);
@@ -482,11 +506,9 @@ op_cp(ops_t *ops, void *data, const char src[], const char dst[],
 		.arg1.src = src,
 		.arg2.dst = dst,
 		.arg3.crs = ca_to_crs(conflict_action),
-		.arg4.fast_file_cloning = cfg.fast_file_cloning,
-
-		.cancellable = data == NULL,
+		.arg4.fast_file_cloning = fast_file_cloning,
 	};
-	return exec_io_op(ops, &ior_cp, &args);
+	return exec_io_op(ops, &ior_cp, &args, data == NULL);
 }
 
 /* OP_MOVE operation handler.  Moves file/directory without overwriting
@@ -522,7 +544,7 @@ op_mv(ops_t *ops, void *data, const char src[], const char dst[],
 {
 	int result;
 
-	if(!cfg.use_system_calls)
+	if(!ops_uses_syscalls(ops))
 	{
 #ifndef _WIN32
 		struct stat st;
@@ -552,7 +574,7 @@ op_mv(ops_t *ops, void *data, const char src[], const char dst[],
 		free(escaped_src);
 
 		LOG_INFO_MSG("Running mv command: \"%s\"", cmd);
-		result = background_and_wait_for_errors(cmd, cancellable);
+		result = run_operation_command(ops, cmd, cancellable);
 		if(result != 0)
 		{
 			return result;
@@ -586,10 +608,8 @@ op_mv(ops_t *ops, void *data, const char src[], const char dst[],
 			.arg3.crs = ca_to_crs(conflict_action),
 			/* It's safe to always use fast file cloning on moving files. */
 			.arg4.fast_file_cloning = 1,
-
-			.cancellable = data == NULL,
 		};
-		result = exec_io_op(ops, &ior_mv, &args);
+		result = exec_io_op(ops, &ior_mv, &args, data == NULL);
 	}
 
 	if(result == 0)
@@ -629,7 +649,7 @@ op_chown(ops_t *ops, void *data, const char *src, const char *dst)
 	free(escaped);
 
 	LOG_INFO_MSG("Running chown command: \"%s\"", cmd);
-	return background_and_wait_for_errors(cmd, 1);
+	return run_operation_command(ops, cmd, 1);
 #else
 	return -1;
 #endif
@@ -648,7 +668,7 @@ op_chgrp(ops_t *ops, void *data, const char *src, const char *dst)
 	free(escaped);
 
 	LOG_INFO_MSG("Running chgrp command: \"%s\"", cmd);
-	return background_and_wait_for_errors(cmd, 1);
+	return run_operation_command(ops, cmd, 1);
 #else
 	return -1;
 #endif
@@ -666,7 +686,7 @@ op_chmod(ops_t *ops, void *data, const char *src, const char *dst)
 	free(escaped);
 
 	LOG_INFO_MSG("Running chmod command: \"%s\"", cmd);
-	return background_and_wait_for_errors(cmd, 1);
+	return run_operation_command(ops, cmd, 1);
 }
 
 static int
@@ -680,7 +700,7 @@ op_chmodr(ops_t *ops, void *data, const char *src, const char *dst)
 	free(escaped);
 
 	LOG_INFO_MSG("Running chmodr command: \"%s\"", cmd);
-	return background_and_wait_for_errors(cmd, 1);
+	return run_operation_command(ops, cmd, 1);
 }
 #else
 static int
@@ -731,7 +751,7 @@ op_subattr(ops_t *ops, void *data, const char *src, const char *dst)
 static int
 op_symlink(ops_t *ops, void *data, const char *src, const char *dst)
 {
-	if(!cfg.use_system_calls)
+	if(!ops_uses_syscalls(ops))
 	{
 		char *escaped_src, *escaped_dst;
 		char cmd[6 + PATH_MAX*2 + 1];
@@ -752,7 +772,7 @@ op_symlink(ops_t *ops, void *data, const char *src, const char *dst)
 #ifndef _WIN32
 		snprintf(cmd, sizeof(cmd), "ln -s %s %s", escaped_src, escaped_dst);
 		LOG_INFO_MSG("Running ln command: \"%s\"", cmd);
-		result = background_and_wait_for_errors(cmd, 1);
+		result = run_operation_command(ops, cmd, 1);
 #else
 		if(get_exe_dir(exe_dir, ARRAY_LEN(exe_dir)) != 0)
 		{
@@ -776,13 +796,13 @@ op_symlink(ops_t *ops, void *data, const char *src, const char *dst)
 		.arg2.target = dst,
 		.arg3.crs = IO_CRS_REPLACE_FILES,
 	};
-	return exec_io_op(ops, &iop_ln, &args);
+	return exec_io_op(ops, &iop_ln, &args, 0);
 }
 
 static int
 op_mkdir(ops_t *ops, void *data, const char *src, const char *dst)
 {
-	if(!cfg.use_system_calls)
+	if(!ops_uses_syscalls(ops))
 	{
 #ifndef _WIN32
 		char cmd[128 + PATH_MAX];
@@ -793,7 +813,7 @@ op_mkdir(ops_t *ops, void *data, const char *src, const char *dst)
 				escaped);
 		free(escaped);
 		LOG_INFO_MSG("Running mkdir command: \"%s\"", cmd);
-		return background_and_wait_for_errors(cmd, 1);
+		return run_operation_command(ops, cmd, 1);
 #else
 		if(data == NULL)
 		{
@@ -833,13 +853,13 @@ op_mkdir(ops_t *ops, void *data, const char *src, const char *dst)
 		.arg2.process_parents = data != NULL,
 		.arg3.mode = 0755,
 	};
-	return exec_io_op(ops, &iop_mkdir, &args);
+	return exec_io_op(ops, &iop_mkdir, &args, 0);
 }
 
 static int
 op_rmdir(ops_t *ops, void *data, const char *src, const char *dst)
 {
-	if(!cfg.use_system_calls)
+	if(!ops_uses_syscalls(ops))
 	{
 #ifndef _WIN32
 		char cmd[128 + PATH_MAX];
@@ -849,7 +869,7 @@ op_rmdir(ops_t *ops, void *data, const char *src, const char *dst)
 		snprintf(cmd, sizeof(cmd), "rmdir %s", escaped);
 		free(escaped);
 		LOG_INFO_MSG("Running rmdir command: \"%s\"", cmd);
-		return background_and_wait_for_errors(cmd, 1);
+		return run_operation_command(ops, cmd, 1);
 #else
 		wchar_t *const utf16_path = utf8_to_utf16(src);
 		const BOOL r = RemoveDirectoryW(utf16_path);
@@ -861,13 +881,13 @@ op_rmdir(ops_t *ops, void *data, const char *src, const char *dst)
 	io_args_t args = {
 		.arg1.path = src,
 	};
-	return exec_io_op(ops, &iop_rmdir, &args);
+	return exec_io_op(ops, &iop_rmdir, &args, 0);
 }
 
 static int
 op_mkfile(ops_t *ops, void *data, const char *src, const char *dst)
 {
-	if(!cfg.use_system_calls)
+	if(!ops_uses_syscalls(ops))
 	{
 #ifndef _WIN32
 		char cmd[128 + PATH_MAX];
@@ -877,7 +897,7 @@ op_mkfile(ops_t *ops, void *data, const char *src, const char *dst)
 		snprintf(cmd, sizeof(cmd), "touch %s", escaped);
 		free(escaped);
 		LOG_INFO_MSG("Running touch command: \"%s\"", cmd);
-		return background_and_wait_for_errors(cmd, 1);
+		return run_operation_command(ops, cmd, 1);
 #else
 		HANDLE hfile;
 
@@ -898,13 +918,22 @@ op_mkfile(ops_t *ops, void *data, const char *src, const char *dst)
 	io_args_t args = {
 		.arg1.path = src,
 	};
-	return exec_io_op(ops, &iop_mkfile, &args);
+	return exec_io_op(ops, &iop_mkfile, &args, 0);
+}
+
+/* Checks whether specific operation should use system calls.  Returns non-zero
+ * if so, otherwise zero is returned. */
+static int
+ops_uses_syscalls(const ops_t *ops)
+{
+	return ops == NULL ? cfg.use_system_calls : ops->use_system_calls;
 }
 
 /* Executes i/o operation with some predefined pre/post actions.  Returns exit
  * code of i/o operation. */
 static int
-exec_io_op(ops_t *ops, int (*func)(io_args_t *const), io_args_t *const args)
+exec_io_op(ops_t *ops, int (*func)(io_args_t *const), io_args_t *const args,
+		int cancellable)
 {
 	int result;
 
@@ -921,16 +950,27 @@ exec_io_op(ops_t *ops, int (*func)(io_args_t *const), io_args_t *const args)
 		ioe_errlst_init(&args->result.errors);
 	}
 
-	if(args->cancellable)
+	if(cancellable)
 	{
-		ui_cancellation_enable();
+		if(ops != NULL && ops->bg)
+		{
+			args->cancellation.arg = ops->bg_op;
+			args->cancellation.hook = &bg_cancellation_hook;
+		}
+		else
+		{
+			/* ui_cancellation_reset() should be called outside this unit to allow
+			 * bulking several operations together. */
+			ui_cancellation_enable();
+			args->cancellation.hook = &ui_cancellation_hook;
+		}
 	}
 
 	curr_ops = ops;
 	result = func(args);
 	curr_ops = NULL;
 
-	if(args->cancellable)
+	if(cancellable && (ops == NULL || !ops->bg))
 	{
 		ui_cancellation_disable();
 	}
@@ -940,6 +980,10 @@ exec_io_op(ops_t *ops, int (*func)(io_args_t *const), io_args_t *const args)
 		size_t len = (ops->errors == NULL) ? 0U : strlen(ops->errors);
 		char *const suffix = ioe_errlst_to_str(&args->result.errors);
 
+		if(len != 0U)
+		{
+			(void)strappend(&ops->errors, &len, "\n");
+		}
 		(void)strappend(&ops->errors, &len, suffix);
 
 		free(suffix);
@@ -1073,12 +1117,12 @@ prompt_user(const io_args_t *args, const char title[], const char msg[],
 
 	/* Active cancellation conflicts with input processing by putting terminal in
 	 * a cooked mode. */
-	if(args->cancellable)
+	if(args->cancellation.hook != NULL)
 	{
 		raw();
 	}
 	response = prompt_msg_custom(title, msg, variants);
-	if(args->cancellable)
+	if(args->cancellation.hook != NULL)
 	{
 		noraw();
 	}
@@ -1086,5 +1130,54 @@ prompt_user(const io_args_t *args, const char title[], const char msg[],
 	return response;
 }
 
+/* Implementation of cancellation hook for I/O unit. */
+static int
+ui_cancellation_hook(void *arg)
+{
+	return ui_cancellation_requested();
+}
+
+#ifndef _WIN32
+
+/* Runs command in background and displays its errors to a user.  To determine
+ * an error uses both stderr stream and exit status.  Returns zero on success,
+ * otherwise non-zero is returned. */
+static int
+run_operation_command(ops_t *ops, char cmd[], int cancellable)
+{
+	if(!cancellable)
+	{
+		return bg_and_wait_for_errors(cmd, &no_cancellation);
+	}
+
+	if(ops != NULL && ops->bg)
+	{
+		const cancellation_t bg_cancellation_info = {
+			.arg = ops->bg_op,
+			.hook = &bg_cancellation_hook,
+		};
+		return bg_and_wait_for_errors(cmd, &bg_cancellation_info);
+	}
+	else
+	{
+		int result;
+		/* ui_cancellation_reset() should be called outside this unit to allow
+		 * bulking several operations together. */
+		ui_cancellation_enable();
+		result = bg_and_wait_for_errors(cmd, &ui_cancellation_info);
+		ui_cancellation_disable();
+		return result;
+	}
+}
+
+#endif
+
+/* Implementation of cancellation hook for background tasks. */
+static int
+bg_cancellation_hook(void *arg)
+{
+	return bg_op_cancelled(arg);
+}
+
 /* vim: set tabstop=2 softtabstop=2 shiftwidth=2 noexpandtab cinoptions-=(0 : */
 /* vim: set cinoptions+=t0 filetype=c : */
