@@ -23,7 +23,7 @@
 
 #include <assert.h> /* assert() */
 #include <ctype.h> /* isdigit() */
-#include <limits.h> /* INT_MIN */
+#include <limits.h> /* INT_MAX INT_MIN */
 #include <stddef.h> /* NULL size_t */
 #include <stdio.h> /* snprintf() */
 #include <stdlib.h> /* abs() free() */
@@ -35,7 +35,6 @@
 #include "engine/text_buffer.h"
 #include "int/term_title.h"
 #include "modes/view.h"
-#include "ui/column_view.h"
 #include "ui/fileview.h"
 #include "ui/quickview.h"
 #include "ui/statusbar.h"
@@ -49,6 +48,7 @@
 #include "utils/string_array.h"
 #include "utils/utils.h"
 #include "filelist.h"
+#include "flist_hist.h"
 #include "search.h"
 #include "sort.h"
 #include "status.h"
@@ -80,6 +80,7 @@ typedef struct
 }
 optinit_t;
 
+static void uni_handler(const char name[], optval_t val, OPT_SCOPE scope);
 static void init_classify(optval_t *val);
 static char * double_commas(const char str[]);
 static void init_cpoptions(optval_t *val);
@@ -87,8 +88,10 @@ static void init_dirsize(optval_t *val);
 static const char * to_endpoint(int i, char buffer[]);
 static void init_timefmt(optval_t *val);
 static void init_trashdir(optval_t *val);
+static void init_dotfiles(optval_t *val);
 static void init_lsview(optval_t *val);
 static void init_shortmess(optval_t *val);
+static void init_sizefmt(optval_t *val);
 static void init_iooptions(optval_t *val);
 static void init_number(optval_t *val);
 static void init_numberwidth(optval_t *val);
@@ -102,6 +105,7 @@ static void add_options(void);
 static void load_sort_option_inner(FileView *view, char sort_keys[]);
 static void aproposprg_handler(OPT_OP op, optval_t val);
 static void autochpos_handler(OPT_OP op, optval_t val);
+static void caseoptions_handler(OPT_OP op, optval_t val);
 static void cdpath_handler(OPT_OP op, optval_t val);
 static void chaselinks_handler(OPT_OP op, optval_t val);
 static void classify_handler(OPT_OP op, optval_t val);
@@ -109,6 +113,7 @@ static int str_to_classify(const char str[], char type_decs[FT_COUNT][2][9]);
 static const char * pick_out_decoration(char classify_item[], FileType *type,
 		const char **expr);
 static int validate_decorations(const char prefix[], const char suffix[]);
+static void free_file_decs(file_dec_t *name_decs, int count);
 static void columns_handler(OPT_OP op, optval_t val);
 static void confirm_handler(OPT_OP op, optval_t val);
 static void cpoptions_handler(OPT_OP op, optval_t val);
@@ -118,7 +123,7 @@ static void dirsize_handler(OPT_OP op, optval_t val);
 static void dotdirs_handler(OPT_OP op, optval_t val);
 static void fastrun_handler(OPT_OP op, optval_t val);
 static void fillchars_handler(OPT_OP op, optval_t val);
-static void reset_fillchars(void);
+static void load_fillchars(void);
 static void findprg_handler(OPT_OP op, optval_t val);
 static void followlinks_handler(OPT_OP op, optval_t val);
 static void fusehome_handler(OPT_OP op, optval_t val);
@@ -130,8 +135,6 @@ static void iec_handler(OPT_OP op, optval_t val);
 static void ignorecase_handler(OPT_OP op, optval_t val);
 static void incsearch_handler(OPT_OP op, optval_t val);
 static void iooptions_handler(OPT_OP op, optval_t val);
-static int parse_range(const char range[], int *from, int *to);
-static int parse_endpoint(const char **str, int *endpoint);
 static void laststatus_handler(OPT_OP op, optval_t val);
 static void lines_handler(OPT_OP op, optval_t val);
 static void locateprg_handler(OPT_OP op, optval_t val);
@@ -143,11 +146,15 @@ static void scrollbind_handler(OPT_OP op, optval_t val);
 static void scrolloff_handler(OPT_OP op, optval_t val);
 static void shell_handler(OPT_OP op, optval_t val);
 static void shortmess_handler(OPT_OP op, optval_t val);
+static void sizefmt_handler(OPT_OP op, optval_t val);
+static optval_t make_sizefmt_value(void);
 #ifndef _WIN32
 static void slowfs_handler(OPT_OP op, optval_t val);
 #endif
 static void smartcase_handler(OPT_OP op, optval_t val);
 static void sortnumbers_handler(OPT_OP op, optval_t val);
+static void dotfiles_global(OPT_OP op, optval_t val);
+static void dotfiles_local(OPT_OP op, optval_t val);
 static void lsview_global(OPT_OP op, optval_t val);
 static void lsview_local(OPT_OP op, optval_t val);
 static void number_global(OPT_OP op, optval_t val);
@@ -174,7 +181,7 @@ static void viewcolumns_local(OPT_OP op, optval_t val);
 static void set_viewcolumns(FileView *view, const char view_columns[]);
 static void set_view_columns_option(FileView *view, const char value[],
 		int update_ui);
-static void add_column(columns_t columns, column_info_t column_info);
+static void add_column(columns_t *columns, column_info_t column_info);
 static int map_name(const char name[], void *arg);
 static void resort_view(FileView * view);
 static void statusline_handler(OPT_OP op, optval_t val);
@@ -197,6 +204,8 @@ static void vimhelp_handler(OPT_OP op, optval_t val);
 static void wildmenu_handler(OPT_OP op, optval_t val);
 static void wildstyle_handler(OPT_OP op, optval_t val);
 static void wordchars_handler(OPT_OP op, optval_t val);
+static int parse_range(const char range[], int *from, int *to);
+static int parse_endpoint(const char **str, int *endpoint);
 static void wrap_handler(OPT_OP op, optval_t val);
 static void text_option_changed(void);
 static void wrapscan_handler(OPT_OP op, optval_t val);
@@ -230,6 +239,16 @@ static const char *sort_enum[] = {
 };
 ARRAY_GUARD(sort_enum, 1 + SK_COUNT);
 
+/* Possible values of 'caseoptions'. */
+static const char *caseoptions_vals[][2] = {
+	{ "pPgG", "all caseoptions values" },
+	{ "p", "always ignore case of paths during completion" },
+	{ "P", "always match case of paths during completion" },
+	{ "g", "always ignore case of characters for f/F/;/," },
+	{ "G", "always match case of characters for f/F/;/," },
+};
+ARRAY_GUARD(caseoptions_vals, 1 + NUM_CASE_OPTS*2);
+
 /* Possible values of 'cpoptions'. */
 static const char *cpoptions_vals[][2] = {
 	{ "fst", "all cpoptions values" },
@@ -301,6 +320,12 @@ static const char *fillchars_enum[][2] = {
 	{ "vborder:", "filler of vertical borders" },
 };
 
+/* Possible keys of 'sizefmt' option. */
+static const char *sizefmt_enum[][2] = {
+	{ "units:",     "iec (1024) or si (1000) unit size" },
+	{ "precision:", "maximum width of fraction part" },
+};
+
 /* Possible values of 'sort' option. */
 static const char *sort_types[][2] = {
 	{ "ext",  "by file/directory extension" },
@@ -454,6 +479,11 @@ options[] = {
 	  OPT_BOOL, 0, NULL, &autochpos_handler, NULL,
 	  { .ref.bool_val = &cfg.auto_ch_pos },
 	},
+	{ "caseoptions", "", "case sensitivity overrides",
+	  OPT_CHARSET, ARRAY_LEN(caseoptions_vals), caseoptions_vals,
+		&caseoptions_handler, NULL,
+	  { .ref.str_val = &empty },
+	},
 	{ "cdpath", "cd", "list of prefixes for relative cd",
 	  OPT_STRLIST, 0, NULL, &cdpath_handler, NULL,
 	  { .ref.str_val = &cfg.cd_path },
@@ -501,8 +531,8 @@ options[] = {
 	  { .ref.bool_val = &cfg.fast_run },
 	},
 	{ "fillchars", "fcs", "fillers for interface parts",
-		OPT_STRLIST, ARRAY_LEN(fillchars_enum), fillchars_enum, &fillchars_handler,
-		NULL,
+	  OPT_STRLIST, ARRAY_LEN(fillchars_enum), fillchars_enum, &fillchars_handler,
+	  NULL,
 	  { .ref.str_val = &empty },
 	},
 	{ "findprg", "", ":find invocation format",
@@ -535,7 +565,7 @@ options[] = {
 	},
 	{ "iec", "", "use IEC notation for sizes",
 	  OPT_BOOL, 0, NULL, &iec_handler, NULL,
-	  { .ref.bool_val = &cfg.use_iec_prefixes },
+	  { .ref.bool_val = &cfg.sizefmt.ieci_prefixes },
 	},
 	{ "ignorecase", "ic", "ignore case on searches",
 	  OPT_BOOL, 0, NULL, &ignorecase_handler, NULL,
@@ -591,6 +621,10 @@ options[] = {
 		NULL,
 	  { .init = &init_shortmess },
 	},
+	{ "sizefmt", "", "human-friendly size format",
+	  OPT_STRLIST, ARRAY_LEN(sizefmt_enum), sizefmt_enum, &sizefmt_handler, NULL,
+	  { .init = &init_sizefmt },
+	},
 #ifndef _WIN32
 	{ "slowfs", "", "list of slow filesystem types",
 	  OPT_STRLIST, 0, NULL, &slowfs_handler, NULL,
@@ -689,7 +723,11 @@ options[] = {
 	  { .ref.bool_val = &cfg.wrap_scan },
 	},
 
-	/* Local options. */
+	/* Local options must be grouped here. */
+	{ "dotfiles", "", "show dot files",
+	  OPT_BOOL, 0, NULL, &dotfiles_global, &dotfiles_local,
+	  { .init = &init_dotfiles },
+	},
 	{ "lsview", "", "display items in a table",
 	  OPT_BOOL, 0, NULL, &lsview_global, &lsview_local,
 	  { .init = &init_lsview },
@@ -731,11 +769,80 @@ void
 init_option_handlers(void)
 {
 	static int opt_changed;
-	init_options(&opt_changed);
+	init_options(&opt_changed, &uni_handler);
 	load_options_defaults();
 	add_options();
 }
 
+/* Additional handler for processing of global-local options, namely for
+ * updating the other view. */
+static void
+uni_handler(const char name[], optval_t val, OPT_SCOPE scope)
+{
+	/* += and similar operators act in their way only for the current view, the
+	 * other view just gets value resulted from the operation.  The behaviour is
+	 * fine and doesn't seem to be a bug. */
+
+	static size_t first_local = (size_t)-1;
+
+	size_t i;
+
+	if(!curr_stats.global_local_settings)
+	{
+		return;
+	}
+
+	/* Find first local option once. */
+	if(first_local == (size_t)-1)
+	{
+		for(first_local = 0U; first_local < ARRAY_LEN(options); ++first_local)
+		{
+			if(options[first_local].local_handler != NULL)
+			{
+				break;
+			}
+		}
+	}
+
+	/* Look up option name and update it in the other view if found. */
+	for(i = first_local; i < ARRAY_LEN(options); ++i)
+	{
+		if(strcmp(options[i].name, name) == 0)
+		{
+			FileView *const tmp_view = curr_view;
+			curr_view = other_view;
+
+			/* Make sure option value remains valid even if updated in the handler.
+			 * Do this before calling load_view_options(), because it can change the
+			 * value. */
+			if(ONE_OF(options[i].type, OPT_STR, OPT_STRLIST, OPT_CHARSET))
+			{
+				val.str_val = strdup(val.str_val);
+			}
+
+			load_view_options(curr_view);
+
+			if(scope == OPT_LOCAL)
+			{
+				options[i].local_handler(OP_SET, val);
+			}
+			else
+			{
+				options[i].global_handler(OP_SET, val);
+			}
+
+			if(ONE_OF(options[i].type, OPT_STR, OPT_STRLIST, OPT_CHARSET))
+			{
+				free(val.str_val);
+			}
+
+			curr_view = tmp_view;
+			load_view_options(curr_view);
+			break;
+		}
+	}
+}
+
 /* Composes the default value for the 'classify' option. */
 static void
 init_classify(optval_t *val)
@@ -860,6 +967,13 @@ init_trashdir(optval_t *val)
 	val->str_val = cfg.trash_dir;
 }
 
+/* Initializes 'dotfiles' option from global value. */
+static void
+init_dotfiles(optval_t *val)
+{
+	val->bool_val = !curr_view->hide_dot_g;
+}
+
 static void
 init_lsview(optval_t *val)
 {
@@ -876,6 +990,13 @@ init_shortmess(optval_t *val)
 	val->str_val = buf;
 }
 
+/* Initializes 'sizefmt' from current configuration state. */
+static void
+init_sizefmt(optval_t *val)
+{
+	*val = make_sizefmt_value();
+}
+
 /* Initializes value of 'iooptions' from configuration. */
 static void
 init_iooptions(optval_t *val)
@@ -995,7 +1116,7 @@ load_options_defaults(void)
 		{
 			options[i].initializer.init(&options[i].val);
 		}
-		else if(options[i].type == OPT_STR || options[i].type == OPT_STRLIST)
+		else if(ONE_OF(options[i].type, OPT_STR, OPT_STRLIST, OPT_CHARSET))
 		{
 			options[i].val.str_val = *options[i].initializer.ref.str_val;
 		}
@@ -1034,6 +1155,10 @@ reset_local_options(FileView *view)
 	memcpy(view->sort, view->sort_g, sizeof(view->sort));
 	load_sort_option_inner(view, view->sort);
 
+	view->hide_dot = view->hide_dot_g;
+	val.int_val = !view->hide_dot_g;
+	set_option("dotfiles", val, OPT_LOCAL);
+
 	fview_set_lsview(view, view->ls_view_g);
 	val.int_val = view->ls_view_g;
 	set_option("lsview", val, OPT_LOCAL);
@@ -1075,6 +1200,11 @@ load_view_options(FileView *view)
 	val.str_val = view->sort_groups_g;
 	set_option("sortgroups", val, OPT_GLOBAL);
 
+	val.bool_val = !view->hide_dot;
+	set_option("dotfiles", val, OPT_LOCAL);
+	val.bool_val = !view->hide_dot_g;
+	set_option("dotfiles", val, OPT_GLOBAL);
+
 	val.bool_val = view->ls_view;
 	set_option("lsview", val, OPT_LOCAL);
 	val.bool_val = view->ls_view_g;
@@ -1110,10 +1240,13 @@ clone_local_options(const FileView *from, FileView *to, int defer_slow)
 	to->num_type = from->num_type;
 	to->num_type_g = from->num_type_g;
 
+	to->hide_dot = from->hide_dot;
+	to->hide_dot_g = from->hide_dot_g;
+
 	replace_string(&to->sort_groups, from->sort_groups);
 	replace_string(&to->sort_groups_g, from->sort_groups_g);
 
-	sort = (flist_custom_active(from) && from->custom.unsorted)
+	sort = (flist_custom_active(from) && ui_view_unsorted(from))
 	     ? (char *)from->custom.sort
 	     : (char *)from->sort;
 	memcpy(to->sort, sort, sizeof(to->sort));
@@ -1131,9 +1264,12 @@ load_sort_option(FileView *view)
 	load_sort_option_inner(view, view->sort_g);
 
 	/* The check is to skip this in tests which don't need columns. */
-	if(view->columns != NULL_COLUMNS)
+	if(view->columns != NULL)
 	{
-		set_viewcolumns(view, view->view_columns);
+		if(!cv_compare(view->custom.type))
+		{
+			set_viewcolumns(view, view->view_columns);
+		}
 	}
 }
 
@@ -1152,6 +1288,13 @@ load_sort_option_inner(FileView *view, char sort_keys[])
 	size_t opt_val_len = 0U;
 	OPT_SCOPE scope = (sort_keys == view->sort) ? OPT_LOCAL : OPT_GLOBAL;
 
+	if(sort_keys == view->custom.sort)
+	{
+		val.str_val = "";
+		set_option("sort", val, OPT_LOCAL);
+		return;
+	}
+
 	opt_val[0] = '\0';
 
 	ui_view_sort_list_ensure_well_formed(view, sort_keys);
@@ -1165,9 +1308,9 @@ load_sort_option_inner(FileView *view, char sort_keys[])
 		const char option_mark = (sort_option < 0) ? '-' : '+';
 		const char *const option_name = sort_enum[abs(sort_option)];
 
-		opt_val_len += snprintf(opt_val + opt_val_len,
-				sizeof(opt_val) - opt_val_len, "%s%c%s", comma, option_mark,
-				option_name);
+		snprintf(opt_val + opt_val_len, sizeof(opt_val) - opt_val_len, "%s%c%s",
+				comma, option_mark, option_name);
+		opt_val_len += strlen(opt_val + opt_val_len);
 	}
 
 	val.str_val = opt_val;
@@ -1212,6 +1355,7 @@ process_set_args(const char args[], int global, int local)
 	{
 		status_bar_message(text_buffer);
 	}
+
 	return error ? -1 : (text_buffer[0] != '\0');
 }
 
@@ -1227,11 +1371,57 @@ autochpos_handler(OPT_OP op, optval_t val)
 	cfg.auto_ch_pos = val.bool_val;
 	if(curr_stats.load_stage < 2)
 	{
-		clean_positions_in_history(curr_view);
-		clean_positions_in_history(other_view);
+		flist_hist_clear(curr_view);
+		flist_hist_clear(other_view);
 	}
 }
 
+/* Handles changes of 'caseoptions' option.  Updates configuration and
+ * normalizes option value. */
+static void
+caseoptions_handler(OPT_OP op, optval_t val)
+{
+	char valid_val[32];
+	const char *p;
+
+	cfg.case_override = 0;
+	cfg.case_ignore = 0;
+
+	p = val.str_val;
+	while(*p != '\0')
+	{
+		if(*p == 'P' || *p == 'p')
+		{
+			cfg.case_override |= CO_PATH_COMPL;
+			cfg.case_ignore = (*p == 'p')
+			                ? (cfg.case_ignore | CO_PATH_COMPL)
+			                : (cfg.case_ignore & ~CO_PATH_COMPL);
+		}
+		else if(*p == 'g' || *p == 'G')
+		{
+			cfg.case_override |= CO_GOTO_FILE;
+			cfg.case_ignore = (*p == 'g')
+			                ? (cfg.case_ignore | CO_GOTO_FILE)
+			                : (cfg.case_ignore & ~CO_GOTO_FILE);
+		}
+
+		++p;
+	}
+
+	valid_val[0] = '\0';
+	if(cfg.case_override & CO_PATH_COMPL)
+	{
+		strcat(valid_val, (cfg.case_ignore & CO_PATH_COMPL) ? "p" : "P");
+	}
+	if(cfg.case_override & CO_GOTO_FILE)
+	{
+		strcat(valid_val, (cfg.case_ignore & CO_GOTO_FILE) ? "g" : "G");
+	}
+	val.str_val = valid_val;
+
+	set_option("caseoptions", val, OPT_GLOBAL);
+}
+
 /* Specifies directories to check on cding by relative path. */
 static void
 cdpath_handler(OPT_OP op, optval_t val)
@@ -1300,7 +1490,6 @@ str_to_classify(const char str[], char type_decs[FT_COUNT][2][9])
 	int error_encountered;
 	file_dec_t *name_decs = NULL;
 	DA_INSTANCE(name_decs);
-	int i;
 
 	str_copy = strdup(str);
 	if(str_copy == NULL)
@@ -1372,14 +1561,16 @@ str_to_classify(const char str[], char type_decs[FT_COUNT][2][9])
 	}
 	free(str_copy);
 
-	for(i = 0; i < cfg.name_dec_count; ++i)
+	if(error_encountered)
 	{
-		matchers_free(cfg.name_decs[i].matchers);
+		free_file_decs(name_decs, DA_SIZE(name_decs));
+	}
+	else
+	{
+		free_file_decs(cfg.name_decs, cfg.name_dec_count);
+		cfg.name_decs = name_decs;
+		cfg.name_dec_count = DA_SIZE(name_decs);
 	}
-	free(cfg.name_decs);
-
-	cfg.name_decs = name_decs;
-	cfg.name_dec_count = DA_SIZE(name_decs);
 
 	return error_encountered;
 }
@@ -1438,6 +1629,19 @@ validate_decorations(const char prefix[], const char suffix[])
 	return error;
 }
 
+/* Frees array of file declarations of length count (both elements and array
+ * itself). */
+static void
+free_file_decs(file_dec_t *name_decs, int count)
+{
+	int i;
+	for(i = 0; i < count; ++i)
+	{
+		matchers_free(name_decs[i].matchers);
+	}
+	free(name_decs);
+}
+
 /* Handles updates of the global 'columns' option, which reflects width of
  * terminal. */
 static void
@@ -1555,8 +1759,9 @@ fillchars_handler(OPT_OP op, optval_t val)
 	char *new_val = strdup(val.str_val);
 	char *part = new_val, *state = NULL;
 
-	/* XXX: we shouldn't reset this value unless format is correct. */
-	(void)replace_string(&cfg.border_filler, " ");
+	/* Save current state. */
+	char *const old_border = cfg.border_filler;
+	cfg.border_filler = NULL;
 
 	while((part = split_and_get(part, ',', &state)) != NULL)
 	{
@@ -1574,21 +1779,34 @@ fillchars_handler(OPT_OP op, optval_t val)
 	}
 	free(new_val);
 
-	if(part != NULL)
+	/* Restore previous state of unset elements or drop saved state. */
+	if(cfg.border_filler == NULL)
 	{
-		reset_fillchars();
+		cfg.border_filler = old_border;
+	}
+	else
+	{
+		free(old_border);
 	}
 
-	curr_stats.need_update = UT_REDRAW;
+	/* In case of error, restore previous value, otherwise reload it anyway to
+	 * remove any duplicates. */
+	load_fillchars();
+
+	if(part == NULL)
+	{
+		/* Schedule a redraw if option was set successfully. */
+		curr_stats.need_update = UT_REDRAW;
+	}
 }
 
-/* Resets value of 'fillchars' option by composing it from current
+/* Sets value of 'fillchars' option by composing it from current
  * configuration. */
 static void
-reset_fillchars(void)
+load_fillchars(void)
 {
-	optval_t val;
 	char value[128];
+	const optval_t val = { .str_val = value };
 
 	value[0] = '\0';
 
@@ -1597,7 +1815,6 @@ reset_fillchars(void)
 		snprintf(value, sizeof(value), "vborder:%s", cfg.border_filler);
 	}
 
-	val.str_val = value;
 	set_option("fillchars", val, OPT_GLOBAL);
 }
 
@@ -1659,7 +1876,7 @@ hlsearch_handler(OPT_OP op, optval_t val)
 static void
 iec_handler(OPT_OP op, optval_t val)
 {
-	cfg.use_iec_prefixes = val.bool_val;
+	cfg.sizefmt.ieci_prefixes = val.bool_val;
 
 	redraw_lists();
 }
@@ -1683,72 +1900,6 @@ iooptions_handler(OPT_OP op, optval_t val)
 	cfg.fast_file_cloning = ((val.set_items & 1) != 0);
 }
 
-/* Parses range, which can be shortened to single endpoint if first element
- * matches last one.  Returns non-zero on error, otherwise zero is returned. */
-static int
-parse_range(const char range[], int *from, int *to)
-{
-	const char *p = range;
-
-	if(parse_endpoint(&p, from) != 0)
-	{
-		vle_tb_append_linef(vle_err, "Wrong range: %s", range);
-		return 1;
-	}
-	if(*p == '-')
-	{
-		++p;
-		if(parse_endpoint(&p, to) != 0)
-		{
-			vle_tb_append_linef(vle_err, "Wrong range: %s", range);
-			return 1;
-		}
-	}
-	else
-	{
-		*to = *from;
-	}
-
-	if(*p != '\0')
-	{
-		vle_tb_append_linef(vle_err, "Wrong range: %s", range);
-		return 1;
-	}
-
-	if(*from > *to)
-	{
-		vle_tb_append_linef(vle_err, "Inversed range: %s", range);
-		return 1;
-	}
-
-	return 0;
-}
-
-/* Parses single endpoint of a range.  It's either a number or a character.
- * Returns non-zero on error, otherwise zero is returned. */
-static int
-parse_endpoint(const char **str, int *endpoint)
-{
-	if(isdigit(**str))
-	{
-		char *endptr;
-		const long int val = strtol(*str, &endptr, 10);
-		if(val < 0 || val >= 256)
-		{
-			vle_tb_append_linef(vle_err, "Wrong value: %ld", val);
-			return 1;
-		}
-		*str = endptr;
-		*endpoint = val;
-	}
-	else
-	{
-		*endpoint = **str;
-		++*str;
-	}
-	return 0;
-}
-
 static void
 laststatus_handler(OPT_OP op, optval_t val)
 {
@@ -1905,6 +2056,90 @@ shortmess_handler(OPT_OP op, optval_t val)
 	ui_views_update_titles();
 }
 
+/* Handles changes of 'sizefmt' option by parsing string value and updating
+ * configuration values. */
+static void
+sizefmt_handler(OPT_OP op, optval_t val)
+{
+	char *const new_val = strdup(val.str_val);
+	char *part = new_val, *state = NULL;
+
+	int base = -1, precision = 0;
+
+	while((part = split_and_get(part, ',', &state)) != NULL)
+	{
+		if(starts_with_lit(part, "units:"))
+		{
+			const char *const val = after_first(part, ':');
+			if(strcmp(val, "iec") == 0)
+			{
+				base = 1024;
+			}
+			else if(strcmp(val, "si") == 0)
+			{
+				base = 1000;
+			}
+			else
+			{
+				vle_tb_append_linef(vle_err, "Invalid units value: %s", val);
+				break;
+			}
+		}
+		else if(starts_with_lit(part, "precision:"))
+		{
+			const char *const val = after_first(part, ':');
+			if(!read_int(val, &precision) || precision <= 0)
+			{
+				vle_tb_append_linef(vle_err, "Invalid precision value: %s", val);
+				break;
+			}
+		}
+		else
+		{
+			break_at(part, ':');
+			vle_tb_append_linef(vle_err, "Unknown key for 'sizefmt' option: %s",
+					part);
+			break;
+		}
+	}
+	free(new_val);
+
+	if(part == NULL && base != -1)
+	{
+		cfg.sizefmt.base = base;
+		cfg.sizefmt.precision = precision;
+
+		curr_stats.need_update = UT_REDRAW;
+	}
+	else if(base == -1)
+	{
+		vle_tb_append_line(vle_err, "'sizefmt' value is missing units type");
+	}
+
+	/* In case of error, restore previous value, otherwise reload it anyway to
+	 * remove any duplicates. */
+	set_option("sizefmt", make_sizefmt_value(), OPT_GLOBAL);
+}
+
+/* Makes string value describing 'sizefmt' from current configuration state.
+ * The value is statically allocated.  Returns the value. */
+static optval_t
+make_sizefmt_value(void)
+{
+	static char value[128];
+	static const optval_t val = { .str_val = value };
+
+	const int len = snprintf(value, sizeof(value), "units:%s",
+			cfg.sizefmt.base == 1024 ? "iec" : "si");
+	if(cfg.sizefmt.precision != 0)
+	{
+		snprintf(value + len, sizeof(value) - len, ",precision:%d",
+				cfg.sizefmt.precision);
+	}
+
+	return val;
+}
+
 #ifndef _WIN32
 static void
 slowfs_handler(OPT_OP op, optval_t val)
@@ -1928,15 +2163,26 @@ sortnumbers_handler(OPT_OP op, optval_t val)
 	redraw_lists();
 }
 
+/* Handles switch that controls visibility of dot files globally. */
+static void
+dotfiles_global(OPT_OP op, optval_t val)
+{
+	curr_view->hide_dot_g = !val.bool_val;
+}
+
+/* Handles switch that controls visibility of dot files locally. */
+static void
+dotfiles_local(OPT_OP op, optval_t val)
+{
+	curr_view->hide_dot = !val.bool_val;
+	ui_view_schedule_reload(curr_view);
+}
+
 /* Handles switch that controls column vs. ls-like view in global option. */
 static void
 lsview_global(OPT_OP op, optval_t val)
 {
 	curr_view->ls_view_g = val.bool_val;
-	if(curr_stats.global_local_settings)
-	{
-		other_view->ls_view_g = val.bool_val;
-	}
 }
 
 /* Handles switch that controls column vs. ls-like view in local option. */
@@ -1944,10 +2190,6 @@ static void
 lsview_local(OPT_OP op, optval_t val)
 {
 	fview_set_lsview(curr_view, val.bool_val);
-	if(curr_stats.global_local_settings)
-	{
-		fview_set_lsview(other_view, val.bool_val);
-	}
 }
 
 /* Handles file numbers displaying toggle in global option. */
@@ -1955,10 +2197,6 @@ static void
 number_global(OPT_OP op, optval_t val)
 {
 	update_num_type(curr_view, &curr_view->num_type_g, NT_SEQ, val.bool_val);
-	if(curr_stats.global_local_settings)
-	{
-		update_num_type(other_view, &other_view->num_type_g, NT_SEQ, val.bool_val);
-	}
 }
 
 /* Handles file numbers displaying toggle in local option. */
@@ -1966,10 +2204,6 @@ static void
 number_local(OPT_OP op, optval_t val)
 {
 	update_num_type(curr_view, &curr_view->num_type, NT_SEQ, val.bool_val);
-	if(curr_stats.global_local_settings)
-	{
-		update_num_type(other_view, &other_view->num_type, NT_SEQ, val.bool_val);
-	}
 }
 
 /* Handles changes of minimum width of file number field in global option. */
@@ -1977,10 +2211,6 @@ static void
 numberwidth_global(OPT_OP op, optval_t val)
 {
 	set_numberwidth(curr_view, &curr_view->num_width_g, val.int_val);
-	if(curr_stats.global_local_settings)
-	{
-		set_numberwidth(other_view, &other_view->num_width_g, val.int_val);
-	}
 }
 
 /* Handles changes of minimum width of file number field in local option. */
@@ -1988,10 +2218,6 @@ static void
 numberwidth_local(OPT_OP op, optval_t val)
 {
 	set_numberwidth(curr_view, &curr_view->num_width, val.int_val);
-	if(curr_stats.global_local_settings)
-	{
-		set_numberwidth(other_view, &other_view->num_width, val.int_val);
-	}
 }
 
 /* Sets number width for the view. */
@@ -2011,10 +2237,6 @@ static void
 relativenumber_global(OPT_OP op, optval_t val)
 {
 	update_num_type(curr_view, &curr_view->num_type_g, NT_REL, val.bool_val);
-	if(curr_stats.global_local_settings)
-	{
-		update_num_type(other_view, &other_view->num_type_g, NT_REL, val.bool_val);
-	}
 }
 
 /* Handles relative file numbers displaying toggle in local option. */
@@ -2022,10 +2244,6 @@ static void
 relativenumber_local(OPT_OP op, optval_t val)
 {
 	update_num_type(curr_view, &curr_view->num_type, NT_REL, val.bool_val);
-	if(curr_stats.global_local_settings)
-	{
-		update_num_type(other_view, &other_view->num_type, NT_REL, val.bool_val);
-	}
 }
 
 /* Handles toggling of boolean number related option and updates current view if
@@ -2051,15 +2269,7 @@ update_num_type(FileView *view, NumberingType *num_type, NumberingType type,
 static void
 sort_global(OPT_OP op, optval_t val)
 {
-	char *const value = strdup(val.str_val);
-
-	set_sort(curr_view, curr_view->sort_g, value);
-	if(curr_stats.global_local_settings)
-	{
-		set_sort(other_view, other_view->sort_g, value);
-	}
-
-	free(value);
+	set_sort(curr_view, curr_view->sort_g, val.str_val);
 }
 
 /* Handler for local 'sort' option, parses the value and checks it for
@@ -2067,23 +2277,13 @@ sort_global(OPT_OP op, optval_t val)
 static void
 sort_local(OPT_OP op, optval_t val)
 {
-	char *const value = strdup(val.str_val);
-
-	/* Make sure we don't sort unsorted custom view on :restart. */
-	char *const sort = curr_stats.restart_in_progress
-	                 ? ui_view_sort_list_get(curr_view)
+	/* Make sure we don't sort unsorted custom view on :restart or when it's a
+	 * compare view. */
+	char *const sort = (curr_stats.restart_in_progress ||
+	                    cv_compare(curr_view->custom.type))
+	                 ? ui_view_sort_list_get(curr_view, curr_view->sort)
 	                 : curr_view->sort;
-	set_sort(curr_view, sort, value);
-	if(curr_stats.global_local_settings)
-	{
-		/* Make sure we don't sort unsorted custom view on :restart. */
-		char *const sort = curr_stats.restart_in_progress
-		                 ? ui_view_sort_list_get(other_view)
-		                 : other_view->sort;
-		set_sort(other_view, sort, value);
-	}
-
-	free(value);
+	set_sort(curr_view, sort, val.str_val);
 }
 
 /* Sets sorting value for the view. */
@@ -2153,32 +2353,15 @@ set_sort(FileView *view, char sort_keys[], char order[])
 static void
 sortgroups_global(OPT_OP op, optval_t val)
 {
-	char *const value = strdup(val.str_val);
-
-	set_sortgroups(curr_view, &curr_view->sort_groups_g, value);
-	if(curr_stats.global_local_settings)
-	{
-		set_sortgroups(other_view, &other_view->sort_groups_g, value);
-	}
-
-	free(value);
+	set_sortgroups(curr_view, &curr_view->sort_groups_g, val.str_val);
 }
 
 /* Handles local 'sortgroup' option changes. */
 static void
 sortgroups_local(OPT_OP op, optval_t val)
 {
-	char *const value = strdup(val.str_val);
-
-	set_sortgroups(curr_view, &curr_view->sort_groups, value);
+	set_sortgroups(curr_view, &curr_view->sort_groups, val.str_val);
 	sorting_changed(curr_view, 0);
-	if(curr_stats.global_local_settings)
-	{
-		set_sortgroups(other_view, &other_view->sort_groups, value);
-		sorting_changed(other_view, 0);
-	}
-
-	free(value);
 }
 
 /* Sets sort_groups fields (*opt) of the view to the value handling malformed
@@ -2251,10 +2434,6 @@ static void
 sortorder_global(OPT_OP op, optval_t val)
 {
 	set_sortorder(curr_view, (val.enum_item == 1) ? 0 : 1, curr_view->sort_g);
-	if(curr_stats.global_local_settings)
-	{
-		set_sortorder(other_view, (val.enum_item == 1) ? 0 : 1, other_view->sort_g);
-	}
 }
 
 /* Handles local 'sortorder' option and corrects ordering for primary sorting
@@ -2263,10 +2442,6 @@ static void
 sortorder_local(OPT_OP op, optval_t val)
 {
 	set_sortorder(curr_view, (val.enum_item == 1) ? 0 : 1, curr_view->sort);
-	if(curr_stats.global_local_settings)
-	{
-		set_sortorder(other_view, (val.enum_item == 1) ? 0 : 1, other_view->sort);
-	}
 }
 
 /* Updates sorting order for the view. */
@@ -2289,30 +2464,14 @@ set_sortorder(FileView *view, int ascending, char sort_keys[])
 static void
 viewcolumns_global(OPT_OP op, optval_t val)
 {
-	char *const value = strdup(val.str_val);
-
-	replace_string(&curr_view->view_columns_g, value);
-	if(curr_stats.global_local_settings)
-	{
-		replace_string(&other_view->view_columns_g, value);
-	}
-
-	free(value);
+	replace_string(&curr_view->view_columns_g, val.str_val);
 }
 
 /* Handler of local 'viewcolumns' option, which defines custom view columns. */
 static void
 viewcolumns_local(OPT_OP op, optval_t val)
 {
-	char *const value = strdup(val.str_val);
-
-	set_viewcolumns(curr_view, value);
-	if(curr_stats.global_local_settings)
-	{
-		set_viewcolumns(other_view, value);
-	}
-
-	free(value);
+	set_viewcolumns(curr_view, val.str_val);
 }
 
 /* Setups view columns for the view. */
@@ -2323,6 +2482,14 @@ set_viewcolumns(FileView *view, const char view_columns[])
 	set_view_columns_option(view, view_columns, update_columns_ui);
 }
 
+void
+load_dot_filter_option(const FileView *view)
+{
+	const optval_t val = { .bool_val = !view->hide_dot };
+	set_option("dotfiles", val, OPT_GLOBAL);
+	set_option("dotfiles", val, OPT_LOCAL);
+}
+
 void
 load_view_columns_option(FileView *view, const char value[])
 {
@@ -2336,7 +2503,7 @@ static void
 set_view_columns_option(FileView *view, const char value[], int update_ui)
 {
 	const char *new_value = (value[0] == '\0') ? DEFAULT_VIEW_COLUMNS : value;
-	const columns_t columns = update_ui ? view->columns : NULL;
+	columns_t *columns = update_ui ? view->columns : NULL;
 
 	if(update_ui)
 	{
@@ -2376,7 +2543,7 @@ set_view_columns_option(FileView *view, const char value[], int update_ui)
 
 /* Adds new column to view columns. */
 static void
-add_column(columns_t columns, column_info_t column_info)
+add_column(columns_t *columns, column_info_t column_info)
 {
 	/* Handle dry run mode, when we don't actually update column view. */
 	if(columns != NULL)
@@ -2395,20 +2562,7 @@ map_name(const char name[], void *arg)
 	if(*name == '\0')
 	{
 		const FileView *const view = arg;
-
-		const char *sort;
-
-		if(curr_stats.restart_in_progress)
-		{
-			sort = ui_view_sort_list_get(view);
-		}
-		else
-		{
-			sort = (flist_custom_active(view) && view->custom.unsorted)
-			      ? (char *)view->custom.sort
-			      : (char *)view->sort;
-		}
-
+		const char *const sort = ui_view_sort_list_get(view, view->sort);
 		return (int)get_secondary_key((SortingKey)abs(sort[0]));
 	}
 
@@ -2516,8 +2670,8 @@ suggestoptions_handler(OPT_OP op, optval_t val)
 	}
 }
 
-/* Converts line to a number.  Handles overflow/underflow.  Returns non-zero on
- * success and zero otherwise. */
+/* Converts line to a number.  Handles overflow/underflow by saturating
+ * resulting value.  Returns non-zero on success and zero otherwise. */
 static int
 read_int(const char line[], int *i)
 {
@@ -2754,7 +2908,7 @@ wordchars_handler(OPT_OP op, optval_t val)
 			break;
 		}
 
-		while(from < to)
+		while(from <= to)
 		{
 			word_chars[from++] = 1;
 		}
@@ -2767,6 +2921,78 @@ wordchars_handler(OPT_OP op, optval_t val)
 	}
 }
 
+/* Parses range, which can be shortened to single endpoint if first element
+ * matches last one.  Returns non-zero on error, otherwise zero is returned. */
+static int
+parse_range(const char range[], int *from, int *to)
+{
+	const char *p = range;
+
+	if(parse_endpoint(&p, from) != 0)
+	{
+		vle_tb_append_linef(vle_err, "Wrong range: %s", range);
+		return 1;
+	}
+	if(*p == '-')
+	{
+		++p;
+		if(parse_endpoint(&p, to) != 0)
+		{
+			vle_tb_append_linef(vle_err, "Wrong range: %s", range);
+			return 1;
+		}
+	}
+	else
+	{
+		*to = *from;
+	}
+
+	if(*p != '\0')
+	{
+		vle_tb_append_linef(vle_err, "Wrong range: %s", range);
+		return 1;
+	}
+
+	if(*from > *to)
+	{
+		vle_tb_append_linef(vle_err, "Inversed range: %s", range);
+		return 1;
+	}
+
+	return 0;
+}
+
+/* Parses single endpoint of a range.  It's either a number or a character.
+ * Returns non-zero on error, otherwise zero is returned. */
+static int
+parse_endpoint(const char **str, int *endpoint)
+{
+	if(**str == '\0')
+	{
+		vle_tb_append_line(vle_err, "Range bound can't be empty");
+		return 1;
+	}
+
+	if(isdigit(**str))
+	{
+		char *endptr;
+		const long int val = strtol(*str, &endptr, 10);
+		if(val < 0 || val >= 256)
+		{
+			vle_tb_append_linef(vle_err, "Wrong value: %ld", val);
+			return 1;
+		}
+		*str = endptr;
+		*endpoint = val;
+	}
+	else
+	{
+		*endpoint = (unsigned char)**str;
+		++*str;
+	}
+	return 0;
+}
+
 static void
 wrap_handler(OPT_OP op, optval_t val)
 {
@@ -2781,7 +3007,7 @@ text_option_changed(void)
 {
 	if(curr_stats.view)
 	{
-		quick_view_file(curr_view);
+		qv_draw(curr_view);
 	}
 	else if(other_view->explore_mode)
 	{
