@@ -41,8 +41,6 @@ enum
 	VIFMINFO_FHISTORY  = 1 << 15,
 };
 
-const char * cursorline_enum[3];
-
 void init_option_handlers(void);
 
 /* Resets local options so that they use values of global options. */
@@ -62,8 +60,12 @@ void clone_local_options(const FileView *from, FileView *to, int defer_slow);
  * message. */
 int process_set_args(const char args[], int global, int local);
 
+/* Loads current sorting of the view into an option. */
 void load_sort_option(FileView *view);
 
+/* Loads current value of dot filter into an option. */
+void load_dot_filter_option(const FileView *view);
+
 /* Updates view columns value as if 'viewcolumns' option has been changed.
  * Doesn't change actual value of the option, which is important for setting
  * sorting order via sort dialog. */
