@@ -51,7 +51,9 @@
 #include "int/vim.h"
 #include "menus/users_menu.h"
 #include "modes/dialogs/msg_dialog.h"
+#include "modes/view.h"
 #include "ui/statusbar.h"
+#include "ui/quickview.h"
 #include "ui/ui.h"
 #include "utils/env.h"
 #include "utils/fs.h"
@@ -65,6 +67,9 @@
 #include "background.h"
 #include "filelist.h"
 #include "filetype.h"
+#include "flist_hist.h"
+#include "flist_pos.h"
+#include "flist_sel.h"
 #include "macros.h"
 #include "opt_handlers.h"
 #include "status.h"
@@ -93,7 +98,6 @@ static int run_win_executable_as_evaluated(const char full_path[]);
 static int selection_is_consistent(FileView *view);
 static void execute_file(const char full_path[], int elevate);
 static void run_selection(FileView *view, int dont_execute);
-static void run_file(FileView *view, int dont_execute);
 static void run_with_defaults(FileView *view);
 static void run_selection_separately(FileView *view, int dont_execute);
 static int is_multi_run_compat(FileView *view, const char prog_cmd[]);
@@ -116,6 +120,7 @@ static void set_pwd_in_screen(const char path[]);
 static int try_run_with_filetype(FileView *view, const assoc_records_t assocs,
 		const char start[], int background);
 static void output_to_statusbar(const char cmd[]);
+static int output_to_preview(const char cmd[]);
 static void output_to_nowhere(const char cmd[]);
 static void run_in_split(const FileView *view, const char cmd[]);
 static void path_handler(const char line[], void *arg);
@@ -136,14 +141,17 @@ follow_file(FileView *view)
 {
 	if(flist_custom_active(view))
 	{
-		/* Entry might be freed on navigation, so make sure name and origin will
-		 * remain available for the call. */
-		const dir_entry_t *const entry = &view->dir_entry[view->list_pos];
-		char *const name = strdup(entry->name);
-		char *const origin = strdup(entry->origin);
-		navigate_to_file(view, origin, name, 0);
-		free(origin);
-		free(name);
+		const dir_entry_t *const curr = get_current_entry(view);
+		if(!fentry_is_fake(curr))
+		{
+			/* Entry might be freed on navigation, so make sure name and origin will
+			 * remain available for the call. */
+			char *const name = strdup(curr->name);
+			char *const origin = strdup(curr->origin);
+			navigate_to_file(view, origin, name, 0);
+			free(origin);
+			free(name);
+		}
 		return;
 	}
 
@@ -156,7 +164,12 @@ handle_file(FileView *view, FileHandleExec exec, FileHandleLink follow)
 	char full_path[PATH_MAX];
 	int executable;
 	int runnable;
-	const dir_entry_t *const curr = &view->dir_entry[view->list_pos];
+	const dir_entry_t *const curr = get_current_entry(view);
+
+	if(fentry_is_fake(curr))
+	{
+		return;
+	}
 
 	get_full_path_of(curr, sizeof(full_path), full_path);
 
@@ -375,21 +388,7 @@ execute_file(const char full_path[], int elevate)
 static void
 run_selection(FileView *view, int dont_execute)
 {
-	if(selection_is_consistent(view))
-	{
-		run_file(view, dont_execute);
-	}
-	else
-	{
-		show_error_msg("Selection error",
-				"Selection cannot contain files and directories at the same time");
-	}
-}
-
-static void
-run_file(FileView *view, int dont_execute)
-{
-	/* TODO: refactor this function run_file() */
+	/* TODO: refactor this function run_selection() */
 
 	char *typed_fname;
 	const char *multi_prog_cmd;
@@ -398,9 +397,16 @@ run_file(FileView *view, int dont_execute)
 	dir_entry_t *entry;
 	int no_multi_run;
 
-	if(!view->dir_entry[view->list_pos].selected)
+	if(!selection_is_consistent(view))
 	{
-		clean_selected_files(view);
+		show_error_msg("Selection error",
+				"Selection cannot contain files and directories at the same time");
+		return;
+	}
+
+	if(!get_current_entry(view)->selected)
+	{
+		flist_sel_stash(view);
 	}
 
 	typed_fname = get_typed_entry_fpath(get_current_entry(view));
@@ -478,7 +484,7 @@ run_file(FileView *view, int dont_execute)
 static void
 run_with_defaults(FileView *view)
 {
-	if(view->dir_entry[view->list_pos].type == FT_DIR)
+	if(get_current_entry(view)->type == FT_DIR)
 	{
 		open_dir(view);
 	}
@@ -543,10 +549,10 @@ void
 run_using_prog(FileView *view, const char prog_spec[], int dont_execute,
 		int force_bg)
 {
-	const dir_entry_t *const entry = &view->dir_entry[view->list_pos];
+	const dir_entry_t *const curr = get_current_entry(view);
 	const int pause = skip_prefix(&prog_spec, "!!");
 
-	if(!path_exists_at(entry->origin, entry->name, DEREF))
+	if(!path_exists_at(curr->origin, curr->name, DEREF))
 	{
 		show_error_msg("Access Error", "File doesn't exist.");
 		return;
@@ -601,7 +607,7 @@ run_explicit_prog(const char prog_spec[], int pause, int force_bg)
 	else if(bg)
 	{
 		assert(flags != MF_IGNORE && "This case is for run_ext_command()");
-		(void)start_background_job(cmd, flags == MF_IGNORE);
+		(void)bg_run_external(cmd, flags == MF_IGNORE);
 	}
 	else
 	{
@@ -622,7 +628,7 @@ run_implicit_prog(FileView *view, const char prog_spec[], int pause,
 	char cmd[NAME_MAX + 1 + NAME_MAX + 1];
 	const char *name_macro;
 	char *file_name;
-	char spec[strlen(prog_spec) + 1];
+	char spec[strlen(prog_spec) + 1U];
 
 	strcpy(spec, prog_spec);
 	bg = cut_suffix(spec, " &") || force_bg;
@@ -642,7 +648,7 @@ run_implicit_prog(FileView *view, const char prog_spec[], int pause,
 
 	if(bg)
 	{
-		(void)start_background_job(cmd, 0);
+		(void)bg_run_external(cmd, 0);
 	}
 	else
 	{
@@ -668,11 +674,11 @@ follow_link(FileView *view, int follow_dirs)
 	char *dir, *file;
 	char full_path[PATH_MAX];
 	char linkto[PATH_MAX + NAME_MAX];
-	dir_entry_t *const entry = &curr_view->dir_entry[curr_view->list_pos];
+	const dir_entry_t *const curr = get_current_entry(curr_view);
 
-	get_full_path_of(entry, sizeof(full_path), full_path);
+	get_full_path_of(curr, sizeof(full_path), full_path);
 
-	if(get_link_target_abs(full_path, entry->origin, linkto, sizeof(linkto)) != 0)
+	if(get_link_target_abs(full_path, curr->origin, linkto, sizeof(linkto)) != 0)
 	{
 		show_error_msg("Error", "Can't read link.");
 		return;
@@ -689,7 +695,7 @@ follow_link(FileView *view, int follow_dirs)
 
 	if(is_dir(linkto) && !follow_dirs)
 	{
-		dir = strdup(entry->name);
+		dir = strdup(curr->name);
 		file = NULL;
 	}
 	else
@@ -754,7 +760,7 @@ cd_updir(FileView *view, int levels)
 	}
 
 	curr_stats.drop_new_dir_hist = 0;
-	save_view_history(view, NULL, NULL, -1);
+	flist_hist_save(view, NULL, NULL, -1);
 }
 
 /* Goes one directory up from current location.  Returns zero unless it won't
@@ -1116,9 +1122,18 @@ set_pwd_in_screen(const char path[])
 int
 run_with_filetype(FileView *view, const char beginning[], int background)
 {
-	char *const typed_fname = get_typed_entry_fpath(get_current_entry(view));
-	assoc_records_t ft = ft_get_all_programs(typed_fname);
-	assoc_records_t magic = get_magic_handlers(typed_fname);
+	dir_entry_t *const curr = get_current_entry(view);
+	assoc_records_t ft, magic;
+	char *typed_fname;
+
+	if(fentry_is_fake(curr))
+	{
+		return 1;
+	}
+
+	typed_fname = get_typed_entry_fpath(curr);
+	ft = ft_get_all_programs(typed_fname);
+	magic = get_magic_handlers(typed_fname);
 	free(typed_fname);
 
 	if(try_run_with_filetype(view, ft, beginning, background))
@@ -1167,6 +1182,11 @@ run_ext_command(const char cmd[], MacroFlags flags, int bg, int *save_msg)
 		*save_msg = 1;
 		return -1;
 	}
+	else if(flags == MF_PREVIEW_OUTPUT)
+	{
+		*save_msg = output_to_preview(cmd);
+		return -1;
+	}
 	else if(flags == MF_IGNORE)
 	{
 		*save_msg = 0;
@@ -1175,7 +1195,7 @@ run_ext_command(const char cmd[], MacroFlags flags, int bg, int *save_msg)
 			int error;
 
 			setup_shellout_env();
-			error = (start_background_job(cmd, 1) != 0);
+			error = (bg_run_external(cmd, 1) != 0);
 			cleanup_shellout_env();
 
 			if(error)
@@ -1201,10 +1221,14 @@ run_ext_command(const char cmd[], MacroFlags flags, int bg, int *save_msg)
 	{
 		run_in_split(curr_view, cmd);
 	}
-	else if(flags == MF_CUSTOMVIEW_OUTPUT || flags == MF_VERYCUSTOMVIEW_OUTPUT)
+	else if(ONE_OF(flags, MF_CUSTOMVIEW_OUTPUT, MF_VERYCUSTOMVIEW_OUTPUT,
+				MF_CUSTOMVIEW_IOUTPUT, MF_VERYCUSTOMVIEW_IOUTPUT))
 	{
-		const int very = flags == MF_VERYCUSTOMVIEW_OUTPUT;
-		output_to_custom_flist(curr_view, cmd, very);
+		const int very =
+			ONE_OF(flags, MF_VERYCUSTOMVIEW_OUTPUT, MF_VERYCUSTOMVIEW_IOUTPUT);
+		const int interactive =
+			ONE_OF(flags, MF_CUSTOMVIEW_IOUTPUT, MF_VERYCUSTOMVIEW_IOUTPUT);
+		output_to_custom_flist(curr_view, cmd, very, interactive);
 	}
 	else
 	{
@@ -1224,7 +1248,7 @@ output_to_statusbar(const char cmd[])
 	int error;
 
 	setup_shellout_env();
-	error = (background_and_capture((char *)cmd, 1, &file, &err) == (pid_t)-1);
+	error = (bg_run_and_capture((char *)cmd, 1, &file, &err) == (pid_t)-1);
 	cleanup_shellout_env();
 	if(error)
 	{
@@ -1254,6 +1278,19 @@ output_to_statusbar(const char cmd[])
 	free(lines);
 }
 
+/* Runs the command and captures its output into abandoned preview.  Returns new
+ * value for the save_msg flag. */
+static int
+output_to_preview(const char cmd[])
+{
+	if(qv_ensure_is_shown() != 0)
+	{
+		return 1;
+	}
+	make_abandoned_view(other_view, cmd);
+	return 0;
+}
+
 /* Executes the cmd ignoring its output. */
 static void
 output_to_nowhere(const char cmd[])
@@ -1262,7 +1299,7 @@ output_to_nowhere(const char cmd[])
 	int error;
 
 	setup_shellout_env();
-	error = (background_and_capture((char *)cmd, 1, &file, &err) == (pid_t)-1);
+	error = (bg_run_and_capture((char *)cmd, 1, &file, &err) == (pid_t)-1);
 	cleanup_shellout_env();
 	if(error)
 	{
@@ -1321,7 +1358,8 @@ run_in_split(const FileView *view, const char cmd[])
 }
 
 int
-output_to_custom_flist(FileView *view, const char cmd[], int very)
+output_to_custom_flist(FileView *view, const char cmd[], int very,
+		int interactive)
 {
 	char *title;
 	int error;
@@ -1330,9 +1368,14 @@ output_to_custom_flist(FileView *view, const char cmd[], int very)
 	flist_custom_start(view, title);
 	free(title);
 
+	if(interactive && curr_stats.load_stage != 0)
+	{
+		endwin();
+	}
+
 	setup_shellout_env();
-	error = (process_cmd_output("Loading custom view", cmd, 1, &path_handler,
-				view) != 0);
+	error = (process_cmd_output("Loading custom view", cmd, 1, interactive,
+				&path_handler, view) != 0);
 	cleanup_shellout_env();
 
 	if(error)
@@ -1341,7 +1384,7 @@ output_to_custom_flist(FileView *view, const char cmd[], int very)
 		return 1;
 	}
 
-	flist_end_custom(view, very);
+	flist_custom_end(view, very);
 	return 0;
 }
 
@@ -1351,7 +1394,7 @@ static void
 path_handler(const char line[], void *arg)
 {
 	FileView *view = arg;
-	flist_add_custom_line(view, line);
+	flist_custom_add_spec(view, line);
 }
 
 int
@@ -1361,7 +1404,7 @@ run_cmd_for_output(const char cmd[], char ***files, int *nfiles)
 	strlist_t list = {};
 
 	setup_shellout_env();
-	error = (process_cmd_output("Loading list", cmd, 1, &line_handler,
+	error = (process_cmd_output("Loading list", cmd, 1, 0, &line_handler,
 				&list) != 0);
 	cleanup_shellout_env();
 
