@@ -35,8 +35,8 @@
 #include "utils/str.h"
 #include "utils/utils.h"
 #include "filelist.h"
+#include "flist_sel.h"
 
-static int find_and_goto_pattern(FileView *view, int wrap_start, int backward);
 static int find_and_goto_match(FileView *view, int start, int backward);
 static void print_result(const FileView *const view, int found, int backward);
 
@@ -44,15 +44,6 @@ int
 goto_search_match(FileView *view, int backward)
 {
 	const int wrap_start = backward ? view->list_rows : -1;
-	return find_and_goto_pattern(view, wrap_start, backward);
-}
-
-/* Looks for a search match in specified direction navigates to it if match is
- * found.  Wraps search around specified wrap start position, when requested.
- * Returns non-zero if something was found, otherwise zero is returned. */
-static int
-find_and_goto_pattern(FileView *view, int wrap_start, int backward)
-{
 	if(!find_and_goto_match(view, view->list_pos, backward))
 	{
 		if(!cfg.wrap_scan || !find_and_goto_match(view, wrap_start, backward))
@@ -60,7 +51,12 @@ find_and_goto_pattern(FileView *view, int wrap_start, int backward)
 			return 0;
 		}
 	}
+
+	/* Redraw the cursor which also might synchronize cursors of two views. */
 	fview_cursor_redraw(view);
+	/* Schedule redraw of the view to highlight search matches. */
+	ui_view_schedule_redraw(view);
+
 	return 1;
 }
 
@@ -115,11 +111,15 @@ find_pattern(FileView *view, const char pattern[], int backward, int move,
 
 	if(move && cfg.hl_search)
 	{
-		clean_selected_files(view);
+		flist_sel_stash(view);
 	}
 
 	reset_search_results(view);
 
+	/* We at least could wipe out previous search results, so schedule a
+	 * redraw. */
+	ui_view_schedule_redraw(view);
+
 	if(pattern[0] == '\0')
 	{
 		*found = 1;
@@ -136,16 +136,26 @@ find_pattern(FileView *view, const char pattern[], int backward, int move,
 		{
 			regmatch_t matches[1];
 			dir_entry_t *const entry = &view->dir_entry[i];
+			const char *name = entry->name;
+			char *free_this = NULL;
 
-			if(is_parent_dir(entry->name))
+			if(is_parent_dir(name))
 			{
 				continue;
 			}
 
-			if(regexec(&re, entry->name, 1, matches, 0) != 0)
+			if(fentry_is_dir(entry))
 			{
+				free_this = format_str("%s/", name);
+				name = free_this;
+			}
+
+			if(regexec(&re, name, 1, matches, 0) != 0)
+			{
+				free(free_this);
 				continue;
 			}
+			free(free_this);
 
 			entry->search_match = nmatches + 1;
 			entry->match_left = matches[0].rm_so;
@@ -178,21 +188,12 @@ find_pattern(FileView *view, const char pattern[], int backward, int move,
 	view->matches = nmatches;
 	copy_str(view->last_search, sizeof(view->last_search), pattern);
 
-	/* Need to redraw the list so that the matching files are highlighted */
-	draw_dir_list(view);
-
 	view->matches = nmatches;
 	if(nmatches > 0)
 	{
 		const int was_found = move ? goto_search_match(view, backward) : 1;
 		*found = was_found;
 
-		if(cfg.hl_search && !was_found)
-		{
-			/* Update the view.  It look might have changed, because of selection. */
-			fview_cursor_redraw(view);
-		}
-
 		if(!cfg.hl_search)
 		{
 			if(print_errors)
@@ -205,7 +206,6 @@ find_pattern(FileView *view, const char pattern[], int backward, int move,
 	}
 	else
 	{
-		fview_cursor_redraw(view);
 		if(print_errors)
 		{
 			print_search_fail_msg(view, backward);
@@ -239,8 +239,7 @@ print_search_msg(const FileView *view, int backward)
 	else
 	{
 		status_bar_messagef("%d of %d matching file%s for: %s",
-				view->dir_entry[view->list_pos].search_match,
-				view->matches,
+				get_current_entry(view)->search_match, view->matches,
 				(view->matches == 1) ? "" : "s", cfg_get_last_search_pattern());
 	}
 }
@@ -248,8 +247,10 @@ print_search_msg(const FileView *view, int backward)
 void
 print_search_next_msg(const FileView *view, int backward)
 {
-	status_bar_messagef("(%d of %d) %c%s", view->dir_entry[view->list_pos].search_match,
-			view->matches, backward ? '?' : '/', cfg_get_last_search_pattern());
+	const int match_number = get_current_entry(view)->search_match;
+	const char search_type = backward ? '?' : '/';
+	status_bar_messagef("(%d of %d) %c%s", match_number, view->matches,
+			search_type, cfg_get_last_search_pattern());
 }
 
 void
