@@ -26,8 +26,6 @@
 #undef MIN
 #endif
 
-#include <pthread.h> /* PTHREAD_* pthread_* */
-
 #include <assert.h> /* assert() */
 #include <limits.h> /* INT_MIN */
 #include <stddef.h> /* NULL */
@@ -36,6 +34,7 @@
 
 #include "cfg/config.h"
 #include "compat/fs_limits.h"
+#include "compat/pthread.h"
 #include "ui/colors.h"
 #include "ui/ui.h"
 #include "utils/env.h"
@@ -104,6 +103,8 @@ init_status(config_t *config)
 static void
 load_def_values(status_t *stats, config_t *config)
 {
+	size_t i;
+
 	pending_redraw = 0;
 
 	stats->need_update = UT_NONE;
@@ -132,6 +133,10 @@ load_def_values(status_t *stats, config_t *config)
 	stats->msg_tail = 0;
 	stats->save_msg_in_list = 1;
 	stats->allow_sb_msg_truncation = 1;
+	for(i = 0U; i < ARRAY_LEN(stats->msgs); ++i)
+	{
+		update_string(&stats->msgs[i], NULL);
+	}
 
 	stats->scroll_bind_off = 0;
 	stats->split = VSPLIT;
@@ -215,10 +220,10 @@ static int
 reset_dircache(void)
 {
 	fsdata_free(dcache_size);
-	dcache_size = fsdata_create(0);
+	dcache_size = fsdata_create(0, 1);
 
 	fsdata_free(dcache_nitems);
-	dcache_nitems = fsdata_create(0);
+	dcache_nitems = fsdata_create(0, 1);
 
 	return (dcache_size == NULL || dcache_nitems == NULL);
 }
@@ -335,6 +340,29 @@ stats_file_choose_action_set(void)
 	    || !is_null_or_empty(curr_stats.on_choose);
 }
 
+void
+stats_save_msg(const char msg[])
+{
+	if(!curr_stats.save_msg_in_list || msg[0] == '\0')
+	{
+		return;
+	}
+
+	if(curr_stats.msg_tail != curr_stats.msg_head &&
+			strcmp(curr_stats.msgs[curr_stats.msg_tail], msg) == 0)
+	{
+		return;
+	}
+
+	curr_stats.msg_tail = (curr_stats.msg_tail + 1)%ARRAY_LEN(curr_stats.msgs);
+	if(curr_stats.msg_tail == curr_stats.msg_head)
+	{
+		free(curr_stats.msgs[curr_stats.msg_head]);
+		curr_stats.msg_head = (curr_stats.msg_head + 1)%ARRAY_LEN(curr_stats.msgs);
+	}
+	curr_stats.msgs[curr_stats.msg_tail] = strdup(msg);
+}
+
 void
 dcache_get_at(const char path[], uint64_t *size, uint64_t *nitems)
 {
