@@ -9,6 +9,8 @@ const char *tags[] = {
 	"vifm-%C",
 	"vifm-%D",
 	"vifm-%F",
+	"vifm-%IU",
+	"vifm-%Iu",
 	"vifm-%M",
 	"vifm-%S",
 	"vifm-%U",
@@ -25,12 +27,14 @@ const char *tags[] = {
 	"vifm-%pw",
 	"vifm-%px",
 	"vifm-%py",
+	"vifm-%q",
 	"vifm-%r",
 	"vifm-%s",
 	"vifm-%u",
 	"vifm-'",
 	"vifm-'aproposprg'",
 	"vifm-'autochpos'",
+	"vifm-'caseoptions'",
 	"vifm-'cd'",
 	"vifm-'cdpath'",
 	"vifm-'cf'",
@@ -45,6 +49,7 @@ const char *tags[] = {
 	"vifm-'deleteprg'",
 	"vifm-'dirsize'",
 	"vifm-'dotdirs'",
+	"vifm-'dotfiles'",
 	"vifm-'fastrun'",
 	"vifm-'fcs'",
 	"vifm-'fillchars'",
@@ -87,6 +92,7 @@ const char *tags[] = {
 	"vifm-'shell'",
 	"vifm-'shm'",
 	"vifm-'shortmess'",
+	"vifm-'sizefmt'",
 	"vifm-'slowfs'",
 	"vifm-'smartcase'",
 	"vifm-'so'",
@@ -177,6 +183,9 @@ const char *tags[] = {
 	"vifm-:comc",
 	"vifm-:comclear",
 	"vifm-:command",
+	"vifm-:compare",
+	"vifm-:cope",
+	"vifm-:copen",
 	"vifm-:copy",
 	"vifm-:cq",
 	"vifm-:cquit",
@@ -194,6 +203,12 @@ const char *tags[] = {
 	"vifm-:di",
 	"vifm-:dirs",
 	"vifm-:display",
+	"vifm-:dm",
+	"vifm-:dmap",
+	"vifm-:dn",
+	"vifm-:dnoremap",
+	"vifm-:du",
+	"vifm-:dunmap",
 	"vifm-:e",
 	"vifm-:ec",
 	"vifm-:echo",
@@ -300,6 +315,8 @@ const char *tags[] = {
 	"vifm-:setlocal",
 	"vifm-:sh",
 	"vifm-:shell",
+	"vifm-:siblnext",
+	"vifm-:siblprev",
 	"vifm-:so",
 	"vifm-:sor",
 	"vifm-:sort",
@@ -311,6 +328,7 @@ const char *tags[] = {
 	"vifm-:touch",
 	"vifm-:tr",
 	"vifm-:trashes",
+	"vifm-:tree",
 	"vifm-:undol",
 	"vifm-:undolist",
 	"vifm-:unl",
@@ -409,7 +427,15 @@ const char *tags[] = {
 	"vifm-Y",
 	"vifm-ZQ",
 	"vifm-ZZ",
+	"vifm-[c",
 	"vifm-[count]",
+	"vifm-[d",
+	"vifm-[s",
+	"vifm-[z",
+	"vifm-]c",
+	"vifm-]d",
+	"vifm-]s",
+	"vifm-]z",
 	"vifm-^",
 	"vifm-al",
 	"vifm-app.txt",
@@ -477,6 +503,7 @@ const char *tags[] = {
 	"vifm-commands",
 	"vifm-commands-and-selection",
 	"vifm-commands-bg",
+	"vifm-compare-views",
 	"vifm-configure",
 	"vifm-count",
 	"vifm-cp",
@@ -487,6 +514,8 @@ const char *tags[] = {
 	"vifm-cw",
 	"vifm-d",
 	"vifm-dd",
+	"vifm-do",
+	"vifm-dp",
 	"vifm-e",
 	"vifm-env-vars",
 	"vifm-executable()",
@@ -508,6 +537,7 @@ const char *tags[] = {
 	"vifm-expr-unary-!",
 	"vifm-expr-unary-+",
 	"vifm-expr-unary--",
+	"vifm-expr-variable",
 	"vifm-expr1",
 	"vifm-expr2",
 	"vifm-expr3",
@@ -673,6 +703,7 @@ const char *tags[] = {
 	"vifm-scripts",
 	"vifm-see-also",
 	"vifm-selectors",
+	"vifm-servername-variable",
 	"vifm-set-options",
 	"vifm-startup",
 	"vifm-system()",
@@ -682,6 +713,7 @@ const char *tags[] = {
 	"vifm-trash",
 	"vifm-u",
 	"vifm-v",
+	"vifm-v:servername",
 	"vifm-v_:",
 	"vifm-v_CTRL-C",
 	"vifm-v_CTRL-G",
@@ -711,6 +743,8 @@ const char *tags[] = {
 	"vifm-zb",
 	"vifm-zd",
 	"vifm-zf",
+	"vifm-zj",
+	"vifm-zk",
 	"vifm-zm",
 	"vifm-zo",
 	"vifm-zr",
