@@ -58,6 +58,14 @@ typedef enum
 }
 TrashResidentType;
 
+/* Type of path check. */
+typedef enum
+{
+	PREFIXED_WITH, /* Checks that path starts with a prefix. */
+	SAME_AS,       /* Checks that path is equal to the other one. */
+}
+PathCheckType;
+
 /* Client of the traverse_specs() function.  Should return non-zero to stop
  * traversal. */
 typedef int (*traverser)(const char base_path[], const char trash_dir[],
@@ -90,6 +98,7 @@ static trashes_list get_list_of_trashes(void);
 static int get_list_of_trashes_traverser(struct mntent *entry, void *arg);
 static int is_trash_valid(const char trash_dir[]);
 static void remove_from_trash(const char trash_name[]);
+static void free_entry(const trash_entry_t *entry);
 static int pick_trash_dir_traverser(const char base_path[],
 		const char trash_dir[], int user_specific, void *arg);
 static int is_rooted_trash_dir(const char spec[]);
@@ -98,6 +107,10 @@ static int get_resident_type_traverser(const char path[],
 		const char trash_dir[], int user_specific, void *arg);
 static int is_trash_directory_traverser(const char path[],
 		const char trash_dir[], int user_specific, void *arg);
+static int path_is(PathCheckType check, const char path[], const char other[]);
+static int entry_is(PathCheckType check, trash_entry_t *entry,
+		const char other[]);
+static void make_real_path(const char path[], char buf[], size_t buf_len);
 static void traverse_specs(const char base_path[], traverser client, void *arg);
 static char * expand_uid(const char spec[], int *expanded);
 static char * get_rooted_trash_dir(const char base_path[], const char spec[]);
@@ -235,7 +248,7 @@ trash_empty_all(void)
 {
 	regs_remove_trashed_files(NULL);
 	empty_trash_dirs();
-	clean_cmds_with_trash(NULL);
+	un_clear_cmds_with_trash(NULL);
 	remove_trash_entries(NULL);
 }
 
@@ -259,7 +272,7 @@ trash_empty(const char trash_dir[])
 {
 	regs_remove_trashed_files(trash_dir);
 	empty_trash_dir(trash_dir);
-	clean_cmds_with_trash(trash_dir);
+	un_clear_cmds_with_trash(trash_dir);
 	remove_trash_entries(trash_dir);
 }
 
@@ -305,11 +318,9 @@ remove_trash_entries(const char trash_dir[])
 
 	for(i = 0; i < nentries; ++i)
 	{
-		if(trash_dir == NULL ||
-				path_starts_with(trash_list[i].trash_name, trash_dir))
+		if(trash_dir == NULL || entry_is(PREFIXED_WITH, &trash_list[i], trash_dir))
 		{
-			free(trash_list[i].path);
-			free(trash_list[i].trash_name);
+			free_entry(&trash_list[i]);
 			continue;
 		}
 
@@ -342,12 +353,7 @@ add_to_trash(const char path[], const char trash_name[])
 {
 	void *p;
 
-	if(!exists_in_trash(trash_name))
-	{
-		return -1;
-	}
-
-	if(is_in_trash(trash_name))
+	if(trash_includes(path))
 	{
 		return 0;
 	}
@@ -361,11 +367,11 @@ add_to_trash(const char path[], const char trash_name[])
 
 	trash_list[nentries].path = strdup(path);
 	trash_list[nentries].trash_name = strdup(trash_name);
+	trash_list[nentries].real_trash_name = NULL;
 	if(trash_list[nentries].path == NULL ||
 			trash_list[nentries].trash_name == NULL)
 	{
-		free(trash_list[nentries].path);
-		free(trash_list[nentries].trash_name);
+		free_entry(&trash_list[nentries]);
 		return -1;
 	}
 
@@ -374,13 +380,16 @@ add_to_trash(const char path[], const char trash_name[])
 }
 
 int
-is_in_trash(const char trash_name[])
+trash_includes(const char original_path[])
 {
 	int i;
-	for(i = 0; i < nentries; i++)
+	for(i = 0; i < nentries; ++i)
 	{
-		if(stroscmp(trash_list[i].trash_name, trash_name) == 0)
+		/* Assuming canonicalized paths for this unit. */
+		if(stroscmp(original_path, trash_list[i].path) == 0)
+		{
 			return 1;
+		}
 	}
 	return 0;
 }
@@ -466,19 +475,23 @@ restore_from_trash(const char trash_name[])
 {
 	int i;
 	char full[PATH_MAX];
-	char buf[PATH_MAX];
+	char path[PATH_MAX];
 
-	for(i = 0; i < nentries; i++)
+	for(i = 0; i < nentries; ++i)
 	{
-		if(stroscmp(trash_list[i].trash_name, trash_name) == 0)
+		if(entry_is(SAME_AS, &trash_list[i], trash_name))
+		{
 			break;
+		}
 	}
 	if(i >= nentries)
+	{
 		return -1;
+	}
 
-	copy_str(buf, sizeof(buf), trash_list[i].path);
+	copy_str(path, sizeof(path), trash_list[i].path);
 	copy_str(full, sizeof(full), trash_list[i].trash_name);
-	if(perform_operation(OP_MOVE, NULL, NULL, full, trash_list[i].path) == 0)
+	if(perform_operation(OP_MOVE, NULL, NULL, full, path) == 0)
 	{
 		char *msg, *p;
 		size_t len;
@@ -498,7 +511,7 @@ restore_from_trash(const char trash_name[])
 		replace_group_msg(msg);
 		free(msg);
 
-		add_operation(OP_MOVE, NULL, NULL, full, buf);
+		add_operation(OP_MOVE, NULL, NULL, full, path);
 		cmd_group_end();
 		remove_from_trash(trash_name);
 		return 0;
@@ -512,24 +525,34 @@ static void
 remove_from_trash(const char trash_name[])
 {
 	int i;
-	for(i = 0; i < nentries; i++)
+	for(i = 0; i < nentries; ++i)
 	{
-		if(stroscmp(trash_list[i].trash_name, trash_name) == 0)
+		if(entry_is(SAME_AS, &trash_list[i], trash_name))
+		{
 			break;
+		}
 	}
 	if(i >= nentries)
 	{
 		return;
 	}
 
-	free(trash_list[i].path);
-	free(trash_list[i].trash_name);
+	free_entry(&trash_list[i]);
 	memmove(trash_list + i, trash_list + i + 1,
 			sizeof(*trash_list)*((nentries - 1) - i));
 
 	--nentries;
 }
 
+/* Frees memory allocated by given trash entry. */
+static void
+free_entry(const trash_entry_t *entry)
+{
+	free(entry->path);
+	free(entry->trash_name);
+	free(entry->real_trash_name);
+}
+
 char *
 gen_trash_name(const char base_path[], const char name[])
 {
@@ -615,7 +638,7 @@ trash_contains(const char trash_dir[], const char path[])
 		return is_under_trash(path);
 	}
 
-	return path_starts_with(path, trash_dir);
+	return path_is(PREFIXED_WITH, path, trash_dir);
 }
 
 /* Gets status of file relative to trash directories.  Returns the status. */
@@ -633,13 +656,13 @@ static int
 get_resident_type_traverser(const char path[], const char trash_dir[],
 		int user_specific, void *arg)
 {
-	if(path_starts_with(path, trash_dir))
+	if(path_is(PREFIXED_WITH, path, trash_dir))
 	{
 		TrashResidentType *const result = arg;
 		char *const parent_dir = strdup(path);
 
 		remove_last_path_component(parent_dir);
-		*result = paths_are_equal(parent_dir, trash_dir)
+		*result = path_is(SAME_AS, parent_dir, trash_dir)
 		        ? TRT_IN_TRASH
 		        : TRT_IN_REMOVED_DIRECTORY;
 
@@ -664,7 +687,7 @@ static int
 is_trash_directory_traverser(const char path[], const char trash_dir[],
 		int user_specific, void *arg)
 {
-	if(stroscmp(path, trash_dir) == 0)
+	if(path_is(SAME_AS, path, trash_dir))
 	{
 		int *const result = arg;
 		*result = 1;
@@ -673,6 +696,87 @@ is_trash_directory_traverser(const char path[], const char trash_dir[],
 	return 0;
 }
 
+/* Performs specified check on a path.  The check is constructed in such a way
+ * that all path components except for the last one are expanded.  Returns
+ * non-zero if path passes the check, otherwise zero is returned. */
+static int
+path_is(PathCheckType check, const char path[], const char other[])
+{
+	char path_real[PATH_MAX*2], other_real[PATH_MAX*2];
+
+	make_real_path(path, path_real, sizeof(path_real));
+	make_real_path(other, other_real, sizeof(other_real));
+
+	return (check == PREFIXED_WITH)
+	     ? path_starts_with(path_real, other_real)
+	     : paths_are_same(path_real, other_real);
+}
+
+/* An optimized for entries version of path_is(). */
+static int
+entry_is(PathCheckType check, trash_entry_t *entry, const char other[])
+{
+	char real[PATH_MAX*2];
+
+	if(entry->real_trash_name == NULL)
+	{
+		char real[PATH_MAX*2];
+		make_real_path(entry->trash_name, real, sizeof(real));
+		entry->real_trash_name = strdup(real);
+	}
+
+	make_real_path(other, real, sizeof(real));
+
+	return (check == PREFIXED_WITH)
+	     ? path_starts_with(entry->real_trash_name, real)
+	     : paths_are_same(entry->real_trash_name, real);
+}
+
+/* Resolves all but last path components in the path.  Permanently caches
+ * results of previous invocations. */
+static void
+make_real_path(const char path[], char buf[], size_t buf_len)
+{
+	/* This cache only grows, but as number of distinct trash directories is
+	 * likely to be limited, not much space should be wasted. */
+	static char **link, **real;
+	static int nrecords;
+
+	int pos;
+	char copy[PATH_MAX*2];
+	char real_dir[PATH_MAX*2];
+
+	copy_str(copy, sizeof(copy), path);
+	remove_last_path_component(copy);
+
+	pos = string_array_pos(link, nrecords, copy);
+	if(pos != -1)
+	{
+		copy_str(real_dir, sizeof(real_dir), real[pos]);
+	}
+	else if(os_realpath(copy, real_dir) != real_dir)
+	{
+		copy_str(real_dir, sizeof(real_dir), path);
+	}
+	else
+	{
+		if(add_to_string_array(&link, nrecords, 1, copy) == nrecords + 1)
+		{
+			if(add_to_string_array(&real, nrecords, 1, real_dir) == nrecords + 1)
+			{
+				++nrecords;
+			}
+			else
+			{
+				free(link[nrecords]);
+			}
+		}
+	}
+
+	snprintf(buf, buf_len, "%s%s%s", real_dir,
+			ends_with_slash(real_dir) ? "" : "/", get_last_path_component(path));
+}
+
 /* Calls client traverser for each trash directory specification defined by
  * specs array. */
 static void
