@@ -67,6 +67,7 @@ char *HI_GROUPS[] = {
 	[OTHER_LINE_COLOR]   = "OtherLine",
 	[JOB_LINE_COLOR]     = "JobLine",
 	[SUGGEST_BOX_COLOR]  = "SuggestBox",
+	[MISMATCH_COLOR]     = "CmpMismatch",
 };
 ARRAY_GUARD(HI_GROUPS, MAXNUM_COLOR);
 
@@ -91,6 +92,7 @@ const char *HI_GROUPS_DESCR[] = {
 	[OTHER_LINE_COLOR]   = "cursor in inactive pane",
 	[JOB_LINE_COLOR]     = "job bar",
 	[SUGGEST_BOX_COLOR]  = "suggestion box",
+	[MISMATCH_COLOR]     = "mismatched diff entries",
 };
 ARRAY_GUARD(HI_GROUPS_DESCR, ARRAY_LEN(HI_GROUPS));
 
@@ -387,6 +389,7 @@ static const col_attr_t default_cs[] = {
 	[OTHER_LINE_COLOR]   = { -1,            -1,          -1                      },
 	[JOB_LINE_COLOR]     = { COLOR_BLACK,   COLOR_WHITE, A_BOLD | A_REVERSE      },
 	[SUGGEST_BOX_COLOR]  = { -1,            -1,          A_BOLD                  },
+	[MISMATCH_COLOR]     = { COLOR_WHITE,   COLOR_RED,   A_BOLD                  },
 };
 ARRAY_GUARD(default_cs, MAXNUM_COLOR);
 
@@ -398,6 +401,7 @@ static file_hi_t * clone_cs_highlights(const col_scheme_t *from);
 static void reset_cs_colors(col_scheme_t *cs);
 static int source_cs(const char name[]);
 static void get_cs_path(const char name[], char buf[], size_t buf_size);
+static const char * get_global_colors_dir(void);
 static void check_cs(col_scheme_t *cs);
 static void load_color_pairs(col_scheme_t *cs);
 static void ensure_dir_map_exists(void);
@@ -439,8 +443,8 @@ cs_rename_all(void)
 
 	while((d = os_readdir(dir)) != NULL)
 	{
-		char full_old_path[PATH_MAX];
-		char full_new_path[PATH_MAX];
+		char full_old_path[PATH_MAX + 16 + NAME_MAX];
+		char full_new_path[PATH_MAX + 16 + NAME_MAX];
 		snprintf(full_old_path, sizeof(full_old_path), "%s/%s", cfg.colors_dir,
 				d->d_name);
 		snprintf(full_new_path, sizeof(full_new_path), "%s/%s.vifm", cfg.colors_dir,
@@ -504,7 +508,7 @@ list_cs_files(int *len)
 	*len = 0;
 
 	list = list_regular_files(cfg.colors_dir, list, len);
-	list = list_regular_files(GLOBAL_COLORS_DIR, list, len);
+	list = list_regular_files(get_global_colors_dir(), list, len);
 
 	return list;
 }
@@ -521,7 +525,7 @@ void
 cs_write(void)
 {
 	FILE *fp;
-	char def_cs_path[PATH_MAX];
+	char def_cs_path[PATH_MAX + 32];
 	size_t i;
 
 	if(create_path(cfg.colors_dir, S_IRWXU) != 0)
@@ -855,7 +859,8 @@ get_cs_path(const char name[], char buf[], size_t buf_size)
 		return;
 	}
 
-	snprintf(buf, buf_size, GLOBAL_COLORS_DIR "/%s.vifm", name);
+	snprintf(buf, buf_size, "%s/%s.vifm", get_global_colors_dir(),
+			name);
 	if(is_regular_file(buf))
 	{
 		return;
@@ -864,6 +869,23 @@ get_cs_path(const char name[], char buf[], size_t buf_size)
 	(void)cut_suffix(buf, ".vifm");
 }
 
+/* Retrieves path to global directory containing color schemes.  Returns the
+ * path. */
+static const char *
+get_global_colors_dir(void)
+{
+#ifndef _WIN32
+	return GLOBAL_COLORS_DIR;
+#else
+	static char dir_path[PATH_MAX];
+	if(dir_path[0] == '\0')
+	{
+		snprintf(dir_path, sizeof(dir_path), "%s/colors", get_installed_data_dir());
+	}
+	return dir_path;
+#endif
+}
+
 /* Checks whether colorscheme is in unusable state and resets it to normal
  * state. */
 static void
