@@ -39,9 +39,11 @@
 #include "../utils/path.h"
 #include "../utils/regexp.h"
 #include "../utils/str.h"
+#include "../utils/test_helpers.h"
 #include "../utils/utf8.h"
 #include "../utils/utils.h"
 #include "../filelist.h"
+#include "../flist_pos.h"
 #include "../opt_handlers.h"
 #include "../sort.h"
 #include "color_manager.h"
@@ -63,6 +65,10 @@ typedef struct
 
 	size_t current_line;  /* Line of the cell. */
 	size_t column_offset; /* Offset in characters of the column. */
+
+	size_t *prefix_len; /* Data prefix length (should be drawn in neutral color).
+	                     * A pointer to allow changing value in const struct.
+	                     * Should be zero first time, then auto reset. */
 }
 column_data_t;
 
@@ -75,12 +81,14 @@ static size_t calculate_print_width(const FileView *view, int i,
 		size_t max_width);
 static void draw_cell(const FileView *view, const column_data_t *cdt,
 		size_t col_width, size_t print_width);
+static columns_t * get_view_columns(const FileView *view);
 static void consider_scroll_bind(FileView *view);
+static void put_inactive_mark(FileView *view);
 static int prepare_inactive_color(FileView *view, dir_entry_t *entry,
 		int line_color);
-static int clear_current_line_bar(FileView *view, int is_current);
+static void clear_current_line_bar(FileView *view, int is_current);
 static size_t get_effective_scroll_offset(const FileView *view);
-static void column_line_print(const void *data, int column_id, const char *buf,
+static void column_line_print(const void *data, int column_id, const char buf[],
 		size_t offset, AlignType align, const char full_column[]);
 static void draw_line_number(const column_data_t *cdt, int column);
 static void highlight_search(FileView *view, dir_entry_t *entry,
@@ -92,7 +100,7 @@ static void mix_in_file_hi(const FileView *view, dir_entry_t *entry,
 		int type_hi, col_attr_t *col);
 static void mix_in_file_name_hi(const FileView *view, dir_entry_t *entry,
 		col_attr_t *col);
-static void format_name(int id, const void *data, size_t buf_len, char buf[]);
+TSTATIC void format_name(int id, const void *data, size_t buf_len, char buf[]);
 static void format_size(int id, const void *data, size_t buf_len, char buf[]);
 static void format_nitems(int id, const void *data, size_t buf_len, char buf[]);
 static void format_primary_group(int id, const void *data, size_t buf_len,
@@ -111,6 +119,7 @@ static void format_owner(int id, const void *data, size_t buf_len, char buf[]);
 static void format_perms(int id, const void *data, size_t buf_len, char buf[]);
 static void format_nlinks(int id, const void *data, size_t buf_len, char buf[]);
 #endif
+static void format_id(int id, const void *data, size_t buf_len, char buf[]);
 static size_t calculate_column_width(FileView *view);
 static size_t get_max_filename_width(const FileView *view);
 static size_t get_filename_width(const FileView *view, int i);
@@ -162,6 +171,7 @@ fview_init(void)
 	{
 		columns_add_column_desc(sort_to_func[i].key, sort_to_func[i].func);
 	}
+	columns_add_column_desc(SK_BY_ID, &format_id);
 }
 
 void
@@ -261,9 +271,10 @@ draw_dir_list_only(FileView *view)
 	ui_view_erase(view);
 
 	cell = 0U;
-	coll_pad = (view->ls_view && cfg.extra_padding) ? 1 : 0;
+	coll_pad = (!ui_view_displays_columns(view) && cfg.extra_padding) ? 1 : 0;
 	for(x = top; x < view->list_rows; ++x)
 	{
+		size_t prefix_len = 0U;
 		const column_data_t cdt = {
 			.view = view,
 			.line_pos = x,
@@ -271,6 +282,7 @@ draw_dir_list_only(FileView *view)
 			.is_current = (view == curr_view) ? x == view->list_pos : 0,
 			.current_line = cell/col_count,
 			.column_offset = (cell%col_count)*col_width,
+			.prefix_len = &prefix_len,
 		};
 
 		const size_t print_width = calculate_print_width(view, x, col_width);
@@ -301,16 +313,16 @@ calculate_table_conf(FileView *view, size_t *count, size_t *width)
 {
 	calculate_number_width(view);
 
-	if(view->ls_view)
-	{
-		*count = calculate_columns_count(view);
-		*width = calculate_column_width(view);
-	}
-	else
+	if(ui_view_displays_columns(view))
 	{
 		*count = 1;
 		*width = MAX(0, ui_view_available_width(view) - view->real_num_width);
 	}
+	else
+	{
+		*count = calculate_columns_count(view);
+		*width = calculate_column_width(view);
+	}
 
 	view->column_count = *count;
 	view->window_cells = *count*(view->window_rows + 1);
@@ -401,7 +413,7 @@ get_line_color(const FileView *view, int pos)
 
 				/* Assume that targets on slow file system are not broken as actual
 				 * check might take long time. */
-				if(is_on_slow_fs(full))
+				if(is_on_slow_fs(full, cfg.slow_fs_list))
 				{
 					return LINK_COLOR;
 				}
@@ -427,7 +439,7 @@ get_line_color(const FileView *view, int pos)
 static size_t
 calculate_print_width(const FileView *view, int i, size_t max_width)
 {
-	if(view->ls_view)
+	if(!ui_view_displays_columns(view))
 	{
 		const size_t raw_name_width = get_filename_width(view, i);
 		return MIN(max_width - 1, raw_name_width);
@@ -446,7 +458,7 @@ draw_cell(const FileView *view, const column_data_t *cdt, size_t col_width,
 		column_line_print(cdt, FILL_COLUMN_ID, " ", -1, AT_LEFT, " ");
 	}
 
-	columns_format_line(view->columns, cdt, col_width);
+	columns_format_line(get_view_columns(view), cdt, col_width);
 
 	if(cfg.extra_padding)
 	{
@@ -454,22 +466,67 @@ draw_cell(const FileView *view, const column_data_t *cdt, size_t col_width,
 	}
 }
 
+/* Retrieves active view columns handle of the view considering 'lsview' option
+ * status.  Returns the handle. */
+static columns_t *
+get_view_columns(const FileView *view)
+{
+	/* Note that columns_t performs some caching, so we might want to keep one
+	 * handle per view rather than sharing one. */
+
+	static const column_info_t name_column = {
+		.column_id = SK_BY_NAME, .full_width = 0UL,    .text_width = 0UL,
+		.align = AT_LEFT,        .sizing = ST_AUTO,    .cropping = CT_ELLIPSIS,
+	};
+	static const column_info_t id_column = {
+		.column_id = SK_BY_ID,  .full_width = 7UL,     .text_width = 7UL,
+		.align = AT_LEFT,       .sizing = ST_ABSOLUTE, .cropping = CT_ELLIPSIS,
+	};
+
+	static columns_t *comparison_columns;
+	static columns_t *ls_columns;
+
+	if(!ui_view_displays_columns(view))
+	{
+		if(ls_columns == NULL)
+		{
+			ls_columns = columns_create();
+			columns_add_column(ls_columns, name_column);
+		}
+		return ls_columns;
+	}
+
+	if(cv_compare(view->custom.type))
+	{
+		if(comparison_columns == NULL)
+		{
+			comparison_columns = columns_create();
+			columns_add_column(comparison_columns, name_column);
+			columns_add_column(comparison_columns, id_column);
+		}
+		return comparison_columns;
+	}
+
+	return view->columns;
+}
+
 /* Corrects top of the other view to synchronize it with the current view if
- * 'scrollbind' option is set. */
+ * 'scrollbind' option is set or view is in the compare mode. */
 static void
 consider_scroll_bind(FileView *view)
 {
-	if(cfg.scroll_bind)
+	if(cfg.scroll_bind || view->custom.type == CV_DIFF)
 	{
-		FileView *other = (view == &lwin) ? &rwin : &lwin;
+		FileView *const other = (view == &lwin) ? &rwin : &lwin;
+		const int bind_off = cfg.scroll_bind ? curr_stats.scroll_bind_off : 0;
 		other->top_line = view->top_line/view->column_count;
 		if(view == &lwin)
 		{
-			other->top_line += curr_stats.scroll_bind_off;
+			other->top_line += bind_off;
 		}
 		else
 		{
-			other->top_line -= curr_stats.scroll_bind_off;
+			other->top_line -= bind_off;
 		}
 		other->top_line *= other->column_count;
 		other->top_line = calculate_top_position(other, other->top_line);
@@ -485,8 +542,11 @@ consider_scroll_bind(FileView *view)
 
 		other->curr_line = other->list_pos - other->top_line;
 
-		draw_dir_list(other);
-		refresh_view_win(other);
+		if(window_shows_dirlist(other))
+		{
+			draw_dir_list(other);
+			refresh_view_win(other);
+		}
 	}
 }
 
@@ -525,10 +585,14 @@ redraw_current_view(void)
 void
 erase_current_line_bar(FileView *view)
 {
-	if(clear_current_line_bar(view, 0) && view == other_view)
+	if(view == other_view)
 	{
 		put_inactive_mark(view);
 	}
+	else
+	{
+		clear_current_line_bar(view, 0);
+	}
 }
 
 void
@@ -550,7 +614,8 @@ fview_cursor_redraw(FileView *view)
 	}
 }
 
-void
+/* Adds inactive cursor mark to the view. */
+static void
 put_inactive_mark(FileView *view)
 {
 	size_t col_width;
@@ -558,7 +623,7 @@ put_inactive_mark(FileView *view)
 	int line_attrs;
 	int line, column;
 
-	(void)clear_current_line_bar(view, 1);
+	clear_current_line_bar(view, 1);
 
 	if(!cfg.extra_padding)
 	{
@@ -567,7 +632,7 @@ put_inactive_mark(FileView *view)
 
 	calculate_table_conf(view, &col_count, &col_width);
 
-	line_attrs = prepare_inactive_color(view, &view->dir_entry[view->list_pos],
+	line_attrs = prepare_inactive_color(view, get_current_entry(view),
 			get_line_color(view, view->list_pos));
 
 	line = view->curr_line/col_count;
@@ -575,6 +640,7 @@ put_inactive_mark(FileView *view)
 	checked_wmove(view->win, line, column);
 
 	wprinta(view->win, INACTIVE_CURSOR_MARK, line_attrs);
+	ui_view_win_changed(view);
 }
 
 int
@@ -650,9 +716,8 @@ prepare_inactive_color(FileView *view, dir_entry_t *entry, int line_color)
 
 /* Redraws directory list without any extra actions that are performed in
  * erase_current_line_bar().  is_current defines whether element under the
- * cursor is being erased.  Returns non-zero if something was actually redrawn,
- * otherwise zero is returned. */
-static int
+ * cursor is being erased. */
+static void
 clear_current_line_bar(FileView *view, int is_current)
 {
 	const int old_cursor = view->curr_line;
@@ -661,26 +726,28 @@ clear_current_line_bar(FileView *view, int is_current)
 	size_t col_count;
 	size_t print_width;
 
+	size_t prefix_len = 0U;
 	column_data_t cdt = {
 		.view = view,
 		.line_pos = old_pos,
 		.is_current = is_current,
+		.prefix_len = &prefix_len,
 	};
 
 	if(curr_stats.load_stage < 2)
 	{
-		return 0;
+		return;
 	}
 
 	if(old_cursor < 0)
 	{
-		return 0;
+		return;
 	}
 
 	if(old_pos < 0 || old_pos >= view->list_rows)
 	{
 		/* The entire list is going to be redrawn so just return. */
-		return 0;
+		return;
 	}
 
 	cdt.line_hi_group = get_line_color(view, old_pos),
@@ -692,22 +759,23 @@ clear_current_line_bar(FileView *view, int is_current)
 
 	print_width = calculate_print_width(view, old_pos, col_width);
 
-	if(is_current)
+	if(!ui_view_displays_columns(view))
 	{
-		/* When this function is used to draw cursor position in inactive view, only
-		 * name width should be updated. */
-		col_width = print_width;
-	}
-	else if(view->ls_view && cfg.extra_padding)
-	{
-		/* Padding in ls-like view adds additional empty single character between
-		 * columns, on which we shouldn't draw anything here. */
-		--col_width;
+		if(is_current)
+		{
+			/* When this function is used to draw cursor position in inactive ls-like
+			 * view, only name width should be updated. */
+			col_width = print_width;
+		}
+		else if(cfg.extra_padding)
+		{
+			/* Padding in ls-like view adds additional empty single character between
+			 * columns, on which we shouldn't draw anything here. */
+			--col_width;
+		}
 	}
 
 	draw_cell(view, &cdt, col_width, print_width);
-
-	return 1;
 }
 
 int
@@ -834,41 +902,74 @@ update_scroll_bind_offset(void)
 
 /* Print callback for column_view unit. */
 static void
-column_line_print(const void *data, int column_id, const char *buf,
+column_line_print(const void *data, int column_id, const char buf[],
 		size_t offset, AlignType align, const char full_column[])
 {
-	const int padding = (cfg.extra_padding != 0);
-
-	int primary;
-	int line_attrs;
 	char print_buf[strlen(buf) + 1];
-	size_t width_left;
-	size_t trim_pos;
+	size_t prefix_len, final_offset;
+	size_t width_left, trim_pos;
 	int reserved_width;
 
 	const column_data_t *const cdt = data;
-	const size_t i = cdt->line_pos;
 	FileView *view = cdt->view;
-	dir_entry_t *entry = &view->dir_entry[i];
+	dir_entry_t *entry = &view->dir_entry[cdt->line_pos];
 
 	const int numbers_visible = (offset == 0 && ui_view_displays_numbers(view));
+	const int padding = (cfg.extra_padding != 0);
 
-	const size_t prefix_len = padding + view->real_num_width;
-	const size_t final_offset = prefix_len + cdt->column_offset + offset;
+	const int primary = (column_id == SK_BY_NAME || column_id == SK_BY_INAME);
+	const int line_attrs = prepare_col_color(view, entry, primary,
+			cdt->line_hi_group, cdt->is_current);
 
-	primary = (column_id == SK_BY_NAME || column_id == SK_BY_INAME);
-	line_attrs = prepare_col_color(view, entry, primary, cdt->line_hi_group,
-			cdt->is_current);
+	size_t extra_prefix = primary ? *cdt->prefix_len : 0U;
+
+	if(extra_prefix != 0U && align == AT_RIGHT)
+	{
+		/* Prefix length requires correction if left hand side of file name is
+		 * trimmed. */
+		size_t width = utf8_strsw(buf);
+		if(utf8_strsw(full_column) > width)
+		{
+			/* As left side is trimmed and might contain ellipsis calculate offsets
+			 * according to the right side. */
+			width -= utf8_strsw(full_column + extra_prefix);
+			extra_prefix = utf8_strsnlen(buf, width);
+		}
+	}
+
+	prefix_len = padding + view->real_num_width + extra_prefix;
+	final_offset = prefix_len + cdt->column_offset + offset;
 
 	if(numbers_visible)
 	{
-		const int column = final_offset - view->real_num_width;
+		const int column = final_offset - extra_prefix - view->real_num_width;
 		draw_line_number(cdt, column);
 	}
 
+	if(extra_prefix != 0U)
+	{
+		/* Copy prefix part into working buffer. */
+		strncpy(print_buf, buf, extra_prefix);
+		print_buf[extra_prefix] = '\0';
+		buf += extra_prefix;
+		full_column += extra_prefix;
+
+		checked_wmove(view->win, cdt->current_line, final_offset - extra_prefix);
+		wprinta(view->win, print_buf,
+				prepare_col_color(view, entry, 0, cdt->line_hi_group, cdt->is_current));
+	}
+
 	checked_wmove(view->win, cdt->current_line, final_offset);
 
-	strcpy(print_buf, buf);
+	if(fentry_is_fake(entry))
+	{
+		memset(print_buf, '.', sizeof(print_buf) - 1U);
+		print_buf[sizeof(print_buf) - 1U] = '\0';
+	}
+	else
+	{
+		strcpy(print_buf, buf);
+	}
 	reserved_width = cfg.extra_padding ? (column_id != FILL_COLUMN_ID) : 0;
 	width_left = padding + ui_view_available_width(view)
 	           - reserved_width - offset;
@@ -929,6 +1030,17 @@ highlight_search(FileView *view, dir_entry_t *entry, const char full_column[],
 	lo = name_offset + entry->match_left;
 	ro = name_offset + entry->match_right;
 
+	if((size_t)entry->match_right > strlen(fname) - strlen(suffix))
+	{
+		/* Don't highlight anything past the end of file name except for single
+		 * trailing slash. */
+		ro -= entry->match_right - (strlen(fname) - strlen(suffix));
+		if(suffix[0] == '/')
+		{
+			++ro;
+		}
+	}
+
 	if(align == AT_LEFT && buf_len < ro)
 	{
 		/* Right end of the match isn't visible. */
@@ -988,6 +1100,7 @@ static int
 prepare_col_color(const FileView *view, dir_entry_t *entry, int primary,
 		int line_color, int current)
 {
+	FileView *const other = (view == &lwin) ? &rwin : &lwin;
 	const col_scheme_t *const cs = ui_view_get_cs(view);
 	col_attr_t col = cs->color[WIN_COLOR];
 
@@ -998,6 +1111,14 @@ prepare_col_color(const FileView *view, dir_entry_t *entry, int primary,
 		mix_in_file_hi(view, entry, line_color, &col);
 	}
 
+	/* If two files on the same line in side-by-side comparison have different
+	 * ids, that's a mismatch. */
+	if(view->custom.type == CV_DIFF &&
+			other->dir_entry[entry_to_pos(view, entry)].id != entry->id)
+	{
+		cs_mix_colors(&col, &cs->color[MISMATCH_COLOR]);
+	}
+
 	if(entry->selected)
 	{
 		cs_mix_colors(&col, &cs->color[SELECTED_COLOR]);
@@ -1023,6 +1144,11 @@ static void
 mix_in_file_hi(const FileView *view, dir_entry_t *entry, int type_hi,
 		col_attr_t *col)
 {
+	if(fentry_is_fake(entry))
+	{
+		return;
+	}
+
 	/* Apply file name specific highlights. */
 	mix_in_file_name_hi(view, entry, col);
 
@@ -1040,7 +1166,9 @@ static void
 mix_in_file_name_hi(const FileView *view, dir_entry_t *entry, col_attr_t *col)
 {
 	const col_scheme_t *const cs = ui_view_get_cs(view);
-	const col_attr_t *color = cs_get_file_hi(cs, entry->name, &entry->hi_num);
+	char *const typed_fname = get_typed_entry_fpath(entry);
+	const col_attr_t *color = cs_get_file_hi(cs, typed_fname, &entry->hi_num);
+	free(typed_fname);
 	if(color != NULL)
 	{
 		cs_mix_colors(col, color);
@@ -1048,20 +1176,62 @@ mix_in_file_name_hi(const FileView *view, dir_entry_t *entry, col_attr_t *col)
 }
 
 /* File name format callback for column_view unit. */
-static void
+TSTATIC void
 format_name(int id, const void *data, size_t buf_len, char buf[])
 {
+	size_t len, i;
+	dir_entry_t *child, *parent;
+
 	const column_data_t *cdt = data;
 	FileView *view = cdt->view;
 	dir_entry_t *const entry = &view->dir_entry[cdt->line_pos];
-	if(flist_custom_active(view))
-	{
-		get_short_path_of(view, entry, 1, buf_len + 1, buf);
-	}
-	else
+
+	if(!flist_custom_active(view))
 	{
+		/* Just file name. */
 		format_entry_name(entry, buf_len + 1, buf);
+		return;
 	}
+
+	if(!ui_view_displays_columns(view) || view->custom.type != CV_TREE)
+	{
+		/* File name possibly with path prefix. */
+		get_short_path_of(view, entry, 1, 0, buf_len + 1U, buf);
+		return;
+	}
+
+	/* File name possibly with path and tree prefixes. */
+	len = 0U;
+	child = entry;
+	parent = child - child->child_pos;
+	while(parent != child)
+	{
+		const char *prefix;
+		/* To avoid prepending, strings are reversed here and whole tree prefix is
+		 * reversed below to compensate for it. */
+		if(parent->child_count == child->child_pos + child->child_count)
+		{
+			prefix = (child == entry ? " --`" : "    ");
+		}
+		else
+		{
+			prefix = (child == entry ? " --|" : "   |");
+		}
+		(void)sstrappend(buf, &len, buf_len + 1U, prefix);
+
+		child = parent;
+		parent -= parent->child_pos;
+	}
+
+	for(i = 0U; i < len/2U; ++i)
+	{
+		const char t = buf[i];
+		buf[i] = buf[len - 1U - i];
+		buf[len - 1U - i] = t;
+	}
+
+	get_short_path_of(view, entry, 1, 1, buf_len + 1U - len, buf + len);
+	*cdt->prefix_len = len;
 }
 
 /* Primary name group format (first value of 'sortgroups' option) callback for
@@ -1082,13 +1252,13 @@ format_primary_group(int id, const void *data, size_t buf_len, char buf[])
 static void
 format_size(int id, const void *data, size_t buf_len, char buf[])
 {
-	char str[24];
+	char str[64];
 	const column_data_t *cdt = data;
 	const FileView *view = cdt->view;
 	const dir_entry_t *const entry = &view->dir_entry[cdt->line_pos];
 	uint64_t size = DCACHE_UNKNOWN;
 
-	if(is_directory_entry(entry))
+	if(fentry_is_dir(entry))
 	{
 		uint64_t nitems;
 		dcache_get_of(entry, &size, &nitems);
@@ -1125,7 +1295,7 @@ format_nitems(int id, const void *data, size_t buf_len, char buf[])
 	const dir_entry_t *const entry = &view->dir_entry[cdt->line_pos];
 	uint64_t nitems;
 
-	if(!is_directory_entry(entry))
+	if(!fentry_is_dir(entry))
 	{
 		copy_str(buf, buf_len + 1, " 0");
 		return;
@@ -1194,7 +1364,7 @@ format_fileext(int id, const void *data, size_t buf_len, char buf[])
 	const column_data_t *cdt = data;
 	dir_entry_t *entry = &cdt->view->dir_entry[cdt->line_pos];
 
-	if(!is_directory_entry(entry))
+	if(!fentry_is_dir(entry))
 	{
 		format_ext(id, data, buf_len, buf);
 	}
@@ -1247,7 +1417,7 @@ format_dir(int id, const void *data, size_t buf_len, char buf[])
 {
 	const column_data_t *cdt = data;
 	dir_entry_t *entry = &cdt->view->dir_entry[cdt->line_pos];
-	const char *type = is_directory_entry(entry) ? "dir" : "file";
+	const char *type = fentry_is_dir(entry) ? "dir" : "file";
 	snprintf(buf, buf_len, " %s", type);
 }
 
@@ -1305,37 +1475,29 @@ format_nlinks(int id, const void *data, size_t buf_len, char buf[])
 
 #endif
 
+/* File identifier on comparisons format callback for column_view unit. */
+static void
+format_id(int id, const void *data, size_t buf_len, char buf[])
+{
+	const column_data_t *cdt = data;
+	dir_entry_t *entry = &cdt->view->dir_entry[cdt->line_pos];
+	snprintf(buf, buf_len, "#%d", entry->id);
+}
+
 void
 fview_set_lsview(FileView *view, int enabled)
 {
-	if(view->ls_view == enabled)
+	if(view->ls_view != enabled)
 	{
-		return;
-	}
-
-	view->ls_view = enabled;
-
-	if(view->ls_view)
-	{
-		column_info_t column_info = {
-			.column_id = SK_BY_NAME, .full_width = 0UL, .text_width = 0UL,
-			.align = AT_LEFT,        .sizing = ST_AUTO, .cropping = CT_ELLIPSIS,
-		};
-
-		columns_clear(view->columns);
-		columns_add_column(view->columns, column_info);
+		view->ls_view = enabled;
 		ui_view_schedule_redraw(view);
 	}
-	else
-	{
-		load_view_columns_option(view, view->view_columns);
-	}
 }
 
 size_t
 calculate_columns_count(FileView *view)
 {
-	if(view->ls_view)
+	if(!ui_view_displays_columns(view))
 	{
 		const size_t column_width = calculate_column_width(view);
 		return ui_view_available_width(view)/column_width;
@@ -1396,8 +1558,9 @@ get_filename_width(const FileView *view, int i)
 	size_t name_len;
 	if(flist_custom_active(view))
 	{
-		char name[NAME_MAX];
-		get_short_path_of(view, entry, 0, sizeof(name), name);
+		char name[NAME_MAX + 1];
+		/* XXX: should this be formatted name?. */
+		get_short_path_of(view, entry, 0, 0, sizeof(name), name);
 		name_len = utf8_strsw(name);
 	}
 	else
@@ -1424,9 +1587,11 @@ fview_position_updated(FileView *view)
 	size_t col_width;
 	size_t col_count;
 	size_t print_width;
+	size_t prefix_len = 0U;
 	column_data_t cdt = {
 		.view = view,
 		.is_current = 1,
+		.prefix_len = &prefix_len,
 	};
 
 	if(view->curr_line > view->list_rows - 1)
@@ -1434,12 +1599,25 @@ fview_position_updated(FileView *view)
 		view->curr_line = view->list_rows - 1;
 	}
 
-	if(curr_stats.load_stage < 1 ||
-			(!curr_stats.view && !window_shows_dirlist(view)))
+	if(curr_stats.load_stage < 1 || !window_shows_dirlist(view))
 	{
 		return;
 	}
 
+	if(view == other_view)
+	{
+		clear_current_line_bar(view, 0);
+		if(move_curr_line(view))
+		{
+			draw_dir_list(view);
+		}
+		else
+		{
+			put_inactive_mark(view);
+		}
+		return;
+	}
+
 	erase_current_line_bar(view);
 
 	redraw = move_curr_line(view);
@@ -1468,9 +1646,19 @@ fview_position_updated(FileView *view)
 	refresh_view_win(view);
 	update_stat_window(view, 0);
 
-	if(curr_stats.view)
+	if(view == curr_view)
 	{
-		quick_view_file(view);
+		/* We're updating view non-lazily above, so doing the same with the
+		 * ruler. */
+		ui_ruler_update(view, 0);
+
+		checked_wmove(view->win, cdt.current_line,
+				cdt.column_offset + prefix_len + (cfg.extra_padding != 0));
+
+		if(curr_stats.view)
+		{
+			qv_draw(view);
+		}
 	}
 }
 
@@ -1532,9 +1720,9 @@ fview_sorting_updated(FileView *view)
 static void
 reset_view_columns(FileView *view)
 {
-	if(view->ls_view ||
+	if(!ui_view_displays_columns(view) ||
 			(curr_stats.restart_in_progress && flist_custom_active(view) &&
-			 view->custom.unsorted))
+			 ui_view_unsorted(view)))
 	{
 		return;
 	}
