@@ -21,19 +21,24 @@
 
 #include <curses.h> /* mvwin() wbkgdset() werase() */
 
+#include <assert.h> /* assert() */
 #include <ctype.h> /* isdigit() */
 #include <stddef.h> /* NULL size_t */
-#include <stdlib.h> /* RAND_MAX rand() */
+#include <stdlib.h> /* RAND_MAX free() rand() */
 #include <string.h> /* strcat() strdup() strlen() */
 #include <time.h> /* time() */
 #include <unistd.h>
 
 #include "../cfg/config.h"
+#include "../compat/pthread.h"
 #include "../compat/reallocarray.h"
 #include "../engine/mode.h"
 #include "../modes/modes.h"
+#include "../utils/fs.h"
 #include "../utils/log.h"
 #include "../utils/macros.h"
+#include "../utils/path.h"
+#include "../utils/str.h"
 #include "../utils/string_array.h"
 #include "../utils/test_helpers.h"
 #include "../utils/utf8.h"
@@ -42,15 +47,16 @@
 #include "../filelist.h"
 #include "ui.h"
 
-#include "../utils/str.h"
-
 static void update_stat_window_old(FileView *view, int lazy_redraw);
+static void refresh_window(WINDOW *win, int lazily);
 TSTATIC char * expand_status_line_macros(FileView *view, const char format[]);
 static char * parse_view_macros(FileView *view, const char **format,
 		const char macros[], int opt);
 static int expand_num(char buf[], size_t buf_len, int val);
 static const char * get_tip(void);
 static void check_expanded_str(const char buf[], int skip, int *nexpansions);
+static pthread_spinlock_t * get_job_bar_changed_lock(void);
+static void init_job_bar_changed_lock(void);
 static int is_job_bar_visible(void);
 static void update_job_bar(void);
 static const char * format_job_bar(void);
@@ -62,6 +68,8 @@ static size_t nbar_jobs;
 static bg_op_t **bar_jobs;
 /* Whether list of jobs needs to be redrawn. */
 static int job_bar_changed;
+/* Protects accesses to job_bar_changed variable. */
+static pthread_spinlock_t job_bar_changed_lock;
 
 void
 update_stat_window(FileView *view, int lazy_redraw)
@@ -69,7 +77,7 @@ update_stat_window(FileView *view, int lazy_redraw)
 	int x;
 	char *buf;
 
-	if(!cfg.display_statusline)
+	if(!cfg.display_statusline || view != curr_view)
 	{
 		return;
 	}
@@ -99,17 +107,9 @@ update_stat_window(FileView *view, int lazy_redraw)
 	werase(stat_win);
 	checked_wmove(stat_win, 0, 0);
 	wprint(stat_win, buf);
-
-	if(lazy_redraw)
-	{
-		wnoutrefresh(stat_win);
-	}
-	else
-	{
-		wrefresh(stat_win);
-	}
-
 	free(buf);
+
+	refresh_window(stat_win, lazy_redraw);
 }
 
 /* Formats status line in the "old way" (before introduction of 'statusline'
@@ -117,16 +117,23 @@ update_stat_window(FileView *view, int lazy_redraw)
 static void
 update_stat_window_old(FileView *view, int lazy_redraw)
 {
-	const dir_entry_t *const entry = &view->dir_entry[view->list_pos];
+	const dir_entry_t *const curr = get_current_entry(view);
 	char name_buf[160*2 + 1];
 	char perm_buf[26];
-	char size_buf[56];
+	char size_buf[64];
 	char id_buf[52];
 	int x;
 	int cur_x;
 	size_t print_width;
 	char *filename;
 
+	if(curr == NULL || fentry_is_fake(curr))
+	{
+		werase(stat_win);
+		refresh_window(stat_win, lazy_redraw);
+		return;
+	}
+
 	x = getmaxx(stdscr);
 	wresize(stat_win, 1, x);
 	wbkgdset(stat_win, COLOR_PAIR(cfg.cs.pair[STATUS_LINE_COLOR]) |
@@ -135,17 +142,17 @@ update_stat_window_old(FileView *view, int lazy_redraw)
 	filename = get_current_file_name(view);
 	print_width = utf8_strsnlen(filename, 20 + MAX(0, x - 83));
 	snprintf(name_buf, MIN(sizeof(name_buf), print_width + 1), "%s", filename);
-	friendly_size_notation(entry->size, sizeof(size_buf), size_buf);
+	friendly_size_notation(curr->size, sizeof(size_buf), size_buf);
 
-	get_uid_string(entry, 0, sizeof(id_buf), id_buf);
+	get_uid_string(curr, 0, sizeof(id_buf), id_buf);
 	if(id_buf[0] != '\0')
 		strcat(id_buf, ":");
-	get_gid_string(entry, 0, sizeof(id_buf) - strlen(id_buf),
+	get_gid_string(curr, 0, sizeof(id_buf) - strlen(id_buf),
 			id_buf + strlen(id_buf));
 #ifndef _WIN32
-	get_perm_string(perm_buf, sizeof(perm_buf), entry->mode);
+	get_perm_string(perm_buf, sizeof(perm_buf), curr->mode);
 #else
-	snprintf(perm_buf, sizeof(perm_buf), "%s", attr_str_long(entry->attrs));
+	snprintf(perm_buf, sizeof(perm_buf), "%s", attr_str_long(curr->attrs));
 #endif
 
 	werase(stat_win);
@@ -171,7 +178,14 @@ update_stat_window_old(FileView *view, int lazy_redraw)
 		id_buf[0] = '\0';
 	mvwaddstr(stat_win, 0, cur_x, id_buf);
 
-	if(lazy_redraw)
+	refresh_window(stat_win, lazy_redraw);
+}
+
+/* Refreshes given window, possibly lazily. */
+static void
+refresh_window(WINDOW *win, int lazily)
+{
+	if(lazily)
 	{
 		wnoutrefresh(stat_win);
 	}
@@ -187,7 +201,7 @@ update_stat_window_old(FileView *view, int lazy_redraw)
 TSTATIC char *
 expand_status_line_macros(FileView *view, const char format[])
 {
-	return expand_view_macros(view, format, "tAugsEd-lLSz%[]");
+	return expand_view_macros(view, format, "tTfaAugsEdD-lLSz%[]");
 }
 
 /* Expands possibly limited set of view macros.  Returns newly allocated string,
@@ -206,17 +220,22 @@ static char *
 parse_view_macros(FileView *view, const char **format, const char macros[],
 		int opt)
 {
-	const dir_entry_t *const entry = &view->dir_entry[view->list_pos];
+	const dir_entry_t *const curr = get_current_entry(view);
 	char *result = strdup("");
 	size_t len = 0;
 	char c;
 	int nexpansions = 0;
 
+	if(curr == NULL || fentry_is_fake(curr))
+	{
+		return result;
+	}
+
 	while((c = **format) != '\0')
 	{
 		size_t width = 0;
 		int left_align = 0;
-		char buf[PATH_MAX];
+		char buf[PATH_MAX + 1];
 		const char *const next = ++*format;
 		int skip, ok;
 
@@ -243,26 +262,45 @@ parse_view_macros(FileView *view, const char **format, const char macros[],
 
 		skip = 0;
 		ok = 1;
+		buf[0] = '\0';
 		switch(c)
 		{
+			case 'a':
+				friendly_size_notation(get_free_space(curr_view->curr_dir), sizeof(buf),
+						buf);
+				break;
 			case 't':
-				format_entry_name(entry, sizeof(buf), buf);
+				format_entry_name(curr, sizeof(buf), buf);
+				break;
+			case 'T':
+				if(curr->type == FT_LINK)
+				{
+					char full_path[PATH_MAX + 1];
+					get_full_path_of(curr, sizeof(full_path), full_path);
+					if(get_link_target(full_path, buf, sizeof(buf)) != 0)
+					{
+						copy_str(buf, sizeof(buf), "Failed to resolve link");
+					}
+				}
+				break;
+			case 'f':
+				get_short_path_of(view, curr, 1, 0, sizeof(buf), buf);
 				break;
 			case 'A':
 #ifndef _WIN32
-				get_perm_string(buf, sizeof(buf), entry->mode);
+				get_perm_string(buf, sizeof(buf), curr->mode);
 #else
-				copy_str(buf, sizeof(buf), attr_str_long(entry->attrs));
+				copy_str(buf, sizeof(buf), attr_str_long(curr->attrs));
 #endif
 				break;
 			case 'u':
-				get_uid_string(entry, 0, sizeof(buf), buf);
+				get_uid_string(curr, 0, sizeof(buf), buf);
 				break;
 			case 'g':
-				get_gid_string(entry, 0, sizeof(buf), buf);
+				get_gid_string(curr, 0, sizeof(buf), buf);
 				break;
 			case 's':
-				friendly_size_notation(entry->size, sizeof(buf), buf);
+				friendly_size_notation(curr->size, sizeof(buf), buf);
 				break;
 			case 'E':
 				{
@@ -289,7 +327,7 @@ parse_view_macros(FileView *view, const char **format, const char macros[],
 				break;
 			case 'd':
 				{
-					struct tm *tm_ptr = localtime(&entry->mtime);
+					struct tm *tm_ptr = localtime(&curr->mtime);
 					strftime(buf, sizeof(buf), cfg.time_format, tm_ptr);
 				}
 				break;
@@ -311,6 +349,13 @@ parse_view_macros(FileView *view, const char **format, const char macros[],
 			case 'z':
 				copy_str(buf, sizeof(buf), get_tip());
 				break;
+			case 'D':
+				if(curr_stats.number_of_windows == 1)
+				{
+					FileView *const other = (view == curr_view) ? other_view : curr_view;
+					copy_str(buf, sizeof(buf), replace_home_part(other->curr_dir));
+				}
+				break;
 			case '[':
 				{
 					char *const opt_str = parse_view_macros(view, format, macros, 1);
@@ -327,11 +372,9 @@ parse_view_macros(FileView *view, const char **format, const char macros[],
 					}
 					return result;
 				}
-				else
-				{
-					LOG_INFO_MSG("Unmatched %%]");
-					ok = 0;
-				}
+
+				LOG_INFO_MSG("Unmatched %%]");
+				ok = 0;
 				break;
 
 			default:
@@ -340,6 +383,11 @@ parse_view_macros(FileView *view, const char **format, const char macros[],
 				break;
 		}
 
+		if(char_is_one_of("tTAugsEd", c) && fentry_is_fake(curr))
+		{
+			buf[0] = '\0';
+		}
+
 		if(!ok)
 		{
 			*format = next;
@@ -541,7 +589,11 @@ ui_stat_job_bar_remove(bg_op_t *bg_op)
 void
 ui_stat_job_bar_changed(bg_op_t *bg_op)
 {
+	pthread_spinlock_t *const lock = get_job_bar_changed_lock();
+
+	pthread_spin_lock(lock);
 	job_bar_changed = 1;
+	pthread_spin_unlock(lock);
 }
 
 void
@@ -555,21 +607,49 @@ ui_stat_job_bar_check_for_updates(void)
 {
 	static int prev_width;
 
-	if(job_bar_changed || getmaxx(job_bar) != prev_width)
+	pthread_spinlock_t *const lock = get_job_bar_changed_lock();
+	int job_bar_changed_value;
+
+	pthread_spin_lock(lock);
+	job_bar_changed_value = job_bar_changed;
+	job_bar_changed = 0;
+	pthread_spin_unlock(lock);
+
+	if(job_bar_changed_value || getmaxx(job_bar) != prev_width)
 	{
-		job_bar_changed = 0;
 		update_job_bar();
 	}
 
 	prev_width = getmaxx(job_bar);
 }
 
+/* Gets spinlock for the job_bar_changed variable in thread-safe way.  Returns
+ * the lock. */
+static pthread_spinlock_t *
+get_job_bar_changed_lock(void)
+{
+	static pthread_once_t once = PTHREAD_ONCE_INIT;
+	pthread_once(&once, &init_job_bar_changed_lock);
+	return &job_bar_changed_lock;
+}
+
+/* Performs spinlock initialization at most once. */
+static void
+init_job_bar_changed_lock(void)
+{
+	int ret = pthread_spin_init(&job_bar_changed_lock, PTHREAD_PROCESS_PRIVATE);
+	assert(ret == 0 && "Failed to initialize spinlock!");
+	(void)ret;
+}
+
 /* Checks whether job bar is visible.  Returns non-zero if so, and zero
  * otherwise. */
 static int
 is_job_bar_visible(void)
 {
-	return ui_stat_job_bar_height() != 0 && !is_in_menu_like_mode();
+	/* Pretend that bar isn't visible in tests. */
+	return curr_stats.load_stage >= 2
+	    && ui_stat_job_bar_height() != 0 && !is_in_menu_like_mode();
 }
 
 /* Fills job bar with up-to-date content. */
@@ -612,7 +692,8 @@ format_job_bar(void)
 	bar_text[0] = '\0';
 	text_width = 0U;
 
-	max_width = getmaxx(job_bar);
+	/* The check of stage is for tests. */
+	max_width = (curr_stats.load_stage < 2) ? 80 : getmaxx(job_bar);
 	width_used = 0U;
 	for(i = 0U; i < nbar_jobs; ++i)
 	{
