@@ -26,7 +26,6 @@
 #include <termios.h> /* struct winsize */
 #endif
 #include <sys/time.h> /* gettimeofday() */
-#include <pthread.h> /* PTHREAD_* pthread_* */
 #include <unistd.h>
 
 #include <assert.h> /* assert() */
@@ -44,6 +43,7 @@
 #include "../cfg/info.h"
 #include "../compat/curses.h"
 #include "../compat/fs_limits.h"
+#include "../compat/pthread.h"
 #include "../engine/mode.h"
 #include "../int/term_title.h"
 #include "../modes/dialogs/msg_dialog.h"
@@ -60,6 +60,7 @@
 #include "../utils/utils.h"
 #include "../event_loop.h"
 #include "../filelist.h"
+#include "../flist_sel.h"
 #include "../opt_handlers.h"
 #include "../status.h"
 #include "../vifm.h"
@@ -93,6 +94,7 @@ static void create_windows(void);
 static void update_geometry(void);
 static int get_working_area_height(void);
 static void clear_border(WINDOW *border);
+static int middle_border_is_visible(void);
 static void update_views(int reload);
 static void reload_lists(void);
 static void reload_list(FileView *view);
@@ -113,16 +115,29 @@ static void fixup_titles_attributes(const FileView *view, int active_view);
 static uint64_t get_updated_time(uint64_t prev);
 
 void
-ui_ruler_update(FileView *view)
+ui_ruler_update(FileView *view, int lazy_redraw)
 {
+	const int ruler_is_visible = !vle_mode_is(CMDLINE_MODE)
+	                          && !is_status_bar_multiline();
 	char *expanded;
 
+	if(!ruler_is_visible)
+	{
+		/* Do nothing, especially don't update layout because it might be a custom
+		 * layout at the moment. */
+		return;
+	}
+
 	update_statusbar_layout();
 
 	expanded = expand_ruler_macros(view, cfg.ruler_format);
 	expanded = break_in_two(expanded, getmaxx(ruler_win));
 
 	ui_ruler_set(expanded);
+	if(!lazy_redraw)
+	{
+		wrefresh(ruler_win);
+	}
 
 	free(expanded);
 }
@@ -174,8 +189,11 @@ setup_ncurses_interface(void)
 
 #if defined(NCURSES_EXT_FUNCS) && NCURSES_EXT_FUNCS >= 20081102
 #ifdef HAVE_SET_ESCDELAY_FUNC
-	/* Use ncurses specific function to disable delay after pressing escape key */
-	set_escdelay(0);
+	/* Use ncurses specific function to make delay after pressing escape key
+	 * unnoticeable.  Used to be zero, but in some corner cases multiple bytes
+	 * composing a functional key code might be handled to the application with a
+	 * delay. */
+	set_escdelay(5);
 #endif
 #endif
 
@@ -272,7 +290,7 @@ correct_size(FileView *view)
 /* Updates TUI elements sizes and coordinates for single window
  * configuration. */
 static void
-only_layout(FileView *view, int screen_x, int screen_y)
+only_layout(FileView *view, int screen_x)
 {
 	const int vborder_pos_correction = cfg.side_borders_visible ? 1 : 0;
 	const int vborder_size_correction = cfg.side_borders_visible ? -2 : 0;
@@ -288,7 +306,7 @@ only_layout(FileView *view, int screen_x, int screen_y)
 /* Updates TUI elements sizes and coordinates for vertical configuration of
  * panes: left one and right one. */
 static void
-vertical_layout(int screen_x, int screen_y)
+vertical_layout(int screen_x)
 {
 	const int vborder_pos_correction = cfg.side_borders_visible ? 1 : 0;
 	const int vborder_size_correction = cfg.side_borders_visible ? -1 : 0;
@@ -452,8 +470,8 @@ resize_all(void)
 
 	if(curr_stats.number_of_windows == 1)
 	{
-		only_layout(&lwin, screen_x, screen_y);
-		only_layout(&rwin, screen_x, screen_y);
+		only_layout(&lwin, screen_x);
+		only_layout(&rwin, screen_x);
 
 		mvwin(ltop_line1, 0, 0);
 		mvwin(ltop_line2, 0, 0);
@@ -465,7 +483,7 @@ resize_all(void)
 		if(curr_stats.split == HSPLIT)
 			horizontal_layout(screen_x, screen_y);
 		else
-			vertical_layout(screen_x, screen_y);
+			vertical_layout(screen_x);
 	}
 
 	correct_size(&lwin);
@@ -502,8 +520,8 @@ update_geometry(void)
 	update_term_size();
 
 #ifdef _WIN32
-	getmaxyx(stdscr, screen_y, screen_x);
-	resize_term(screen_y, screen_x);
+	/* This resizes vifm to window size. */
+	resize_term(0, 0);
 #endif
 
 	getmaxyx(stdscr, screen_y, screen_x);
@@ -521,6 +539,18 @@ update_geometry(void)
 	load_geometry();
 }
 
+int
+cv_unsorted(CVType type)
+{
+	return type == CV_VERY || cv_compare(type);
+}
+
+int
+cv_compare(CVType type)
+{
+	return type == CV_COMPARE || type == CV_DIFF;
+}
+
 void
 update_screen(UpdateType update_kind)
 {
@@ -544,7 +574,10 @@ update_screen(UpdateType update_kind)
 		clear_border(lborder);
 		clear_border(rborder);
 	}
-	clear_border(mborder);
+	if(middle_border_is_visible())
+	{
+		clear_border(mborder);
+	}
 
 	if(curr_stats.term_state != TS_NORMAL)
 	{
@@ -572,7 +605,7 @@ update_screen(UpdateType update_kind)
 		}
 		else
 		{
-			clean_status_bar();
+			ui_sb_clear();
 		}
 
 		if(vle_mode_is(VIEW_MODE))
@@ -581,7 +614,7 @@ update_screen(UpdateType update_kind)
 		}
 		else
 		{
-			ui_ruler_update(curr_view);
+			ui_ruler_update(curr_view, 1);
 		}
 	}
 
@@ -633,6 +666,14 @@ clear_border(WINDOW *border)
 	wnoutrefresh(border);
 }
 
+/* Checks whether mborder should be displayed/updated.  Returns non-zero if so,
+ * otherwise zero is returned. */
+static int
+middle_border_is_visible(void)
+{
+	return (curr_stats.number_of_windows == 2 && curr_stats.split == VSPLIT);
+}
+
 /* Updates (redraws or reloads) views. */
 static void
 update_views(int reload)
@@ -654,7 +695,7 @@ reload_lists(void)
 		ui_view_title_update(other_view);
 		if(curr_stats.view)
 		{
-			quick_view_file(curr_view);
+			qv_draw(curr_view);
 		}
 		else if(!other_view->explore_mode)
 		{
@@ -681,19 +722,22 @@ change_window(void)
 
 	load_view_options(curr_view);
 
-	if(curr_stats.number_of_windows != 1)
+	if(window_shows_dirlist(other_view))
 	{
-		if(!other_view->explore_mode)
-		{
-			put_inactive_mark(other_view);
-			erase_current_line_bar(other_view);
-		}
+		erase_current_line_bar(other_view);
 	}
 
 	if(curr_stats.view && !is_dir_list_loaded(curr_view))
 	{
+		/* This view hasn't been loaded since startup yet, do it now. */
 		navigate_to(curr_view, curr_view->curr_dir);
 	}
+	else
+	{
+		/* Change working directory, so that %c macro and other cwd-sensitive things
+		 * work as expected. */
+		(void)vifm_chdir(flist_get_dir(curr_view));
+	}
 
 	if(window_shows_dirlist(&lwin) && window_shows_dirlist(&rwin))
 	{
@@ -846,7 +890,7 @@ clear_num_window(void)
 void
 show_progress(const char msg[], int period)
 {
-	static char marks[] = {'|', '/', '-', '\\'};
+	static char marks[] = { '|', '/', '-', '\\' };
 	static int mark = 0;
 	static int count = 1;
 	static int total = 0;
@@ -870,15 +914,15 @@ show_progress(const char msg[], int period)
 	++total;
 
 	/* Skip intermediate updates to do not hammer UI with refreshes. */
-	if(period != 1 && count%period != 1)
+	if(abs(period) != 1 && count%abs(period) != 1)
 	{
 		return;
 	}
 	count = 1;
 
-	/* Assume that period equal to one means that message already contains counter
-	 * (maybe along with total number). */
-	if(period == 1)
+	/* Assume that period equal to or less than one means that message already
+	 * contains counter (maybe along with total number) or doesn't need one. */
+	if(period <= 1)
 	{
 		ui_sb_quick_msgf("%s %c", msg, marks[mark]);
 	}
@@ -899,7 +943,7 @@ redraw_lists(void)
 	{
 		if(curr_stats.view)
 		{
-			quick_view_file(curr_view);
+			qv_draw(curr_view);
 			refresh_view_win(other_view);
 		}
 		else if(!other_view->explore_mode)
@@ -1061,17 +1105,6 @@ update_term_size(void)
 #endif
 }
 
-int
-ui_term_is_alive(void)
-{
-#ifndef _WIN32
-	struct winsize ws;
-	return (ioctl(0, TIOCGWINSZ, &ws) != -1);
-#else
-	return 1;
-#endif
-}
-
 /* Re-layouts windows located on status bar (status bar itself, input and the
  * ruler). */
 static void
@@ -1354,6 +1387,10 @@ void
 ui_get_decors(const dir_entry_t *entry, const char **prefix,
 		const char **suffix)
 {
+	/* The check of actual file type can be relatively slow in some cases, so make
+	 * sure we do it only when needed and at most once. */
+	FileType type = FT_UNK;
+
 	if(entry->name_dec_num == -1)
 	{
 		/* Find a match and cache the result. */
@@ -1364,18 +1401,20 @@ ui_get_decors(const dir_entry_t *entry, const char **prefix,
 			char full_path[PATH_MAX];
 			int i;
 
-			get_full_path_of(entry, sizeof(full_path), full_path);
+			get_full_path_of(entry, sizeof(full_path) - 1U, full_path);
+			type = ui_view_entry_target_type(entry);
+			if(type == FT_DIR)
+			{
+				strcat(full_path, "/");
+			}
 
 			for(i = 0; i < cfg.name_dec_count; ++i)
 			{
 				const file_dec_t *const file_dec = &cfg.name_decs[i];
 				if(matchers_match(file_dec->matchers, full_path))
 				{
-					if(matchers_match(file_dec->matchers, full_path))
-					{
-						((dir_entry_t *)entry)->name_dec_num = i + 1;
-						break;
-					}
+					((dir_entry_t *)entry)->name_dec_num = i + 1;
+					break;
 				}
 			}
 		}
@@ -1383,7 +1422,7 @@ ui_get_decors(const dir_entry_t *entry, const char **prefix,
 
 	if(entry->name_dec_num == 0)
 	{
-		const FileType type = ui_view_entry_target_type(entry);
+		type = (type == FT_UNK) ? ui_view_entry_target_type(entry) : type;
 		*prefix = cfg.type_decs[type][DECORATION_PREFIX];
 		*suffix = cfg.type_decs[type][DECORATION_SUFFIX];
 	}
@@ -1429,7 +1468,7 @@ ui_view_win_changed(FileView *view)
 void
 ui_view_reset_selection_and_reload(FileView *view)
 {
-	clean_selected_files(view);
+	flist_sel_stash(view);
 	load_saving_pos(view, 1);
 }
 
@@ -1537,6 +1576,10 @@ static void
 print_view_title(const FileView *view, int active_view, char title[])
 {
 	const size_t title_width = getmaxx(view->title);
+	if(title_width == (size_t)-1)
+	{
+		return;
+	}
 
 	fixup_titles_attributes(view, active_view);
 	werase(view->title);
@@ -1618,7 +1661,7 @@ ui_view_sort_list_ensure_well_formed(FileView *view, char sort_keys[])
 
 	if(!found_name_key && i < SK_COUNT &&
 			(!flist_custom_active(view) ||
-			 (sort_keys == view->sort && !view->custom.unsorted)))
+			 (sort_keys == view->sort && !ui_view_unsorted(view))))
 	{
 		sort_keys[i++] = SK_DEFAULT;
 	}
@@ -1630,17 +1673,17 @@ ui_view_sort_list_ensure_well_formed(FileView *view, char sort_keys[])
 }
 
 char *
-ui_view_sort_list_get(const FileView *view)
+ui_view_sort_list_get(const FileView *view, const char sort[])
 {
-	return (flist_custom_active(view) && view->custom.unsorted)
+	return (flist_custom_active(view) && ui_view_unsorted(view))
 	     ? (char *)view->custom.sort
-	     : (char *)view->sort_g;
+	     : (char *)sort;
 }
 
 int
 ui_view_displays_numbers(const FileView *const view)
 {
-	return view->num_type != NT_NONE && !view->ls_view;
+	return view->num_type != NT_NONE && ui_view_displays_columns(view);
 }
 
 int
@@ -1660,7 +1703,9 @@ ui_view_clear_history(FileView *const view)
 int
 ui_view_displays_columns(const FileView *const view)
 {
-	return !view->ls_view;
+	return !view->ls_view
+	    || (flist_custom_active(view) &&
+					(view->custom.type == CV_TREE || cv_compare(view->custom.type)));
 }
 
 FileType
@@ -1750,6 +1795,12 @@ ui_view_wipe(FileView *view)
 	wrefresh(view->win);
 }
 
+int
+ui_view_unsorted(const FileView *view)
+{
+	return cv_unsorted(view->custom.type);
+}
+
 void
 ui_view_schedule_redraw(FileView *view)
 {
