@@ -24,7 +24,6 @@
 
 #include <curses.h>
 #include <regex.h> /* regex_t */
-#include <pthread.h> /* pthread_mutex_t */
 
 #include <stddef.h> /* size_t wchar_t */
 #include <stdint.h> /* uint64_t uint32_t */
@@ -33,13 +32,12 @@
 #include <wchar.h> /* wint_t */
 
 #include "../compat/fs_limits.h"
+#include "../compat/pthread.h"
 #include "../utils/filter.h"
 #include "../utils/fswatch.h"
-#include "../utils/trie.h"
 #include "../status.h"
 #include "../types.h"
 #include "color_scheme.h"
-#include "column_view.h"
 
 #define SORT_WIN_WIDTH 32
 
@@ -110,7 +108,10 @@ enum
 	SK_COUNT = SK_LAST,
 
 	/* Special value to use for unset options. */
-	SK_NONE = SK_LAST + 1
+	SK_NONE = SK_LAST + 1,
+
+	/* Id ordering.  Defined here to do not add it to sorting keys. */
+	SK_BY_ID = SK_NONE + 1
 };
 
 /* Type of file numbering. */
@@ -123,6 +124,26 @@ typedef enum
 }
 NumberingType;
 
+/* Variants of custom view. */
+typedef enum
+{
+	CV_REGULAR, /* Sorted list of files. */
+	CV_VERY,    /* No initial sorting of file list is enforced. */
+	CV_TREE,    /* Files of a file system sub-tree. */
+	CV_COMPARE, /* Directory comparison pane. */
+	CV_DIFF,    /* One of two directory comparison panes. */
+}
+CVType;
+
+/* Type of file comparison. */
+typedef enum
+{
+	CT_NAME,     /* Compare just names. */
+	CT_SIZE,     /* Compare file sizes. */
+	CT_CONTENTS, /* Compare file contents by combining size and hash. */
+}
+CompareType;
+
 /* Type of scheduled view update event. */
 typedef enum
 {
@@ -141,6 +162,7 @@ typedef struct
 }
 history_t;
 
+/* Description of a single directory entry. */
 typedef struct dir_entry_t
 {
 	char *name;
@@ -159,23 +181,41 @@ typedef struct dir_entry_t
 	FileType type;
 	int nlinks;       /* Number of hard links to the entry. */
 
-	int list_num;     /* Used by sorting comparer to perform stable sort. */
+	int id;           /* File uniqueness identifier. */
+
+	int tag;          /* Used to hold temporary data associated with the item,
+	                     e.g. by sorting comparer to perform stable sort or item
+	                     mapping during tree filtering. */
 
 	int hi_num;       /* File highlighting parameters cache (initially -1). */
 	int name_dec_num; /* File decoration parameters cache (initially -1).  The
 	                     value is shifted by one, 0 means type decoration. */
 
+	int child_count; /* Number of child entries (all, not just direct). */
+	int child_pos;   /* Position of this entry in among children of its parent.
+	                    Zero for top-level entries. */
+
 	int search_match;      /* Non-zero if the item matches last search.  Equals to
 	                          search match number (top to bottom order). */
 	short int match_left;  /* Starting position of search match. */
 	short int match_right; /* Ending position of search match. */
 
-	unsigned int selected : 1;
+	unsigned int selected : 1;     /* Whether file is selected. */
 	unsigned int was_selected : 1; /* Previous selection state for Visual mode. */
 	unsigned int marked : 1;       /* Whether file should be processed. */
+	unsigned int temporary : 1;    /* Whether this is temporary node. */
+	unsigned int dir_link : 1;     /* Whether this is symlink to a directory. */
 }
 dir_entry_t;
 
+/* List of entries bundled with its size. */
+typedef struct
+{
+	dir_entry_t *entries; /* List of entries. */
+	int nentries;         /* Number entries in the list. */
+}
+entries_t;
+
 typedef struct
 {
 	WINDOW *win;
@@ -184,28 +224,40 @@ typedef struct
 	/* Directory we're currently in. */
 	char curr_dir[PATH_MAX];
 
-	/* Parameters related to custom filling. */
+	/* Data related to custom filling. */
 	struct
 	{
+		/* Type of the custom view. */
+		CVType type;
+
+		/* Additional data about CV_DIFF type. */
+		CompareType diff_cmp_type; /* Type of comparison. */
+		int diff_path_group;       /* Whether entries are grouped by paths. */
+
 		/* This is temporary storage for custom list entries used during its
-		 * construction as well as storage for unfiltered custom list if local
-		 * filter is not empty. */
+		 * construction. */
 		dir_entry_t *entries; /* File entries. */
 		int entry_count;      /* Number of file entries. */
 
+		/* Title of the custom view being constructed.  Discarded if finishing
+		 * fails. */
+		char *next_title;
+
 		/* Directory we were in before custom view activation. */
 		char *orig_dir;
 		/* Title for the custom view. */
 		char *title;
 
-		/* Whether list is unsorted. */
-		int unsorted;
 		/* Previous sorting value, before unsorted custom view was loaded. */
 		char sort[SK_COUNT];
 
+		/* List of paths that should be ignored (including all nested paths).  Used
+		 * by tree-view. */
+		struct trie_t *excluded_paths;
+
 		/* Names of files in custom view while it's being composed.  Used for
 		 * duplicate elimination during construction of custom list. */
-		trie_t paths_cache;
+		struct trie_t *paths_cache;
 	}
 	custom;
 
@@ -218,9 +270,9 @@ typedef struct
 	/* Number of files that match current search pattern. */
 	int matches;
 	/* Last used search pattern, empty if none. */
-	char last_search[NAME_MAX];
+	char last_search[NAME_MAX + 1];
 
-	int hide_dot;
+	int hide_dot, hide_dot_g; /* Whether dot files are hidden. */
 	int prev_invert;
 	int invert; /* whether to invert the filename pattern */
 	int curr_line; /* current line # of the window  */
@@ -247,7 +299,7 @@ typedef struct
 	/* Filter which is controlled by user. */
 	filter_t manual_filter;
 	/* Stores previous raw value of the manual_filter to make filter restoring
-	 * possible.  Not NULL. */
+	 * possible.  Always not NULL. */
 	char *prev_manual_filter;
 
 	/* Filter which is controlled automatically and never filled by user. */
@@ -259,6 +311,10 @@ typedef struct
 	/* Various parameters related to local filter. */
 	struct
 	{
+		/* Original list of custom entries saved because otherwise we lose it. */
+		dir_entry_t *entries; /* File entries. */
+		int entry_count;      /* Number of file entries. */
+
 		/* Local filename filter. */
 		filter_t filter;
 		/* Whether interactive filtering in progress. */
@@ -289,14 +345,17 @@ typedef struct
 	/* Primary group in compiled form. */
 	regex_t primary_group;
 
-	int history_num;
-	int history_pos;
-	history_t *history;
+	int history_num;    /* Number of used history elements. */
+	int history_pos;    /* Current position in history. */
+	history_t *history; /* Directory history itself. */
 
 	col_scheme_t cs;
 
-	columns_t columns; /* handle for column_view unit */
-	char *view_columns, *view_columns_g; /* Format string of columns. */
+	/* Handle for column_view unit.  Contains view columns configuration even when
+	 * 'lsview' is on. */
+	struct columns_t *columns;
+	/* Format string that specifies view columns. */
+	char *view_columns, *view_columns_g;
 
 	/* ls-like view related fields */
 	int ls_view, ls_view_g; /* Non-zero if ls-like view is enabled. */
@@ -350,8 +409,8 @@ WINDOW *lborder;
 WINDOW *mborder;
 WINDOW *rborder;
 
-/* Updates the ruler with information from the view. */
-void ui_ruler_update(FileView *view);
+/* Updates the ruler with information from the view (possibly lazily). */
+void ui_ruler_update(FileView *view, int lazy_redraw);
 
 /* Sets text to be displayed on the ruler.  Real window update is postponed for
  * efficiency reasons. */
@@ -362,16 +421,18 @@ void ui_ruler_set(const char val[]);
  * term_state in status structure. */
 void ui_update_term_state(void);
 
-/* Performs a quick check whether terminal is functional.  Returns non-zero if
- * so, otherwise zero is returned. */
-int ui_term_is_alive(void);
-
 /* Checks whether given character was pressed, ignores any other characters. */
 int ui_char_pressed(wint_t c);
 
 int setup_ncurses_interface(void);
 
-float get_splitter_pos(int max);
+/* Checks whether custom view of specified type is unsorted.  Returns non-zero
+ * if so, otherwise zero is returned. */
+int cv_unsorted(CVType type);
+
+/* Checks whether custom view of specified type is a compare or diff view.
+ * Returns non-zero if so, otherwise zero is returned. */
+int cv_compare(CVType type);
 
 /* Redraws whole screen with possible reloading of file lists (depends on
  * argument). */
@@ -396,7 +457,8 @@ void clear_num_window(void);
 
 /* Displays progress on the status bar, not updating it frequently.  msg can't
  * be NULL.  period - how often status bar should be updated.  If period equals
- * 0 inner counter is reset, do this on start of operation. */
+ * 0 inner counter is reset, do this on start of operation.  For period <= 1,
+ * its absolute value is used and count is not printed. */
 void show_progress(const char msg[], int period);
 
 void redraw_lists(void);
@@ -494,9 +556,10 @@ int ui_view_sort_list_contains(const char sort[SK_COUNT], char key);
  * "iname" for views, except for unsorted custom view). */
 void ui_view_sort_list_ensure_well_formed(FileView *view, char sort_keys[]);
 
-/* Picks sort array for the view taking custom view into account.  Returns
- * pointer to the array. */
-char * ui_view_sort_list_get(const FileView *view);
+/* Picks sort array for the view taking custom view into account.  sort should
+ * point to sorting array preferred by default.  Returns pointer to the
+ * array. */
+char * ui_view_sort_list_get(const FileView *view, const char sort[]);
 
 /* Checks whether file numbers should be displayed for the view.  Returns
  * non-zero if so, otherwise zero is returned. */
@@ -506,7 +569,7 @@ int ui_view_displays_numbers(const FileView *const view);
  * otherwise zero is returned. */
 int ui_view_is_visible(const FileView *const view);
 
-/* Cleans directory history of the view. */
+/* Clears directory history of the view. */
 void ui_view_clear_history(FileView *const view);
 
 /* Checks whether view displays column view.  Returns non-zero if so, otherwise
@@ -546,6 +609,11 @@ void ui_view_erase(FileView *view);
  * screen (e.g. to clear anything put there by other programs as well). */
 void ui_view_wipe(FileView *view);
 
+/* Checks whether custom view type of specified view is unsorted.  It doesn't
+ * need the view to be custom, checks just the type.  Returns non-zero if so,
+ * otherwise zero is returned. */
+int ui_view_unsorted(const FileView *view);
+
 /* View update scheduling. */
 
 /* Schedules redraw of the view for the future.  Doesn't perform any actual
