@@ -31,12 +31,12 @@
 
 #include <sys/stat.h> /* S_* statbuf */
 #include <sys/types.h> /* size_t mode_t */
-#include <unistd.h> /* getcwd() pathconf() readlink() */
+#include <unistd.h> /* pathconf() readlink() */
 
 #include <errno.h> /* errno */
 #include <stddef.h> /* NULL */
 #include <stdio.h> /* snprintf() remove() */
-#include <stdlib.h> /* free() */
+#include <stdlib.h> /* free() qsort() */
 #include <string.h> /* strcpy() strdup() strlen() strncmp() strncpy() */
 
 #include "../compat/fs_limits.h"
@@ -51,7 +51,7 @@
 static int is_dir_fast(const char path[]);
 static int path_exists_internal(const char path[], const char filename[],
 		int deref);
-static int case_sensitive_paths(const char at[]);
+static int path_sorter(const void *first, const void *second);
 
 #ifndef _WIN32
 static int is_directory(const char path[], int dereference_links);
@@ -153,6 +153,11 @@ path_exists(const char path[], int deref)
 int
 path_exists_at(const char path[], const char filename[], int deref)
 {
+	if(is_path_absolute(filename))
+	{
+		LOG_ERROR_MSG("Passed absolute path where relative one is expected: %s",
+				filename);
+	}
 	return path_exists_internal(path, filename, deref);
 }
 
@@ -161,24 +166,29 @@ path_exists_at(const char path[], const char filename[], int deref)
 static int
 path_exists_internal(const char path[], const char filename[], int deref)
 {
-	const char *path_to_check;
 	char full[PATH_MAX];
 	if(path == NULL)
 	{
-		path_to_check = filename;
+		copy_str(full, sizeof(full), filename);
 	}
 	else
 	{
 		snprintf(full, sizeof(full), "%s/%s", path, filename);
-		path_to_check = full;
+	}
+
+	/* At least on Windows extra trailing slash can mess up the check, so get rid
+	 * of it. */
+	if(!is_root_dir(full))
+	{
+		chosp(full);
 	}
 
 	if(!deref)
 	{
 		struct stat st;
-		return os_lstat(path_to_check, &st) == 0;
+		return os_lstat(full, &st) == 0;
 	}
-	return os_access(path_to_check, F_OK) == 0;
+	return os_access(full, F_OK) == 0;
 }
 
 int
@@ -187,13 +197,9 @@ paths_are_same(const char s[], const char t[])
 	char s_real[PATH_MAX];
 	char t_real[PATH_MAX];
 
-	if(os_realpath(s, s_real) != s_real)
+	if(os_realpath(s, s_real) != s_real || os_realpath(t, t_real) != t_real)
 	{
-		return 0;
-	}
-	if(os_realpath(t, t_real) != t_real)
-	{
-		return 0;
+		return (stroscmp(s, t) == 0);
 	}
 	return (stroscmp(s_real, t_real) == 0);
 }
@@ -533,6 +539,53 @@ list_regular_files(const char path[], char *list[], int *len)
 	return list;
 }
 
+char **
+list_all_files(const char path[], int *len)
+{
+	DIR *dir;
+	struct dirent *d;
+	char **list = NULL;
+
+	dir = os_opendir(path);
+	if(dir == NULL)
+	{
+		*len = -1;
+		return NULL;
+	}
+
+	*len = 0;
+	while((d = os_readdir(dir)) != NULL)
+	{
+		if(!is_builtin_dir(d->d_name))
+		{
+			*len = add_to_string_array(&list, *len, 1, d->d_name);
+		}
+	}
+	os_closedir(dir);
+
+	return list;
+}
+
+char **
+list_sorted_files(const char path[], int *len)
+{
+	char **const list = list_all_files(path, len);
+	if(*len > 0)
+	{
+		qsort(list, *len, sizeof(*list), &path_sorter);
+	}
+	return list;
+}
+
+/* Wraps stroscmp() for use with qsort(). */
+static int
+path_sorter(const void *first, const void *second)
+{
+	const char *const *const a = first;
+	const char *const *const b = second;
+	return stroscmp(*a, *b);
+}
+
 int
 is_regular_file(const char path[])
 {
@@ -680,9 +733,7 @@ is_case_change(const char src[], const char dst[])
 	return strcasecmp(src, dst) == 0 && strcmp(src, dst) != 0;
 }
 
-/* Whether paths are case sensitive at specified location.  Returns non-zero if
- * so, otherwise zero is returned. */
-static int
+int
 case_sensitive_paths(const char at[])
 {
 #if HAVE_DECL__PC_CASE_SENSITIVE
@@ -777,13 +828,10 @@ count_dir_items(const char path[])
 char *
 get_cwd(char buf[], size_t size)
 {
-	if(getcwd(buf, size) == NULL)
+	if(os_getcwd(buf, size) == NULL)
 	{
 		return NULL;
 	}
-#ifdef _WIN32
-	to_forward_slash(buf);
-#endif
 	return buf;
 }
 
@@ -791,7 +839,7 @@ char *
 save_cwd(void)
 {
 	char cwd[PATH_MAX];
-	if(getcwd(cwd, sizeof(cwd)) != cwd)
+	if(os_getcwd(cwd, sizeof(cwd)) != cwd)
 	{
 		return NULL;
 	}
@@ -866,7 +914,7 @@ readlink(const char *path, char *buf, size_t len)
 int
 is_on_fat_volume(const char *path)
 {
-	char buf[NAME_MAX];
+	char buf[NAME_MAX + 1];
 	char fs[16];
 	if(is_unc_path(path))
 	{
