@@ -911,18 +911,23 @@ split_and_get(char str[], char sep, char **state)
 char *
 split_and_get_dc(char str[], char **state)
 {
-	if(str[0] == '\0')
+	if(*state == NULL)
 	{
-		return NULL;
+		/* Do nothing on empty input. */
+		if(str[0] == '\0')
+		{
+			return NULL;
+		}
 	}
-
-	if(*state != NULL)
+	else
 	{
+		/* Check if we reached end of input. */
 		if(**state == '\0')
 		{
 			return NULL;
 		}
 
+		/* Process next item of the list. */
 		str = *state;
 	}
 
@@ -1035,6 +1040,20 @@ count_lines(const char text[], int max_width)
 	return nlines;
 }
 
+size_t
+chars_in_str(const char s[], char c)
+{
+	size_t char_count = 0;
+	while(*s != '\0')
+	{
+		if(*s++ == c)
+		{
+			++char_count;
+		}
+	}
+	return char_count;
+}
+
 #ifdef _WIN32
 
 char *
