@@ -248,7 +248,7 @@ size_t copy_str(char dst[], size_t dst_len, const char src[]);
 /* Copies characters from the string pointed to by src and terminated by the
  * terminator to piece of memory of size dst_len pointed to by dst.  Ensures
  * that copied string ends with null character.  Does nothing for zero
- * dst_len. Returns number of characters written, including terminating null
+ * dst_len.  Returns number of characters written, including terminating null
  * character. */
 size_t copy_substr(char dst[], size_t dst_len, const char src[],
 		char terminator);
@@ -286,6 +286,9 @@ char * split_and_get_dc(char str[], char **state);
  * it's equal to INT_MAX. */
 int count_lines(const char text[], int max_width);
 
+/* Counts number of c char occurrences in the s string.  Returns the number. */
+size_t chars_in_str(const char s[], char c);
+
 #ifdef _WIN32
 
 /* Same as strstr(), but in case insensitive way. */
