@@ -32,6 +32,7 @@
 
 #include <ctype.h> /* isalnum() isalpha() */
 #include <errno.h> /* errno */
+#include <math.h> /* modf() pow() */
 #include <stddef.h> /* size_t */
 #include <stdio.h> /* snprintf() */
 #include <stdlib.h> /* free() malloc() */
@@ -42,6 +43,7 @@
 #include "../compat/fs_limits.h"
 #include "../compat/os.h"
 #include "../engine/keys.h"
+#include "../engine/variables.h"
 #include "../int/fuse.h"
 #include "../modes/dialogs/msg_dialog.h"
 #include "../ui/cancellation.h"
@@ -57,11 +59,28 @@
 #include "str.h"
 #include "string_array.h"
 
+static void show_progress_cb(const void *descr);
+static const char ** get_size_suffixes(void);
+static double split_size_double(double d, unsigned long long *ifraction,
+		int *fraction_width);
 #ifdef _WIN32
 static void unquote(char quoted[]);
 #endif
 static int is_line_spec(const char str[]);
 
+/* Size suffixes of different kinds. */
+static const char *iec_i_units[] = {
+	"  B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB"
+};
+static const char *iec_units[] = {
+	"B", "K", "M", "G", "T", "P", "E", "Z", "Y"
+};
+static const char *si_units[] = {
+	" B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"
+};
+ARRAY_GUARD(iec_units, ARRAY_LEN(iec_i_units));
+ARRAY_GUARD(si_units, ARRAY_LEN(iec_units));
+
 int
 vifm_system(char command[])
 {
@@ -78,7 +97,7 @@ vifm_system(char command[])
 
 int
 process_cmd_output(const char descr[], const char cmd[], int user_sh,
-		cmd_output_handler handler, void *arg)
+		int interactive, cmd_output_handler handler, void *arg)
 {
 	FILE *file, *err;
 	pid_t pid;
@@ -88,7 +107,7 @@ process_cmd_output(const char descr[], const char cmd[], int user_sh,
 
 	LOG_INFO_MSG("Capturing output of the command: %s", cmd);
 
-	pid = background_and_capture((char *)cmd, user_sh, &file, &err);
+	pid = bg_run_and_capture((char *)cmd, user_sh, &file, &err);
 	if(pid == (pid_t)-1)
 	{
 		return 1;
@@ -97,8 +116,14 @@ process_cmd_output(const char descr[], const char cmd[], int user_sh,
 	ui_cancellation_reset();
 	ui_cancellation_enable();
 
-	wait_for_data_from(pid, file, 0);
-	lines = read_stream_lines(file, &nlines, 1);
+	if(!interactive)
+	{
+		show_progress("", 0);
+	}
+
+	wait_for_data_from(pid, file, 0, &ui_cancellation_info);
+	lines = read_stream_lines(file, &nlines, 1,
+			interactive ? NULL : &show_progress_cb, descr);
 
 	ui_cancellation_disable();
 	fclose(file);
@@ -113,6 +138,13 @@ process_cmd_output(const char descr[], const char cmd[], int user_sh,
 	return 0;
 }
 
+/* Callback implementation for read_stream_lines(). */
+static void
+show_progress_cb(const void *descr)
+{
+	show_progress(descr, -250);
+}
+
 int
 vifm_chdir(const char path[])
 {
@@ -145,7 +177,7 @@ expand_envvars(const char str[], int escape_vals)
 	{
 		if(!prev_slash && *str == '$' && isalpha(str[1]))
 		{
-			char var_name[NAME_MAX];
+			char var_name[NAME_MAX + 1];
 			const char *p = str + 1;
 			char *q = var_name;
 			const char *var_value;
@@ -157,8 +189,8 @@ expand_envvars(const char str[], int escape_vals)
 			}
 			*q = '\0';
 
-			var_value = env_get(var_name);
-			if(var_value != NULL)
+			var_value = local_getenv(var_name);
+			if(!is_null_or_empty(var_value))
 			{
 				char *escaped_var_value = NULL;
 				if(escape_vals)
@@ -196,59 +228,106 @@ expand_envvars(const char str[], int escape_vals)
 }
 
 int
-friendly_size_notation(uint64_t num, int str_size, char *str)
+friendly_size_notation(uint64_t num, int str_size, char str[])
 {
-	static const char* iec_units[] = {
-		"  B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB"
-	};
-	static const char* si_units[] = {
-		"B", "K", "M", "G", "T", "P", "E", "Z", "Y"
-	};
-	ARRAY_GUARD(iec_units, ARRAY_LEN(si_units));
-
-	const char** units;
-	size_t u;
+	unsigned long long fraction = 0ULL;
+	int fraction_width;
+	const char **const units = get_size_suffixes();
+	size_t u = 0U;
 	double d = num;
 
-	units = cfg.use_iec_prefixes ? iec_units : si_units;
-
-	u = 0;
-	while(d >= 1023.5 && u < ARRAY_LEN(iec_units) - 1)
+	while(d >= cfg.sizefmt.base - 0.5 && u < ARRAY_LEN(iec_i_units) - 1U)
 	{
-		d /= 1024.0;
-		u++;
+		d /= cfg.sizefmt.base;
+		++u;
 	}
-	if(u == 0)
+
+	/* Fractional part is ignored when it's absent (we didn't divide anything) and
+	 * when we format size in previously used manner and hide it for values
+	 * greater than 9. */
+	if(u != 0U && !(cfg.sizefmt.precision == 0 && d > 9.0))
+	{
+		d = split_size_double(d, &fraction, &fraction_width);
+	}
+
+	if(fraction == 0)
 	{
 		snprintf(str, str_size, "%.0f %s", d, units[u]);
 	}
 	else
 	{
-		if(d > 9)
-			snprintf(str, str_size, "%.0f %s", d, units[u]);
-		else
-		{
-			size_t len = snprintf(str, str_size, "%.1f %s", d, units[u]);
-			if(str[len - strlen(units[u]) - 1 - 1] == '0')
-				snprintf(str, str_size, "%.0f %s", d, units[u]);
-		}
+		snprintf(str, str_size, "%.0f.%0*" PRINTF_ULL " %s", d, fraction_width, fraction,
+				units[u]);
 	}
 
-	return u > 0;
+	return u > 0U;
+}
+
+/* Picks size suffixes as per configuration.  Returns one of *_units arrays. */
+static const char **
+get_size_suffixes(void)
+{
+	return (cfg.sizefmt.base == 1000) ? si_units :
+	       (cfg.sizefmt.ieci_prefixes ? iec_i_units : iec_units);
+}
+
+/* Breaks size (floating point, not integer) into integer and fractional parts
+ * rounding and truncating them as necessary.  Sets *ifraction and
+ * *fraction_width.  Returns new value for the size (truncated and/or
+ * rounded). */
+static double
+split_size_double(double size, unsigned long long *ifraction, int *fraction_width)
+{
+	double ten_power, dfraction, integer;
+
+	*fraction_width = (cfg.sizefmt.precision == 0) ? 1 : cfg.sizefmt.precision;
+
+	ten_power = pow(10, *fraction_width + 1);
+	while(ten_power > ULLONG_MAX)
+	{
+		ten_power /= 10;
+		--*fraction_width;
+	}
+
+	dfraction = modf(size, &integer);
+	*ifraction = DIV_ROUND_UP(ten_power*dfraction, 10);
+
+	if(*ifraction >= ten_power/10.0)
+	{
+		/* Overflow into integer part during rounding. */
+		integer += 1.0;
+		*ifraction -= ten_power/10.0;
+	}
+	else if(*ifraction == 0 && dfraction != 0.0)
+	{
+		/* Fractional part is too small, "round up" to make it visible. */
+		*ifraction = 1;
+	}
+
+	/* Skip trailing zeroes. */
+	for(; *fraction_width != 0 && *ifraction%10 == 0; --*fraction_width)
+	{
+		*ifraction /= 10;
+	}
+
+	return (*ifraction == 0) ? size : integer;
 }
 
 const char *
 enclose_in_dquotes(const char str[])
 {
-	static char buf[PATH_MAX];
+	static char buf[1 + PATH_MAX*2 + 1 + 1];
 	char *p;
 
 	p = buf;
 	*p++ = '"';
 	while(*str != '\0')
 	{
-		if(*str == '\\' || *str == '"')
+		if(*str == '\\' || *str == '"' ||
+				(curr_stats.shell_type == ST_NORMAL && (*str == '$' || *str == '`')))
+		{
 			*p++ = '\\';
+		}
 		*p++ = *str;
 
 		str++;
@@ -266,12 +345,13 @@ make_name_unique(const char filename[])
 	int i;
 
 #ifndef _WIN32
-	len = snprintf(unique, sizeof(unique), "%s_%u%u_00", filename, getppid(),
+	snprintf(unique, sizeof(unique), "%s_%u%u_00", filename, getppid(),
 			get_pid());
 #else
 	/* TODO: think about better name uniqualization on Windows. */
-	len = snprintf(unique, sizeof(unique), "%s_%u_00", filename, get_pid());
+	snprintf(unique, sizeof(unique), "%s_%u_00", filename, get_pid());
 #endif
+	len = strlen(unique);
 	i = 0;
 
 	while(path_exists(unique, DEREF))
@@ -535,11 +615,12 @@ def_count(int count)
 }
 
 char *
-parse_file_spec(const char spec[], int *line_num)
+parse_file_spec(const char spec[], int *line_num, const char cwd[])
 {
 	char *path_buf;
 	const char *colon;
-	const size_t bufs_len = 2 + strlen(spec) + 1 + 1;
+	const size_t bufs_len = strlen(cwd) + 1U + strlen(spec) + 1U + 1U;
+	char canonicalized[PATH_MAX];
 
 	path_buf = malloc(bufs_len);
 	if(path_buf == NULL)
@@ -553,7 +634,7 @@ parse_file_spec(const char spec[], int *line_num)
 	}
 	else
 	{
-		copy_str(path_buf, bufs_len, "./");
+		snprintf(path_buf, bufs_len, "%s/", cwd);
 	}
 
 #ifdef _WIN32
@@ -598,12 +679,21 @@ parse_file_spec(const char spec[], int *line_num)
 	}
 
 	chomp(path_buf);
+	canonicalize_path(path_buf, canonicalized, sizeof(canonicalized));
 
 #ifdef _WIN32
-	to_forward_slash(path_buf);
+	to_forward_slash(canonicalized);
 #endif
 
-	return replace_tilde(path_buf);
+	if(!ends_with_slash(path_buf) && !is_root_dir(canonicalized) &&
+			strcmp(canonicalized, "./") != 0)
+	{
+		chosp(canonicalized);
+	}
+
+	free(path_buf);
+
+	return replace_tilde(strdup(canonicalized));
 }
 
 /* Checks whether str points to a valid line number.  Returns non-zero if so,
