@@ -28,11 +28,13 @@
 
 #include <sys/select.h> /* select() FD_SET FD_ZERO */
 #include <sys/stat.h> /* O_* S_* */
+#include <sys/statvfs.h> /* statvfs statvfs() */
 #include <sys/time.h> /* timeval futimens() utimes() */
 #include <sys/types.h> /* gid_t mode_t pid_t uid_t */
 #include <sys/wait.h> /* waitpid */
 #include <fcntl.h> /* open() close() */
 #include <grp.h> /* getgrnam() getgrgid_r() */
+#include <pthread.h> /* pthread_sigmask() */
 #include <pwd.h> /* getpwnam() getpwuid_r() */
 #include <unistd.h> /* X_OK dup() dup2() getpid() isatty() pause() sysconf()
                        ttyname() */
@@ -40,9 +42,8 @@
 #include <assert.h> /* assert() */
 #include <ctype.h> /* isdigit() */
 #include <errno.h> /* EINTR errno */
-#include <signal.h> /* SIGINT SIGTSTP SIGCHLD SIG_DFL SIG_BLOCK SIG_UNBLOCK
-                       sigset_t kill() sigaddset() sigemptyset() signal()
-                       sigprocmask() */
+#include <signal.h> /* SIG* SIG_* sigset_t kill() sigaddset() sigemptyset()
+                       sigfillset() signal() sigprocmask() */
 #include <stddef.h> /* NULL size_t */
 #include <stdio.h> /* FILE stderr fdopen() fprintf() snprintf() */
 #include <stdlib.h> /* atoi() free() */
@@ -53,9 +54,9 @@
 #include "../compat/mntent.h" /* mntent setmntent() getmntent() endmntent() */
 #include "../compat/os.h"
 #include "../compat/reallocarray.h"
-#include "../ui/cancellation.h"
 #include "../running.h"
 #include "../status.h"
+#include "cancellation.h"
 #include "env.h"
 #include "filemon.h"
 #include "fs.h"
@@ -154,7 +155,8 @@ recover_after_shellout(void)
 }
 
 void
-wait_for_data_from(pid_t pid, FILE *f, int fd)
+wait_for_data_from(pid_t pid, FILE *f, int fd,
+		const struct cancellation_t *cancellation)
 {
 	const struct timeval ts_init = { .tv_sec = 0, .tv_usec = 1000 };
 	struct timeval ts;
@@ -167,7 +169,7 @@ wait_for_data_from(pid_t pid, FILE *f, int fd)
 
 	do
 	{
-		process_cancel_request(pid);
+		process_cancel_request(pid, cancellation);
 		ts = ts_init;
 		FD_SET(fd, &read_ready);
 		select_result = select(fd + 1, &read_ready, NULL, NULL, &ts);
@@ -187,9 +189,17 @@ set_sigchld(int block)
 }
 
 void
-process_cancel_request(pid_t pid)
+block_all_thread_signals(void)
 {
-	if(ui_cancellation_requested())
+	sigset_t set;
+	sigfillset(&set);
+	pthread_sigmask(SIG_SETMASK, &set, NULL);
+}
+
+void
+process_cancel_request(pid_t pid, const struct cancellation_t *cancellation)
+{
+	if(cancellation_requested(cancellation))
 	{
 		if(kill(pid, SIGINT) != 0)
 		{
@@ -226,39 +236,36 @@ run_from_fork(int pipe[2], int err_only, char cmd[])
 {
 	int nullfd;
 
+	/* Close read end of the pipe. */
+	(void)close(pipe[0]);
+
 	/* Redirect stderr and maybe stdout to write end of the pipe. */
 	if(dup2(pipe[1], STDERR_FILENO) == -1)
 	{
 		_Exit(EXIT_FAILURE);
 	}
-	if(err_only)
+	if(!err_only && dup2(pipe[1], STDOUT_FILENO) == -1)
 	{
-		close(STDOUT_FILENO);
-	}
-	else
-	{
-		if(dup2(pipe[1], STDOUT_FILENO) == -1)
-		{
-			_Exit(EXIT_FAILURE);
-		}
+		_Exit(EXIT_FAILURE);
 	}
 
-	/* Close read end of pipe. */
-	close(pipe[0]);
+	/* Close write end of the pipe (already duplicated it). */
+	(void)close(pipe[1]);
 
-	close(STDIN_FILENO);
-
-	/* Send stdout, stdin to /dev/null */
-	if((nullfd = open("/dev/null", O_RDONLY)) != -1)
+	/* Send stdin and maybe stdout to /dev/null */
+	nullfd = open("/dev/null", O_RDWR);
+	if(nullfd == -1)
 	{
-		if(dup2(nullfd, STDIN_FILENO) == -1)
-		{
-			_Exit(EXIT_FAILURE);
-		}
-		if(err_only && dup2(nullfd, STDOUT_FILENO) == -1)
-		{
-			_Exit(EXIT_FAILURE);
-		}
+		_Exit(EXIT_FAILURE);
+	}
+
+	if(dup2(nullfd, STDIN_FILENO) == -1)
+	{
+		_Exit(EXIT_FAILURE);
+	}
+	if(err_only && dup2(nullfd, STDOUT_FILENO) == -1)
+	{
+		_Exit(EXIT_FAILURE);
 	}
 
 	execvp(get_execv_path(cfg.shell), make_execv_array(cfg.shell, cmd));
@@ -268,7 +275,7 @@ run_from_fork(int pipe[2], int err_only, char cmd[])
 char *
 get_execv_path(char shell[])
 {
-	static char name[NAME_MAX];
+	static char name[NAME_MAX + 1];
 	(void)extract_cmd_name(shell, 0, sizeof(name), name);
 	return name;
 }
@@ -291,7 +298,7 @@ make_execv_array(char shell[], char cmd[])
 	const size_t npieces = 1U;
 #endif
 
-	char name[NAME_MAX];
+	char name[NAME_MAX + 1];
 
 	char **args = reallocarray(NULL, 4 + npieces + 1, sizeof(*args));
 	char *eval_cmd;
@@ -402,7 +409,7 @@ refers_to_slower_fs(const char from[], const char to[])
 }
 
 int
-is_on_slow_fs(const char full_path[])
+is_on_slow_fs(const char full_path[], const char slowfs_specs[])
 {
 	char fs_name[PATH_MAX];
 	get_mount_point_traverser_state state = {
@@ -414,7 +421,7 @@ is_on_slow_fs(const char full_path[])
 	};
 
 	/* Empty list optimization. */
-	if(cfg.slow_fs_list[0] == '\0')
+	if(slowfs_specs[0] == '\0')
 	{
 		return 0;
 	}
@@ -422,7 +429,7 @@ is_on_slow_fs(const char full_path[])
 	/* If slowfs equals "*" then all file systems are considered slow.  On cygwin
 	 * obtaining list of mounts from /etc/mtab, which is linked to /proc/mounts,
 	 * is very slow in presence of network drives. */
-	if(strcmp(cfg.slow_fs_list, "*") == 0)
+	if(strcmp(slowfs_specs, "*") == 0)
 	{
 		return 1;
 	}
@@ -431,14 +438,14 @@ is_on_slow_fs(const char full_path[])
 	{
 		if(state.curr_len > 0)
 		{
-			if(starts_with_list_item(fs_name, cfg.slow_fs_list))
+			if(starts_with_list_item(fs_name, slowfs_specs))
 			{
 				return 1;
 			}
 		}
 	}
 
-	return find_path_prefix_index(full_path, cfg.slow_fs_list) != -1;
+	return find_path_prefix_index(full_path, slowfs_specs) != -1;
 }
 
 int
@@ -1027,5 +1034,17 @@ clone_timestamps(const char path[], const char from[], const struct stat *st)
 	utimes(path, tv);
 }
 
+uint64_t
+get_free_space(const char at[])
+{
+	struct statvfs st;
+	if(statvfs(at, &st) != 0)
+	{
+		return 0;
+	}
+
+	return (uint64_t)st.f_bsize*st.f_bfree;
+}
+
 /* vim: set tabstop=2 softtabstop=2 shiftwidth=2 noexpandtab cinoptions-=(0 : */
 /* vim: set cinoptions+=t0 filetype=c : */
