@@ -38,6 +38,7 @@
 
 #include "cfg/config.h"
 #include "cfg/info.h"
+#include "compat/reallocarray.h"
 #include "engine/autocmds.h"
 #include "compat/fs_limits.h"
 #include "engine/cmds.h"
@@ -65,6 +66,7 @@
 #include "utils/path.h"
 #include "utils/str.h"
 #include "utils/string_array.h"
+#include "utils/utf8.h"
 #include "utils/utils.h"
 #include "args.h"
 #include "background.h"
@@ -76,8 +78,10 @@
 #include "dir_stack.h"
 #include "event_loop.h"
 #include "filelist.h"
-#include "fileops.h"
 #include "filetype.h"
+#include "flist_hist.h"
+#include "flist_pos.h"
+#include "fops_common.h"
 #include "ipc.h"
 #include "marks.h"
 #include "ops.h"
@@ -89,12 +93,13 @@
 #include "trash.h"
 #include "undo.h"
 
+static int vifm_main(int argc, char *argv[]);
 static int pair_in_use(short int pair);
 static void move_pair(short int from, short int to);
 static int undo_perform_func(OPS op, void *data, const char src[],
 		const char dst[]);
 static void parse_received_arguments(char *args[]);
-static void remote_cd(FileView *view, const char *path, int handle);
+static void remote_cd(FileView *view, const char path[], int handle);
 static void check_path_for_file(FileView *view, const char path[], int handle);
 static int need_to_switch_active_pane(const char lwin_path[],
 		const char rwin_path[]);
@@ -105,16 +110,54 @@ static void _gnuc_noreturn vifm_leave(int exit_code, int cquit);
 /* Command-line arguments in parsed form. */
 static args_t vifm_args;
 
+#ifndef _WIN32
+
 int
 main(int argc, char *argv[])
 {
-	/* TODO: refactor main() function */
+	return vifm_main(argc, argv);
+}
+
+#else
+
+int
+main()
+{
+	extern int _CRT_glob;
+	extern void __wgetmainargs(int *, wchar_t ***, wchar_t ***, int, int *);
+
+	char **utf8_argv;
+	int i;
+
+	wchar_t **envp, **argv;
+	int argc, si = 0;
+	__wgetmainargs(&argc, &argv, &envp, _CRT_glob, &si);
+
+	utf8_argv = reallocarray(NULL, argc + 1, sizeof(*utf8_argv));
+
+	for(i = 0; i < argc; ++i)
+	{
+		utf8_argv[i] = utf8_from_utf16(argv[i]);
+	}
+	utf8_argv[i] = NULL;
+
+	return vifm_main(argc, utf8_argv);
+}
+
+#endif
+
+/* Entry-point.  Has same semantics as main(). */
+static int
+vifm_main(int argc, char *argv[])
+{
+	/* TODO: refactor vifm_main() function */
 
 	static const int quit = 0;
 
 	char dir[PATH_MAX];
 	char **files = NULL;
 	int nfiles = 0;
+	int lwin_cv, rwin_cv;
 
 	if(get_cwd(dir, sizeof(dir)) == NULL)
 	{
@@ -126,12 +169,14 @@ main(int argc, char *argv[])
 	args_parse(&vifm_args, argc, argv, dir);
 	args_process(&vifm_args, 1);
 
-	if(strcmp(vifm_args.lwin_path, "-") == 0 ||
-			strcmp(vifm_args.rwin_path, "-") == 0)
+	lwin_cv = (strcmp(vifm_args.lwin_path, "-") == 0 && vifm_args.lwin_handle);
+	rwin_cv = (strcmp(vifm_args.rwin_path, "-") == 0 && vifm_args.rwin_handle);
+	if(lwin_cv || rwin_cv)
 	{
-		files = read_stream_lines(stdin, &nfiles, 1);
+		files = read_stream_lines(stdin, &nfiles, 1, NULL, NULL);
 		if(reopen_term_stdin() != 0)
 		{
+			free_string_array(files, nfiles);
 			return EXIT_FAILURE;
 		}
 	}
@@ -159,6 +204,8 @@ main(int argc, char *argv[])
 
 	if(init_status(&cfg) != 0)
 	{
+		free_string_array(files, nfiles);
+
 		puts("Error during session status initialization.");
 		return -1;
 	}
@@ -179,11 +226,19 @@ main(int argc, char *argv[])
 	}
 
 	ipc_init(vifm_args.server_name, &parse_received_arguments);
+	/* Export chosen server name to parsing unit. */
+	{
+		var_val_t value = { .string = (char *)ipc_get_name() };
+		var_t var = var_new(VTYPE_STRING, value);
+		setvar("v:servername", var);
+		var_free(var);
+	}
+
 	args_process(&vifm_args, 0);
 
-	init_background();
+	bg_init();
 
-	init_fileops(&enter_prompt_mode, &prompt_msg_custom);
+	fops_init(&enter_prompt_mode, &prompt_msg_custom);
 
 	set_view_path(&lwin, vifm_args.lwin_path);
 	set_view_path(&rwin, vifm_args.rwin_path);
@@ -206,11 +261,13 @@ main(int argc, char *argv[])
 	curr_stats.original_stdout = reopen_term_stdout();
 	if(curr_stats.original_stdout == NULL)
 	{
+		free_string_array(files, nfiles);
 		return -1;
 	}
 
 	if(!setup_ncurses_interface())
 	{
+		free_string_array(files, nfiles);
 		return -1;
 	}
 
@@ -237,19 +294,19 @@ main(int argc, char *argv[])
 	{
 		load_scheme();
 		cfg_load();
-
-		if(strcmp(vifm_args.lwin_path, "-") == 0)
-		{
-			flist_set(&lwin, "-", dir, files, nfiles);
-		}
-		else if(strcmp(vifm_args.rwin_path, "-") == 0)
-		{
-			flist_set(&rwin, "-", dir, files, nfiles);
-		}
 	}
-	/* Load colors in any case to load color pairs. */
-	cs_load_pairs();
 
+	if(lwin_cv)
+	{
+		flist_custom_set(&lwin, "-", dir, files, nfiles);
+	}
+	else if(rwin_cv)
+	{
+		flist_custom_set(&rwin, "-", dir, files, nfiles);
+	}
+	free_string_array(files, nfiles);
+
+	cs_load_pairs();
 	cs_write();
 	setup_signals();
 
@@ -267,18 +324,22 @@ main(int argc, char *argv[])
 	 * history.  This is not done automatically as history manipulation should be
 	 * postponed until views are fully loaded, otherwise there is no correct
 	 * information about current file and relative cursor position. */
-	save_view_history(&lwin, NULL, NULL, -1);
-	save_view_history(&rwin, NULL, NULL, -1);
+	flist_hist_save(&lwin, NULL, NULL, -1);
+	flist_hist_save(&rwin, NULL, NULL, -1);
 
 	/* Trigger auto-commands for initial directories. */
-	vle_aucmd_execute("DirEnter", lwin.curr_dir, &lwin);
-	vle_aucmd_execute("DirEnter", rwin.curr_dir, &rwin);
-
-	exec_startup_commands(&vifm_args);
+	(void)vifm_chdir(flist_get_dir(&lwin));
+	vle_aucmd_execute("DirEnter", flist_get_dir(&lwin), &lwin);
+	(void)vifm_chdir(flist_get_dir(&rwin));
+	vle_aucmd_execute("DirEnter", flist_get_dir(&rwin), &rwin);
 
 	update_screen(UT_FULL);
 	modes_update();
 
+	/* Run startup commands after loading file lists into views, so that commands
+	 * like +1 work. */
+	exec_startup_commands(&vifm_args);
+
 	curr_stats.load_stage = 3;
 
 	event_loop(&quit);
@@ -380,12 +441,12 @@ parse_received_arguments(char *argv[])
 		change_window();
 	}
 
-	clean_status_bar();
+	ui_sb_clear();
 	curr_stats.save_msg = 0;
 }
 
 static void
-remote_cd(FileView *view, const char *path, int handle)
+remote_cd(FileView *view, const char path[], int handle)
 {
 	char buf[PATH_MAX];
 
@@ -399,9 +460,9 @@ remote_cd(FileView *view, const char *path, int handle)
 		leave_view_mode();
 	}
 
-	if(curr_stats.view)
+	if(curr_stats.view && (handle || view == other_view))
 	{
-		toggle_quick_view();
+		qv_toggle();
 	}
 
 	copy_str(buf, sizeof(buf), path);
@@ -416,7 +477,7 @@ remote_cd(FileView *view, const char *path, int handle)
 static void
 check_path_for_file(FileView *view, const char path[], int handle)
 {
-	if(path[0] == '\0' || is_dir(path) || strcmp(path, "-") == 0)
+	if(path[0] == '\0' || is_dir(path) || (handle && strcmp(path, "-") == 0))
 	{
 		return;
 	}
@@ -505,7 +566,7 @@ vifm_restart(void)
 	reset_undo_list();
 
 	/* Directory stack. */
-	clean_stack();
+	dir_stack_clear();
 
 	/* Registers. */
 	regs_reset();
@@ -515,15 +576,15 @@ vifm_restart(void)
 	bmarks_clear();
 
 	/* Reset variables. */
-	clear_variables();
+	clear_envvars();
 	init_variables();
 	/* This update is needed as clear_variables() will reset $PATH. */
 	update_path_env(1);
 
 	reset_views();
 	read_info_file(1);
-	save_view_history(&lwin, NULL, NULL, -1);
-	save_view_history(&rwin, NULL, NULL, -1);
+	flist_hist_save(&lwin, NULL, NULL, -1);
+	flist_hist_save(&rwin, NULL, NULL, -1);
 
 	/* Color schemes. */
 	if(stroscmp(curr_stats.color_scheme, DEF_CS_NAME) != 0 &&
@@ -546,8 +607,10 @@ vifm_restart(void)
 	curr_stats.restart_in_progress = 0;
 
 	/* Trigger auto-commands for initial directories. */
-	vle_aucmd_execute("DirEnter", lwin.curr_dir, &lwin);
-	vle_aucmd_execute("DirEnter", rwin.curr_dir, &rwin);
+	(void)vifm_chdir(flist_get_dir(&lwin));
+	vle_aucmd_execute("DirEnter", flist_get_dir(&lwin), &lwin);
+	(void)vifm_chdir(flist_get_dir(&rwin));
+	vle_aucmd_execute("DirEnter", flist_get_dir(&rwin), &rwin);
 
 	update_screen(UT_REDRAW);
 }
@@ -559,6 +622,9 @@ exec_startup_commands(const args_t *args)
 	size_t i;
 	for(i = 0; i < args->ncmds; ++i)
 	{
+		/* Make sure we're executing commands in correct directory. */
+		(void)vifm_chdir(flist_get_dir(curr_view));
+
 		(void)exec_commands(args->cmds[i], curr_view, CIT_COMMAND);
 	}
 }
@@ -586,12 +652,12 @@ vifm_try_leave(int write_info, int cquit, int force)
 	{
 		vim_write_empty_file_list();
 	}
-
 #ifdef _WIN32
-	system("cls");
+	erase();
+	refresh();
 #endif
-
 	endwin();
+
 	vifm_leave(EXIT_SUCCESS, cquit);
 }
 
