@@ -1,24 +1,27 @@
 #include <stic.h>
 
+#include <sys/stat.h> /* chmod() */
+
 #include <sys/types.h> /* stat */
 #include <sys/stat.h> /* stat */
 #include <unistd.h> /* lstat() */
 
+#include "../../src/compat/fs_limits.h"
 #include "../../src/compat/os.h"
 #include "../../src/io/iop.h"
 #include "../../src/utils/fs.h"
-#include "../../src/utils/fs_limits.h"
 #include "../../src/utils/utils.h"
 
 #include "utils.h"
 
+static void file_is_copied(const char original[]);
 static int not_windows(void);
 
 TEST(dir_is_not_copied)
 {
 	io_args_t args = {
-		.arg1.src = "../existing-files",
-		.arg2.dst = "existing-files",
+		.arg1.src = TEST_DATA_PATH "/existing-files",
+		.arg2.dst = SANDBOX_PATH "/existing-files",
 	};
 	ioe_errlst_init(&args.result.errors);
 
@@ -30,12 +33,12 @@ TEST(dir_is_not_copied)
 
 TEST(empty_file_is_copied)
 {
-	create_test_file("empty");
+	create_test_file(SANDBOX_PATH "/empty");
 
 	{
 		io_args_t args = {
-			.arg1.src = "empty",
-			.arg2.dst = "empty-copy",
+			.arg1.src = SANDBOX_PATH "/empty",
+			.arg2.dst = SANDBOX_PATH "/empty-copy",
 		};
 		ioe_errlst_init(&args.result.errors);
 
@@ -44,21 +47,22 @@ TEST(empty_file_is_copied)
 		assert_int_equal(0, args.result.errors.error_count);
 	}
 
-	assert_true(files_are_identical("empty", "empty-copy"));
+	assert_true(files_are_identical(SANDBOX_PATH "/empty",
+				SANDBOX_PATH "/empty-copy"));
 
-	delete_test_file("empty");
-	delete_test_file("empty-copy");
+	delete_test_file(SANDBOX_PATH "/empty");
+	delete_test_file(SANDBOX_PATH "/empty-copy");
 }
 
 TEST(file_is_not_overwritten_if_not_asked)
 {
-	create_test_file("empty");
-	clone_test_file("empty", "empty-copy");
+	create_test_file(SANDBOX_PATH "/empty");
+	clone_test_file(SANDBOX_PATH "/empty", SANDBOX_PATH "/empty-copy");
 
 	{
 		io_args_t args = {
-			.arg1.src = "../read/two-lines",
-			.arg2.dst = "empty-copy",
+			.arg1.src = TEST_DATA_PATH "/read/two-lines",
+			.arg2.dst = SANDBOX_PATH "/empty-copy",
 			.arg3.crs = IO_CRS_FAIL,
 		};
 		ioe_errlst_init(&args.result.errors);
@@ -69,19 +73,21 @@ TEST(file_is_not_overwritten_if_not_asked)
 		ioe_errlst_free(&args.result.errors);
 	}
 
-	assert_true(files_are_identical("empty", "empty-copy"));
-	assert_false(files_are_identical("../read/two-lines", "empty-copy"));
+	assert_true(files_are_identical(SANDBOX_PATH "/empty",
+				SANDBOX_PATH "/empty-copy"));
+	assert_false(files_are_identical(SANDBOX_PATH "/read/two-lines",
+				SANDBOX_PATH "/empty-copy"));
 
-	delete_test_file("empty");
-	delete_test_file("empty-copy");
+	delete_test_file(SANDBOX_PATH "/empty");
+	delete_test_file(SANDBOX_PATH "/empty-copy");
 }
 
 TEST(file_is_overwritten_if_asked)
 {
 	{
 		io_args_t args = {
-			.arg1.src = "../read/two-lines",
-			.arg2.dst = "two-lines",
+			.arg1.src = TEST_DATA_PATH "/read/two-lines",
+			.arg2.dst = SANDBOX_PATH "/two-lines",
 			.arg3.crs = IO_CRS_FAIL,
 		};
 		ioe_errlst_init(&args.result.errors);
@@ -91,12 +97,13 @@ TEST(file_is_overwritten_if_asked)
 		assert_int_equal(0, args.result.errors.error_count);
 	}
 
-	assert_false(files_are_identical("../read/binary-data", "two-lines"));
+	assert_false(files_are_identical(TEST_DATA_PATH "/read/binary-data",
+				SANDBOX_PATH "/two-lines"));
 
 	{
 		io_args_t args = {
-			.arg1.src = "../read/binary-data",
-			.arg2.dst = "two-lines",
+			.arg1.src = TEST_DATA_PATH "/read/binary-data",
+			.arg2.dst = SANDBOX_PATH "/two-lines",
 			.arg3.crs = IO_CRS_REPLACE_FILES,
 		};
 		ioe_errlst_init(&args.result.errors);
@@ -106,9 +113,42 @@ TEST(file_is_overwritten_if_asked)
 		assert_int_equal(0, args.result.errors.error_count);
 	}
 
-	assert_true(files_are_identical("../read/binary-data", "two-lines"));
+	assert_true(files_are_identical(TEST_DATA_PATH "/read/binary-data",
+				SANDBOX_PATH "/two-lines"));
 
-	delete_test_file("two-lines");
+	delete_test_file(SANDBOX_PATH "/two-lines");
+}
+
+TEST(block_size_file_is_copied)
+{
+	file_is_copied(TEST_DATA_PATH "/various-sizes/block-size-file");
+}
+
+TEST(block_size_minus_one_file_is_copied)
+{
+	file_is_copied(TEST_DATA_PATH "/various-sizes/block-size-minus-one-file");
+}
+
+TEST(block_size_plus_one_file_is_copied)
+{
+	file_is_copied(TEST_DATA_PATH "/various-sizes/block-size-plus-one-file");
+}
+
+TEST(double_block_size_file_is_copied)
+{
+	file_is_copied(TEST_DATA_PATH "/various-sizes/double-block-size-file");
+}
+
+TEST(double_block_size_minus_one_file_is_copied)
+{
+	file_is_copied(TEST_DATA_PATH
+			"/various-sizes/double-block-size-minus-one-file");
+}
+
+TEST(double_block_size_plus_one_file_is_copied)
+{
+	file_is_copied(TEST_DATA_PATH
+			"/various-sizes/double-block-size-plus-one-file");
 }
 
 static void
@@ -117,7 +157,7 @@ file_is_copied(const char original[])
 	{
 		io_args_t args = {
 			.arg1.src = original,
-			.arg2.dst = "copy",
+			.arg2.dst = SANDBOX_PATH "/copy",
 		};
 		ioe_errlst_init(&args.result.errors);
 
@@ -126,53 +166,25 @@ file_is_copied(const char original[])
 		assert_int_equal(0, args.result.errors.error_count);
 	}
 
-	assert_true(files_are_identical("copy", original));
+	assert_true(files_are_identical(SANDBOX_PATH "/copy", original));
 
-	delete_test_file("copy");
-}
-
-TEST(block_size_file_is_copied)
-{
-	file_is_copied("../various-sizes/block-size-file");
-}
-
-TEST(block_size_minus_one_file_is_copied)
-{
-	file_is_copied("../various-sizes/block-size-minus-one-file");
-}
-
-TEST(block_size_plus_one_file_is_copied)
-{
-	file_is_copied("../various-sizes/block-size-plus-one-file");
-}
-
-TEST(double_block_size_file_is_copied)
-{
-	file_is_copied("../various-sizes/double-block-size-file");
-}
-
-TEST(double_block_size_minus_one_file_is_copied)
-{
-	file_is_copied("../various-sizes/double-block-size-minus-one-file");
-}
-
-TEST(double_block_size_plus_one_file_is_copied)
-{
-	file_is_copied("../various-sizes/double-block-size-plus-one-file");
+	delete_test_file(SANDBOX_PATH "/copy");
 }
 
 TEST(appending_works_for_files)
 {
 	uint64_t size;
 
-	clone_test_file("../various-sizes/block-size-minus-one-file", "appending");
+	clone_test_file(TEST_DATA_PATH "/various-sizes/block-size-minus-one-file",
+			SANDBOX_PATH "/appending");
+	assert_success(chmod(SANDBOX_PATH "/appending", 0700));
 
-	size = get_file_size("appending");
+	size = get_file_size(SANDBOX_PATH "/appending");
 
 	{
 		io_args_t args = {
-			.arg1.src = "../various-sizes/block-size-file",
-			.arg2.dst = "appending",
+			.arg1.src = TEST_DATA_PATH "/various-sizes/block-size-file",
+			.arg2.dst = SANDBOX_PATH "/appending",
 			.arg3.crs = IO_CRS_APPEND_TO_FILES,
 		};
 		ioe_errlst_init(&args.result.errors);
@@ -182,12 +194,13 @@ TEST(appending_works_for_files)
 		assert_int_equal(0, args.result.errors.error_count);
 	}
 
-	assert_int_equal(size + 1, get_file_size("appending"));
+	assert_int_equal(size + 1, get_file_size(SANDBOX_PATH "/appending"));
+	assert_success(chmod(SANDBOX_PATH "/appending", 0700));
 
 	{
 		io_args_t args = {
-			.arg1.src = "../various-sizes/block-size-plus-one-file",
-			.arg2.dst = "appending",
+			.arg1.src = TEST_DATA_PATH "/various-sizes/block-size-plus-one-file",
+			.arg2.dst = SANDBOX_PATH "/appending",
 			.arg3.crs = IO_CRS_APPEND_TO_FILES,
 		};
 		ioe_errlst_init(&args.result.errors);
@@ -197,23 +210,24 @@ TEST(appending_works_for_files)
 		assert_int_equal(0, args.result.errors.error_count);
 	}
 
-	assert_int_equal(size + 2, get_file_size("appending"));
+	assert_int_equal(size + 2, get_file_size(SANDBOX_PATH "/appending"));
 
-	delete_test_file("appending");
+	delete_test_file(SANDBOX_PATH "/appending");
 }
 
 TEST(appending_does_not_shrink_files)
 {
 	uint64_t size;
 
-	clone_test_file("../read/two-lines", "two-lines");
+	clone_test_file(TEST_DATA_PATH "/read/two-lines", SANDBOX_PATH "/two-lines");
+	assert_success(chmod(SANDBOX_PATH "/two-lines", 0700));
 
-	size = get_file_size("two-lines");
+	size = get_file_size(SANDBOX_PATH "/two-lines");
 
 	{
 		io_args_t args = {
-			.arg1.src = "../existing-files/a",
-			.arg2.dst = "two-lines",
+			.arg1.src = TEST_DATA_PATH "/existing-files/a",
+			.arg2.dst = SANDBOX_PATH "/two-lines",
 			.arg3.crs = IO_CRS_APPEND_TO_FILES,
 		};
 		ioe_errlst_init(&args.result.errors);
@@ -223,11 +237,11 @@ TEST(appending_does_not_shrink_files)
 		assert_int_equal(0, args.result.errors.error_count);
 	}
 
-	assert_int_equal(size, get_file_size("two-lines"));
+	assert_int_equal(size, get_file_size(SANDBOX_PATH "/two-lines"));
 
 	{
 		io_args_t args = {
-			.arg1.path = "two-lines",
+			.arg1.path = SANDBOX_PATH "/two-lines",
 		};
 		ioe_errlst_init(&args.result.errors);
 
@@ -243,19 +257,19 @@ TEST(file_permissions_are_preserved, IF(not_windows))
 	struct stat src;
 	struct stat dst;
 
-	create_test_file("file");
+	create_test_file(SANDBOX_PATH "/file");
 
-	assert_int_equal(0, chmod("file", 0200));
+	assert_int_equal(0, chmod(SANDBOX_PATH "/file", 0200));
 
-	assert_int_equal(0, lstat("file", &src));
+	assert_int_equal(0, lstat(SANDBOX_PATH "/file", &src));
 	assert_false((src.st_mode & 0777) == 0600);
 
-	assert_int_equal(0, chmod("file", 0600));
+	assert_int_equal(0, chmod(SANDBOX_PATH "/file", 0600));
 
 	{
 		io_args_t args = {
-			.arg1.src = "file",
-			.arg2.dst = "file-copy",
+			.arg1.src = SANDBOX_PATH "/file",
+			.arg2.dst = SANDBOX_PATH "/file-copy",
 		};
 		ioe_errlst_init(&args.result.errors);
 
@@ -264,12 +278,12 @@ TEST(file_permissions_are_preserved, IF(not_windows))
 		assert_int_equal(0, args.result.errors.error_count);
 	}
 
-	assert_int_equal(0, lstat("file", &src));
-	assert_int_equal(0, lstat("file-copy", &dst));
+	assert_int_equal(0, lstat(SANDBOX_PATH "/file", &src));
+	assert_int_equal(0, lstat(SANDBOX_PATH "/file-copy", &dst));
 	assert_int_equal(src.st_mode & 0777, dst.st_mode & 0777);
 
-	delete_test_file("file");
-	delete_test_file("file-copy");
+	delete_test_file(SANDBOX_PATH "/file");
+	delete_test_file(SANDBOX_PATH "/file-copy");
 }
 
 /* Creating symbolic links on Windows requires administrator rights. */
@@ -280,8 +294,8 @@ TEST(file_symlink_copy_is_symlink, IF(not_windows))
 
 	{
 		io_args_t args = {
-			.arg1.path = "../read/two-lines",
-			.arg2.target = "sym-link",
+			.arg1.path = TEST_DATA_PATH "/read/two-lines",
+			.arg2.target = SANDBOX_PATH "/sym-link",
 		};
 		ioe_errlst_init(&args.result.errors);
 
@@ -290,12 +304,12 @@ TEST(file_symlink_copy_is_symlink, IF(not_windows))
 		assert_int_equal(0, args.result.errors.error_count);
 	}
 
-	assert_true(is_symlink("sym-link"));
+	assert_true(is_symlink(SANDBOX_PATH "/sym-link"));
 
 	{
 		io_args_t args = {
-			.arg1.src = "sym-link",
-			.arg2.dst = "sym-link-copy",
+			.arg1.src = SANDBOX_PATH "/sym-link",
+			.arg2.dst = SANDBOX_PATH "/sym-link-copy",
 		};
 		ioe_errlst_init(&args.result.errors);
 
@@ -304,29 +318,56 @@ TEST(file_symlink_copy_is_symlink, IF(not_windows))
 		assert_int_equal(0, args.result.errors.error_count);
 	}
 
-	assert_true(is_symlink("sym-link"));
-	assert_true(is_symlink("sym-link-copy"));
+	assert_true(is_symlink(SANDBOX_PATH "/sym-link"));
+	assert_true(is_symlink(SANDBOX_PATH "/sym-link-copy"));
 
-	assert_success(get_link_target("sym-link", old_target, sizeof(old_target)));
-	assert_success(get_link_target("sym-link-copy", new_target,
+	assert_success(get_link_target(SANDBOX_PATH "/sym-link", old_target,
+				sizeof(old_target)));
+	assert_success(get_link_target(SANDBOX_PATH "/sym-link-copy", new_target,
 				sizeof(new_target)));
 
 	assert_string_equal(new_target, old_target);
 
-	delete_test_file("sym-link");
-	delete_test_file("sym-link-copy");
+	delete_test_file(SANDBOX_PATH "/sym-link");
+	delete_test_file(SANDBOX_PATH "/sym-link-copy");
+}
+
+TEST(timestamps_are_preserved)
+{
+	struct stat src;
+	struct stat dst;
+
+	assert_int_equal(0, lstat(TEST_DATA_PATH "/existing-files/a", &src));
+
+	{
+		io_args_t args = {
+			.arg1.src = TEST_DATA_PATH "/existing-files/a",
+			.arg2.dst = SANDBOX_PATH "/a-copy",
+		};
+		ioe_errlst_init(&args.result.errors);
+
+		assert_success(iop_cp(&args));
+
+		assert_int_equal(0, args.result.errors.error_count);
+	}
+
+	assert_int_equal(0, lstat(SANDBOX_PATH "/a-copy", &dst));
+	assert_ulong_equal(src.st_atime, dst.st_atime);
+	assert_ulong_equal(src.st_mtime, dst.st_mtime);
+
+	delete_test_file(SANDBOX_PATH "/a-copy");
 }
 
 /* Creating symbolic links on Windows requires administrator rights. */
 TEST(dir_symlink_copy_is_symlink, IF(not_windows))
 {
-	os_mkdir("dir", 0700);
-	assert_true(is_dir("dir"));
+	os_mkdir(SANDBOX_PATH "/dir", 0700);
+	assert_true(is_dir(SANDBOX_PATH "/dir"));
 
 	{
 		io_args_t args = {
 			.arg1.path = "dir",
-			.arg2.target = "dir-sym-link",
+			.arg2.target = SANDBOX_PATH "/dir-sym-link",
 		};
 		ioe_errlst_init(&args.result.errors);
 
@@ -335,13 +376,13 @@ TEST(dir_symlink_copy_is_symlink, IF(not_windows))
 		assert_int_equal(0, args.result.errors.error_count);
 	}
 
-	assert_true(is_symlink("dir-sym-link"));
-	assert_true(is_dir("dir-sym-link"));
+	assert_true(is_symlink(SANDBOX_PATH "/dir-sym-link"));
+	assert_true(is_dir(SANDBOX_PATH "/dir-sym-link"));
 
 	{
 		io_args_t args = {
-			.arg1.src = "dir-sym-link",
-			.arg2.dst = "dir-sym-link-copy",
+			.arg1.src = SANDBOX_PATH "/dir-sym-link",
+			.arg2.dst = SANDBOX_PATH "/dir-sym-link-copy",
 		};
 		ioe_errlst_init(&args.result.errors);
 
@@ -350,17 +391,17 @@ TEST(dir_symlink_copy_is_symlink, IF(not_windows))
 		assert_int_equal(0, args.result.errors.error_count);
 	}
 
-	assert_true(is_symlink("dir-sym-link"));
-	assert_true(is_dir("dir-sym-link"));
-	assert_true(is_symlink("dir-sym-link-copy"));
-	assert_true(is_dir("dir-sym-link-copy"));
+	assert_true(is_symlink(SANDBOX_PATH "/dir-sym-link"));
+	assert_true(is_dir(SANDBOX_PATH "/dir-sym-link"));
+	assert_true(is_symlink(SANDBOX_PATH "/dir-sym-link-copy"));
+	assert_true(is_dir(SANDBOX_PATH "/dir-sym-link-copy"));
 
-	delete_test_file("dir-sym-link");
-	delete_test_file("dir-sym-link-copy");
+	delete_test_file(SANDBOX_PATH "/dir-sym-link");
+	delete_test_file(SANDBOX_PATH "/dir-sym-link-copy");
 
 	{
 		io_args_t args = {
-			.arg1.path = "dir",
+			.arg1.path = SANDBOX_PATH "/dir",
 		};
 		ioe_errlst_init(&args.result.errors);
 
@@ -370,7 +411,7 @@ TEST(dir_symlink_copy_is_symlink, IF(not_windows))
 	}
 }
 
-/* Windows lacks definitions some declarations. */
+/* Windows lacks definitions of some declarations. */
 #ifndef _WIN32
 
 /* No named fifo in file systems on Windows. */
@@ -379,8 +420,8 @@ TEST(fifo_is_copied, IF(not_windows))
 	struct stat st;
 
 	io_args_t args = {
-		.arg1.src = "fifo-src",
-		.arg2.dst = "fifo-dst",
+		.arg1.src = SANDBOX_PATH "/fifo-src",
+		.arg2.dst = SANDBOX_PATH "/fifo-dst",
 
 		.result.errors = IOE_ERRLST_INIT,
 	};
@@ -390,26 +431,37 @@ TEST(fifo_is_copied, IF(not_windows))
 	assert_success(iop_cp(&args));
 	assert_int_equal(0, args.result.errors.error_count);
 
-	assert_success(lstat("fifo-dst", &st));
+	assert_success(lstat(SANDBOX_PATH "/fifo-dst", &st));
 	assert_true(S_ISFIFO(st.st_mode));
 
-	args.arg1.path = "fifo-src";
+	args.arg1.path = SANDBOX_PATH "/fifo-src";
 	assert_success(iop_rmfile(&args));
 	assert_int_equal(0, args.result.errors.error_count);
 
-	args.arg1.path = "fifo-dst";
+	args.arg1.path = SANDBOX_PATH "/fifo-dst";
 	assert_success(iop_rmfile(&args));
 	assert_int_equal(0, args.result.errors.error_count);
 }
 
-/* No named sockets in file systems on Windows. */
-TEST(socket_is_copied, IF(not_windows))
+static int
+can_create_sockets(void)
+{
+	if(mknod(SANDBOX_PATH "/sock", S_IFSOCK | 0755, 0) != 0)
+	{
+		return 0;
+	}
+	remove(SANDBOX_PATH "/sock");
+	return 1;
+}
+
+/* Socket creation might not be available on some file systems. */
+TEST(socket_is_copied, IF(can_create_sockets))
 {
 	struct stat st;
 
 	io_args_t args = {
-		.arg1.src = "sock-src",
-		.arg2.dst = "sock-dst",
+		.arg1.src = SANDBOX_PATH "/sock-src",
+		.arg2.dst = SANDBOX_PATH "/sock-dst",
 
 		.result.errors = IOE_ERRLST_INIT,
 	};
@@ -419,14 +471,14 @@ TEST(socket_is_copied, IF(not_windows))
 	assert_success(iop_cp(&args));
 	assert_int_equal(0, args.result.errors.error_count);
 
-	assert_success(lstat("sock-dst", &st));
+	assert_success(lstat(SANDBOX_PATH "/sock-dst", &st));
 	assert_true(S_ISSOCK(st.st_mode));
 
-	args.arg1.path = "sock-src";
+	args.arg1.path = SANDBOX_PATH "/sock-src";
 	assert_success(iop_rmfile(&args));
 	assert_int_equal(0, args.result.errors.error_count);
 
-	args.arg1.path = "sock-dst";
+	args.arg1.path = SANDBOX_PATH "/sock-dst";
 	assert_success(iop_rmfile(&args));
 	assert_int_equal(0, args.result.errors.error_count);
 }
@@ -440,4 +492,4 @@ not_windows(void)
 }
 
 /* vim: set tabstop=2 softtabstop=2 shiftwidth=2 noexpandtab cinoptions-=(0 : */
-/* vim: set cinoptions+=t0 : */
+/* vim: set cinoptions+=t0 filetype=c : */
