@@ -1,7 +1,5 @@
 #include <stic.h>
 
-#include <stdio.h> /* FILE fopen() fclose() */
-
 #include <unistd.h> /* F_OK access() rmdir() */
 
 #include "../../src/compat/os.h"
@@ -9,16 +7,16 @@
 #include "../../src/utils/fs.h"
 #include "../../src/utils/utils.h"
 
-static int not_windows(void);
+#include "utils.h"
 
-static const char *const FILE_NAME = "file-to-remove";
-static const char *const DIRECTORY_NAME = "directory-to-remove";
+#define DIRECTORY_NAME SANDBOX_PATH "/directory-to-remove"
+#define FILE_NAME SANDBOX_PATH "/file-to-remove"
+
+static int not_windows(void);
 
 TEST(file_is_removed)
 {
-	FILE *const f = fopen(FILE_NAME, "w");
-	fclose(f);
-	assert_success(access(FILE_NAME, F_OK));
+	create_test_file(FILE_NAME);
 
 	{
 		io_args_t args = {
@@ -58,9 +56,7 @@ TEST(directory_is_not_removed)
 /* Creating symbolic links on Windows requires administrator rights. */
 TEST(symlink_is_removed_but_not_its_target, IF(not_windows))
 {
-	FILE *const f = fopen(FILE_NAME, "w");
-	fclose(f);
-	assert_success(access(FILE_NAME, F_OK));
+	create_test_file(FILE_NAME);
 
 	{
 		io_args_t args = {
@@ -107,4 +103,4 @@ not_windows(void)
 }
 
 /* vim: set tabstop=2 softtabstop=2 shiftwidth=2 noexpandtab cinoptions-=(0 : */
-/* vim: set cinoptions+=t0 : */
+/* vim: set cinoptions+=t0 filetype=c : */
