@@ -1,6 +1,6 @@
 #include <stic.h>
 
-#include <unistd.h> /* F_OK access() chdir() rmdir() symlink() */
+#include <unistd.h> /* F_OK access() chdir() rmdir() symlink() unlink() */
 
 #include <stdio.h> /* remove() */
 #include <string.h> /* strcpy() strdup() */
@@ -10,28 +10,27 @@
 #include "../../src/cfg/config.h"
 #include "../../src/engine/cmds.h"
 #include "../../src/engine/functions.h"
-#include "../../src/int/file_magic.h"
+#include "../../src/engine/keys.h"
 #include "../../src/modes/modes.h"
+#include "../../src/ui/statusbar.h"
+#include "../../src/ui/ui.h"
 #include "../../src/utils/dynarray.h"
 #include "../../src/utils/env.h"
 #include "../../src/utils/fs.h"
+#include "../../src/utils/matcher.h"
 #include "../../src/utils/path.h"
 #include "../../src/utils/str.h"
 #include "../../src/builtin_functions.h"
 #include "../../src/cmd_core.h"
 #include "../../src/filelist.h"
-#include "../../src/filetype.h"
+#include "../../src/filtering.h"
 #include "../../src/ops.h"
+#include "../../src/registers.h"
 #include "../../src/undo.h"
 
 #include "utils.h"
 
 static int builtin_cmd(const cmd_info_t* cmd_info);
-static int exec_func(OPS op, void *data, const char *src, const char *dst);
-static int op_avail(OPS op);
-static void check_filetype(void);
-static int prog_exists(const char name[]);
-static int has_mime_type_detection(void);
 
 static const cmd_add_t commands[] = {
 	{ .name = "builtin",       .abbr = NULL,  .id = -1,      .descr = "descr",
@@ -45,38 +44,22 @@ static const cmd_add_t commands[] = {
 static int called;
 static int bg;
 static char *arg;
+static char *saved_cwd;
 
-static char cwd[PATH_MAX];
-static char sandbox[PATH_MAX];
-static char test_data[PATH_MAX];
+static char cwd[PATH_MAX + 1];
+static char sandbox[PATH_MAX + 1];
+static char test_data[PATH_MAX + 1];
 
 SETUP_ONCE()
 {
 	assert_non_null(get_cwd(cwd, sizeof(cwd)));
 
-	if(is_path_absolute(SANDBOX_PATH))
-	{
-		snprintf(sandbox, sizeof(sandbox), "%s", SANDBOX_PATH);
-	}
-	else
-	{
-		snprintf(sandbox, sizeof(sandbox), "%s/%s", cwd, SANDBOX_PATH);
-	}
-
-	if(is_path_absolute(TEST_DATA_PATH))
-	{
-		snprintf(test_data, sizeof(test_data), "%s", TEST_DATA_PATH);
-	}
-	else
-	{
-		snprintf(test_data, sizeof(test_data), "%s/%s", cwd, TEST_DATA_PATH);
-	}
+	make_abs_path(sandbox, sizeof(sandbox), SANDBOX_PATH, "", cwd);
+	make_abs_path(test_data, sizeof(test_data), TEST_DATA_PATH, "", cwd);
 }
 
 SETUP()
 {
-	static int max_undo_levels = 0;
-
 	view_setup(&lwin);
 	view_setup(&rwin);
 
@@ -107,11 +90,15 @@ SETUP()
 
 	called = 0;
 
-	init_undo_list(&exec_func, &op_avail, NULL, &max_undo_levels);
+	undo_setup();
+
+	saved_cwd = save_cwd();
 }
 
 TEARDOWN()
 {
+	restore_cwd(saved_cwd);
+
 	update_string(&cfg.cd_path, NULL);
 	update_string(&cfg.fuse_home, NULL);
 	update_string(&cfg.slow_fs_list, NULL);
@@ -123,7 +110,8 @@ TEARDOWN()
 	view_teardown(&rwin);
 
 	reset_cmds();
-	reset_undo_list();
+
+	undo_teardown();
 }
 
 static int
@@ -140,18 +128,6 @@ builtin_cmd(const cmd_info_t* cmd_info)
 	return 0;
 }
 
-static int
-exec_func(OPS op, void *data, const char *src, const char *dst)
-{
-	return 0;
-}
-
-static int
-op_avail(OPS op)
-{
-	return op == OP_MOVE;
-}
-
 TEST(space_amp)
 {
 	assert_success(exec_commands("builtin &", &lwin, CIT_COMMAND));
@@ -276,6 +252,19 @@ TEST(double_cd_uses_same_base_for_rel_paths)
 	assert_true(paths_are_equal(rwin.curr_dir, path));
 }
 
+TEST(envvars_of_commands_come_from_variables_unit)
+{
+	assert_success(chdir(test_data));
+
+	strcpy(lwin.curr_dir, test_data);
+
+	assert_false(is_root_dir(lwin.curr_dir));
+	assert_success(exec_commands("let $ABCDE = '/'", &lwin, CIT_COMMAND));
+	env_set("ABCDE", SANDBOX_PATH);
+	assert_success(exec_commands("cd $ABCDE", &lwin, CIT_COMMAND));
+	assert_true(is_root_dir(lwin.curr_dir));
+}
+
 TEST(cpmv_does_not_crash_on_wrong_list_access)
 {
 	char path[PATH_MAX];
@@ -377,6 +366,29 @@ TEST(tr_extends_second_field)
 	assert_success(remove(path));
 }
 
+TEST(putting_files_works)
+{
+	char path[PATH_MAX + 1];
+
+	regs_init();
+
+	assert_success(os_mkdir(SANDBOX_PATH "/empty-dir", 0700));
+	assert_success(flist_load_tree(&lwin, sandbox));
+
+	make_abs_path(path, sizeof(path), TEST_DATA_PATH, "read/binary-data", cwd);
+	assert_success(regs_append(DEFAULT_REG_NAME, path));
+	lwin.list_pos = 1;
+
+	assert_true(exec_commands("put", &lwin, CIT_COMMAND) != 0);
+	restore_cwd(saved_cwd);
+	saved_cwd = save_cwd();
+
+	assert_success(unlink(SANDBOX_PATH "/empty-dir/binary-data"));
+	assert_success(rmdir(SANDBOX_PATH "/empty-dir"));
+
+	regs_reset();
+}
+
 TEST(put_bg_cmd_is_parsed_correctly)
 {
 	/* Simulate custom view to force failure of the command. */
@@ -425,6 +437,7 @@ TEST(wincmd_can_switch_views)
 
 	opt_handlers_teardown();
 	assert_success(reset_status(&cfg));
+	vle_keys_reset();
 }
 
 TEST(yank_works_with_ranges)
@@ -437,7 +450,7 @@ TEST(yank_works_with_ranges)
 	flist_custom_start(&lwin, "test");
 	snprintf(path, sizeof(path), "%s/%s", test_data, "existing-files/a");
 	flist_custom_add(&lwin, path);
-	assert_true(flist_custom_finish(&lwin, 0) == 0);
+	assert_true(flist_custom_finish(&lwin, CV_REGULAR, 0) == 0);
 
 	reg = regs_find(DEFAULT_REG_NAME);
 	assert_non_null(reg);
@@ -449,135 +462,60 @@ TEST(yank_works_with_ranges)
 	regs_reset();
 }
 
-TEST(filetype_accepts_negated_patterns)
-{
-	ft_init(&prog_exists);
-
-	assert_success(exec_commands("filetype !{*.tar} prog", &lwin, CIT_COMMAND));
-
-	check_filetype();
-
-	ft_reset(0);
-}
-
-TEST(filextype_accepts_negated_patterns)
-{
-	ft_init(&prog_exists);
-	curr_stats.exec_env_type = EET_EMULATOR_WITH_X;
-
-	assert_success(exec_commands("filextype !{*.tar} prog", &lwin, CIT_COMMAND));
-
-	check_filetype();
-
-	curr_stats.exec_env_type = EET_LINUX_NATIVE;
-	ft_reset(1);
-}
-
-TEST(fileviewer_accepts_negated_patterns)
-{
-	ft_init(&prog_exists);
-
-	assert_success(exec_commands("fileviewer !{*.tar} view", &lwin, CIT_COMMAND));
-	assert_string_equal("view", ft_get_viewer("file.version.tar.bz2"));
-
-	ft_reset(0);
-}
-
-TEST(pattern_anding_and_orring_failures)
-{
-	/* No matching is performed, so we can use application/octet-stream. */
-	assert_failure(exec_commands("filetype /*/,"
-				"<application/octet-stream>{binary-data} app", &lwin, CIT_COMMAND));
-	assert_failure(exec_commands("fileviewer /*/,"
-				"<application/octet-stream>{binary-data} viewer", &lwin, CIT_COMMAND));
-}
-
-TEST(pattern_anding_and_orring, IF(has_mime_type_detection))
-{
-	char cmd[1024];
-	assoc_records_t ft;
-
-	ft_init(&prog_exists);
-
-	snprintf(cmd, sizeof(cmd),
-			"filetype {two-lines}<text/plain>,<%s>{binary-data} app",
-			get_mimetype(TEST_DATA_PATH "/read/binary-data"));
-	assert_success(exec_commands(cmd, &lwin, CIT_COMMAND));
-	snprintf(cmd, sizeof(cmd),
-			"fileviewer {two-lines}<text/plain>,<%s>{binary-data} viewer",
-			get_mimetype(TEST_DATA_PATH "/read/binary-data"));
-	assert_success(exec_commands(cmd, &lwin, CIT_COMMAND));
-
-	ft = ft_get_all_programs(TEST_DATA_PATH "/read/two-lines");
-	assert_int_equal(1, ft.count);
-	if(ft.count == 1)
-	{
-		assert_string_equal("app", ft.list[0].command);
-	}
-	ft_assoc_records_free(&ft);
-
-	ft = ft_get_all_programs(TEST_DATA_PATH "/read/binary-data");
-	assert_int_equal(1, ft.count);
-	if(ft.count == 1)
-	{
-		assert_string_equal("app", ft.list[0].command);
-	}
-	ft_assoc_records_free(&ft);
-
-	ft = ft_get_all_programs(TEST_DATA_PATH "/read/utf8-bom");
-	assert_int_equal(0, ft.count);
-	ft_assoc_records_free(&ft);
-
-	assert_string_equal("viewer",
-			ft_get_viewer(TEST_DATA_PATH "/read/two-lines"));
-	assert_string_equal("viewer",
-			ft_get_viewer(TEST_DATA_PATH "/read/binary-data"));
-	assert_string_equal(NULL, ft_get_viewer(TEST_DATA_PATH "/read/utf8-bom"));
-
-	ft_reset(0);
-}
-
 TEST(symlinks_in_paths_are_not_resolved, IF(not_windows))
 {
-	char canonic_path[PATH_MAX];
+	char canonic_path[PATH_MAX + 1];
+	char buf[PATH_MAX + 1];
 
 	assert_success(os_mkdir(SANDBOX_PATH "/dir1", 0700));
 	assert_success(os_mkdir(SANDBOX_PATH "/dir1/dir2", 0700));
 
 	/* symlink() is not available on Windows, but the rest of the code is fine. */
 #ifndef _WIN32
-	assert_success(symlink(SANDBOX_PATH "/dir1/dir2", SANDBOX_PATH "/dir-link"));
+	{
+		char src[PATH_MAX + 1], dst[PATH_MAX + 1];
+		make_abs_path(src, sizeof(src), SANDBOX_PATH, "dir1/dir2", saved_cwd);
+		make_abs_path(dst, sizeof(dst), SANDBOX_PATH, "dir-link", saved_cwd);
+		assert_success(symlink(src, dst));
+	}
 #endif
 
 	assert_success(chdir(SANDBOX_PATH "/dir-link"));
-	to_canonic_path(SANDBOX_PATH "/dir-link", "/fake-root", lwin.curr_dir,
+	make_abs_path(buf, sizeof(buf), SANDBOX_PATH, "dir-link", saved_cwd);
+	to_canonic_path(buf, "/fake-root", lwin.curr_dir,
 			sizeof(lwin.curr_dir));
-	to_canonic_path(SANDBOX_PATH, "/fake-root", canonic_path,
+	to_canonic_path(sandbox, "/fake-root", canonic_path,
 			sizeof(canonic_path));
 
 	/* :mkdir */
 	(void)exec_commands("mkdir ../dir", &lwin, CIT_COMMAND);
-	assert_true(is_dir(SANDBOX_PATH "/dir"));
+	restore_cwd(saved_cwd);
+	saved_cwd = save_cwd();
 	assert_success(rmdir(SANDBOX_PATH "/dir"));
 
 	/* :clone file name. */
 	create_file(SANDBOX_PATH "/dir-link/file");
 	populate_dir_list(&lwin, 1);
 	(void)exec_commands("clone ../file-clone", &lwin, CIT_COMMAND);
+	restore_cwd(saved_cwd);
+	saved_cwd = save_cwd();
 	assert_success(remove(SANDBOX_PATH "/file-clone"));
 	assert_success(remove(SANDBOX_PATH "/dir-link/file"));
 
 	/* :colorscheme */
-	strcpy(cfg.colors_dir, TEST_DATA_PATH "/scripts/");
-	assert_success(
-			exec_commands("colorscheme set-env " SANDBOX_PATH "/../dir-link/..",
-				&lwin, CIT_COMMAND));
+	make_abs_path(cfg.colors_dir, sizeof(cfg.colors_dir), TEST_DATA_PATH,
+			"scripts/", saved_cwd);
+	snprintf(buf, sizeof(buf), "colorscheme set-env %s/../dir-link/..",
+			sandbox);
+	assert_success(exec_commands(buf, &lwin, CIT_COMMAND));
 	cs_load_defaults();
 
 	/* :cd */
 	assert_success(exec_commands("cd ../dir-link/..", &lwin, CIT_COMMAND));
 	assert_string_equal(canonic_path, lwin.curr_dir);
 
+	restore_cwd(saved_cwd);
+	saved_cwd = save_cwd();
 	assert_success(remove(SANDBOX_PATH "/dir-link"));
 	assert_success(rmdir(SANDBOX_PATH "/dir1/dir2"));
 	assert_success(rmdir(SANDBOX_PATH "/dir1"));
@@ -590,7 +528,7 @@ TEST(find_command, IF(not_windows))
 	replace_string(&cfg.shell, "/bin/sh");
 
 	assert_success(chdir(TEST_DATA_PATH));
-	strcpy(lwin.curr_dir, TEST_DATA_PATH);
+	strcpy(lwin.curr_dir, test_data);
 
 	assert_success(exec_commands("set findprg='find %s %a %u'", &lwin,
 				CIT_COMMAND));
@@ -599,7 +537,7 @@ TEST(find_command, IF(not_windows))
 	assert_failure(exec_commands("find", &lwin, CIT_COMMAND));
 
 	assert_success(exec_commands("find a", &lwin, CIT_COMMAND));
-	assert_int_equal(2, lwin.list_rows);
+	assert_int_equal(3, lwin.list_rows);
 
 	assert_success(exec_commands("find . -name aaa", &lwin, CIT_COMMAND));
 	assert_int_equal(1, lwin.list_rows);
@@ -611,7 +549,7 @@ TEST(find_command, IF(not_windows))
 	view_setup(&lwin);
 
 	/* Repeat last search. */
-	strcpy(lwin.curr_dir, TEST_DATA_PATH);
+	strcpy(lwin.curr_dir, test_data);
 	assert_success(exec_commands("find", &lwin, CIT_COMMAND));
 	assert_int_equal(4, lwin.list_rows);
 
@@ -625,7 +563,7 @@ TEST(grep_command, IF(not_windows))
 	replace_string(&cfg.shell, "/bin/sh");
 
 	assert_success(chdir(TEST_DATA_PATH "/scripts"));
-	strcpy(lwin.curr_dir, TEST_DATA_PATH "/scripts");
+	assert_non_null(get_cwd(lwin.curr_dir, sizeof(lwin.curr_dir)));
 
 	assert_success(exec_commands("set grepprg='grep -n -H -r %i %a %s %u'", &lwin,
 				CIT_COMMAND));
@@ -643,31 +581,204 @@ TEST(grep_command, IF(not_windows))
 	opt_handlers_teardown();
 }
 
-static void
-check_filetype(void)
+TEST(touch)
 {
-	assoc_records_t ft;
+	to_canonic_path(SANDBOX_PATH, cwd, lwin.curr_dir, sizeof(lwin.curr_dir));
+	(void)exec_commands("touch file", &lwin, CIT_COMMAND);
 
-	ft = ft_get_all_programs("file.version.tar");
-	assert_int_equal(0, ft.count);
-	ft_assoc_records_free(&ft);
-
-	ft = ft_get_all_programs("file.version.tar.bz");
-	assert_int_equal(1, ft.count);
-	assert_string_equal("prog", ft.list[0].command);
-	ft_assoc_records_free(&ft);
+	assert_success(remove(SANDBOX_PATH "/file"));
 }
 
-static int
-prog_exists(const char name[])
+TEST(compare)
 {
-	return 1;
+	opt_handlers_setup();
+
+	create_file(SANDBOX_PATH "/file");
+
+	to_canonic_path(SANDBOX_PATH, cwd, lwin.curr_dir, sizeof(lwin.curr_dir));
+
+	/* The file is empty so nothing to do when "skipempty" is specified. */
+	assert_success(exec_commands("compare ofone skipempty", &lwin, CIT_COMMAND));
+	assert_false(flist_custom_active(&lwin));
+
+	(void)exec_commands("compare byname bysize bycontents listall listdups "
+			"listunique ofboth ofone groupids grouppaths", &lwin, CIT_COMMAND);
+	assert_true(flist_custom_active(&lwin));
+	assert_int_equal(CV_REGULAR, lwin.custom.type);
+
+	assert_success(remove(SANDBOX_PATH "/file"));
+
+	opt_handlers_teardown();
 }
 
-static int
-has_mime_type_detection(void)
+TEST(screen)
 {
-	return get_mimetype(TEST_DATA_PATH "/read/dos-line-endings") != NULL;
+	assert_false(cfg.use_term_multiplexer);
+
+	/* :screen toggles the option. */
+	assert_success(exec_commands("screen", &lwin, CIT_COMMAND));
+	assert_true(cfg.use_term_multiplexer);
+	assert_success(exec_commands("screen", &lwin, CIT_COMMAND));
+	assert_false(cfg.use_term_multiplexer);
+
+	/* :screen! sets it to on. */
+	assert_success(exec_commands("screen!", &lwin, CIT_COMMAND));
+	assert_true(cfg.use_term_multiplexer);
+	assert_success(exec_commands("screen!", &lwin, CIT_COMMAND));
+	assert_true(cfg.use_term_multiplexer);
+
+	cfg.use_term_multiplexer = 0;
+}
+
+TEST(map_commands_count_arguments_correctly)
+{
+	init_modes();
+
+	/* Each map command below should receive two arguments: "\\" and "j". */
+	/* Each unmap command below should receive single argument: "\\". */
+	assert_success(exec_commands("cmap \\ j", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("cnoremap \\ j", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("cunmap \\", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("dmap \\ j", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("dnoremap \\ j", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("dunmap \\", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("mmap \\ j", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("mnoremap \\ j", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("munmap \\", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("nmap \\ j", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("nnoremap \\ j", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("nunmap \\", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("map \\ j", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("noremap \\ j", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("unmap \\", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("map! \\ j", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("noremap! \\ j", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("unmap! \\", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("qmap \\ j", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("qnoremap \\ j", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("qunmap \\", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("vmap \\ j", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("vnoremap \\ j", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("vunmap \\", &lwin, CIT_COMMAND));
+
+	vle_keys_reset();
+}
+
+TEST(filter_prints_empty_filters_correctly)
+{
+	const char *expected = "Filter -- Flags -- Value\n"
+	                       "Local              \n"
+	                       "Name               \n"
+	                       "Auto               ";
+
+	status_bar_message("");
+	assert_failure(exec_commands("filter?", &lwin, CIT_COMMAND));
+	assert_string_equal(expected, get_last_message());
+}
+
+TEST(filter_prints_non_empty_filters)
+{
+	const char *expected = "Filter -- Flags -- Value\n"
+	                       "Local     I        local\n"
+#ifndef _WIN32
+	                       "Name      I        abc\n"
+#else
+	                       "Name      i        abc\n"
+#endif
+	                       "Auto               ";
+
+	assert_success(exec_commands("filter abc", &lwin, CIT_COMMAND));
+	local_filter_apply(&lwin, "local");
+
+	status_bar_message("");
+	assert_failure(exec_commands("filter?", &lwin, CIT_COMMAND));
+	assert_string_equal(expected, get_last_message());
+}
+
+TEST(filter_with_empty_value_reuses_last_search)
+{
+#ifndef _WIN32
+#define I "I"
+#else
+#define I "i"
+#endif
+
+	const char *expected = "Filter -- Flags -- Value\n"
+	                       "Local              \n"
+	                       "Name      " I "        pattern\n"
+	                       "Auto               ";
+
+	cfg_resize_histories(5);
+	cfg_save_search_history("pattern");
+
+	assert_success(exec_commands("filter //" I, &lwin, CIT_COMMAND));
+	status_bar_message("");
+	assert_failure(exec_commands("filter?", &lwin, CIT_COMMAND));
+	assert_string_equal(expected, get_last_message());
+
+	assert_success(exec_commands("filter ''", &lwin, CIT_COMMAND));
+	status_bar_message("");
+	assert_failure(exec_commands("filter?", &lwin, CIT_COMMAND));
+	assert_string_equal(expected, get_last_message());
+
+	assert_success(exec_commands("filter \"\"", &lwin, CIT_COMMAND));
+	status_bar_message("");
+	assert_failure(exec_commands("filter?", &lwin, CIT_COMMAND));
+	assert_string_equal(expected, get_last_message());
+
+#undef I
+}
+
+TEST(filter_without_args_resets_manual_filter)
+{
+	const char *expected = "Filter -- Flags -- Value\n"
+	                       "Local              \n"
+	                       "Name               \n"
+	                       "Auto               ";
+
+	assert_success(exec_commands("filter this", &lwin, CIT_COMMAND));
+	assert_success(exec_commands("filter", &lwin, CIT_COMMAND));
+
+	status_bar_message("");
+	assert_failure(exec_commands("filter?", &lwin, CIT_COMMAND));
+	assert_string_equal(expected, get_last_message());
+}
+
+TEST(filter_can_affect_both_views)
+{
+	assert_string_equal("", curr_view->manual_filter.raw);
+	assert_string_equal("", other_view->manual_filter.raw);
+	curr_view->invert = 1;
+	other_view->invert = 1;
+
+	curr_stats.global_local_settings = 1;
+	assert_success(exec_commands("filter /x/", &lwin, CIT_COMMAND));
+	curr_stats.global_local_settings = 0;
+
+	assert_string_equal("x", curr_view->manual_filter.raw);
+	assert_string_equal("x", other_view->manual_filter.raw);
+	assert_false(curr_view->invert);
+	assert_false(other_view->invert);
+}
+
+TEST(filter_can_setup_inverted_filter)
+{
+	assert_string_equal("", curr_view->manual_filter.raw);
+	curr_view->invert = 0;
+
+	assert_success(exec_commands("filter! /x/", &lwin, CIT_COMMAND));
+
+	assert_string_equal("x", curr_view->manual_filter.raw);
+	assert_true(curr_view->invert);
+}
+
+TEST(filter_can_invert_manual_filter)
+{
+	curr_view->invert = 0;
+	assert_success(exec_commands("filter!", &lwin, CIT_COMMAND));
+	assert_true(curr_view->invert);
+	assert_success(exec_commands("filter!", &lwin, CIT_COMMAND));
+	assert_false(curr_view->invert);
 }
 
 /* vim: set tabstop=2 softtabstop=2 shiftwidth=2 noexpandtab cinoptions-=(0 : */
