@@ -7,6 +7,7 @@
 #include "../../src/ui/statusline.h"
 #include "../../src/utils/dynarray.h"
 #include "../../src/utils/str.h"
+#include "../../src/status.h"
 
 /* Checks that expanded string isn't equal to format string. */
 #define ASSERT_EXPANDED(format) \
@@ -77,6 +78,24 @@ TEST(t_macro_expanded)
 	ASSERT_EXPANDED("%t");
 }
 
+TEST(T_macro_expanded)
+{
+	lwin.dir_entry[0].type = FT_LINK;
+	ASSERT_EXPANDED("%T");
+}
+
+TEST(f_macro_expanded)
+{
+	ASSERT_EXPANDED("%f");
+}
+
+TEST(a_macro_expanded)
+{
+	ASSERT_EXPANDED("%a");
+	strcpy(lwin.curr_dir, SANDBOX_PATH);
+	ASSERT_EXPANDED("%a");
+}
+
 TEST(A_macro_expanded)
 {
 	ASSERT_EXPANDED("%A");
@@ -107,6 +126,14 @@ TEST(d_macro_expanded)
 	ASSERT_EXPANDED("%d");
 }
 
+TEST(D_macro_expanded)
+{
+	curr_stats.number_of_windows = 1;
+	ASSERT_EXPANDED("%D");
+	curr_stats.number_of_windows = 2;
+	ASSERT_EXPANDED("%D");
+}
+
 TEST(l_macro_expanded)
 {
 	ASSERT_EXPANDED("%l");
@@ -125,6 +152,7 @@ TEST(z_macro_expanded)
 TEST(dash_macro_expanded)
 {
 	ASSERT_EXPANDED("%-t");
+	ASSERT_EXPANDED("%-f");
 	ASSERT_EXPANDED("%-A");
 	ASSERT_EXPANDED("%-u");
 	ASSERT_EXPANDED("%-g");
@@ -149,7 +177,7 @@ TEST(percent_macro_expanded)
 
 TEST(wrong_macros_ignored)
 {
-	static const char STATUS_CHARS[] = "tAugsEd-lLS%[]z";
+	static const char STATUS_CHARS[] = "tTfaAugsEdD-lLS%[]z";
 	int i;
 
 	for(i = 1; i <= 255; ++i)
@@ -164,7 +192,7 @@ TEST(wrong_macros_ignored)
 
 TEST(wrong_macros_with_width_field_ignored)
 {
-	static const char STATUS_CHARS[] = "tAugsEd-lLS%[]z";
+	static const char STATUS_CHARS[] = "tTfaAugsEdD-lLS%[]z";
 	int i;
 
 	for(i = 1; i <= 255; ++i)
