@@ -12,6 +12,7 @@
 #include "../../src/cfg/info_chars.h"
 #include "../../src/compat/fs_limits.h"
 #include "../../src/compat/os.h"
+#include "../../src/ui/column_view.h"
 #include "../../src/ui/fileview.h"
 #include "../../src/ui/ui.h"
 #include "../../src/utils/dynarray.h"
@@ -22,6 +23,7 @@
 #include "../../src/cmd_core.h"
 #include "../../src/filelist.h"
 #include "../../src/filtering.h"
+#include "../../src/flist_hist.h"
 #include "../../src/macros.h"
 #include "../../src/registers.h"
 #include "../../src/running.h"
@@ -34,24 +36,17 @@ static void column_line_print(const void *data, int column_id, const char buf[],
 static void setup_custom_view(FileView *view, int very);
 static int filenames_can_include_newline(void);
 
-static char test_data[PATH_MAX];
+static char cwd[PATH_MAX + 1];
+static char test_data[PATH_MAX + 1];
 static const size_t MAX_WIDTH = 20;
 static char buf1[80 + 1];
 static char buf2[80 + 1];
 
 SETUP_ONCE()
 {
-	char cwd[PATH_MAX];
 	assert_non_null(get_cwd(cwd, sizeof(cwd)));
 
-	if(is_path_absolute(TEST_DATA_PATH))
-	{
-		snprintf(test_data, sizeof(test_data), "%s", TEST_DATA_PATH);
-	}
-	else
-	{
-		snprintf(test_data, sizeof(test_data), "%s/%s", cwd, TEST_DATA_PATH);
-	}
+	make_abs_path(test_data, sizeof(test_data), TEST_DATA_PATH, "", cwd);
 }
 
 SETUP()
@@ -68,6 +63,8 @@ SETUP()
 	other_view = &lwin;
 
 	snprintf(lwin.curr_dir, sizeof(lwin.curr_dir), "%s/..", test_data);
+
+	init_commands();
 }
 
 TEARDOWN()
@@ -76,12 +73,14 @@ TEARDOWN()
 	update_string(&cfg.fuse_home, NULL);
 
 	view_teardown(&lwin);
+
+	reset_cmds();
 }
 
 TEST(empty_list_is_not_accepted)
 {
 	flist_custom_start(&lwin, "test");
-	assert_false(flist_custom_finish(&lwin, 0) == 0);
+	assert_false(flist_custom_finish(&lwin, CV_REGULAR, 0) == 0);
 }
 
 TEST(duplicates_are_not_added)
@@ -89,7 +88,7 @@ TEST(duplicates_are_not_added)
 	flist_custom_start(&lwin, "test");
 	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/a");
 	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/a");
-	assert_true(flist_custom_finish(&lwin, 0) == 0);
+	assert_true(flist_custom_finish(&lwin, CV_REGULAR, 0) == 0);
 	assert_int_equal(1, lwin.list_rows);
 }
 
@@ -100,7 +99,7 @@ TEST(parent_dir_is_not_added_to_very_custom_view)
 	cfg.dot_dirs = DD_NONROOT_PARENT;
 	flist_custom_start(&lwin, "test");
 	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/a");
-	assert_true(flist_custom_finish(&lwin, 1) == 0);
+	assert_true(flist_custom_finish(&lwin, CV_VERY, 0) == 0);
 	assert_int_equal(1, lwin.list_rows);
 	cfg.dot_dirs = 0;
 
@@ -113,14 +112,14 @@ TEST(custom_view_replaces_custom_view_fine)
 
 	flist_custom_start(&lwin, "test");
 	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/a");
-	assert_true(flist_custom_finish(&lwin, 0) == 0);
+	assert_true(flist_custom_finish(&lwin, CV_REGULAR, 0) == 0);
 	assert_int_equal(1, lwin.list_rows);
 
 	assert_true(flist_custom_active(&lwin));
 
 	flist_custom_start(&lwin, "test");
 	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/b");
-	assert_true(flist_custom_finish(&lwin, 0) == 0);
+	assert_true(flist_custom_finish(&lwin, CV_REGULAR, 0) == 0);
 	assert_int_equal(1, lwin.list_rows);
 
 	assert_true(flist_custom_active(&lwin));
@@ -135,7 +134,7 @@ TEST(reload_considers_local_filter)
 	flist_custom_start(&lwin, "test");
 	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/a");
 	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/b");
-	assert_true(flist_custom_finish(&lwin, 0) == 0);
+	assert_true(flist_custom_finish(&lwin, CV_REGULAR, 0) == 0);
 
 	local_filter_apply(&lwin, "b");
 
@@ -163,11 +162,11 @@ TEST(reload_does_not_remove_broken_symlinks, IF(not_windows))
 
 	assert_false(flist_custom_active(&lwin));
 
-	copy_str(lwin.curr_dir, sizeof(lwin.curr_dir), SANDBOX_PATH);
+	assert_non_null(get_cwd(lwin.curr_dir, sizeof(lwin.curr_dir)));
 	flist_custom_start(&lwin, "test");
 	flist_custom_add(&lwin, test_file);
 	flist_custom_add(&lwin, "./broken-link");
-	assert_true(flist_custom_finish(&lwin, 0) == 0);
+	assert_true(flist_custom_finish(&lwin, CV_REGULAR, 0) == 0);
 
 	assert_int_equal(2, lwin.list_rows);
 	load_dir_list(&lwin, 1);
@@ -192,13 +191,13 @@ TEST(symlinks_to_dirs_are_recognized_as_dirs, IF(not_windows))
 
 	assert_false(flist_custom_active(&lwin));
 
-	copy_str(lwin.curr_dir, sizeof(lwin.curr_dir), SANDBOX_PATH);
+	assert_non_null(get_cwd(lwin.curr_dir, sizeof(lwin.curr_dir)));
 	flist_custom_start(&lwin, "test");
 	flist_custom_add(&lwin, "./dir-link");
-	assert_true(flist_custom_finish(&lwin, 0) == 0);
+	assert_true(flist_custom_finish(&lwin, CV_REGULAR, 0) == 0);
 
 	assert_int_equal(1, lwin.list_rows);
-	assert_true(is_directory_entry(&lwin.dir_entry[0]));
+	assert_true(fentry_is_dir(&lwin.dir_entry[0]));
 
 	assert_success(remove("dir-link"));
 }
@@ -212,7 +211,7 @@ TEST(locally_filtered_files_are_not_lost_on_reload)
 	flist_custom_start(&lwin, "test");
 	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/a");
 	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/b");
-	assert_true(flist_custom_finish(&lwin, 0) == 0);
+	assert_true(flist_custom_finish(&lwin, CV_REGULAR, 0) == 0);
 
 	local_filter_apply(&lwin, "b");
 
@@ -280,7 +279,7 @@ TEST(files_are_sorted_undecorated)
 	flist_custom_add(&lwin, "foo");
 	flist_custom_add(&lwin, "foo-");
 	flist_custom_add(&lwin, "foo0");
-	assert_success(flist_custom_finish(&lwin, 0));
+	assert_success(flist_custom_finish(&lwin, CV_REGULAR, 0));
 
 	assert_string_equal("foo", lwin.dir_entry[0].name);
 	assert_string_equal("foo-", lwin.dir_entry[1].name);
@@ -299,7 +298,7 @@ TEST(unsorted_custom_view_does_not_change_order_of_files)
 	flist_custom_start(&lwin, "test");
 	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/b");
 	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/a");
-	assert_true(flist_custom_finish(&lwin, 1) == 0);
+	assert_true(flist_custom_finish(&lwin, CV_VERY, 0) == 0);
 
 	assert_string_equal("b", lwin.dir_entry[0].name);
 	assert_string_equal("a", lwin.dir_entry[1].name);
@@ -319,12 +318,12 @@ TEST(sorted_custom_view_after_unsorted)
 	flist_custom_start(&lwin, "test");
 	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/b");
 	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/a");
-	assert_true(flist_custom_finish(&lwin, 1) == 0);
+	assert_true(flist_custom_finish(&lwin, CV_VERY, 0) == 0);
 
 	flist_custom_start(&lwin, "test");
 	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/b");
 	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/a");
-	assert_true(flist_custom_finish(&lwin, 0) == 0);
+	assert_true(flist_custom_finish(&lwin, CV_REGULAR, 0) == 0);
 
 	assert_string_equal("a", lwin.dir_entry[0].name);
 	assert_string_equal("b", lwin.dir_entry[1].name);
@@ -344,8 +343,8 @@ TEST(unsorted_view_remains_one_on_vifminfo_reread_on_restart)
 	flist_custom_start(&lwin, "test");
 	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/b");
 	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/a");
-	assert_true(flist_custom_finish(&lwin, 1) == 0);
-	assert_true(lwin.custom.unsorted);
+	assert_true(flist_custom_finish(&lwin, CV_VERY, 0) == 0);
+	assert_true(lwin.custom.type == CV_VERY);
 	assert_int_equal(SK_NONE, lwin.sort[0]);
 
 	/* ls-like view blocks view column updates. */
@@ -354,7 +353,7 @@ TEST(unsorted_view_remains_one_on_vifminfo_reread_on_restart)
 	read_info_file(1);
 	curr_stats.restart_in_progress = 0;
 
-	assert_true(lwin.custom.unsorted);
+	assert_true(lwin.custom.type == CV_VERY);
 	assert_int_equal(SK_NONE, lwin.sort[0]);
 
 	opt_handlers_teardown();
@@ -378,7 +377,7 @@ TEST(location_is_saved_on_entering_custom_view)
 	copy_str(lwin.curr_dir, sizeof(lwin.curr_dir), cwd);
 
 	/* Put specific history entry and make sure it's used. */
-	save_view_history(&lwin, lwin.curr_dir, "b", 1);
+	flist_hist_save(&lwin, lwin.curr_dir, "b", 1);
 	load_dir_list(&lwin, 0);
 	assert_string_equal(lwin.curr_dir, cwd);
 	assert_string_equal("b", lwin.dir_entry[lwin.list_pos].name);
@@ -429,7 +428,8 @@ TEST(custom_view_does_not_reset_local_state)
 
 	for(very = 0; very < 2; ++very)
 	{
-		column_data_t cdt = { .view = &lwin };
+		size_t prefix_len = 0U;
+		column_data_t cdt = { .view = &lwin, .prefix_len = &prefix_len };
 
 		lwin.columns = columns_create();
 
@@ -473,7 +473,7 @@ TEST(custom_view_does_not_reset_local_state)
 		}
 
 		columns_free(lwin.columns);
-		lwin.columns = NULL_COLUMNS;
+		lwin.columns = NULL;
 	}
 
 	opt_handlers_teardown();
@@ -482,8 +482,13 @@ TEST(custom_view_does_not_reset_local_state)
 
 TEST(files_with_newline_in_names, IF(filenames_can_include_newline))
 {
-	FILE *const f = fopen(SANDBOX_PATH "/list", "w");
-	fprintf(f, "%s%c", SANDBOX_PATH "/a\nb", '\0');
+	char path[PATH_MAX + 1];
+	FILE *f;
+
+	make_abs_path(path, sizeof(path), SANDBOX_PATH, "a\nb", cwd);
+
+	f = fopen(SANDBOX_PATH "/list", "w");
+	fprintf(f, "%s%c", path, '\0');
 	fclose(f);
 
 	assert_success(chdir(SANDBOX_PATH));
@@ -493,7 +498,7 @@ TEST(files_with_newline_in_names, IF(filenames_can_include_newline))
 
 	create_file("a\nb");
 	assert_non_null(get_cwd(lwin.curr_dir, sizeof(lwin.curr_dir)));
-	assert_success(output_to_custom_flist(&lwin, "cat list", 0));
+	assert_success(output_to_custom_flist(&lwin, "cat list", 0, 0));
 	assert_success(unlink("a\nb"));
 	assert_success(unlink("list"));
 
@@ -520,10 +525,10 @@ TEST(can_set_very_cv_twice_in_a_row)
 	setup_custom_view(&lwin, 1);
 	flist_custom_start(&lwin, "test");
 	flist_custom_add(&lwin, "a");
-	assert_true(flist_custom_finish(&lwin, 1) == 0);
+	assert_true(flist_custom_finish(&lwin, CV_VERY, 0) == 0);
 
 	columns_free(lwin.columns);
-	lwin.columns = NULL_COLUMNS;
+	lwin.columns = NULL;
 
 	opt_handlers_teardown();
 	columns_clear_column_descs();
@@ -538,7 +543,7 @@ TEST(renaming_dir_in_cv_adjust_its_children_entries)
 	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files");
 	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/a");
 	copy_str(lwin.curr_dir, sizeof(lwin.curr_dir), test_data);
-	assert_true(flist_custom_finish(&lwin, 0) == 0);
+	assert_true(flist_custom_finish(&lwin, CV_REGULAR, 0) == 0);
 	assert_int_equal(2, lwin.list_rows);
 
 	fentry_rename(&lwin, &lwin.dir_entry[0], "existing_files");
@@ -548,22 +553,220 @@ TEST(renaming_dir_in_cv_adjust_its_children_entries)
 
 TEST(symlinks_are_not_resolved_in_origins, IF(not_windows))
 {
+	char path[PATH_MAX + 1];
+
+	/* symlink() is not available on Windows, but the rest of the code is fine. */
 #ifndef _WIN32
-	assert_success(symlink(TEST_DATA_PATH "/existing-files",
-				SANDBOX_PATH "/link"));
+	{
+		char src[PATH_MAX + 1], dst[PATH_MAX + 1];
+		make_abs_path(src, sizeof(src), TEST_DATA_PATH, "existing-files", cwd);
+		make_abs_path(dst, sizeof(dst), SANDBOX_PATH, "link", cwd);
+		assert_success(symlink(src, dst));
+	}
 #endif
 
 	assert_success(chdir(SANDBOX_PATH "/link"));
-	copy_str(lwin.curr_dir, sizeof(lwin.curr_dir), SANDBOX_PATH "/link");
+	make_abs_path(lwin.curr_dir, sizeof(lwin.curr_dir), SANDBOX_PATH, "link",
+			cwd);
 	flist_custom_start(&lwin, "test");
-	flist_custom_add(&lwin, SANDBOX_PATH "/link"); /* Absolute path. */
-	flist_custom_add(&lwin, "a");                  /* Relative path. */
-	assert_true(flist_custom_finish(&lwin, 0) == 0);
+	make_abs_path(path, sizeof(path), SANDBOX_PATH, "link", cwd);
+	flist_custom_add(&lwin, path); /* Absolute path. */
+	flist_custom_add(&lwin, "a");  /* Relative path. */
+	assert_true(flist_custom_finish(&lwin, CV_REGULAR, 0) == 0);
 
-	assert_true(paths_are_equal(SANDBOX_PATH, lwin.dir_entry[0].origin));
-	assert_true(paths_are_equal(SANDBOX_PATH "/link", lwin.dir_entry[1].origin));
+	make_abs_path(path, sizeof(path), SANDBOX_PATH, "", cwd);
+	assert_true(paths_are_equal(path, lwin.dir_entry[0].origin));
+	make_abs_path(path, sizeof(path), SANDBOX_PATH, "link", cwd);
+	assert_true(paths_are_equal(path, lwin.dir_entry[1].origin));
 
-	assert_success(remove(SANDBOX_PATH "/link"));
+	assert_success(remove(path));
+}
+
+TEST(files_are_excluded_from_custom_view)
+{
+	setup_custom_view(&lwin, 0);
+
+	lwin.dir_entry[0].selected = 1;
+	lwin.selected_files = 1;
+	flist_custom_exclude(&lwin, 1);
+
+	assert_int_equal(0, lwin.selected_files);
+}
+
+TEST(cv_flist_title_is_preserved_on_next_cv_load_failure)
+{
+	setup_custom_view(&lwin, 0);
+	assert_string_equal("test", lwin.custom.title);
+
+	flist_custom_start(&lwin, "test-2");
+	assert_false(flist_custom_finish(&lwin, CV_REGULAR, 0) == 0);
+
+	assert_string_equal("test", lwin.custom.title);
+}
+
+TEST(applying_local_filter_saves_custom_list)
+{
+	filters_view_reset(&lwin);
+
+	assert_false(flist_custom_active(&lwin));
+
+	flist_custom_start(&lwin, "test");
+	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/a");
+	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/b");
+	assert_true(flist_custom_finish(&lwin, CV_REGULAR, 0) == 0);
+
+	local_filter_apply(&lwin, "b");
+	load_dir_list(&lwin, 1);
+	assert_int_equal(1, lwin.list_rows);
+
+	local_filter_remove(&lwin);
+	load_dir_list(&lwin, 1);
+	assert_int_equal(2, lwin.list_rows);
+
+	filter_dispose(&lwin.manual_filter);
+	filter_dispose(&lwin.auto_filter);
+}
+
+TEST(excluded_entries_do_not_return)
+{
+	filters_view_reset(&lwin);
+
+	assert_false(flist_custom_active(&lwin));
+
+	flist_custom_start(&lwin, "test");
+	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/a");
+	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/b");
+	assert_true(flist_custom_finish(&lwin, CV_REGULAR, 0) == 0);
+
+	local_filter_apply(&lwin, ".");
+	load_dir_list(&lwin, 1);
+	assert_int_equal(2, lwin.list_rows);
+
+	flist_custom_exclude(&lwin, 0);
+	assert_int_equal(1, lwin.list_rows);
+
+	load_dir_list(&lwin, 1);
+	assert_int_equal(1, lwin.list_rows);
+
+	filter_dispose(&lwin.manual_filter);
+	filter_dispose(&lwin.auto_filter);
+}
+
+TEST(excluding_entries_does_not_affect_local_filter_list)
+{
+	filters_view_reset(&lwin);
+
+	assert_false(flist_custom_active(&lwin));
+
+	flist_custom_start(&lwin, "test");
+	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/a");
+	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/b");
+	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/c");
+	assert_true(flist_custom_finish(&lwin, CV_REGULAR, 0) == 0);
+
+	local_filter_apply(&lwin, "[ab]");
+	load_dir_list(&lwin, 1);
+	assert_int_equal(2, lwin.list_rows);
+
+	flist_custom_exclude(&lwin, 0);
+	assert_int_equal(1, lwin.list_rows);
+
+	local_filter_remove(&lwin);
+	load_dir_list(&lwin, 1);
+	assert_int_equal(2, lwin.list_rows);
+
+	filter_dispose(&lwin.manual_filter);
+	filter_dispose(&lwin.auto_filter);
+}
+
+TEST(failed_loadin_of_cv_does_not_override_saved_list)
+{
+	filters_view_reset(&lwin);
+
+	assert_false(flist_custom_active(&lwin));
+
+	flist_custom_start(&lwin, "test");
+	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/a");
+	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/b");
+	assert_true(flist_custom_finish(&lwin, CV_REGULAR, 0) == 0);
+
+	local_filter_apply(&lwin, "b");
+	load_dir_list(&lwin, 1);
+	assert_int_equal(1, lwin.list_rows);
+
+	flist_custom_start(&lwin, "test-2");
+	assert_false(flist_custom_finish(&lwin, CV_REGULAR, 0) == 0);
+
+	local_filter_remove(&lwin);
+	load_dir_list(&lwin, 1);
+	assert_int_equal(2, lwin.list_rows);
+
+	filter_dispose(&lwin.manual_filter);
+	filter_dispose(&lwin.auto_filter);
+}
+
+TEST(loading_cv_resets_search_results)
+{
+	lwin.matches = 100;
+
+	flist_custom_start(&lwin, "test");
+	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/a");
+	assert_true(flist_custom_finish(&lwin, CV_REGULAR, 0) == 0);
+
+	/* We need to reset the count as its non-zero value implies that search has
+	 * already been done. */
+	assert_int_equal(0, lwin.matches);
+}
+
+TEST(cursor_is_positioned_close_to_disappeared_file)
+{
+	char *const saved_cwd = save_cwd();
+
+	make_abs_path(lwin.curr_dir, sizeof(lwin.curr_dir), SANDBOX_PATH, "", cwd);
+	assert_success(chdir(SANDBOX_PATH));
+	create_file("1");
+	create_file("2");
+	create_file("3");
+	create_file("4");
+	create_file("5");
+	create_file("6");
+
+	filters_view_reset(&lwin);
+
+	flist_custom_start(&lwin, "test");
+	flist_custom_add(&lwin, "1");
+	flist_custom_add(&lwin, "2");
+	flist_custom_add(&lwin, "3");
+	flist_custom_add(&lwin, "4");
+	flist_custom_add(&lwin, "5");
+	flist_custom_add(&lwin, "6");
+	assert_true(flist_custom_finish(&lwin, CV_REGULAR, 0) == 0);
+	assert_int_equal(6, lwin.list_rows);
+
+	local_filter_apply(&lwin, "[4-6]");
+	load_dir_list(&lwin, 1);
+	assert_int_equal(3, lwin.list_rows);
+
+	lwin.list_pos = 2;
+	assert_success(unlink("6"));
+	/* This reloads view zipping removed and filtered-out files.  It should try to
+	 * stay close to original position. */
+	load_dir_list(&lwin, 1);
+	assert_int_equal(2, lwin.list_rows);
+
+	/* Both 2 and 1 are valid answers, list_pos is automatically corrected. */
+	assert_true(lwin.list_pos == 1 || lwin.list_pos == 2);
+
+	filter_dispose(&lwin.manual_filter);
+	filter_dispose(&lwin.auto_filter);
+
+	restore_cwd(saved_cwd);
+	assert_success(chdir(SANDBOX_PATH));
+	assert_success(unlink("1"));
+	assert_success(unlink("2"));
+	assert_success(unlink("3"));
+	assert_success(unlink("4"));
+	assert_success(unlink("5"));
 }
 
 static void
@@ -583,7 +786,7 @@ setup_custom_view(FileView *view, int very)
 	flist_custom_add(view, TEST_DATA_PATH "/existing-files/a");
 	snprintf(view->curr_dir, sizeof(view->curr_dir), "%s/existing-files",
 			test_data);
-	assert_true(flist_custom_finish(view, very) == 0);
+	assert_true(flist_custom_finish(view, very ? CV_VERY : CV_REGULAR, 0) == 0);
 }
 
 static int
