@@ -5,8 +5,11 @@
 
 #include "../../src/cfg/config.h"
 #include "../../src/utils/dynarray.h"
+#include "../../src/utils/str.h"
 #include "../../src/cmd_core.h"
-#include "../../src/filelist.h"
+#include "../../src/flist_hist.h"
+
+#include "utils.h"
 
 #define INITIAL_SIZE 10
 
@@ -24,6 +27,10 @@
 
 SETUP()
 {
+	update_string(&cfg.slow_fs_list, "");
+
+	view_setup(&lwin);
+
 	/* lwin */
 	strcpy(lwin.curr_dir, "/lwin");
 
@@ -50,17 +57,19 @@ SETUP()
 	cfg_resize_histories(0);
 
 	cfg_resize_histories(INITIAL_SIZE);
+
+	curr_view = NULL;
 }
 
 TEARDOWN()
 {
 	int i;
 
+	update_string(&cfg.slow_fs_list, NULL);
+
 	cfg_resize_histories(0);
 
-	for(i = 0; i < lwin.list_rows; i++)
-		free(lwin.dir_entry[i].name);
-	dynarray_free(lwin.dir_entry);
+	view_teardown(&lwin);
 
 	for(i = 0; i < rwin.list_rows; i++)
 		free(rwin.dir_entry[i].name);
@@ -74,8 +83,8 @@ save_to_history(const char str[])
 	cfg_save_search_history(str);
 	cfg_save_prompt_history(str);
 
-	save_view_history(&lwin, str, str + 1, 0);
-	save_view_history(&rwin, str, str + 1, 0);
+	flist_hist_save(&lwin, str, str + 1, 0);
+	flist_hist_save(&rwin, str, str + 1, 0);
 }
 
 TEST(view_history_after_reset_contains_valid_data)
@@ -92,8 +101,8 @@ TEST(view_history_avoids_duplicates)
 	assert_int_equal(1, lwin.history_num);
 	assert_int_equal(1, rwin.history_num);
 
-	save_view_history(&lwin, NULL, NULL, -1);
-	save_view_history(&rwin, NULL, NULL, -1);
+	flist_hist_save(&lwin, NULL, NULL, -1);
+	flist_hist_save(&rwin, NULL, NULL, -1);
 
 	assert_int_equal(1, lwin.history_num);
 	assert_int_equal(1, rwin.history_num);
@@ -142,5 +151,47 @@ TEST(add_after_increasing_ok)
 	}
 }
 
+TEST(navigating_within_history)
+{
+	save_to_history(SANDBOX_PATH);
+	save_to_history(TEST_DATA_PATH);
+
+	flist_hist_go_back(&lwin);
+
+	/* This overwrites previously active history entries and should not result in
+	 * memory leaks. */
+	save_to_history("somewhere");
+}
+
+TEST(specified_file_position_is_unaffected_by_top_line)
+{
+	lwin.top_line = 3;
+	flist_hist_save(&lwin, "/dir", "file", 0);
+	assert_string_equal("/dir", lwin.history[1].dir);
+	assert_string_equal("file", lwin.history[1].file);
+	assert_int_equal(0, lwin.history[1].rel_pos);
+}
+
+TEST(history_size_reduction_leaves_correct_number_of_elements)
+{
+	assert_int_equal(1, lwin.history_num);
+	flist_hist_save(&lwin, "/dir1", "file1", 1);
+	flist_hist_save(&lwin, "/dir2", "file2", 2);
+	assert_int_equal(2, lwin.history_pos);
+	assert_int_equal(3, lwin.history_num);
+
+	cfg_resize_histories(2);
+
+	assert_int_equal(1, lwin.history_pos);
+	assert_int_equal(2, lwin.history_num);
+
+	assert_string_equal("/dir1", lwin.history[0].dir);
+	assert_string_equal("file1", lwin.history[0].file);
+	assert_int_equal(1, lwin.history[0].rel_pos);
+	assert_string_equal("/dir2", lwin.history[1].dir);
+	assert_string_equal("file2", lwin.history[1].file);
+	assert_int_equal(2, lwin.history[1].rel_pos);
+}
+
 /* vim: set tabstop=2 softtabstop=2 shiftwidth=2 noexpandtab cinoptions-=(0 : */
 /* vim: set cinoptions+=t0 filetype=c : */
