@@ -72,13 +72,13 @@ free_view(FileView *view)
 
 	for(i = 0; i < view->list_rows; ++i)
 	{
-		free_dir_entry(view, &view->dir_entry[i]);
+		fentry_free(view, &view->dir_entry[i]);
 	}
 	dynarray_free(view->dir_entry);
 
 	for(i = 0; i < view->custom.entry_count; ++i)
 	{
-		free_dir_entry(view, &view->custom.entries[i]);
+		fentry_free(view, &view->custom.entries[i]);
 	}
 	dynarray_free(view->custom.entries);
 
@@ -91,7 +91,7 @@ TEST(custom_view_is_preserved_on_goto_mark)
 {
 	flist_custom_start(&lwin, "test");
 	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/a");
-	assert_true(flist_custom_finish(&lwin, 0) == 0);
+	assert_true(flist_custom_finish(&lwin, CV_REGULAR, 0) == 0);
 
 	curr_view = &lwin;
 	assert_success(set_user_mark('a', TEST_DATA_PATH "/existing-files", "a"));
@@ -104,7 +104,7 @@ TEST(local_filter_is_reset_in_cv_to_follow_mark)
 {
 	flist_custom_start(&lwin, "test");
 	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/a");
-	assert_true(flist_custom_finish(&lwin, 0) == 0);
+	assert_true(flist_custom_finish(&lwin, CV_REGULAR, 0) == 0);
 
 	curr_view = &lwin;
 	assert_success(set_user_mark('a', TEST_DATA_PATH "/existing-files", "a"));
@@ -124,7 +124,7 @@ TEST(following_resets_cv)
 {
 	flist_custom_start(&lwin, "test");
 	flist_custom_add(&lwin, TEST_DATA_PATH "/existing-files/a");
-	assert_true(flist_custom_finish(&lwin, 0) == 0);
+	assert_true(flist_custom_finish(&lwin, CV_REGULAR, 0) == 0);
 
 	follow_file(&lwin);
 	assert_false(flist_custom_active(&lwin));
