@@ -4,18 +4,24 @@
 #include <unistd.h> /* access() */
 
 #include <stddef.h> /* NULL */
-#include <stdio.h> /* fclose() fopen() */
+#include <stdio.h> /* FILE fclose() fopen() fread() */
 #include <string.h> /* memset() strcpy() */
 
 #include "../../src/cfg/config.h"
+#include "../../src/compat/os.h"
 #include "../../src/engine/options.h"
 #include "../../src/ui/ui.h"
 #include "../../src/utils/dynarray.h"
+#include "../../src/utils/fswatch.h"
 #include "../../src/utils/path.h"
 #include "../../src/utils/str.h"
 #include "../../src/filelist.h"
 #include "../../src/filtering.h"
 #include "../../src/opt_handlers.h"
+#include "../../src/undo.h"
+
+static int exec_func(OPS op, void *data, const char *src, const char *dst);
+static int op_avail(OPS op);
 
 void
 opt_handlers_setup(void)
@@ -77,6 +83,31 @@ opt_handlers_teardown(void)
 	update_string(&rwin.sort_groups_g, NULL);
 }
 
+void
+undo_setup(void)
+{
+	static int max_undo_levels = 0;
+	init_undo_list(&exec_func, &op_avail, NULL, &max_undo_levels);
+}
+
+static int
+exec_func(OPS op, void *data, const char *src, const char *dst)
+{
+	return 0;
+}
+
+static int
+op_avail(OPS op)
+{
+	return 0;
+}
+
+void
+undo_teardown(void)
+{
+	reset_undo_list();
+}
+
 void
 view_setup(FileView *view)
 {
@@ -84,6 +115,10 @@ view_setup(FileView *view)
 	view->filtered = 0;
 	view->list_pos = 0;
 	view->dir_entry = NULL;
+	view->hide_dot = 0;
+	view->hide_dot_g = 0;
+	view->invert = 1;
+	view->selected_files = 0;
 
 	assert_success(filter_init(&view->local_filter.filter, 1));
 	assert_success(filter_init(&view->manual_filter, 1));
@@ -94,6 +129,9 @@ view_setup(FileView *view)
 
 	view->sort[0] = SK_BY_NAME;
 	memset(&view->sort[1], SK_NONE, sizeof(view->sort) - 1);
+
+	view->custom.entry_count = 0;
+	view->custom.entries = NULL;
 }
 
 void
@@ -103,19 +141,24 @@ view_teardown(FileView *view)
 
 	for(i = 0; i < view->list_rows; ++i)
 	{
-		free_dir_entry(view, &view->dir_entry[i]);
+		fentry_free(view, &view->dir_entry[i]);
 	}
 	dynarray_free(view->dir_entry);
 
 	for(i = 0; i < view->custom.entry_count; ++i)
 	{
-		free_dir_entry(view, &view->custom.entries[i]);
+		fentry_free(view, &view->custom.entries[i]);
 	}
 	dynarray_free(view->custom.entries);
 
 	filter_dispose(&view->local_filter.filter);
 	filter_dispose(&view->auto_filter);
 	filter_dispose(&view->manual_filter);
+
+	view->custom.type = CV_REGULAR;
+
+	fswatch_free(view->watch);
+	view->watch = NULL;
 }
 
 void
@@ -144,14 +187,35 @@ make_abs_path(char buf[], size_t buf_len, const char base[], const char sub[],
 {
 	if(is_path_absolute(base))
 	{
-		snprintf(buf, buf_len, "%s/%s", base, sub);
+		snprintf(buf, buf_len, "%s%s%s", base, (sub[0] == '\0' ? "" : "/"), sub);
 	}
 	else
 	{
-		snprintf(buf, buf_len, "%s/%s/%s", cwd, base, sub);
+		snprintf(buf, buf_len, "%s/%s%s%s", cwd, base, (sub[0] == '\0' ? "" : "/"),
+				sub);
 	}
 }
 
+void
+copy_file(const char src[], const char dst[])
+{
+	char buf[4*1024];
+	size_t nread;
+	FILE *const in = os_fopen(src, "rb");
+	FILE *const out = os_fopen(dst, "wb");
+
+	assert_non_null(in);
+	assert_non_null(out);
+
+	while((nread = fread(&buf, 1, sizeof(buf), in)) != 0U)
+	{
+		assert_int_equal(nread, fwrite(&buf, 1, nread, out));
+	}
+
+	fclose(out);
+	fclose(in);
+}
+
 int
 not_windows(void)
 {
