@@ -18,6 +18,12 @@ void opt_handlers_setup(void);
 /* Cleans up option handlers. */
 void opt_handlers_teardown(void);
 
+/* Prepares undo unit for file operations. */
+void undo_setup(void);
+
+/* Cleans up undo unit. */
+void undo_teardown(void);
+
 /* Initializes view with safe defaults. */
 void view_setup(FileView *view);
 
@@ -34,7 +40,10 @@ void create_executable(const char path[]);
 void make_abs_path(char buf[], size_t buf_len, const char base[],
 		const char sub[], const char cwd[]);
 
-/* Whether running on non-Windowsla.  Returns non-zero if so, otherwise zero is
+/* Copies file at src to dst. */
+void copy_file(const char src[], const char dst[]);
+
+/* Whether running on non-Windows.  Returns non-zero if so, otherwise zero is
  * returned. */
 int not_windows(void);
 
