@@ -86,7 +86,8 @@ static stic_void_void stic_suite_teardown_func = 0;
 static stic_void_void stic_fixture_setup = 0;
 static stic_void_void stic_fixture_teardown = 0;
 
-const char *stic_current_test;
+const char *stic_current_test_name;
+stic_void_void stic_current_test;
 const char *stic_suite_name;
 
 void (*stic_simple_test_result)(int passed, char* reason, const char* function, const char file[], unsigned int line) = stic_simple_test_result_log;
@@ -143,12 +144,12 @@ void stic_suite_teardown( void )
 
 void fixture_setup(void (*setup)( void ))
 {
-	stic_current_test = "<setup>";
+	stic_current_test_name = "<setup>";
 	stic_fixture_setup = setup;
 }
 void fixture_teardown(void (*teardown)( void ))
 {
-	stic_current_test = "<teardown>";
+	stic_current_test_name = "<teardown>";
 	stic_fixture_teardown = teardown;
 }
 
@@ -183,11 +184,11 @@ static int test_had_output(void)
 
 void stic_simple_test_result_log(int passed, char* reason, const char* function, const char file[], unsigned int line)
 {
-	static const char *last_test;
+	static stic_void_void last_test;
 
-	const char *test_name = (stic_current_test == last_test) ? "" : stic_current_test;
+	const char *test_name = (stic_current_test == last_test) ? "" : stic_current_test_name;
 
-	if (stic_current_test != NULL && strcmp(function, stic_current_test) == 0)
+	if (stic_current_test_name != NULL && strcmp(function, stic_current_test_name) == 0)
 	{
 		function = "test body";
 	}
@@ -206,7 +207,7 @@ void stic_simple_test_result_log(int passed, char* reason, const char* function,
 	{
 		if(stic_machine_readable)
 		{
-			printf("%s%s,%s,%u,%s\n", stic_magic_marker, stic_current_fixture_path, stic_current_test, line, reason);
+			printf("%s%s,%s,%u,%s\n", stic_magic_marker, stic_current_fixture_path, stic_current_test_name, line, reason);
 		}
 		else
 		{
@@ -226,7 +227,7 @@ void stic_simple_test_result_log(int passed, char* reason, const char* function,
 		{
 			if(stic_machine_readable)
 			{
-				printf("%s%s,%s,%u,Passed\n", stic_magic_marker, stic_current_fixture_path, stic_current_test, line);
+				printf("%s%s,%s,%u,Passed\n", stic_magic_marker, stic_current_fixture_path, stic_current_test_name, line);
 			}
 			else
 			{
@@ -534,7 +535,7 @@ int run_tests(stic_void_void tests)
 				 (stic_suite_name[0] != '\0') ? " " : "", stic_suite_name);
 		stic_header_printer(s, stic_screen_width, ' ');
 	}
-	sprintf(s,"%d tests run", sea_tests_run);
+	sprintf(s,"%d test%s run", sea_tests_run, sea_tests_run == 1 ? "" : "s");
 	stic_header_printer(s, stic_screen_width, ' ');
 
 	if (end - start == 0)
